/* Generated by wayland-scanner 1.24.0 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include "wayland-util.h"

#ifndef __has_attribute
# define __has_attribute(x) 0  /* Compatibility with non-clang compilers. */
#endif

#if (__has_attribute(visibility) || defined(__GNUC__) && __GNUC__ >= 4)
#define WL_PRIVATE __attribute__ ((visibility("hidden")))
#else
#define WL_PRIVATE
#endif

extern const struct wl_interface wl_surface_interface;

static const struct wl_interface *mutter_x11_interop_types[] = {
	&wl_surface_interface,
	NULL,
};

static const struct wl_message mutter_x11_interop_requests[] = {
	{ "destroy", "", mutter_x11_interop_types + 0 },
	{ "set_x11_parent", "ou", mutter_x11_interop_types + 0 },
};

WL_PRIVATE const struct wl_interface mutter_x11_interop_interface = {
	"mutter_x11_interop", 1,
	2, mutter_x11_interop_requests,
	0, NULL,
};

