/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "shell-dbus.h"

#include <string.h>
#ifdef G_OS_UNIX
#  include <gio/gunixfdlist.h>
#endif

#ifdef G_ENABLE_DEBUG
#define g_marshal_value_peek_boolean(v)  g_value_get_boolean (v)
#define g_marshal_value_peek_char(v)     g_value_get_schar (v)
#define g_marshal_value_peek_uchar(v)    g_value_get_uchar (v)
#define g_marshal_value_peek_int(v)      g_value_get_int (v)
#define g_marshal_value_peek_uint(v)     g_value_get_uint (v)
#define g_marshal_value_peek_long(v)     g_value_get_long (v)
#define g_marshal_value_peek_ulong(v)    g_value_get_ulong (v)
#define g_marshal_value_peek_int64(v)    g_value_get_int64 (v)
#define g_marshal_value_peek_uint64(v)   g_value_get_uint64 (v)
#define g_marshal_value_peek_enum(v)     g_value_get_enum (v)
#define g_marshal_value_peek_flags(v)    g_value_get_flags (v)
#define g_marshal_value_peek_float(v)    g_value_get_float (v)
#define g_marshal_value_peek_double(v)   g_value_get_double (v)
#define g_marshal_value_peek_string(v)   (char*) g_value_get_string (v)
#define g_marshal_value_peek_param(v)    g_value_get_param (v)
#define g_marshal_value_peek_boxed(v)    g_value_get_boxed (v)
#define g_marshal_value_peek_pointer(v)  g_value_get_pointer (v)
#define g_marshal_value_peek_object(v)   g_value_get_object (v)
#define g_marshal_value_peek_variant(v)  g_value_get_variant (v)
#else /* !G_ENABLE_DEBUG */
/* WARNING: This code accesses GValues directly, which is UNSUPPORTED API.
 *          Do not access GValues directly in your code. Instead, use the
 *          g_value_get_*() functions
 */
#define g_marshal_value_peek_boolean(v)  (v)->data[0].v_int
#define g_marshal_value_peek_char(v)     (v)->data[0].v_int
#define g_marshal_value_peek_uchar(v)    (v)->data[0].v_uint
#define g_marshal_value_peek_int(v)      (v)->data[0].v_int
#define g_marshal_value_peek_uint(v)     (v)->data[0].v_uint
#define g_marshal_value_peek_long(v)     (v)->data[0].v_long
#define g_marshal_value_peek_ulong(v)    (v)->data[0].v_ulong
#define g_marshal_value_peek_int64(v)    (v)->data[0].v_int64
#define g_marshal_value_peek_uint64(v)   (v)->data[0].v_uint64
#define g_marshal_value_peek_enum(v)     (v)->data[0].v_long
#define g_marshal_value_peek_flags(v)    (v)->data[0].v_ulong
#define g_marshal_value_peek_float(v)    (v)->data[0].v_float
#define g_marshal_value_peek_double(v)   (v)->data[0].v_double
#define g_marshal_value_peek_string(v)   (v)->data[0].v_pointer
#define g_marshal_value_peek_param(v)    (v)->data[0].v_pointer
#define g_marshal_value_peek_boxed(v)    (v)->data[0].v_pointer
#define g_marshal_value_peek_pointer(v)  (v)->data[0].v_pointer
#define g_marshal_value_peek_object(v)   (v)->data[0].v_pointer
#define g_marshal_value_peek_variant(v)  (v)->data[0].v_pointer
#endif /* !G_ENABLE_DEBUG */

typedef struct
{
  GDBusArgInfo parent_struct;
  gboolean use_gvariant;
} _ExtendedGDBusArgInfo;

typedef struct
{
  GDBusMethodInfo parent_struct;
  const gchar *signal_name;
  gboolean pass_fdlist;
} _ExtendedGDBusMethodInfo;

typedef struct
{
  GDBusSignalInfo parent_struct;
  const gchar *signal_name;
} _ExtendedGDBusSignalInfo;

typedef struct
{
  GDBusPropertyInfo parent_struct;
  const gchar *hyphen_name;
  guint use_gvariant : 1;
  guint emits_changed_signal : 1;
} _ExtendedGDBusPropertyInfo;

typedef struct
{
  GDBusInterfaceInfo parent_struct;
  const gchar *hyphen_name;
} _ExtendedGDBusInterfaceInfo;

typedef struct
{
  const _ExtendedGDBusPropertyInfo *info;
  guint prop_id;
  GValue orig_value; /* the value before the change */
} ChangedProperty;

static void
_changed_property_free (ChangedProperty *data)
{
  g_value_unset (&data->orig_value);
  g_free (data);
}

static gboolean
_g_strv_equal0 (gchar **a, gchar **b)
{
  gboolean ret = FALSE;
  guint n;
  if (a == NULL && b == NULL)
    {
      ret = TRUE;
      goto out;
    }
  if (a == NULL || b == NULL)
    goto out;
  if (g_strv_length (a) != g_strv_length (b))
    goto out;
  for (n = 0; a[n] != NULL; n++)
    if (g_strcmp0 (a[n], b[n]) != 0)
      goto out;
  ret = TRUE;
out:
  return ret;
}

static gboolean
_g_variant_equal0 (GVariant *a, GVariant *b)
{
  gboolean ret = FALSE;
  if (a == NULL && b == NULL)
    {
      ret = TRUE;
      goto out;
    }
  if (a == NULL || b == NULL)
    goto out;
  ret = g_variant_equal (a, b);
out:
  return ret;
}

G_GNUC_UNUSED static gboolean
_g_value_equal (const GValue *a, const GValue *b)
{
  gboolean ret = FALSE;
  g_assert (G_VALUE_TYPE (a) == G_VALUE_TYPE (b));
  switch (G_VALUE_TYPE (a))
    {
      case G_TYPE_BOOLEAN:
        ret = (g_value_get_boolean (a) == g_value_get_boolean (b));
        break;
      case G_TYPE_UCHAR:
        ret = (g_value_get_uchar (a) == g_value_get_uchar (b));
        break;
      case G_TYPE_INT:
        ret = (g_value_get_int (a) == g_value_get_int (b));
        break;
      case G_TYPE_UINT:
        ret = (g_value_get_uint (a) == g_value_get_uint (b));
        break;
      case G_TYPE_INT64:
        ret = (g_value_get_int64 (a) == g_value_get_int64 (b));
        break;
      case G_TYPE_UINT64:
        ret = (g_value_get_uint64 (a) == g_value_get_uint64 (b));
        break;
      case G_TYPE_DOUBLE:
        {
          /* Avoid -Wfloat-equal warnings by doing a direct bit compare */
          gdouble da = g_value_get_double (a);
          gdouble db = g_value_get_double (b);
          ret = memcmp (&da, &db, sizeof (gdouble)) == 0;
        }
        break;
      case G_TYPE_STRING:
        ret = (g_strcmp0 (g_value_get_string (a), g_value_get_string (b)) == 0);
        break;
      case G_TYPE_VARIANT:
        ret = _g_variant_equal0 (g_value_get_variant (a), g_value_get_variant (b));
        break;
      default:
        if (G_VALUE_TYPE (a) == G_TYPE_STRV)
          ret = _g_strv_equal0 (g_value_get_boxed (a), g_value_get_boxed (b));
        else
          g_critical ("_g_value_equal() does not handle type %s", g_type_name (G_VALUE_TYPE (a)));
        break;
    }
  return ret;
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_name,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectBooleanFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gboolean arg_locked,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectBooleanFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectBooleanFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_boolean (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectIntFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint arg_accountType,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectIntFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectIntFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_password,
        const gchar *arg_hint,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_string (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 2);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64 (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectInt64Func)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint64 arg_id,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectInt64Func callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectInt64Func)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int64 (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING_INT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringStringIntFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_name,
        const gchar *arg_fullname,
        gint arg_accountType,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringStringIntFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 5);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringStringIntFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_string (param_values + 3),
              g_marshal_value_peek_int (param_values + 4),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64_BOOLEAN (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectInt64BooleanFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint64 arg_id,
        gboolean arg_removeFiles,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectInt64BooleanFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectInt64BooleanFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int64 (param_values + 2),
              g_marshal_value_peek_boolean (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringStringVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_app_id,
        const gchar *arg_parent_window,
        GVariant *arg_shortcuts,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringStringVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 5);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringStringVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_string (param_values + 3),
              g_marshal_value_peek_variant (param_values + 4),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_app_id,
        GVariant *arg_shortcuts,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_variant (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_BOOLEAN_VARIANT_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintBooleanVariantVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_serial,
        gboolean arg_persistent,
        GVariant *arg_crtcs,
        GVariant *arg_outputs,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintBooleanVariantVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 6);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintBooleanVariantVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_boolean (param_values + 3),
              g_marshal_value_peek_variant (param_values + 4),
              g_marshal_value_peek_variant (param_values + 5),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT_INT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintUintIntFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_serial,
        guint arg_output,
        gint arg_value,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintUintIntFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 5);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintUintIntFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              g_marshal_value_peek_int (param_values + 4),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintUintFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_serial,
        guint arg_crtc,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintUintFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintUintFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT_VARIANT_VARIANT_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintUintVariantVariantVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_serial,
        guint arg_crtc,
        GVariant *arg_red,
        GVariant *arg_green,
        GVariant *arg_blue,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintUintVariantVariantVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 7);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintUintVariantVariantVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              g_marshal_value_peek_variant (param_values + 4),
              g_marshal_value_peek_variant (param_values + 5),
              g_marshal_value_peek_variant (param_values + 6),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT_VARIANT_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintUintVariantVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_serial,
        guint arg_method,
        GVariant *arg_logical_monitors,
        GVariant *arg_properties,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintUintVariantVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 6);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintUintVariantVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              g_marshal_value_peek_variant (param_values + 4),
              g_marshal_value_peek_variant (param_values + 5),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_capabilities,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_VOID__UINT_UINT_VARIANT (
    GClosure     *closure,
    GValue       *return_value G_GNUC_UNUSED,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef void (*_GDbusCodegenMarshalVoid_UintUintVariantFunc)
       (void *data1,
        guint arg_barrier_id,
        guint arg_activation_id,
        GVariant *arg_cursor_position,
        void *data2);
  _GDbusCodegenMarshalVoid_UintUintVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;

  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalVoid_UintUintVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_uint (param_values + 1),
            g_marshal_value_peek_uint (param_values + 2),
            g_marshal_value_peek_variant (param_values + 3),
            data2);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_serial,
        GVariant *arg_position,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_variant (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectObjectFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        GUnixFDList *arg_fd_list,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectObjectFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectObjectFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_object (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        GVariant *arg_options,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_variant (param_values + 2),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_VOID__STRING_UINT (
    GClosure     *closure,
    GValue       *return_value G_GNUC_UNUSED,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef void (*_GDbusCodegenMarshalVoid_StringUintFunc)
       (void *data1,
        const gchar *arg_mime_type,
        guint arg_serial,
        void *data2);
  _GDbusCodegenMarshalVoid_StringUintFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalVoid_StringUintFunc)
    (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_string (param_values + 1),
            g_marshal_value_peek_uint (param_values + 2),
            data2);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_BOOLEAN (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintBooleanFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_keycode,
        gboolean arg_state,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintBooleanFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintBooleanFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_boolean (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_BOOLEAN (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectIntBooleanFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint arg_button,
        gboolean arg_state,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectIntBooleanFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectIntBooleanFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int (param_values + 2),
              g_marshal_value_peek_boolean (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_DOUBLE_DOUBLE_UINT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectDoubleDoubleUintFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gdouble arg_dx,
        gdouble arg_dy,
        guint arg_flags,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectDoubleDoubleUintFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 5);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectDoubleDoubleUintFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_double (param_values + 2),
              g_marshal_value_peek_double (param_values + 3),
              g_marshal_value_peek_uint (param_values + 4),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_INT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectUintIntFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        guint arg_axis,
        gint arg_steps,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectUintIntFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectUintIntFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_uint (param_values + 2),
              g_marshal_value_peek_int (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_DOUBLE_DOUBLE (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectDoubleDoubleFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gdouble arg_dx,
        gdouble arg_dy,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectDoubleDoubleFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectDoubleDoubleFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_double (param_values + 2),
              g_marshal_value_peek_double (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_DOUBLE_DOUBLE (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringDoubleDoubleFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_stream,
        gdouble arg_x,
        gdouble arg_y,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringDoubleDoubleFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 5);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringDoubleDoubleFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_double (param_values + 3),
              g_marshal_value_peek_double (param_values + 4),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_UINT_DOUBLE_DOUBLE (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringUintDoubleDoubleFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_stream,
        guint arg_slot,
        gdouble arg_x,
        gdouble arg_y,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringUintDoubleDoubleFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 6);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringUintDoubleDoubleFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              g_marshal_value_peek_double (param_values + 4),
              g_marshal_value_peek_double (param_values + 5),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT_UINT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectObjectUintFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        GUnixFDList *arg_fd_list,
        guint arg_serial,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectObjectUintFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectObjectUintFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_object (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectObjectStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        GUnixFDList *arg_fd_list,
        const gchar *arg_mime_type,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectObjectStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectObjectStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_object (param_values + 2),
              g_marshal_value_peek_string (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectObjectVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        GUnixFDList *arg_fd_list,
        GVariant *arg_options,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectObjectVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 4);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectObjectVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_object (param_values + 2),
              g_marshal_value_peek_variant (param_values + 3),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_INT_INT_INT_VARIANT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectIntIntIntIntVariantFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint arg_x,
        gint arg_y,
        gint arg_width,
        gint arg_height,
        GVariant *arg_properties,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectIntIntIntIntVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 7);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectIntIntIntIntVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int (param_values + 2),
              g_marshal_value_peek_int (param_values + 3),
              g_marshal_value_peek_int (param_values + 4),
              g_marshal_value_peek_int (param_values + 5),
              g_marshal_value_peek_variant (param_values + 6),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN_BOOLEAN_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectBooleanBooleanStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gboolean arg_include_cursor,
        gboolean arg_flash,
        const gchar *arg_filename,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectBooleanBooleanStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 5);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectBooleanBooleanStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_boolean (param_values + 2),
              g_marshal_value_peek_boolean (param_values + 3),
              g_marshal_value_peek_string (param_values + 4),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN_BOOLEAN_BOOLEAN_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectBooleanBooleanBooleanStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gboolean arg_include_frame,
        gboolean arg_include_cursor,
        gboolean arg_flash,
        const gchar *arg_filename,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectBooleanBooleanBooleanStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 6);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectBooleanBooleanBooleanStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_boolean (param_values + 2),
              g_marshal_value_peek_boolean (param_values + 3),
              g_marshal_value_peek_boolean (param_values + 4),
              g_marshal_value_peek_string (param_values + 5),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_INT_INT_INT_BOOLEAN_STRING (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectIntIntIntIntBooleanStringFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint arg_x,
        gint arg_y,
        gint arg_width,
        gint arg_height,
        gboolean arg_flash,
        const gchar *arg_filename,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectIntIntIntIntBooleanStringFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 8);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectIntIntIntIntBooleanStringFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int (param_values + 2),
              g_marshal_value_peek_int (param_values + 3),
              g_marshal_value_peek_int (param_values + 4),
              g_marshal_value_peek_int (param_values + 5),
              g_marshal_value_peek_boolean (param_values + 6),
              g_marshal_value_peek_string (param_values + 7),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_INT_INT_INT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectIntIntIntIntFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        gint arg_x,
        gint arg_y,
        gint arg_width,
        gint arg_height,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectIntIntIntIntFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 6);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectIntIntIntIntFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_int (param_values + 2),
              g_marshal_value_peek_int (param_values + 3),
              g_marshal_value_peek_int (param_values + 4),
              g_marshal_value_peek_int (param_values + 5),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_VOID__UINT_VARIANT (
    GClosure     *closure,
    GValue       *return_value G_GNUC_UNUSED,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef void (*_GDbusCodegenMarshalVoid_UintVariantFunc)
       (void *data1,
        guint arg_action,
        GVariant *arg_parameters,
        void *data2);
  _GDbusCodegenMarshalVoid_UintVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalVoid_UintVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_uint (param_values + 1),
            g_marshal_value_peek_variant (param_values + 2),
            data2);
}

static void
_g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_UINT_UINT (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef gboolean (*_GDbusCodegenMarshalBoolean_ObjectStringUintUintFunc)
       (void *data1,
        GDBusMethodInvocation *arg_method_invocation,
        const gchar *arg_accelerator,
        guint arg_modeFlags,
        guint arg_grabFlags,
        void *data2);
  _GDbusCodegenMarshalBoolean_ObjectStringUintUintFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;
  gboolean v_return;

  g_return_if_fail (return_value != NULL);
  g_return_if_fail (n_param_values == 5);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalBoolean_ObjectStringUintUintFunc)
    (marshal_data ? marshal_data : cc->callback);

  v_return =
    callback (data1,
              g_marshal_value_peek_object (param_values + 1),
              g_marshal_value_peek_string (param_values + 2),
              g_marshal_value_peek_uint (param_values + 3),
              g_marshal_value_peek_uint (param_values + 4),
              data2);

  g_value_set_boolean (return_value, v_return);
}

static void
_g_dbus_codegen_marshal_VOID__STRING_STRING_STRING_VARIANT_VARIANT (
    GClosure     *closure,
    GValue       *return_value G_GNUC_UNUSED,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint G_GNUC_UNUSED,
    void         *marshal_data)
{
  typedef void (*_GDbusCodegenMarshalVoid_StringStringStringVariantVariantFunc)
       (void *data1,
        const gchar *arg_app_id,
        const gchar *arg_id,
        const gchar *arg_action,
        GVariant *arg_parameter,
        GVariant *arg_platform_data,
        void *data2);
  _GDbusCodegenMarshalVoid_StringStringStringVariantVariantFunc callback;
  GCClosure *cc = (GCClosure*) closure;
  void *data1, *data2;

  g_return_if_fail (n_param_values == 6);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }

  callback = (_GDbusCodegenMarshalVoid_StringStringStringVariantVariantFunc)
    (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_string (param_values + 1),
            g_marshal_value_peek_string (param_values + 2),
            g_marshal_value_peek_string (param_values + 3),
            g_marshal_value_peek_variant (param_values + 4),
            g_marshal_value_peek_variant (param_values + 5),
            data2);
}

/* ------------------------------------------------------------------------
 * Code for interface org.freedesktop.Accounts.User
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgFreedesktopAccountsUser
 * @title: OrgFreedesktopAccountsUser
 * @short_description: Generated C code for the org.freedesktop.Accounts.User D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link> D-Bus interface in C.
 */

enum
{
  _ORG_FREEDESKTOP_ACCOUNTS_USER_CHANGED,
};

static unsigned _ORG_FREEDESKTOP_ACCOUNTS_USER_SIGNALS[1] = { 0 };

/* ---- Introspection data for org.freedesktop.Accounts.User ---- */

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_user_name_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_user_name_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_user_name_IN_ARG_name.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_user_name_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_user_name_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_user_name_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_user_name =
{
  {
    -1,
    (gchar *) "SetUserName",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_user_name_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_user_name_annotation_info_pointers
  },
  "handle-set-user-name",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_real_name_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_real_name_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_real_name_IN_ARG_name.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_real_name_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_real_name_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_real_name_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_real_name =
{
  {
    -1,
    (gchar *) "SetRealName",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_real_name_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_real_name_annotation_info_pointers
  },
  "handle-set-real-name",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_email_IN_ARG_email =
{
  {
    -1,
    (gchar *) "email",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_email_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_email_IN_ARG_email.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_email_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_email_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_email_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_email =
{
  {
    -1,
    (gchar *) "SetEmail",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_email_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_email_annotation_info_pointers
  },
  "handle-set-email",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_language_IN_ARG_language =
{
  {
    -1,
    (gchar *) "language",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_language_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_language_IN_ARG_language.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_language_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_language_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_language_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_language =
{
  {
    -1,
    (gchar *) "SetLanguage",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_language_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_language_annotation_info_pointers
  },
  "handle-set-language",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_xsession_IN_ARG_x_session =
{
  {
    -1,
    (gchar *) "x_session",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_xsession_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_xsession_IN_ARG_x_session.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_xsession_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_xsession_annotation_info_1 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.CSymbol",
  (gchar *) "user_set_x_session",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_xsession_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_xsession_annotation_info_0,
  &_org_freedesktop_accounts_user_method_set_xsession_annotation_info_1,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_xsession =
{
  {
    -1,
    (gchar *) "SetXSession",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_xsession_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_xsession_annotation_info_pointers
  },
  "handle-set-xsession",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_location_IN_ARG_location =
{
  {
    -1,
    (gchar *) "location",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_location_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_location_IN_ARG_location.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_location_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_location_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_location_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_location =
{
  {
    -1,
    (gchar *) "SetLocation",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_location_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_location_annotation_info_pointers
  },
  "handle-set-location",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_home_directory_IN_ARG_homedir =
{
  {
    -1,
    (gchar *) "homedir",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_home_directory_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_home_directory_IN_ARG_homedir.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_home_directory_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_home_directory_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_home_directory_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_home_directory =
{
  {
    -1,
    (gchar *) "SetHomeDirectory",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_home_directory_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_home_directory_annotation_info_pointers
  },
  "handle-set-home-directory",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_shell_IN_ARG_shell =
{
  {
    -1,
    (gchar *) "shell",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_shell_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_shell_IN_ARG_shell.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_shell_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_shell_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_shell_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_shell =
{
  {
    -1,
    (gchar *) "SetShell",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_shell_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_shell_annotation_info_pointers
  },
  "handle-set-shell",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_icon_file_IN_ARG_filename =
{
  {
    -1,
    (gchar *) "filename",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_icon_file_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_icon_file_IN_ARG_filename.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_icon_file_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_icon_file_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_icon_file_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_icon_file =
{
  {
    -1,
    (gchar *) "SetIconFile",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_icon_file_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_icon_file_annotation_info_pointers
  },
  "handle-set-icon-file",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_locked_IN_ARG_locked =
{
  {
    -1,
    (gchar *) "locked",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_locked_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_locked_IN_ARG_locked.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_locked_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_locked_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_locked_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_locked =
{
  {
    -1,
    (gchar *) "SetLocked",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_locked_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_locked_annotation_info_pointers
  },
  "handle-set-locked",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_account_type_IN_ARG_accountType =
{
  {
    -1,
    (gchar *) "accountType",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_account_type_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_account_type_IN_ARG_accountType.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_account_type_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_account_type_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_account_type_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_account_type =
{
  {
    -1,
    (gchar *) "SetAccountType",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_account_type_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_account_type_annotation_info_pointers
  },
  "handle-set-account-type",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_password_mode_IN_ARG_mode =
{
  {
    -1,
    (gchar *) "mode",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_password_mode_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_password_mode_IN_ARG_mode.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_password_mode_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_password_mode_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_password_mode_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_password_mode =
{
  {
    -1,
    (gchar *) "SetPasswordMode",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_password_mode_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_password_mode_annotation_info_pointers
  },
  "handle-set-password-mode",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_password_IN_ARG_password =
{
  {
    -1,
    (gchar *) "password",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_password_IN_ARG_hint =
{
  {
    -1,
    (gchar *) "hint",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_password_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_password_IN_ARG_password.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_password_IN_ARG_hint.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_password_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_password_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_password_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_password =
{
  {
    -1,
    (gchar *) "SetPassword",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_password_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_password_annotation_info_pointers
  },
  "handle-set-password",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_password_hint_IN_ARG_hint =
{
  {
    -1,
    (gchar *) "hint",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_password_hint_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_password_hint_IN_ARG_hint.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_password_hint_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_password_hint_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_password_hint_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_password_hint =
{
  {
    -1,
    (gchar *) "SetPasswordHint",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_password_hint_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_password_hint_annotation_info_pointers
  },
  "handle-set-password-hint",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_user_method_info_set_automatic_login_IN_ARG_enabled =
{
  {
    -1,
    (gchar *) "enabled",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_user_method_info_set_automatic_login_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_automatic_login_IN_ARG_enabled.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_user_method_set_automatic_login_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_user_method_set_automatic_login_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_set_automatic_login_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_user_method_info_set_automatic_login =
{
  {
    -1,
    (gchar *) "SetAutomaticLogin",
    (GDBusArgInfo **) &_org_freedesktop_accounts_user_method_info_set_automatic_login_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_user_method_set_automatic_login_annotation_info_pointers
  },
  "handle-set-automatic-login",
  FALSE
};

static const GDBusMethodInfo * const _org_freedesktop_accounts_user_method_info_pointers[] =
{
  &_org_freedesktop_accounts_user_method_info_set_user_name.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_real_name.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_email.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_language.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_xsession.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_location.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_home_directory.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_shell.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_icon_file.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_locked.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_account_type.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_password_mode.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_password.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_password_hint.parent_struct,
  &_org_freedesktop_accounts_user_method_info_set_automatic_login.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_freedesktop_accounts_user_signal_info_changed =
{
  {
    -1,
    (gchar *) "Changed",
    NULL,
    NULL
  },
  "changed"
};

static const GDBusSignalInfo * const _org_freedesktop_accounts_user_signal_info_pointers[] =
{
  &_org_freedesktop_accounts_user_signal_info_changed.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_uid =
{
  {
    -1,
    (gchar *) "Uid",
    (gchar *) "t",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "uid",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_user_name =
{
  {
    -1,
    (gchar *) "UserName",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "user-name",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_real_name =
{
  {
    -1,
    (gchar *) "RealName",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "real-name",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_account_type =
{
  {
    -1,
    (gchar *) "AccountType",
    (gchar *) "i",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "account-type",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_home_directory =
{
  {
    -1,
    (gchar *) "HomeDirectory",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "home-directory",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_shell =
{
  {
    -1,
    (gchar *) "Shell",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "shell",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_email =
{
  {
    -1,
    (gchar *) "Email",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "email",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_language =
{
  {
    -1,
    (gchar *) "Language",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "language",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_xsession =
{
  {
    -1,
    (gchar *) "XSession",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "xsession",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_location =
{
  {
    -1,
    (gchar *) "Location",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "location",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_login_frequency =
{
  {
    -1,
    (gchar *) "LoginFrequency",
    (gchar *) "t",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "login-frequency",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_login_time =
{
  {
    -1,
    (gchar *) "LoginTime",
    (gchar *) "x",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "login-time",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_login_history =
{
  {
    -1,
    (gchar *) "LoginHistory",
    (gchar *) "a(xxa{sv})",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "login-history",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_icon_file =
{
  {
    -1,
    (gchar *) "IconFile",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "icon-file",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_locked =
{
  {
    -1,
    (gchar *) "Locked",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "locked",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_password_mode =
{
  {
    -1,
    (gchar *) "PasswordMode",
    (gchar *) "i",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "password-mode",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_password_hint =
{
  {
    -1,
    (gchar *) "PasswordHint",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "password-hint",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_automatic_login =
{
  {
    -1,
    (gchar *) "AutomaticLogin",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "automatic-login",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_system_account =
{
  {
    -1,
    (gchar *) "SystemAccount",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "system-account",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_user_property_info_local_account =
{
  {
    -1,
    (gchar *) "LocalAccount",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "local-account",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_freedesktop_accounts_user_property_info_pointers[] =
{
  &_org_freedesktop_accounts_user_property_info_uid.parent_struct,
  &_org_freedesktop_accounts_user_property_info_user_name.parent_struct,
  &_org_freedesktop_accounts_user_property_info_real_name.parent_struct,
  &_org_freedesktop_accounts_user_property_info_account_type.parent_struct,
  &_org_freedesktop_accounts_user_property_info_home_directory.parent_struct,
  &_org_freedesktop_accounts_user_property_info_shell.parent_struct,
  &_org_freedesktop_accounts_user_property_info_email.parent_struct,
  &_org_freedesktop_accounts_user_property_info_language.parent_struct,
  &_org_freedesktop_accounts_user_property_info_xsession.parent_struct,
  &_org_freedesktop_accounts_user_property_info_location.parent_struct,
  &_org_freedesktop_accounts_user_property_info_login_frequency.parent_struct,
  &_org_freedesktop_accounts_user_property_info_login_time.parent_struct,
  &_org_freedesktop_accounts_user_property_info_login_history.parent_struct,
  &_org_freedesktop_accounts_user_property_info_icon_file.parent_struct,
  &_org_freedesktop_accounts_user_property_info_locked.parent_struct,
  &_org_freedesktop_accounts_user_property_info_password_mode.parent_struct,
  &_org_freedesktop_accounts_user_property_info_password_hint.parent_struct,
  &_org_freedesktop_accounts_user_property_info_automatic_login.parent_struct,
  &_org_freedesktop_accounts_user_property_info_system_account.parent_struct,
  &_org_freedesktop_accounts_user_property_info_local_account.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_freedesktop_accounts_user_interface_info =
{
  {
    -1,
    (gchar *) "org.freedesktop.Accounts.User",
    (GDBusMethodInfo **) &_org_freedesktop_accounts_user_method_info_pointers,
    (GDBusSignalInfo **) &_org_freedesktop_accounts_user_signal_info_pointers,
    (GDBusPropertyInfo **) &_org_freedesktop_accounts_user_property_info_pointers,
    NULL
  },
  "org-freedesktop-accounts-user",
};


/**
 * org_freedesktop_accounts_user_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_freedesktop_accounts_user_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_freedesktop_accounts_user_interface_info.parent_struct;
}

/**
 * org_freedesktop_accounts_user_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgFreedesktopAccountsUser interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_freedesktop_accounts_user_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "uid");
  g_object_class_override_property (klass, property_id_begin++, "user-name");
  g_object_class_override_property (klass, property_id_begin++, "real-name");
  g_object_class_override_property (klass, property_id_begin++, "account-type");
  g_object_class_override_property (klass, property_id_begin++, "home-directory");
  g_object_class_override_property (klass, property_id_begin++, "shell");
  g_object_class_override_property (klass, property_id_begin++, "email");
  g_object_class_override_property (klass, property_id_begin++, "language");
  g_object_class_override_property (klass, property_id_begin++, "xsession");
  g_object_class_override_property (klass, property_id_begin++, "location");
  g_object_class_override_property (klass, property_id_begin++, "login-frequency");
  g_object_class_override_property (klass, property_id_begin++, "login-time");
  g_object_class_override_property (klass, property_id_begin++, "login-history");
  g_object_class_override_property (klass, property_id_begin++, "icon-file");
  g_object_class_override_property (klass, property_id_begin++, "locked");
  g_object_class_override_property (klass, property_id_begin++, "password-mode");
  g_object_class_override_property (klass, property_id_begin++, "password-hint");
  g_object_class_override_property (klass, property_id_begin++, "automatic-login");
  g_object_class_override_property (klass, property_id_begin++, "system-account");
  g_object_class_override_property (klass, property_id_begin++, "local-account");
  return property_id_begin - 1;
}


inline static void
org_freedesktop_accounts_user_signal_marshal_changed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_user_name (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_real_name (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_email (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_language (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_xsession (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_location (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_home_directory (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_shell (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_icon_file (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_locked (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_account_type (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_password_mode (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_password (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_password_hint (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_user_method_marshal_set_automatic_login (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgFreedesktopAccountsUser:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>.
 */

/**
 * OrgFreedesktopAccountsUserIface:
 * @parent_iface: The parent interface.
 * @handle_set_account_type: Handler for the #OrgFreedesktopAccountsUser::handle-set-account-type signal.
 * @handle_set_automatic_login: Handler for the #OrgFreedesktopAccountsUser::handle-set-automatic-login signal.
 * @handle_set_email: Handler for the #OrgFreedesktopAccountsUser::handle-set-email signal.
 * @handle_set_home_directory: Handler for the #OrgFreedesktopAccountsUser::handle-set-home-directory signal.
 * @handle_set_icon_file: Handler for the #OrgFreedesktopAccountsUser::handle-set-icon-file signal.
 * @handle_set_language: Handler for the #OrgFreedesktopAccountsUser::handle-set-language signal.
 * @handle_set_location: Handler for the #OrgFreedesktopAccountsUser::handle-set-location signal.
 * @handle_set_locked: Handler for the #OrgFreedesktopAccountsUser::handle-set-locked signal.
 * @handle_set_password: Handler for the #OrgFreedesktopAccountsUser::handle-set-password signal.
 * @handle_set_password_hint: Handler for the #OrgFreedesktopAccountsUser::handle-set-password-hint signal.
 * @handle_set_password_mode: Handler for the #OrgFreedesktopAccountsUser::handle-set-password-mode signal.
 * @handle_set_real_name: Handler for the #OrgFreedesktopAccountsUser::handle-set-real-name signal.
 * @handle_set_shell: Handler for the #OrgFreedesktopAccountsUser::handle-set-shell signal.
 * @handle_set_user_name: Handler for the #OrgFreedesktopAccountsUser::handle-set-user-name signal.
 * @handle_set_xsession: Handler for the #OrgFreedesktopAccountsUser::handle-set-xsession signal.
 * @get_account_type: Getter for the #OrgFreedesktopAccountsUser:account-type property.
 * @get_automatic_login: Getter for the #OrgFreedesktopAccountsUser:automatic-login property.
 * @get_email: Getter for the #OrgFreedesktopAccountsUser:email property.
 * @get_home_directory: Getter for the #OrgFreedesktopAccountsUser:home-directory property.
 * @get_icon_file: Getter for the #OrgFreedesktopAccountsUser:icon-file property.
 * @get_language: Getter for the #OrgFreedesktopAccountsUser:language property.
 * @get_local_account: Getter for the #OrgFreedesktopAccountsUser:local-account property.
 * @get_location: Getter for the #OrgFreedesktopAccountsUser:location property.
 * @get_locked: Getter for the #OrgFreedesktopAccountsUser:locked property.
 * @get_login_frequency: Getter for the #OrgFreedesktopAccountsUser:login-frequency property.
 * @get_login_history: Getter for the #OrgFreedesktopAccountsUser:login-history property.
 * @get_login_time: Getter for the #OrgFreedesktopAccountsUser:login-time property.
 * @get_password_hint: Getter for the #OrgFreedesktopAccountsUser:password-hint property.
 * @get_password_mode: Getter for the #OrgFreedesktopAccountsUser:password-mode property.
 * @get_real_name: Getter for the #OrgFreedesktopAccountsUser:real-name property.
 * @get_shell: Getter for the #OrgFreedesktopAccountsUser:shell property.
 * @get_system_account: Getter for the #OrgFreedesktopAccountsUser:system-account property.
 * @get_uid: Getter for the #OrgFreedesktopAccountsUser:uid property.
 * @get_user_name: Getter for the #OrgFreedesktopAccountsUser:user-name property.
 * @get_xsession: Getter for the #OrgFreedesktopAccountsUser:xsession property.
 * @changed: Handler for the #OrgFreedesktopAccountsUser::changed signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>.
 */

typedef OrgFreedesktopAccountsUserIface OrgFreedesktopAccountsUserInterface;
G_DEFINE_INTERFACE (OrgFreedesktopAccountsUser, org_freedesktop_accounts_user, G_TYPE_OBJECT)

static void
org_freedesktop_accounts_user_default_init (OrgFreedesktopAccountsUserIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgFreedesktopAccountsUser::handle-set-user-name:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_user_name() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-user-name",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_user_name),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_user_name,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-real-name:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_real_name() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-real-name",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_real_name),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_real_name,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-email:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_email: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_email() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-email",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_email),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_email,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-language:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_language: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_language() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-language",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_language),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_language,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-xsession:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_x_session: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_xsession() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-xsession",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_xsession),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_xsession,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-location:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_location: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_location() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-location",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_location),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_location,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-home-directory:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_homedir: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_home_directory() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-home-directory",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_home_directory),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_home_directory,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-shell:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_shell: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_shell() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-shell",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_shell),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_shell,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-icon-file:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_filename: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_icon_file() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-icon-file",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_icon_file),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_icon_file,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-locked:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_locked: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_locked() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-locked",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_locked),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_locked,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN);

  /**
   * OrgFreedesktopAccountsUser::handle-set-account-type:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_accountType: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_account_type() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-account-type",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_account_type),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_account_type,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT);

  /**
   * OrgFreedesktopAccountsUser::handle-set-password-mode:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_mode: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_password_mode() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-password-mode",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_password_mode),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_password_mode,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT);

  /**
   * OrgFreedesktopAccountsUser::handle-set-password:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_password: Argument passed by remote caller.
   * @arg_hint: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_password() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-password",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_password),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_password,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-password-hint:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_hint: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_password_hint() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-password-hint",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_password_hint),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_password_hint,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccountsUser::handle-set-automatic-login:
   * @object: A #OrgFreedesktopAccountsUser.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_enabled: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_user_complete_set_automatic_login() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-automatic-login",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, handle_set_automatic_login),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_user_method_marshal_set_automatic_login,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgFreedesktopAccountsUser::changed:
   * @object: A #OrgFreedesktopAccountsUser.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-freedesktop-Accounts-User.Changed">"Changed"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_FREEDESKTOP_ACCOUNTS_USER_SIGNALS[_ORG_FREEDESKTOP_ACCOUNTS_USER_CHANGED] =
    g_signal_new ("changed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgFreedesktopAccountsUserIface, changed),
      NULL,
      NULL,
      org_freedesktop_accounts_user_signal_marshal_changed,
      G_TYPE_NONE,
      0);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgFreedesktopAccountsUser:uid:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Uid">"Uid"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_uint64 ("uid", "Uid", "Uid", 0, G_MAXUINT64, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:user-name:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("user-name", "UserName", "UserName", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:real-name:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("real-name", "RealName", "RealName", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:account-type:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.AccountType">"AccountType"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int ("account-type", "AccountType", "AccountType", G_MININT32, G_MAXINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:home-directory:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("home-directory", "HomeDirectory", "HomeDirectory", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:shell:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("shell", "Shell", "Shell", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:email:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("email", "Email", "Email", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:language:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("language", "Language", "Language", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:xsession:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("xsession", "XSession", "XSession", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:location:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("location", "Location", "Location", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:login-frequency:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginFrequency">"LoginFrequency"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_uint64 ("login-frequency", "LoginFrequency", "LoginFrequency", 0, G_MAXUINT64, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:login-time:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginTime">"LoginTime"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int64 ("login-time", "LoginTime", "LoginTime", G_MININT64, G_MAXINT64, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:login-history:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_variant ("login-history", "LoginHistory", "LoginHistory", G_VARIANT_TYPE ("a(xxa{sv})"), NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:icon-file:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("icon-file", "IconFile", "IconFile", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:locked:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.Locked">"Locked"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("locked", "Locked", "Locked", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:password-mode:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordMode">"PasswordMode"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int ("password-mode", "PasswordMode", "PasswordMode", G_MININT32, G_MAXINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:password-hint:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("password-hint", "PasswordHint", "PasswordHint", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:automatic-login:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.AutomaticLogin">"AutomaticLogin"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("automatic-login", "AutomaticLogin", "AutomaticLogin", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:system-account:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.SystemAccount">"SystemAccount"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("system-account", "SystemAccount", "SystemAccount", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgFreedesktopAccountsUser:local-account:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts-User.LocalAccount">"LocalAccount"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("local-account", "LocalAccount", "LocalAccount", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_freedesktop_accounts_user_get_uid: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Uid">"Uid"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
guint64 
org_freedesktop_accounts_user_get_uid (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), 0);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_uid (object);
}

/**
 * org_freedesktop_accounts_user_set_uid: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Uid">"Uid"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_uid (OrgFreedesktopAccountsUser *object, guint64 value)
{
  g_object_set (G_OBJECT (object), "uid", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_user_name: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_user_name() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_user_name (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_user_name (object);
}

/**
 * org_freedesktop_accounts_user_dup_user_name: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_user_name (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "user-name", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_user_name: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.UserName">"UserName"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_user_name (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "user-name", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_real_name: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_real_name() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_real_name (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_real_name (object);
}

/**
 * org_freedesktop_accounts_user_dup_real_name: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_real_name (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "real-name", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_real_name: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.RealName">"RealName"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_real_name (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "real-name", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_account_type: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AccountType">"AccountType"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint 
org_freedesktop_accounts_user_get_account_type (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), 0);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_account_type (object);
}

/**
 * org_freedesktop_accounts_user_set_account_type: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AccountType">"AccountType"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_account_type (OrgFreedesktopAccountsUser *object, gint value)
{
  g_object_set (G_OBJECT (object), "account-type", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_home_directory: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_home_directory() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_home_directory (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_home_directory (object);
}

/**
 * org_freedesktop_accounts_user_dup_home_directory: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_home_directory (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "home-directory", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_home_directory: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.HomeDirectory">"HomeDirectory"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_home_directory (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "home-directory", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_shell: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_shell() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_shell (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_shell (object);
}

/**
 * org_freedesktop_accounts_user_dup_shell: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_shell (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "shell", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_shell: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Shell">"Shell"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_shell (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "shell", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_email: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_email() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_email (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_email (object);
}

/**
 * org_freedesktop_accounts_user_dup_email: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_email (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "email", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_email: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Email">"Email"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_email (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "email", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_language: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_language() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_language (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_language (object);
}

/**
 * org_freedesktop_accounts_user_dup_language: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_language (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "language", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_language: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Language">"Language"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_language (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "language", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_xsession: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_xsession() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_xsession (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_xsession (object);
}

/**
 * org_freedesktop_accounts_user_dup_xsession: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_xsession (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "xsession", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_xsession: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.XSession">"XSession"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_xsession (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "xsession", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_location: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_location() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_location (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_location (object);
}

/**
 * org_freedesktop_accounts_user_dup_location: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_location (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "location", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_location: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Location">"Location"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_location (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "location", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_login_frequency: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginFrequency">"LoginFrequency"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
guint64 
org_freedesktop_accounts_user_get_login_frequency (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), 0);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_login_frequency (object);
}

/**
 * org_freedesktop_accounts_user_set_login_frequency: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginFrequency">"LoginFrequency"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_login_frequency (OrgFreedesktopAccountsUser *object, guint64 value)
{
  g_object_set (G_OBJECT (object), "login-frequency", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_login_time: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginTime">"LoginTime"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint64 
org_freedesktop_accounts_user_get_login_time (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), 0);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_login_time (object);
}

/**
 * org_freedesktop_accounts_user_set_login_time: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginTime">"LoginTime"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_login_time (OrgFreedesktopAccountsUser *object, gint64 value)
{
  g_object_set (G_OBJECT (object), "login-time", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_login_history: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_login_history() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
GVariant *
org_freedesktop_accounts_user_get_login_history (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_login_history (object);
}

/**
 * org_freedesktop_accounts_user_dup_login_history: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_variant_unref().
 */
GVariant *
org_freedesktop_accounts_user_dup_login_history (OrgFreedesktopAccountsUser *object)
{
  GVariant *value;
  g_object_get (G_OBJECT (object), "login-history", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_login_history: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LoginHistory">"LoginHistory"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_login_history (OrgFreedesktopAccountsUser *object, GVariant *value)
{
  g_object_set (G_OBJECT (object), "login-history", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_icon_file: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_icon_file() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_icon_file (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_icon_file (object);
}

/**
 * org_freedesktop_accounts_user_dup_icon_file: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_icon_file (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "icon-file", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_icon_file: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.IconFile">"IconFile"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_icon_file (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "icon-file", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_locked: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Locked">"Locked"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_freedesktop_accounts_user_get_locked (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), FALSE);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_locked (object);
}

/**
 * org_freedesktop_accounts_user_set_locked: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.Locked">"Locked"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_locked (OrgFreedesktopAccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "locked", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_password_mode: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordMode">"PasswordMode"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint 
org_freedesktop_accounts_user_get_password_mode (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), 0);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_password_mode (object);
}

/**
 * org_freedesktop_accounts_user_set_password_mode: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordMode">"PasswordMode"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_password_mode (OrgFreedesktopAccountsUser *object, gint value)
{
  g_object_set (G_OBJECT (object), "password-mode", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_password_hint: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_user_dup_password_hint() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_user_get_password_hint (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_password_hint (object);
}

/**
 * org_freedesktop_accounts_user_dup_password_hint: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_user_dup_password_hint (OrgFreedesktopAccountsUser *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "password-hint", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_user_set_password_hint: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.PasswordHint">"PasswordHint"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_password_hint (OrgFreedesktopAccountsUser *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "password-hint", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_automatic_login: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AutomaticLogin">"AutomaticLogin"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_freedesktop_accounts_user_get_automatic_login (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), FALSE);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_automatic_login (object);
}

/**
 * org_freedesktop_accounts_user_set_automatic_login: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.AutomaticLogin">"AutomaticLogin"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_automatic_login (OrgFreedesktopAccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "automatic-login", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_system_account: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.SystemAccount">"SystemAccount"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_freedesktop_accounts_user_get_system_account (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), FALSE);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_system_account (object);
}

/**
 * org_freedesktop_accounts_user_set_system_account: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.SystemAccount">"SystemAccount"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_system_account (OrgFreedesktopAccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "system-account", value, NULL);
}

/**
 * org_freedesktop_accounts_user_get_local_account: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LocalAccount">"LocalAccount"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_freedesktop_accounts_user_get_local_account (OrgFreedesktopAccountsUser *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS_USER (object), FALSE);

  return ORG_FREEDESKTOP_ACCOUNTS_USER_GET_IFACE (object)->get_local_account (object);
}

/**
 * org_freedesktop_accounts_user_set_local_account: (skip)
 * @object: A #OrgFreedesktopAccountsUser.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts-User.LocalAccount">"LocalAccount"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_user_set_local_account (OrgFreedesktopAccountsUser *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "local-account", value, NULL);
}

/**
 * org_freedesktop_accounts_user_emit_changed:
 * @object: A #OrgFreedesktopAccountsUser.
 *
 * Emits the <link linkend="gdbus-signal-org-freedesktop-Accounts-User.Changed">"Changed"</link> D-Bus signal.
 */
void
org_freedesktop_accounts_user_emit_changed (
    OrgFreedesktopAccountsUser *object)
{
  g_signal_emit (object, _ORG_FREEDESKTOP_ACCOUNTS_USER_SIGNALS[_ORG_FREEDESKTOP_ACCOUNTS_USER_CHANGED], 0);
}

/**
 * org_freedesktop_accounts_user_call_set_user_name:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_user_name_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_user_name_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_user_name (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetUserName",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_user_name_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_user_name().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_user_name().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_user_name_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_user_name_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_user_name() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_user_name_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetUserName",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_real_name:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_real_name_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_real_name_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_real_name (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetRealName",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_real_name_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_real_name().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_real_name().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_real_name_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_real_name_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_real_name() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_real_name_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetRealName",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_email:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_email: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_email_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_email_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_email (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_email,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetEmail",
    g_variant_new ("(s)",
                   arg_email),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_email_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_email().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_email().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_email_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_email_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_email: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_email() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_email_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_email,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetEmail",
    g_variant_new ("(s)",
                   arg_email),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_language:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_language: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_language_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_language_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_language (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_language,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetLanguage",
    g_variant_new ("(s)",
                   arg_language),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_language_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_language().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_language().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_language_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_language_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_language: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_language() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_language_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_language,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetLanguage",
    g_variant_new ("(s)",
                   arg_language),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_xsession:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_x_session: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_xsession_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_xsession_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_xsession (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_x_session,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetXSession",
    g_variant_new ("(s)",
                   arg_x_session),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_xsession_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_xsession().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_xsession().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_xsession_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_xsession_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_x_session: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_xsession() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_xsession_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_x_session,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetXSession",
    g_variant_new ("(s)",
                   arg_x_session),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_location:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_location: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_location_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_location_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_location (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_location,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetLocation",
    g_variant_new ("(s)",
                   arg_location),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_location_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_location().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_location().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_location_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_location_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_location: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_location() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_location_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_location,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetLocation",
    g_variant_new ("(s)",
                   arg_location),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_home_directory:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_homedir: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_home_directory_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_home_directory_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_home_directory (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_homedir,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetHomeDirectory",
    g_variant_new ("(s)",
                   arg_homedir),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_home_directory_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_home_directory().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_home_directory().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_home_directory_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_home_directory_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_homedir: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_home_directory() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_home_directory_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_homedir,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetHomeDirectory",
    g_variant_new ("(s)",
                   arg_homedir),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_shell:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_shell: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_shell_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_shell_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_shell (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_shell,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetShell",
    g_variant_new ("(s)",
                   arg_shell),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_shell_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_shell().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_shell().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_shell_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_shell_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_shell: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_shell() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_shell_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_shell,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetShell",
    g_variant_new ("(s)",
                   arg_shell),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_icon_file:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_filename: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_icon_file_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_icon_file_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_icon_file (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_filename,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetIconFile",
    g_variant_new ("(s)",
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_icon_file_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_icon_file().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_icon_file().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_icon_file_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_icon_file_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_filename: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_icon_file() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_icon_file_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_filename,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetIconFile",
    g_variant_new ("(s)",
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_locked:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_locked: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_locked_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_locked_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_locked (
    OrgFreedesktopAccountsUser *proxy,
    gboolean arg_locked,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetLocked",
    g_variant_new ("(b)",
                   arg_locked),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_locked_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_locked().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_locked().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_locked_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_locked_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_locked: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_locked() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_locked_sync (
    OrgFreedesktopAccountsUser *proxy,
    gboolean arg_locked,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetLocked",
    g_variant_new ("(b)",
                   arg_locked),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_account_type:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_accountType: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_account_type_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_account_type_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_account_type (
    OrgFreedesktopAccountsUser *proxy,
    gint arg_accountType,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetAccountType",
    g_variant_new ("(i)",
                   arg_accountType),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_account_type_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_account_type().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_account_type().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_account_type_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_account_type_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_accountType: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_account_type() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_account_type_sync (
    OrgFreedesktopAccountsUser *proxy,
    gint arg_accountType,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetAccountType",
    g_variant_new ("(i)",
                   arg_accountType),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_password_mode:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_mode: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_password_mode_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_password_mode_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_password_mode (
    OrgFreedesktopAccountsUser *proxy,
    gint arg_mode,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetPasswordMode",
    g_variant_new ("(i)",
                   arg_mode),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_password_mode_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_password_mode().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_password_mode().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_password_mode_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_password_mode_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_mode: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_password_mode() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_password_mode_sync (
    OrgFreedesktopAccountsUser *proxy,
    gint arg_mode,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetPasswordMode",
    g_variant_new ("(i)",
                   arg_mode),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_password:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_password: Argument to pass with the method invocation.
 * @arg_hint: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_password_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_password_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_password (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_password,
    const gchar *arg_hint,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetPassword",
    g_variant_new ("(ss)",
                   arg_password,
                   arg_hint),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_password_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_password().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_password().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_password_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_password_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_password: Argument to pass with the method invocation.
 * @arg_hint: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_password() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_password_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_password,
    const gchar *arg_hint,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetPassword",
    g_variant_new ("(ss)",
                   arg_password,
                   arg_hint),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_password_hint:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_hint: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_password_hint_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_password_hint_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_password_hint (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_hint,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetPasswordHint",
    g_variant_new ("(s)",
                   arg_hint),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_password_hint_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_password_hint().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_password_hint().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_password_hint_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_password_hint_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_hint: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_password_hint() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_password_hint_sync (
    OrgFreedesktopAccountsUser *proxy,
    const gchar *arg_hint,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetPasswordHint",
    g_variant_new ("(s)",
                   arg_hint),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_automatic_login:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_enabled: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_call_set_automatic_login_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_call_set_automatic_login_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_user_call_set_automatic_login (
    OrgFreedesktopAccountsUser *proxy,
    gboolean arg_enabled,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetAutomaticLogin",
    g_variant_new ("(b)",
                   arg_enabled),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_user_call_set_automatic_login_finish:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_call_set_automatic_login().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_user_call_set_automatic_login().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_automatic_login_finish (
    OrgFreedesktopAccountsUser *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_call_set_automatic_login_sync:
 * @proxy: A #OrgFreedesktopAccountsUserProxy.
 * @arg_enabled: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_call_set_automatic_login() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_user_call_set_automatic_login_sync (
    OrgFreedesktopAccountsUser *proxy,
    gboolean arg_enabled,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetAutomaticLogin",
    g_variant_new ("(b)",
                   arg_enabled),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_user_complete_set_user_name:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetUserName">SetUserName()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_user_name (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_real_name:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetRealName">SetRealName()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_real_name (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_email:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetEmail">SetEmail()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_email (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_language:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLanguage">SetLanguage()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_language (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_xsession:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetXSession">SetXSession()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_xsession (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_location:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocation">SetLocation()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_location (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_home_directory:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetHomeDirectory">SetHomeDirectory()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_home_directory (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_shell:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetShell">SetShell()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_shell (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_icon_file:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetIconFile">SetIconFile()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_icon_file (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_locked:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetLocked">SetLocked()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_locked (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_account_type:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAccountType">SetAccountType()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_account_type (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_password_mode:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordMode">SetPasswordMode()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_password_mode (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_password:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPassword">SetPassword()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_password (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_password_hint:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetPasswordHint">SetPasswordHint()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_password_hint (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_user_complete_set_automatic_login:
 * @object: A #OrgFreedesktopAccountsUser.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts-User.SetAutomaticLogin">SetAutomaticLogin()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_user_complete_set_automatic_login (
    OrgFreedesktopAccountsUser *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgFreedesktopAccountsUserProxy:
 *
 * The #OrgFreedesktopAccountsUserProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgFreedesktopAccountsUserProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgFreedesktopAccountsUserProxy.
 */

struct _OrgFreedesktopAccountsUserProxyPrivate
{
  GData *qdata;
};

static void org_freedesktop_accounts_user_proxy_iface_init (OrgFreedesktopAccountsUserIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsUserProxy, org_freedesktop_accounts_user_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgFreedesktopAccountsUserProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER, org_freedesktop_accounts_user_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsUserProxy, org_freedesktop_accounts_user_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER, org_freedesktop_accounts_user_proxy_iface_init))

#endif
static void
org_freedesktop_accounts_user_proxy_finalize (GObject *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_freedesktop_accounts_user_proxy_parent_class)->finalize (object);
}

static void
org_freedesktop_accounts_user_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 20);
  info = (const _ExtendedGDBusPropertyInfo *) _org_freedesktop_accounts_user_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_freedesktop_accounts_user_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.freedesktop.Accounts.User: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_freedesktop_accounts_user_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 20);
  info = (const _ExtendedGDBusPropertyInfo *) _org_freedesktop_accounts_user_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.freedesktop.Accounts.User", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_freedesktop_accounts_user_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_freedesktop_accounts_user_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_user_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_freedesktop_accounts_user_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_user_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_user_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static guint64 
org_freedesktop_accounts_user_proxy_get_uid (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  guint64 value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Uid");
  if (variant != NULL)
    {
      value = g_variant_get_uint64 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_user_name (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "UserName");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_real_name (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "RealName");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gint 
org_freedesktop_accounts_user_proxy_get_account_type (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "AccountType");
  if (variant != NULL)
    {
      value = g_variant_get_int32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_home_directory (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "HomeDirectory");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_shell (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Shell");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_email (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Email");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_language (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Language");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_xsession (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "XSession");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_location (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Location");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static guint64 
org_freedesktop_accounts_user_proxy_get_login_frequency (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  guint64 value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LoginFrequency");
  if (variant != NULL)
    {
      value = g_variant_get_uint64 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gint64 
org_freedesktop_accounts_user_proxy_get_login_time (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gint64 value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LoginTime");
  if (variant != NULL)
    {
      value = g_variant_get_int64 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static GVariant *
org_freedesktop_accounts_user_proxy_get_login_history (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  GVariant *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LoginHistory");
  value = variant;
  if (variant != NULL)
    g_variant_unref (variant);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_icon_file (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "IconFile");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
org_freedesktop_accounts_user_proxy_get_locked (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Locked");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gint 
org_freedesktop_accounts_user_proxy_get_password_mode (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "PasswordMode");
  if (variant != NULL)
    {
      value = g_variant_get_int32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_freedesktop_accounts_user_proxy_get_password_hint (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "PasswordHint");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
org_freedesktop_accounts_user_proxy_get_automatic_login (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "AutomaticLogin");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
org_freedesktop_accounts_user_proxy_get_system_account (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "SystemAccount");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
org_freedesktop_accounts_user_proxy_get_local_account (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "LocalAccount");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_freedesktop_accounts_user_proxy_init (OrgFreedesktopAccountsUserProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_freedesktop_accounts_user_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY, OrgFreedesktopAccountsUserProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_freedesktop_accounts_user_interface_info ());
}

static void
org_freedesktop_accounts_user_proxy_class_init (OrgFreedesktopAccountsUserProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_freedesktop_accounts_user_proxy_finalize;
  gobject_class->get_property = org_freedesktop_accounts_user_proxy_get_property;
  gobject_class->set_property = org_freedesktop_accounts_user_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_freedesktop_accounts_user_proxy_g_signal;
  proxy_class->g_properties_changed = org_freedesktop_accounts_user_proxy_g_properties_changed;

  org_freedesktop_accounts_user_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgFreedesktopAccountsUserProxyPrivate));
#endif
}

static void
org_freedesktop_accounts_user_proxy_iface_init (OrgFreedesktopAccountsUserIface *iface)
{
  iface->get_uid = org_freedesktop_accounts_user_proxy_get_uid;
  iface->get_user_name = org_freedesktop_accounts_user_proxy_get_user_name;
  iface->get_real_name = org_freedesktop_accounts_user_proxy_get_real_name;
  iface->get_account_type = org_freedesktop_accounts_user_proxy_get_account_type;
  iface->get_home_directory = org_freedesktop_accounts_user_proxy_get_home_directory;
  iface->get_shell = org_freedesktop_accounts_user_proxy_get_shell;
  iface->get_email = org_freedesktop_accounts_user_proxy_get_email;
  iface->get_language = org_freedesktop_accounts_user_proxy_get_language;
  iface->get_xsession = org_freedesktop_accounts_user_proxy_get_xsession;
  iface->get_location = org_freedesktop_accounts_user_proxy_get_location;
  iface->get_login_frequency = org_freedesktop_accounts_user_proxy_get_login_frequency;
  iface->get_login_time = org_freedesktop_accounts_user_proxy_get_login_time;
  iface->get_login_history = org_freedesktop_accounts_user_proxy_get_login_history;
  iface->get_icon_file = org_freedesktop_accounts_user_proxy_get_icon_file;
  iface->get_locked = org_freedesktop_accounts_user_proxy_get_locked;
  iface->get_password_mode = org_freedesktop_accounts_user_proxy_get_password_mode;
  iface->get_password_hint = org_freedesktop_accounts_user_proxy_get_password_hint;
  iface->get_automatic_login = org_freedesktop_accounts_user_proxy_get_automatic_login;
  iface->get_system_account = org_freedesktop_accounts_user_proxy_get_system_account;
  iface->get_local_account = org_freedesktop_accounts_user_proxy_get_local_account;
}

/**
 * org_freedesktop_accounts_user_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_proxy_new_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_freedesktop_accounts_user_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
}

/**
 * org_freedesktop_accounts_user_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_freedesktop_accounts_user_proxy_new().
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccountsUser *
org_freedesktop_accounts_user_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS_USER (ret);
  else
    return NULL;
}

/**
 * org_freedesktop_accounts_user_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccountsUser *
org_freedesktop_accounts_user_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS_USER (ret);
  else
    return NULL;
}


/**
 * org_freedesktop_accounts_user_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_freedesktop_accounts_user_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_user_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_user_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_freedesktop_accounts_user_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
}

/**
 * org_freedesktop_accounts_user_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_user_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_freedesktop_accounts_user_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccountsUser *
org_freedesktop_accounts_user_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS_USER (ret);
  else
    return NULL;
}

/**
 * org_freedesktop_accounts_user_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_freedesktop_accounts_user_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_user_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsUserProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccountsUser *
org_freedesktop_accounts_user_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts.User", NULL);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS_USER (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgFreedesktopAccountsUserSkeleton:
 *
 * The #OrgFreedesktopAccountsUserSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgFreedesktopAccountsUserSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgFreedesktopAccountsUserSkeleton.
 */

struct _OrgFreedesktopAccountsUserSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_freedesktop_accounts_user_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_freedesktop_accounts_user_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_user_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_freedesktop_accounts_user_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_user_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_freedesktop_accounts_user_skeleton_vtable =
{
  _org_freedesktop_accounts_user_skeleton_handle_method_call,
  _org_freedesktop_accounts_user_skeleton_handle_get_property,
  _org_freedesktop_accounts_user_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_freedesktop_accounts_user_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_freedesktop_accounts_user_interface_info ();
}

static GDBusInterfaceVTable *
org_freedesktop_accounts_user_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_freedesktop_accounts_user_skeleton_vtable;
}

static GVariant *
org_freedesktop_accounts_user_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_freedesktop_accounts_user_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_freedesktop_accounts_user_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_freedesktop_accounts_user_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_freedesktop_accounts_user_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.freedesktop.Accounts.User", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_freedesktop_accounts_user_emit_changed (gpointer user_data);

static void
org_freedesktop_accounts_user_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_freedesktop_accounts_user_emit_changed (skeleton);
}

static void
_org_freedesktop_accounts_user_on_signal_changed (
    OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.freedesktop.Accounts.User", "Changed",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_freedesktop_accounts_user_skeleton_iface_init (OrgFreedesktopAccountsUserIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsUserSkeleton, org_freedesktop_accounts_user_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgFreedesktopAccountsUserSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER, org_freedesktop_accounts_user_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsUserSkeleton, org_freedesktop_accounts_user_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER, org_freedesktop_accounts_user_skeleton_iface_init))

#endif
static void
org_freedesktop_accounts_user_skeleton_finalize (GObject *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  guint n;
  for (n = 0; n < 20; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_freedesktop_accounts_user_skeleton_parent_class)->finalize (object);
}

static void
org_freedesktop_accounts_user_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 20);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_freedesktop_accounts_user_emit_changed (gpointer user_data)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.freedesktop.Accounts.User",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_freedesktop_accounts_user_schedule_emit_changed (OrgFreedesktopAccountsUserSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_freedesktop_accounts_user_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_freedesktop_accounts_user_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_freedesktop_accounts_user_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_freedesktop_accounts_user_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 20);
  info = (const _ExtendedGDBusPropertyInfo *) _org_freedesktop_accounts_user_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_freedesktop_accounts_user_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_freedesktop_accounts_user_skeleton_init (OrgFreedesktopAccountsUserSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_freedesktop_accounts_user_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON, OrgFreedesktopAccountsUserSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 20);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_UINT64);
  g_value_init (&skeleton->priv->properties[1], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[2], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[3], G_TYPE_INT);
  g_value_init (&skeleton->priv->properties[4], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[5], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[6], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[7], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[8], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[9], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[10], G_TYPE_UINT64);
  g_value_init (&skeleton->priv->properties[11], G_TYPE_INT64);
  g_value_init (&skeleton->priv->properties[12], G_TYPE_VARIANT);
  g_value_init (&skeleton->priv->properties[13], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[14], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[15], G_TYPE_INT);
  g_value_init (&skeleton->priv->properties[16], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[17], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[18], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[19], G_TYPE_BOOLEAN);
}

static guint64 
org_freedesktop_accounts_user_skeleton_get_uid (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  guint64 value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_uint64 (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_user_name (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[1]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_real_name (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[2]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gint 
org_freedesktop_accounts_user_skeleton_get_account_type (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  gint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int (&(skeleton->priv->properties[3]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_home_directory (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[4]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_shell (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[5]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_email (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[6]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_language (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[7]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_xsession (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[8]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_location (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[9]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static guint64 
org_freedesktop_accounts_user_skeleton_get_login_frequency (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  guint64 value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_uint64 (&(skeleton->priv->properties[10]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gint64 
org_freedesktop_accounts_user_skeleton_get_login_time (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  gint64 value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int64 (&(skeleton->priv->properties[11]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static GVariant *
org_freedesktop_accounts_user_skeleton_get_login_history (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  GVariant *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_variant (&(skeleton->priv->properties[12]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_icon_file (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[13]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
org_freedesktop_accounts_user_skeleton_get_locked (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[14]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gint 
org_freedesktop_accounts_user_skeleton_get_password_mode (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  gint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int (&(skeleton->priv->properties[15]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_freedesktop_accounts_user_skeleton_get_password_hint (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[16]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
org_freedesktop_accounts_user_skeleton_get_automatic_login (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[17]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
org_freedesktop_accounts_user_skeleton_get_system_account (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[18]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
org_freedesktop_accounts_user_skeleton_get_local_account (OrgFreedesktopAccountsUser *object)
{
  OrgFreedesktopAccountsUserSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[19]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_freedesktop_accounts_user_skeleton_class_init (OrgFreedesktopAccountsUserSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_freedesktop_accounts_user_skeleton_finalize;
  gobject_class->get_property = org_freedesktop_accounts_user_skeleton_get_property;
  gobject_class->set_property = org_freedesktop_accounts_user_skeleton_set_property;
  gobject_class->notify       = org_freedesktop_accounts_user_skeleton_notify;


  org_freedesktop_accounts_user_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_freedesktop_accounts_user_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_freedesktop_accounts_user_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_freedesktop_accounts_user_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_freedesktop_accounts_user_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgFreedesktopAccountsUserSkeletonPrivate));
#endif
}

static void
org_freedesktop_accounts_user_skeleton_iface_init (OrgFreedesktopAccountsUserIface *iface)
{
  iface->changed = _org_freedesktop_accounts_user_on_signal_changed;
  iface->get_uid = org_freedesktop_accounts_user_skeleton_get_uid;
  iface->get_user_name = org_freedesktop_accounts_user_skeleton_get_user_name;
  iface->get_real_name = org_freedesktop_accounts_user_skeleton_get_real_name;
  iface->get_account_type = org_freedesktop_accounts_user_skeleton_get_account_type;
  iface->get_home_directory = org_freedesktop_accounts_user_skeleton_get_home_directory;
  iface->get_shell = org_freedesktop_accounts_user_skeleton_get_shell;
  iface->get_email = org_freedesktop_accounts_user_skeleton_get_email;
  iface->get_language = org_freedesktop_accounts_user_skeleton_get_language;
  iface->get_xsession = org_freedesktop_accounts_user_skeleton_get_xsession;
  iface->get_location = org_freedesktop_accounts_user_skeleton_get_location;
  iface->get_login_frequency = org_freedesktop_accounts_user_skeleton_get_login_frequency;
  iface->get_login_time = org_freedesktop_accounts_user_skeleton_get_login_time;
  iface->get_login_history = org_freedesktop_accounts_user_skeleton_get_login_history;
  iface->get_icon_file = org_freedesktop_accounts_user_skeleton_get_icon_file;
  iface->get_locked = org_freedesktop_accounts_user_skeleton_get_locked;
  iface->get_password_mode = org_freedesktop_accounts_user_skeleton_get_password_mode;
  iface->get_password_hint = org_freedesktop_accounts_user_skeleton_get_password_hint;
  iface->get_automatic_login = org_freedesktop_accounts_user_skeleton_get_automatic_login;
  iface->get_system_account = org_freedesktop_accounts_user_skeleton_get_system_account;
  iface->get_local_account = org_freedesktop_accounts_user_skeleton_get_local_account;
}

/**
 * org_freedesktop_accounts_user_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts-User.top_of_page">org.freedesktop.Accounts.User</link>.
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsUserSkeleton): The skeleton object.
 */
OrgFreedesktopAccountsUser *
org_freedesktop_accounts_user_skeleton_new (void)
{
  return ORG_FREEDESKTOP_ACCOUNTS_USER (g_object_new (TYPE_ORG_FREEDESKTOP_ACCOUNTS_USER_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.freedesktop.Accounts
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgFreedesktopAccounts
 * @title: OrgFreedesktopAccounts
 * @short_description: Generated C code for the org.freedesktop.Accounts D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-freedesktop-Accounts.top_of_page">org.freedesktop.Accounts</link> D-Bus interface in C.
 */

enum
{
  _ORG_FREEDESKTOP_ACCOUNTS_USER_ADDED,
  _ORG_FREEDESKTOP_ACCOUNTS_USER_DELETED,
};

static unsigned _ORG_FREEDESKTOP_ACCOUNTS_SIGNALS[2] = { 0 };

/* ---- Introspection data for org.freedesktop.Accounts ---- */

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_list_cached_users_OUT_ARG_users =
{
  {
    -1,
    (gchar *) "users",
    (gchar *) "ao",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_list_cached_users_OUT_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_list_cached_users_OUT_ARG_users.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_list_cached_users_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_method_list_cached_users_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_method_list_cached_users_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_method_info_list_cached_users =
{
  {
    -1,
    (gchar *) "ListCachedUsers",
    NULL,
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_list_cached_users_OUT_ARG_pointers,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_method_list_cached_users_annotation_info_pointers
  },
  "handle-list-cached-users",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_find_user_by_id_IN_ARG_id =
{
  {
    -1,
    (gchar *) "id",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_find_user_by_id_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_find_user_by_id_IN_ARG_id.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_find_user_by_id_OUT_ARG_user =
{
  {
    -1,
    (gchar *) "user",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_find_user_by_id_OUT_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_find_user_by_id_OUT_ARG_user.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_find_user_by_id_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_method_find_user_by_id_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_method_find_user_by_id_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_method_info_find_user_by_id =
{
  {
    -1,
    (gchar *) "FindUserById",
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_find_user_by_id_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_find_user_by_id_OUT_ARG_pointers,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_method_find_user_by_id_annotation_info_pointers
  },
  "handle-find-user-by-id",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_find_user_by_name_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_find_user_by_name_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_find_user_by_name_IN_ARG_name.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_find_user_by_name_OUT_ARG_user =
{
  {
    -1,
    (gchar *) "user",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_find_user_by_name_OUT_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_find_user_by_name_OUT_ARG_user.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_find_user_by_name_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_method_find_user_by_name_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_method_find_user_by_name_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_method_info_find_user_by_name =
{
  {
    -1,
    (gchar *) "FindUserByName",
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_find_user_by_name_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_find_user_by_name_OUT_ARG_pointers,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_method_find_user_by_name_annotation_info_pointers
  },
  "handle-find-user-by-name",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_create_user_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_create_user_IN_ARG_fullname =
{
  {
    -1,
    (gchar *) "fullname",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_create_user_IN_ARG_accountType =
{
  {
    -1,
    (gchar *) "accountType",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_create_user_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_create_user_IN_ARG_name.parent_struct,
  &_org_freedesktop_accounts_method_info_create_user_IN_ARG_fullname.parent_struct,
  &_org_freedesktop_accounts_method_info_create_user_IN_ARG_accountType.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_create_user_OUT_ARG_user =
{
  {
    -1,
    (gchar *) "user",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_create_user_OUT_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_create_user_OUT_ARG_user.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_create_user_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_create_user_annotation_info_1 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_method_create_user_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_method_create_user_annotation_info_0,
  &_org_freedesktop_accounts_method_create_user_annotation_info_1,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_method_info_create_user =
{
  {
    -1,
    (gchar *) "CreateUser",
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_create_user_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_create_user_OUT_ARG_pointers,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_method_create_user_annotation_info_pointers
  },
  "handle-create-user",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_cache_user_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_cache_user_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_cache_user_IN_ARG_name.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_cache_user_OUT_ARG_user =
{
  {
    -1,
    (gchar *) "user",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_cache_user_OUT_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_cache_user_OUT_ARG_user.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_cache_user_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_method_cache_user_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_method_cache_user_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_method_info_cache_user =
{
  {
    -1,
    (gchar *) "CacheUser",
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_cache_user_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_cache_user_OUT_ARG_pointers,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_method_cache_user_annotation_info_pointers
  },
  "handle-cache-user",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_uncache_user_IN_ARG_name =
{
  {
    -1,
    (gchar *) "name",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_uncache_user_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_uncache_user_IN_ARG_name.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_uncache_user_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_method_uncache_user_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_method_uncache_user_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_method_info_uncache_user =
{
  {
    -1,
    (gchar *) "UncacheUser",
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_uncache_user_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_method_uncache_user_annotation_info_pointers
  },
  "handle-uncache-user",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_delete_user_IN_ARG_id =
{
  {
    -1,
    (gchar *) "id",
    (gchar *) "x",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_method_info_delete_user_IN_ARG_removeFiles =
{
  {
    -1,
    (gchar *) "removeFiles",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_method_info_delete_user_IN_ARG_pointers[] =
{
  &_org_freedesktop_accounts_method_info_delete_user_IN_ARG_id.parent_struct,
  &_org_freedesktop_accounts_method_info_delete_user_IN_ARG_removeFiles.parent_struct,
  NULL
};

static const GDBusAnnotationInfo _org_freedesktop_accounts_method_delete_user_annotation_info_0 =
{
  -1,
  (gchar *) "org.freedesktop.DBus.GLib.Async",
  (gchar *) "",
  NULL
};

static const GDBusAnnotationInfo * const _org_freedesktop_accounts_method_delete_user_annotation_info_pointers[] =
{
  &_org_freedesktop_accounts_method_delete_user_annotation_info_0,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_freedesktop_accounts_method_info_delete_user =
{
  {
    -1,
    (gchar *) "DeleteUser",
    (GDBusArgInfo **) &_org_freedesktop_accounts_method_info_delete_user_IN_ARG_pointers,
    NULL,
    (GDBusAnnotationInfo **) &_org_freedesktop_accounts_method_delete_user_annotation_info_pointers
  },
  "handle-delete-user",
  FALSE
};

static const GDBusMethodInfo * const _org_freedesktop_accounts_method_info_pointers[] =
{
  &_org_freedesktop_accounts_method_info_list_cached_users.parent_struct,
  &_org_freedesktop_accounts_method_info_find_user_by_id.parent_struct,
  &_org_freedesktop_accounts_method_info_find_user_by_name.parent_struct,
  &_org_freedesktop_accounts_method_info_create_user.parent_struct,
  &_org_freedesktop_accounts_method_info_cache_user.parent_struct,
  &_org_freedesktop_accounts_method_info_uncache_user.parent_struct,
  &_org_freedesktop_accounts_method_info_delete_user.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_signal_info_user_added_ARG_user =
{
  {
    -1,
    (gchar *) "user",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_signal_info_user_added_ARG_pointers[] =
{
  &_org_freedesktop_accounts_signal_info_user_added_ARG_user.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_freedesktop_accounts_signal_info_user_added =
{
  {
    -1,
    (gchar *) "UserAdded",
    (GDBusArgInfo **) &_org_freedesktop_accounts_signal_info_user_added_ARG_pointers,
    NULL
  },
  "user-added"
};

static const _ExtendedGDBusArgInfo _org_freedesktop_accounts_signal_info_user_deleted_ARG_user =
{
  {
    -1,
    (gchar *) "user",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_freedesktop_accounts_signal_info_user_deleted_ARG_pointers[] =
{
  &_org_freedesktop_accounts_signal_info_user_deleted_ARG_user.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_freedesktop_accounts_signal_info_user_deleted =
{
  {
    -1,
    (gchar *) "UserDeleted",
    (GDBusArgInfo **) &_org_freedesktop_accounts_signal_info_user_deleted_ARG_pointers,
    NULL
  },
  "user-deleted"
};

static const GDBusSignalInfo * const _org_freedesktop_accounts_signal_info_pointers[] =
{
  &_org_freedesktop_accounts_signal_info_user_added.parent_struct,
  &_org_freedesktop_accounts_signal_info_user_deleted.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_freedesktop_accounts_property_info_daemon_version =
{
  {
    -1,
    (gchar *) "DaemonVersion",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "daemon-version",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_freedesktop_accounts_property_info_pointers[] =
{
  &_org_freedesktop_accounts_property_info_daemon_version.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_freedesktop_accounts_interface_info =
{
  {
    -1,
    (gchar *) "org.freedesktop.Accounts",
    (GDBusMethodInfo **) &_org_freedesktop_accounts_method_info_pointers,
    (GDBusSignalInfo **) &_org_freedesktop_accounts_signal_info_pointers,
    (GDBusPropertyInfo **) &_org_freedesktop_accounts_property_info_pointers,
    NULL
  },
  "org-freedesktop-accounts",
};


/**
 * org_freedesktop_accounts_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-freedesktop-Accounts.top_of_page">org.freedesktop.Accounts</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_freedesktop_accounts_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_freedesktop_accounts_interface_info.parent_struct;
}

/**
 * org_freedesktop_accounts_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgFreedesktopAccounts interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_freedesktop_accounts_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "daemon-version");
  return property_id_begin - 1;
}


inline static void
org_freedesktop_accounts_signal_marshal_user_added (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_signal_marshal_user_deleted (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_method_marshal_list_cached_users (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_method_marshal_find_user_by_id (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64 (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_method_marshal_find_user_by_name (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_method_marshal_create_user (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_method_marshal_cache_user (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_method_marshal_uncache_user (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_freedesktop_accounts_method_marshal_delete_user (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT64_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgFreedesktopAccounts:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts.top_of_page">org.freedesktop.Accounts</link>.
 */

/**
 * OrgFreedesktopAccountsIface:
 * @parent_iface: The parent interface.
 * @handle_cache_user: Handler for the #OrgFreedesktopAccounts::handle-cache-user signal.
 * @handle_create_user: Handler for the #OrgFreedesktopAccounts::handle-create-user signal.
 * @handle_delete_user: Handler for the #OrgFreedesktopAccounts::handle-delete-user signal.
 * @handle_find_user_by_id: Handler for the #OrgFreedesktopAccounts::handle-find-user-by-id signal.
 * @handle_find_user_by_name: Handler for the #OrgFreedesktopAccounts::handle-find-user-by-name signal.
 * @handle_list_cached_users: Handler for the #OrgFreedesktopAccounts::handle-list-cached-users signal.
 * @handle_uncache_user: Handler for the #OrgFreedesktopAccounts::handle-uncache-user signal.
 * @get_daemon_version: Getter for the #OrgFreedesktopAccounts:daemon-version property.
 * @user_added: Handler for the #OrgFreedesktopAccounts::user-added signal.
 * @user_deleted: Handler for the #OrgFreedesktopAccounts::user-deleted signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts.top_of_page">org.freedesktop.Accounts</link>.
 */

typedef OrgFreedesktopAccountsIface OrgFreedesktopAccountsInterface;
G_DEFINE_INTERFACE (OrgFreedesktopAccounts, org_freedesktop_accounts, G_TYPE_OBJECT)

static void
org_freedesktop_accounts_default_init (OrgFreedesktopAccountsIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgFreedesktopAccounts::handle-list-cached-users:
   * @object: A #OrgFreedesktopAccounts.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts.ListCachedUsers">ListCachedUsers()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_complete_list_cached_users() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-list-cached-users",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, handle_list_cached_users),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_method_marshal_list_cached_users,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgFreedesktopAccounts::handle-find-user-by-id:
   * @object: A #OrgFreedesktopAccounts.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_id: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserById">FindUserById()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_complete_find_user_by_id() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-find-user-by-id",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, handle_find_user_by_id),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_method_marshal_find_user_by_id,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT64);

  /**
   * OrgFreedesktopAccounts::handle-find-user-by-name:
   * @object: A #OrgFreedesktopAccounts.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserByName">FindUserByName()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_complete_find_user_by_name() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-find-user-by-name",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, handle_find_user_by_name),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_method_marshal_find_user_by_name,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccounts::handle-create-user:
   * @object: A #OrgFreedesktopAccounts.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   * @arg_fullname: Argument passed by remote caller.
   * @arg_accountType: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts.CreateUser">CreateUser()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_complete_create_user() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-create-user",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, handle_create_user),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_method_marshal_create_user,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

  /**
   * OrgFreedesktopAccounts::handle-cache-user:
   * @object: A #OrgFreedesktopAccounts.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts.CacheUser">CacheUser()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_complete_cache_user() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-cache-user",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, handle_cache_user),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_method_marshal_cache_user,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccounts::handle-uncache-user:
   * @object: A #OrgFreedesktopAccounts.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_name: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts.UncacheUser">UncacheUser()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_complete_uncache_user() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-uncache-user",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, handle_uncache_user),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_method_marshal_uncache_user,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccounts::handle-delete-user:
   * @object: A #OrgFreedesktopAccounts.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_id: Argument passed by remote caller.
   * @arg_removeFiles: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-freedesktop-Accounts.DeleteUser">DeleteUser()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_freedesktop_accounts_complete_delete_user() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-delete-user",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, handle_delete_user),
    g_signal_accumulator_true_handled,
    NULL,
      org_freedesktop_accounts_method_marshal_delete_user,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT64, G_TYPE_BOOLEAN);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgFreedesktopAccounts::user-added:
   * @object: A #OrgFreedesktopAccounts.
   * @arg_user: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-freedesktop-Accounts.UserAdded">"UserAdded"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_FREEDESKTOP_ACCOUNTS_SIGNALS[_ORG_FREEDESKTOP_ACCOUNTS_USER_ADDED] =
    g_signal_new ("user-added",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, user_added),
      NULL,
      NULL,
      org_freedesktop_accounts_signal_marshal_user_added,
      G_TYPE_NONE,
      1, G_TYPE_STRING);

  /**
   * OrgFreedesktopAccounts::user-deleted:
   * @object: A #OrgFreedesktopAccounts.
   * @arg_user: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-freedesktop-Accounts.UserDeleted">"UserDeleted"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_FREEDESKTOP_ACCOUNTS_SIGNALS[_ORG_FREEDESKTOP_ACCOUNTS_USER_DELETED] =
    g_signal_new ("user-deleted",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgFreedesktopAccountsIface, user_deleted),
      NULL,
      NULL,
      org_freedesktop_accounts_signal_marshal_user_deleted,
      G_TYPE_NONE,
      1, G_TYPE_STRING);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgFreedesktopAccounts:daemon-version:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-freedesktop-Accounts.DaemonVersion">"DaemonVersion"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("daemon-version", "DaemonVersion", "DaemonVersion", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_freedesktop_accounts_get_daemon_version: (skip)
 * @object: A #OrgFreedesktopAccounts.
 *
 * Gets the value of the <link linkend="gdbus-property-org-freedesktop-Accounts.DaemonVersion">"DaemonVersion"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_freedesktop_accounts_dup_daemon_version() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_freedesktop_accounts_get_daemon_version (OrgFreedesktopAccounts *object)
{
  g_return_val_if_fail (IS_ORG_FREEDESKTOP_ACCOUNTS (object), NULL);

  return ORG_FREEDESKTOP_ACCOUNTS_GET_IFACE (object)->get_daemon_version (object);
}

/**
 * org_freedesktop_accounts_dup_daemon_version: (skip)
 * @object: A #OrgFreedesktopAccounts.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-freedesktop-Accounts.DaemonVersion">"DaemonVersion"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_freedesktop_accounts_dup_daemon_version (OrgFreedesktopAccounts *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "daemon-version", &value, NULL);
  return value;
}

/**
 * org_freedesktop_accounts_set_daemon_version: (skip)
 * @object: A #OrgFreedesktopAccounts.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-freedesktop-Accounts.DaemonVersion">"DaemonVersion"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_freedesktop_accounts_set_daemon_version (OrgFreedesktopAccounts *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "daemon-version", value, NULL);
}

/**
 * org_freedesktop_accounts_emit_user_added:
 * @object: A #OrgFreedesktopAccounts.
 * @arg_user: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-freedesktop-Accounts.UserAdded">"UserAdded"</link> D-Bus signal.
 */
void
org_freedesktop_accounts_emit_user_added (
    OrgFreedesktopAccounts *object,
    const gchar *arg_user)
{
  g_signal_emit (object, _ORG_FREEDESKTOP_ACCOUNTS_SIGNALS[_ORG_FREEDESKTOP_ACCOUNTS_USER_ADDED], 0, arg_user);
}

/**
 * org_freedesktop_accounts_emit_user_deleted:
 * @object: A #OrgFreedesktopAccounts.
 * @arg_user: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-freedesktop-Accounts.UserDeleted">"UserDeleted"</link> D-Bus signal.
 */
void
org_freedesktop_accounts_emit_user_deleted (
    OrgFreedesktopAccounts *object,
    const gchar *arg_user)
{
  g_signal_emit (object, _ORG_FREEDESKTOP_ACCOUNTS_SIGNALS[_ORG_FREEDESKTOP_ACCOUNTS_USER_DELETED], 0, arg_user);
}

/**
 * org_freedesktop_accounts_call_list_cached_users:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.ListCachedUsers">ListCachedUsers()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_call_list_cached_users_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_call_list_cached_users_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_call_list_cached_users (
    OrgFreedesktopAccounts *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ListCachedUsers",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_call_list_cached_users_finish:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @out_users: (out) (optional) (array zero-terminated=1): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_call_list_cached_users().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_call_list_cached_users().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_list_cached_users_finish (
    OrgFreedesktopAccounts *proxy,
    gchar *** out_users,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(^ao)",
                 out_users);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_list_cached_users_sync:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @out_users: (out) (optional) (array zero-terminated=1): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.ListCachedUsers">ListCachedUsers()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_call_list_cached_users() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_list_cached_users_sync (
    OrgFreedesktopAccounts *proxy,
    gchar *** out_users,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ListCachedUsers",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(^ao)",
                 out_users);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_find_user_by_id:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_id: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserById">FindUserById()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_call_find_user_by_id_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_call_find_user_by_id_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_call_find_user_by_id (
    OrgFreedesktopAccounts *proxy,
    gint64 arg_id,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "FindUserById",
    g_variant_new ("(x)",
                   arg_id),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_call_find_user_by_id_finish:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_call_find_user_by_id().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_call_find_user_by_id().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_find_user_by_id_finish (
    OrgFreedesktopAccounts *proxy,
    gchar ** out_user,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_find_user_by_id_sync:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_id: Argument to pass with the method invocation.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserById">FindUserById()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_call_find_user_by_id() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_find_user_by_id_sync (
    OrgFreedesktopAccounts *proxy,
    gint64 arg_id,
    gchar ** out_user,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "FindUserById",
    g_variant_new ("(x)",
                   arg_id),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_find_user_by_name:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserByName">FindUserByName()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_call_find_user_by_name_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_call_find_user_by_name_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_call_find_user_by_name (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "FindUserByName",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_call_find_user_by_name_finish:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_call_find_user_by_name().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_call_find_user_by_name().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_find_user_by_name_finish (
    OrgFreedesktopAccounts *proxy,
    gchar ** out_user,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_find_user_by_name_sync:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserByName">FindUserByName()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_call_find_user_by_name() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_find_user_by_name_sync (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    gchar ** out_user,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "FindUserByName",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_create_user:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @arg_fullname: Argument to pass with the method invocation.
 * @arg_accountType: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.CreateUser">CreateUser()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_call_create_user_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_call_create_user_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_call_create_user (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    const gchar *arg_fullname,
    gint arg_accountType,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "CreateUser",
    g_variant_new ("(ssi)",
                   arg_name,
                   arg_fullname,
                   arg_accountType),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_call_create_user_finish:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_call_create_user().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_call_create_user().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_create_user_finish (
    OrgFreedesktopAccounts *proxy,
    gchar ** out_user,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_create_user_sync:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @arg_fullname: Argument to pass with the method invocation.
 * @arg_accountType: Argument to pass with the method invocation.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.CreateUser">CreateUser()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_call_create_user() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_create_user_sync (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    const gchar *arg_fullname,
    gint arg_accountType,
    gchar ** out_user,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "CreateUser",
    g_variant_new ("(ssi)",
                   arg_name,
                   arg_fullname,
                   arg_accountType),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_cache_user:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.CacheUser">CacheUser()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_call_cache_user_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_call_cache_user_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_call_cache_user (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "CacheUser",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_call_cache_user_finish:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_call_cache_user().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_call_cache_user().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_cache_user_finish (
    OrgFreedesktopAccounts *proxy,
    gchar ** out_user,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_cache_user_sync:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @out_user: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.CacheUser">CacheUser()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_call_cache_user() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_cache_user_sync (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    gchar ** out_user,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "CacheUser",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_user);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_uncache_user:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.UncacheUser">UncacheUser()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_call_uncache_user_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_call_uncache_user_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_call_uncache_user (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "UncacheUser",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_call_uncache_user_finish:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_call_uncache_user().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_call_uncache_user().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_uncache_user_finish (
    OrgFreedesktopAccounts *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_uncache_user_sync:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_name: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.UncacheUser">UncacheUser()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_call_uncache_user() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_uncache_user_sync (
    OrgFreedesktopAccounts *proxy,
    const gchar *arg_name,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "UncacheUser",
    g_variant_new ("(s)",
                   arg_name),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_delete_user:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_id: Argument to pass with the method invocation.
 * @arg_removeFiles: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.DeleteUser">DeleteUser()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_call_delete_user_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_call_delete_user_sync() for the synchronous, blocking version of this method.
 */
void
org_freedesktop_accounts_call_delete_user (
    OrgFreedesktopAccounts *proxy,
    gint64 arg_id,
    gboolean arg_removeFiles,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "DeleteUser",
    g_variant_new ("(xb)",
                   arg_id,
                   arg_removeFiles),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_freedesktop_accounts_call_delete_user_finish:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_call_delete_user().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_freedesktop_accounts_call_delete_user().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_delete_user_finish (
    OrgFreedesktopAccounts *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_call_delete_user_sync:
 * @proxy: A #OrgFreedesktopAccountsProxy.
 * @arg_id: Argument to pass with the method invocation.
 * @arg_removeFiles: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-freedesktop-Accounts.DeleteUser">DeleteUser()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_call_delete_user() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_freedesktop_accounts_call_delete_user_sync (
    OrgFreedesktopAccounts *proxy,
    gint64 arg_id,
    gboolean arg_removeFiles,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "DeleteUser",
    g_variant_new ("(xb)",
                   arg_id,
                   arg_removeFiles),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_freedesktop_accounts_complete_list_cached_users:
 * @object: A #OrgFreedesktopAccounts.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @users: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts.ListCachedUsers">ListCachedUsers()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_complete_list_cached_users (
    OrgFreedesktopAccounts *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *const *users)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(^ao)",
                   users));
}

/**
 * org_freedesktop_accounts_complete_find_user_by_id:
 * @object: A #OrgFreedesktopAccounts.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @user: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserById">FindUserById()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_complete_find_user_by_id (
    OrgFreedesktopAccounts *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *user)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   user));
}

/**
 * org_freedesktop_accounts_complete_find_user_by_name:
 * @object: A #OrgFreedesktopAccounts.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @user: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts.FindUserByName">FindUserByName()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_complete_find_user_by_name (
    OrgFreedesktopAccounts *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *user)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   user));
}

/**
 * org_freedesktop_accounts_complete_create_user:
 * @object: A #OrgFreedesktopAccounts.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @user: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts.CreateUser">CreateUser()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_complete_create_user (
    OrgFreedesktopAccounts *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *user)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   user));
}

/**
 * org_freedesktop_accounts_complete_cache_user:
 * @object: A #OrgFreedesktopAccounts.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @user: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts.CacheUser">CacheUser()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_complete_cache_user (
    OrgFreedesktopAccounts *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *user)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   user));
}

/**
 * org_freedesktop_accounts_complete_uncache_user:
 * @object: A #OrgFreedesktopAccounts.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts.UncacheUser">UncacheUser()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_complete_uncache_user (
    OrgFreedesktopAccounts *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_freedesktop_accounts_complete_delete_user:
 * @object: A #OrgFreedesktopAccounts.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-freedesktop-Accounts.DeleteUser">DeleteUser()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_freedesktop_accounts_complete_delete_user (
    OrgFreedesktopAccounts *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgFreedesktopAccountsProxy:
 *
 * The #OrgFreedesktopAccountsProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgFreedesktopAccountsProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgFreedesktopAccountsProxy.
 */

struct _OrgFreedesktopAccountsProxyPrivate
{
  GData *qdata;
};

static void org_freedesktop_accounts_proxy_iface_init (OrgFreedesktopAccountsIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsProxy, org_freedesktop_accounts_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgFreedesktopAccountsProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS, org_freedesktop_accounts_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsProxy, org_freedesktop_accounts_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS, org_freedesktop_accounts_proxy_iface_init))

#endif
static void
org_freedesktop_accounts_proxy_finalize (GObject *object)
{
  OrgFreedesktopAccountsProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_freedesktop_accounts_proxy_parent_class)->finalize (object);
}

static void
org_freedesktop_accounts_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_freedesktop_accounts_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_freedesktop_accounts_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.freedesktop.Accounts: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_freedesktop_accounts_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_freedesktop_accounts_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.freedesktop.Accounts", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_freedesktop_accounts_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_freedesktop_accounts_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_FREEDESKTOP_ACCOUNTS);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_FREEDESKTOP_ACCOUNTS);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_freedesktop_accounts_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgFreedesktopAccountsProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static const gchar *
org_freedesktop_accounts_proxy_get_daemon_version (OrgFreedesktopAccounts *object)
{
  OrgFreedesktopAccountsProxy *proxy = ORG_FREEDESKTOP_ACCOUNTS_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "DaemonVersion");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_freedesktop_accounts_proxy_init (OrgFreedesktopAccountsProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_freedesktop_accounts_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_FREEDESKTOP_ACCOUNTS_PROXY, OrgFreedesktopAccountsProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_freedesktop_accounts_interface_info ());
}

static void
org_freedesktop_accounts_proxy_class_init (OrgFreedesktopAccountsProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_freedesktop_accounts_proxy_finalize;
  gobject_class->get_property = org_freedesktop_accounts_proxy_get_property;
  gobject_class->set_property = org_freedesktop_accounts_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_freedesktop_accounts_proxy_g_signal;
  proxy_class->g_properties_changed = org_freedesktop_accounts_proxy_g_properties_changed;

  org_freedesktop_accounts_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgFreedesktopAccountsProxyPrivate));
#endif
}

static void
org_freedesktop_accounts_proxy_iface_init (OrgFreedesktopAccountsIface *iface)
{
  iface->get_daemon_version = org_freedesktop_accounts_proxy_get_daemon_version;
}

/**
 * org_freedesktop_accounts_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts.top_of_page">org.freedesktop.Accounts</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_proxy_new_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_freedesktop_accounts_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_FREEDESKTOP_ACCOUNTS_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts", NULL);
}

/**
 * org_freedesktop_accounts_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_freedesktop_accounts_proxy_new().
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccounts *
org_freedesktop_accounts_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS (ret);
  else
    return NULL;
}

/**
 * org_freedesktop_accounts_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts.top_of_page">org.freedesktop.Accounts</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccounts *
org_freedesktop_accounts_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_FREEDESKTOP_ACCOUNTS_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts", NULL);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS (ret);
  else
    return NULL;
}


/**
 * org_freedesktop_accounts_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_freedesktop_accounts_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_freedesktop_accounts_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_freedesktop_accounts_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_freedesktop_accounts_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_FREEDESKTOP_ACCOUNTS_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts", NULL);
}

/**
 * org_freedesktop_accounts_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_freedesktop_accounts_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_freedesktop_accounts_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccounts *
org_freedesktop_accounts_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS (ret);
  else
    return NULL;
}

/**
 * org_freedesktop_accounts_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_freedesktop_accounts_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_freedesktop_accounts_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgFreedesktopAccounts *
org_freedesktop_accounts_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_FREEDESKTOP_ACCOUNTS_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.freedesktop.Accounts", NULL);
  if (ret != NULL)
    return ORG_FREEDESKTOP_ACCOUNTS (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgFreedesktopAccountsSkeleton:
 *
 * The #OrgFreedesktopAccountsSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgFreedesktopAccountsSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgFreedesktopAccountsSkeleton.
 */

struct _OrgFreedesktopAccountsSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_freedesktop_accounts_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_FREEDESKTOP_ACCOUNTS);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_FREEDESKTOP_ACCOUNTS);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_freedesktop_accounts_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_freedesktop_accounts_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_freedesktop_accounts_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_freedesktop_accounts_skeleton_vtable =
{
  _org_freedesktop_accounts_skeleton_handle_method_call,
  _org_freedesktop_accounts_skeleton_handle_get_property,
  _org_freedesktop_accounts_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_freedesktop_accounts_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_freedesktop_accounts_interface_info ();
}

static GDBusInterfaceVTable *
org_freedesktop_accounts_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_freedesktop_accounts_skeleton_vtable;
}

static GVariant *
org_freedesktop_accounts_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_freedesktop_accounts_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_freedesktop_accounts_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_freedesktop_accounts_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_freedesktop_accounts_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.freedesktop.Accounts", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_freedesktop_accounts_emit_changed (gpointer user_data);

static void
org_freedesktop_accounts_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_freedesktop_accounts_emit_changed (skeleton);
}

static void
_org_freedesktop_accounts_on_signal_user_added (
    OrgFreedesktopAccounts *object,
    const gchar *arg_user)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(o)",
                   arg_user));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.freedesktop.Accounts", "UserAdded",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_freedesktop_accounts_on_signal_user_deleted (
    OrgFreedesktopAccounts *object,
    const gchar *arg_user)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(o)",
                   arg_user));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.freedesktop.Accounts", "UserDeleted",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_freedesktop_accounts_skeleton_iface_init (OrgFreedesktopAccountsIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsSkeleton, org_freedesktop_accounts_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgFreedesktopAccountsSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS, org_freedesktop_accounts_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgFreedesktopAccountsSkeleton, org_freedesktop_accounts_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_FREEDESKTOP_ACCOUNTS, org_freedesktop_accounts_skeleton_iface_init))

#endif
static void
org_freedesktop_accounts_skeleton_finalize (GObject *object)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (object);
  guint n;
  for (n = 0; n < 1; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_freedesktop_accounts_skeleton_parent_class)->finalize (object);
}

static void
org_freedesktop_accounts_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_freedesktop_accounts_emit_changed (gpointer user_data)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.freedesktop.Accounts",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_freedesktop_accounts_schedule_emit_changed (OrgFreedesktopAccountsSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_freedesktop_accounts_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_freedesktop_accounts_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_freedesktop_accounts_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_freedesktop_accounts_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_freedesktop_accounts_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_freedesktop_accounts_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_freedesktop_accounts_skeleton_init (OrgFreedesktopAccountsSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_freedesktop_accounts_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_FREEDESKTOP_ACCOUNTS_SKELETON, OrgFreedesktopAccountsSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 1);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_STRING);
}

static const gchar *
org_freedesktop_accounts_skeleton_get_daemon_version (OrgFreedesktopAccounts *object)
{
  OrgFreedesktopAccountsSkeleton *skeleton = ORG_FREEDESKTOP_ACCOUNTS_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_freedesktop_accounts_skeleton_class_init (OrgFreedesktopAccountsSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_freedesktop_accounts_skeleton_finalize;
  gobject_class->get_property = org_freedesktop_accounts_skeleton_get_property;
  gobject_class->set_property = org_freedesktop_accounts_skeleton_set_property;
  gobject_class->notify       = org_freedesktop_accounts_skeleton_notify;


  org_freedesktop_accounts_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_freedesktop_accounts_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_freedesktop_accounts_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_freedesktop_accounts_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_freedesktop_accounts_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgFreedesktopAccountsSkeletonPrivate));
#endif
}

static void
org_freedesktop_accounts_skeleton_iface_init (OrgFreedesktopAccountsIface *iface)
{
  iface->user_added = _org_freedesktop_accounts_on_signal_user_added;
  iface->user_deleted = _org_freedesktop_accounts_on_signal_user_deleted;
  iface->get_daemon_version = org_freedesktop_accounts_skeleton_get_daemon_version;
}

/**
 * org_freedesktop_accounts_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-freedesktop-Accounts.top_of_page">org.freedesktop.Accounts</link>.
 *
 * Returns: (transfer full) (type OrgFreedesktopAccountsSkeleton): The skeleton object.
 */
OrgFreedesktopAccounts *
org_freedesktop_accounts_skeleton_new (void)
{
  return ORG_FREEDESKTOP_ACCOUNTS (g_object_new (TYPE_ORG_FREEDESKTOP_ACCOUNTS_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Settings.GlobalShortcutsProvider
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeSettingsGlobalShortcutsProvider
 * @title: OrgGnomeSettingsGlobalShortcutsProvider
 * @short_description: Generated C code for the org.gnome.Settings.GlobalShortcutsProvider D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Settings-GlobalShortcutsProvider.top_of_page">org.gnome.Settings.GlobalShortcutsProvider</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gnome.Settings.GlobalShortcutsProvider ---- */

static const _ExtendedGDBusArgInfo _org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_app_id =
{
  {
    -1,
    (gchar *) "app_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_parent_window =
{
  {
    -1,
    (gchar *) "parent_window",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_shortcuts =
{
  {
    -1,
    (gchar *) "shortcuts",
    (gchar *) "a(sa{sv})",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_pointers[] =
{
  &_org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_app_id.parent_struct,
  &_org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_parent_window.parent_struct,
  &_org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_shortcuts.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_OUT_ARG_results =
{
  {
    -1,
    (gchar *) "results",
    (gchar *) "a(sa{sv})",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_OUT_ARG_pointers[] =
{
  &_org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_OUT_ARG_results.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts =
{
  {
    -1,
    (gchar *) "BindShortcuts",
    (GDBusArgInfo **) &_org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts_OUT_ARG_pointers,
    NULL
  },
  "handle-bind-shortcuts",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_settings_global_shortcuts_provider_method_info_pointers[] =
{
  &_org_gnome_settings_global_shortcuts_provider_method_info_bind_shortcuts.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_settings_global_shortcuts_provider_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Settings.GlobalShortcutsProvider",
    (GDBusMethodInfo **) &_org_gnome_settings_global_shortcuts_provider_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "org-gnome-settings-global-shortcuts-provider",
};


/**
 * org_gnome_settings_global_shortcuts_provider_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Settings-GlobalShortcutsProvider.top_of_page">org.gnome.Settings.GlobalShortcutsProvider</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_settings_global_shortcuts_provider_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct;
}

/**
 * org_gnome_settings_global_shortcuts_provider_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeSettingsGlobalShortcutsProvider interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_settings_global_shortcuts_provider_override_properties (GObjectClass *klass G_GNUC_UNUSED, guint property_id_begin)
{
  return property_id_begin - 1;
}


inline static void
org_gnome_settings_global_shortcuts_provider_method_marshal_bind_shortcuts (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeSettingsGlobalShortcutsProvider:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Settings-GlobalShortcutsProvider.top_of_page">org.gnome.Settings.GlobalShortcutsProvider</link>.
 */

/**
 * OrgGnomeSettingsGlobalShortcutsProviderIface:
 * @parent_iface: The parent interface.
 * @handle_bind_shortcuts: Handler for the #OrgGnomeSettingsGlobalShortcutsProvider::handle-bind-shortcuts signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Settings-GlobalShortcutsProvider.top_of_page">org.gnome.Settings.GlobalShortcutsProvider</link>.
 */

typedef OrgGnomeSettingsGlobalShortcutsProviderIface OrgGnomeSettingsGlobalShortcutsProviderInterface;
G_DEFINE_INTERFACE (OrgGnomeSettingsGlobalShortcutsProvider, org_gnome_settings_global_shortcuts_provider, G_TYPE_OBJECT)

static void
org_gnome_settings_global_shortcuts_provider_default_init (OrgGnomeSettingsGlobalShortcutsProviderIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeSettingsGlobalShortcutsProvider::handle-bind-shortcuts:
   * @object: A #OrgGnomeSettingsGlobalShortcutsProvider.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_app_id: Argument passed by remote caller.
   * @arg_parent_window: Argument passed by remote caller.
   * @arg_shortcuts: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Settings-GlobalShortcutsProvider.BindShortcuts">BindShortcuts()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_settings_global_shortcuts_provider_complete_bind_shortcuts() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-bind-shortcuts",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeSettingsGlobalShortcutsProviderIface, handle_bind_shortcuts),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_settings_global_shortcuts_provider_method_marshal_bind_shortcuts,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_VARIANT);

}

/**
 * org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts:
 * @proxy: A #OrgGnomeSettingsGlobalShortcutsProviderProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_parent_window: Argument to pass with the method invocation.
 * @arg_shortcuts: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Settings-GlobalShortcutsProvider.BindShortcuts">BindShortcuts()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts_finish() to get the result of the operation.
 *
 * See org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts (
    OrgGnomeSettingsGlobalShortcutsProvider *proxy,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    GVariant *arg_shortcuts,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "BindShortcuts",
    g_variant_new ("(ss@a(sa{sv}))",
                   arg_app_id,
                   arg_parent_window,
                   arg_shortcuts),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts_finish:
 * @proxy: A #OrgGnomeSettingsGlobalShortcutsProviderProxy.
 * @out_results: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts_finish (
    OrgGnomeSettingsGlobalShortcutsProvider *proxy,
    GVariant ** out_results,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(sa{sv}))",
                 out_results);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts_sync:
 * @proxy: A #OrgGnomeSettingsGlobalShortcutsProviderProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_parent_window: Argument to pass with the method invocation.
 * @arg_shortcuts: Argument to pass with the method invocation.
 * @out_results: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Settings-GlobalShortcutsProvider.BindShortcuts">BindShortcuts()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_settings_global_shortcuts_provider_call_bind_shortcuts_sync (
    OrgGnomeSettingsGlobalShortcutsProvider *proxy,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    GVariant *arg_shortcuts,
    GVariant ** out_results,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "BindShortcuts",
    g_variant_new ("(ss@a(sa{sv}))",
                   arg_app_id,
                   arg_parent_window,
                   arg_shortcuts),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a(sa{sv}))",
                 out_results);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_settings_global_shortcuts_provider_complete_bind_shortcuts:
 * @object: A #OrgGnomeSettingsGlobalShortcutsProvider.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @results: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Settings-GlobalShortcutsProvider.BindShortcuts">BindShortcuts()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_settings_global_shortcuts_provider_complete_bind_shortcuts (
    OrgGnomeSettingsGlobalShortcutsProvider *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GVariant *results)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a(sa{sv}))",
                   results));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeSettingsGlobalShortcutsProviderProxy:
 *
 * The #OrgGnomeSettingsGlobalShortcutsProviderProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeSettingsGlobalShortcutsProviderProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeSettingsGlobalShortcutsProviderProxy.
 */

struct _OrgGnomeSettingsGlobalShortcutsProviderProxyPrivate
{
  GData *qdata;
};

static void org_gnome_settings_global_shortcuts_provider_proxy_iface_init (OrgGnomeSettingsGlobalShortcutsProviderIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeSettingsGlobalShortcutsProviderProxy, org_gnome_settings_global_shortcuts_provider_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeSettingsGlobalShortcutsProviderProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER, org_gnome_settings_global_shortcuts_provider_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeSettingsGlobalShortcutsProviderProxy, org_gnome_settings_global_shortcuts_provider_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER, org_gnome_settings_global_shortcuts_provider_proxy_iface_init))

#endif
static void
org_gnome_settings_global_shortcuts_provider_proxy_finalize (GObject *object)
{
  OrgGnomeSettingsGlobalShortcutsProviderProxy *proxy = ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_settings_global_shortcuts_provider_proxy_parent_class)->finalize (object);
}

static void
org_gnome_settings_global_shortcuts_provider_proxy_get_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  GValue       *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_settings_global_shortcuts_provider_proxy_set_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  const GValue *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_settings_global_shortcuts_provider_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_settings_global_shortcuts_provider_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeSettingsGlobalShortcutsProviderProxy *proxy = ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
org_gnome_settings_global_shortcuts_provider_proxy_init (OrgGnomeSettingsGlobalShortcutsProviderProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_settings_global_shortcuts_provider_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_PROXY, OrgGnomeSettingsGlobalShortcutsProviderProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_settings_global_shortcuts_provider_interface_info ());
}

static void
org_gnome_settings_global_shortcuts_provider_proxy_class_init (OrgGnomeSettingsGlobalShortcutsProviderProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_settings_global_shortcuts_provider_proxy_finalize;
  gobject_class->get_property = org_gnome_settings_global_shortcuts_provider_proxy_get_property;
  gobject_class->set_property = org_gnome_settings_global_shortcuts_provider_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_settings_global_shortcuts_provider_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_settings_global_shortcuts_provider_proxy_g_properties_changed;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeSettingsGlobalShortcutsProviderProxyPrivate));
#endif
}

static void
org_gnome_settings_global_shortcuts_provider_proxy_iface_init (OrgGnomeSettingsGlobalShortcutsProviderIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_settings_global_shortcuts_provider_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Settings-GlobalShortcutsProvider.top_of_page">org.gnome.Settings.GlobalShortcutsProvider</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_settings_global_shortcuts_provider_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_settings_global_shortcuts_provider_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_settings_global_shortcuts_provider_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Settings.GlobalShortcutsProvider", NULL);
}

/**
 * org_gnome_settings_global_shortcuts_provider_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_settings_global_shortcuts_provider_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_settings_global_shortcuts_provider_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeSettingsGlobalShortcutsProviderProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeSettingsGlobalShortcutsProvider *
org_gnome_settings_global_shortcuts_provider_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER (ret);
  else
    return NULL;
}

/**
 * org_gnome_settings_global_shortcuts_provider_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Settings-GlobalShortcutsProvider.top_of_page">org.gnome.Settings.GlobalShortcutsProvider</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_settings_global_shortcuts_provider_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeSettingsGlobalShortcutsProviderProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeSettingsGlobalShortcutsProvider *
org_gnome_settings_global_shortcuts_provider_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Settings.GlobalShortcutsProvider", NULL);
  if (ret != NULL)
    return ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER (ret);
  else
    return NULL;
}


/**
 * org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_settings_global_shortcuts_provider_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Settings.GlobalShortcutsProvider", NULL);
}

/**
 * org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeSettingsGlobalShortcutsProviderProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeSettingsGlobalShortcutsProvider *
org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER (ret);
  else
    return NULL;
}

/**
 * org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_settings_global_shortcuts_provider_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeSettingsGlobalShortcutsProviderProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeSettingsGlobalShortcutsProvider *
org_gnome_settings_global_shortcuts_provider_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Settings.GlobalShortcutsProvider", NULL);
  if (ret != NULL)
    return ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeSettingsGlobalShortcutsProviderSkeleton:
 *
 * The #OrgGnomeSettingsGlobalShortcutsProviderSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeSettingsGlobalShortcutsProviderSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeSettingsGlobalShortcutsProviderSkeleton.
 */

struct _OrgGnomeSettingsGlobalShortcutsProviderSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_settings_global_shortcuts_provider_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeSettingsGlobalShortcutsProviderSkeleton *skeleton = ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_settings_global_shortcuts_provider_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeSettingsGlobalShortcutsProviderSkeleton *skeleton = ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_settings_global_shortcuts_provider_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeSettingsGlobalShortcutsProviderSkeleton *skeleton = ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_settings_global_shortcuts_provider_skeleton_vtable =
{
  _org_gnome_settings_global_shortcuts_provider_skeleton_handle_method_call,
  _org_gnome_settings_global_shortcuts_provider_skeleton_handle_get_property,
  _org_gnome_settings_global_shortcuts_provider_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_settings_global_shortcuts_provider_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_settings_global_shortcuts_provider_skeleton_vtable;
}

static GVariant *
org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeSettingsGlobalShortcutsProviderSkeleton *skeleton = ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_settings_global_shortcuts_provider_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_settings_global_shortcuts_provider_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Settings.GlobalShortcutsProvider", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton G_GNUC_UNUSED)
{
}

static void org_gnome_settings_global_shortcuts_provider_skeleton_iface_init (OrgGnomeSettingsGlobalShortcutsProviderIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeSettingsGlobalShortcutsProviderSkeleton, org_gnome_settings_global_shortcuts_provider_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeSettingsGlobalShortcutsProviderSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER, org_gnome_settings_global_shortcuts_provider_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeSettingsGlobalShortcutsProviderSkeleton, org_gnome_settings_global_shortcuts_provider_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER, org_gnome_settings_global_shortcuts_provider_skeleton_iface_init))

#endif
static void
org_gnome_settings_global_shortcuts_provider_skeleton_finalize (GObject *object)
{
  OrgGnomeSettingsGlobalShortcutsProviderSkeleton *skeleton = ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_settings_global_shortcuts_provider_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_settings_global_shortcuts_provider_skeleton_init (OrgGnomeSettingsGlobalShortcutsProviderSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_settings_global_shortcuts_provider_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_SKELETON, OrgGnomeSettingsGlobalShortcutsProviderSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
org_gnome_settings_global_shortcuts_provider_skeleton_class_init (OrgGnomeSettingsGlobalShortcutsProviderSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_settings_global_shortcuts_provider_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_settings_global_shortcuts_provider_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeSettingsGlobalShortcutsProviderSkeletonPrivate));
#endif
}

static void
org_gnome_settings_global_shortcuts_provider_skeleton_iface_init (OrgGnomeSettingsGlobalShortcutsProviderIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_settings_global_shortcuts_provider_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Settings-GlobalShortcutsProvider.top_of_page">org.gnome.Settings.GlobalShortcutsProvider</link>.
 *
 * Returns: (transfer full) (type OrgGnomeSettingsGlobalShortcutsProviderSkeleton): The skeleton object.
 */
OrgGnomeSettingsGlobalShortcutsProvider *
org_gnome_settings_global_shortcuts_provider_skeleton_new (void)
{
  return ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER (g_object_new (TYPE_ORG_GNOME_SETTINGS_GLOBAL_SHORTCUTS_PROVIDER_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.GlobalShortcutsRebind
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeGlobalShortcutsRebind
 * @title: OrgGnomeGlobalShortcutsRebind
 * @short_description: Generated C code for the org.gnome.GlobalShortcutsRebind D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-GlobalShortcutsRebind.top_of_page">org.gnome.GlobalShortcutsRebind</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gnome.GlobalShortcutsRebind ---- */

static const _ExtendedGDBusArgInfo _org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts_IN_ARG_app_id =
{
  {
    -1,
    (gchar *) "app_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts_IN_ARG_shortcuts =
{
  {
    -1,
    (gchar *) "shortcuts",
    (gchar *) "a(sa{sv})",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts_IN_ARG_pointers[] =
{
  &_org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts_IN_ARG_app_id.parent_struct,
  &_org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts_IN_ARG_shortcuts.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts =
{
  {
    -1,
    (gchar *) "RebindShortcuts",
    (GDBusArgInfo **) &_org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-rebind-shortcuts",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_global_shortcuts_rebind_method_info_pointers[] =
{
  &_org_gnome_global_shortcuts_rebind_method_info_rebind_shortcuts.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_global_shortcuts_rebind_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.GlobalShortcutsRebind",
    (GDBusMethodInfo **) &_org_gnome_global_shortcuts_rebind_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "org-gnome-global-shortcuts-rebind",
};


/**
 * org_gnome_global_shortcuts_rebind_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-GlobalShortcutsRebind.top_of_page">org.gnome.GlobalShortcutsRebind</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_global_shortcuts_rebind_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_global_shortcuts_rebind_interface_info.parent_struct;
}

/**
 * org_gnome_global_shortcuts_rebind_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeGlobalShortcutsRebind interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_global_shortcuts_rebind_override_properties (GObjectClass *klass G_GNUC_UNUSED, guint property_id_begin)
{
  return property_id_begin - 1;
}


inline static void
org_gnome_global_shortcuts_rebind_method_marshal_rebind_shortcuts (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeGlobalShortcutsRebind:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-GlobalShortcutsRebind.top_of_page">org.gnome.GlobalShortcutsRebind</link>.
 */

/**
 * OrgGnomeGlobalShortcutsRebindIface:
 * @parent_iface: The parent interface.
 * @handle_rebind_shortcuts: Handler for the #OrgGnomeGlobalShortcutsRebind::handle-rebind-shortcuts signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-GlobalShortcutsRebind.top_of_page">org.gnome.GlobalShortcutsRebind</link>.
 */

typedef OrgGnomeGlobalShortcutsRebindIface OrgGnomeGlobalShortcutsRebindInterface;
G_DEFINE_INTERFACE (OrgGnomeGlobalShortcutsRebind, org_gnome_global_shortcuts_rebind, G_TYPE_OBJECT)

static void
org_gnome_global_shortcuts_rebind_default_init (OrgGnomeGlobalShortcutsRebindIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeGlobalShortcutsRebind::handle-rebind-shortcuts:
   * @object: A #OrgGnomeGlobalShortcutsRebind.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_app_id: Argument passed by remote caller.
   * @arg_shortcuts: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-GlobalShortcutsRebind.RebindShortcuts">RebindShortcuts()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_global_shortcuts_rebind_complete_rebind_shortcuts() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-rebind-shortcuts",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeGlobalShortcutsRebindIface, handle_rebind_shortcuts),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_global_shortcuts_rebind_method_marshal_rebind_shortcuts,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_VARIANT);

}

/**
 * org_gnome_global_shortcuts_rebind_call_rebind_shortcuts:
 * @proxy: A #OrgGnomeGlobalShortcutsRebindProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_shortcuts: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-GlobalShortcutsRebind.RebindShortcuts">RebindShortcuts()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_global_shortcuts_rebind_call_rebind_shortcuts_finish() to get the result of the operation.
 *
 * See org_gnome_global_shortcuts_rebind_call_rebind_shortcuts_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_global_shortcuts_rebind_call_rebind_shortcuts (
    OrgGnomeGlobalShortcutsRebind *proxy,
    const gchar *arg_app_id,
    GVariant *arg_shortcuts,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RebindShortcuts",
    g_variant_new ("(s@a(sa{sv}))",
                   arg_app_id,
                   arg_shortcuts),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_global_shortcuts_rebind_call_rebind_shortcuts_finish:
 * @proxy: A #OrgGnomeGlobalShortcutsRebindProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_global_shortcuts_rebind_call_rebind_shortcuts().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_global_shortcuts_rebind_call_rebind_shortcuts().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_global_shortcuts_rebind_call_rebind_shortcuts_finish (
    OrgGnomeGlobalShortcutsRebind *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_global_shortcuts_rebind_call_rebind_shortcuts_sync:
 * @proxy: A #OrgGnomeGlobalShortcutsRebindProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_shortcuts: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-GlobalShortcutsRebind.RebindShortcuts">RebindShortcuts()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_global_shortcuts_rebind_call_rebind_shortcuts() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_global_shortcuts_rebind_call_rebind_shortcuts_sync (
    OrgGnomeGlobalShortcutsRebind *proxy,
    const gchar *arg_app_id,
    GVariant *arg_shortcuts,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RebindShortcuts",
    g_variant_new ("(s@a(sa{sv}))",
                   arg_app_id,
                   arg_shortcuts),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_global_shortcuts_rebind_complete_rebind_shortcuts:
 * @object: A #OrgGnomeGlobalShortcutsRebind.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-GlobalShortcutsRebind.RebindShortcuts">RebindShortcuts()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_global_shortcuts_rebind_complete_rebind_shortcuts (
    OrgGnomeGlobalShortcutsRebind *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeGlobalShortcutsRebindProxy:
 *
 * The #OrgGnomeGlobalShortcutsRebindProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeGlobalShortcutsRebindProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeGlobalShortcutsRebindProxy.
 */

struct _OrgGnomeGlobalShortcutsRebindProxyPrivate
{
  GData *qdata;
};

static void org_gnome_global_shortcuts_rebind_proxy_iface_init (OrgGnomeGlobalShortcutsRebindIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeGlobalShortcutsRebindProxy, org_gnome_global_shortcuts_rebind_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeGlobalShortcutsRebindProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND, org_gnome_global_shortcuts_rebind_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeGlobalShortcutsRebindProxy, org_gnome_global_shortcuts_rebind_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND, org_gnome_global_shortcuts_rebind_proxy_iface_init))

#endif
static void
org_gnome_global_shortcuts_rebind_proxy_finalize (GObject *object)
{
  OrgGnomeGlobalShortcutsRebindProxy *proxy = ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_global_shortcuts_rebind_proxy_parent_class)->finalize (object);
}

static void
org_gnome_global_shortcuts_rebind_proxy_get_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  GValue       *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_global_shortcuts_rebind_proxy_set_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  const GValue *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_global_shortcuts_rebind_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_global_shortcuts_rebind_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_global_shortcuts_rebind_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeGlobalShortcutsRebindProxy *proxy = ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_global_shortcuts_rebind_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_global_shortcuts_rebind_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
org_gnome_global_shortcuts_rebind_proxy_init (OrgGnomeGlobalShortcutsRebindProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_global_shortcuts_rebind_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_PROXY, OrgGnomeGlobalShortcutsRebindProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_global_shortcuts_rebind_interface_info ());
}

static void
org_gnome_global_shortcuts_rebind_proxy_class_init (OrgGnomeGlobalShortcutsRebindProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_global_shortcuts_rebind_proxy_finalize;
  gobject_class->get_property = org_gnome_global_shortcuts_rebind_proxy_get_property;
  gobject_class->set_property = org_gnome_global_shortcuts_rebind_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_global_shortcuts_rebind_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_global_shortcuts_rebind_proxy_g_properties_changed;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeGlobalShortcutsRebindProxyPrivate));
#endif
}

static void
org_gnome_global_shortcuts_rebind_proxy_iface_init (OrgGnomeGlobalShortcutsRebindIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_global_shortcuts_rebind_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-GlobalShortcutsRebind.top_of_page">org.gnome.GlobalShortcutsRebind</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_global_shortcuts_rebind_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_global_shortcuts_rebind_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_global_shortcuts_rebind_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.GlobalShortcutsRebind", NULL);
}

/**
 * org_gnome_global_shortcuts_rebind_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_global_shortcuts_rebind_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_global_shortcuts_rebind_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeGlobalShortcutsRebindProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeGlobalShortcutsRebind *
org_gnome_global_shortcuts_rebind_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_GLOBAL_SHORTCUTS_REBIND (ret);
  else
    return NULL;
}

/**
 * org_gnome_global_shortcuts_rebind_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-GlobalShortcutsRebind.top_of_page">org.gnome.GlobalShortcutsRebind</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_global_shortcuts_rebind_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeGlobalShortcutsRebindProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeGlobalShortcutsRebind *
org_gnome_global_shortcuts_rebind_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.GlobalShortcutsRebind", NULL);
  if (ret != NULL)
    return ORG_GNOME_GLOBAL_SHORTCUTS_REBIND (ret);
  else
    return NULL;
}


/**
 * org_gnome_global_shortcuts_rebind_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_global_shortcuts_rebind_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_global_shortcuts_rebind_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_global_shortcuts_rebind_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_global_shortcuts_rebind_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.GlobalShortcutsRebind", NULL);
}

/**
 * org_gnome_global_shortcuts_rebind_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_global_shortcuts_rebind_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_global_shortcuts_rebind_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeGlobalShortcutsRebindProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeGlobalShortcutsRebind *
org_gnome_global_shortcuts_rebind_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_GLOBAL_SHORTCUTS_REBIND (ret);
  else
    return NULL;
}

/**
 * org_gnome_global_shortcuts_rebind_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_global_shortcuts_rebind_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_global_shortcuts_rebind_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeGlobalShortcutsRebindProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeGlobalShortcutsRebind *
org_gnome_global_shortcuts_rebind_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.GlobalShortcutsRebind", NULL);
  if (ret != NULL)
    return ORG_GNOME_GLOBAL_SHORTCUTS_REBIND (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeGlobalShortcutsRebindSkeleton:
 *
 * The #OrgGnomeGlobalShortcutsRebindSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeGlobalShortcutsRebindSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeGlobalShortcutsRebindSkeleton.
 */

struct _OrgGnomeGlobalShortcutsRebindSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_global_shortcuts_rebind_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeGlobalShortcutsRebindSkeleton *skeleton = ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_global_shortcuts_rebind_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeGlobalShortcutsRebindSkeleton *skeleton = ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_global_shortcuts_rebind_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_global_shortcuts_rebind_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeGlobalShortcutsRebindSkeleton *skeleton = ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_global_shortcuts_rebind_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_global_shortcuts_rebind_skeleton_vtable =
{
  _org_gnome_global_shortcuts_rebind_skeleton_handle_method_call,
  _org_gnome_global_shortcuts_rebind_skeleton_handle_get_property,
  _org_gnome_global_shortcuts_rebind_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_global_shortcuts_rebind_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_global_shortcuts_rebind_skeleton_vtable;
}

static GVariant *
org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeGlobalShortcutsRebindSkeleton *skeleton = ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_global_shortcuts_rebind_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_global_shortcuts_rebind_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_global_shortcuts_rebind_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_global_shortcuts_rebind_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.GlobalShortcutsRebind", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton G_GNUC_UNUSED)
{
}

static void org_gnome_global_shortcuts_rebind_skeleton_iface_init (OrgGnomeGlobalShortcutsRebindIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeGlobalShortcutsRebindSkeleton, org_gnome_global_shortcuts_rebind_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeGlobalShortcutsRebindSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND, org_gnome_global_shortcuts_rebind_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeGlobalShortcutsRebindSkeleton, org_gnome_global_shortcuts_rebind_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND, org_gnome_global_shortcuts_rebind_skeleton_iface_init))

#endif
static void
org_gnome_global_shortcuts_rebind_skeleton_finalize (GObject *object)
{
  OrgGnomeGlobalShortcutsRebindSkeleton *skeleton = ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_global_shortcuts_rebind_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_global_shortcuts_rebind_skeleton_init (OrgGnomeGlobalShortcutsRebindSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_global_shortcuts_rebind_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_SKELETON, OrgGnomeGlobalShortcutsRebindSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
org_gnome_global_shortcuts_rebind_skeleton_class_init (OrgGnomeGlobalShortcutsRebindSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_global_shortcuts_rebind_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_global_shortcuts_rebind_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeGlobalShortcutsRebindSkeletonPrivate));
#endif
}

static void
org_gnome_global_shortcuts_rebind_skeleton_iface_init (OrgGnomeGlobalShortcutsRebindIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_global_shortcuts_rebind_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-GlobalShortcutsRebind.top_of_page">org.gnome.GlobalShortcutsRebind</link>.
 *
 * Returns: (transfer full) (type OrgGnomeGlobalShortcutsRebindSkeleton): The skeleton object.
 */
OrgGnomeGlobalShortcutsRebind *
org_gnome_global_shortcuts_rebind_skeleton_new (void)
{
  return ORG_GNOME_GLOBAL_SHORTCUTS_REBIND (g_object_new (TYPE_ORG_GNOME_GLOBAL_SHORTCUTS_REBIND_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.DisplayConfig
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterDisplayConfig
 * @title: OrgGnomeMutterDisplayConfig
 * @short_description: Generated C code for the org.gnome.Mutter.DisplayConfig D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-DisplayConfig.top_of_page">org.gnome.Mutter.DisplayConfig</link> D-Bus interface in C.
 */

enum
{
  _ORG_GNOME_MUTTER_DISPLAY_CONFIG_MONITORS_CHANGED,
};

static unsigned _ORG_GNOME_MUTTER_DISPLAY_CONFIG_SIGNALS[1] = { 0 };

/* ---- Introspection data for org.gnome.Mutter.DisplayConfig ---- */

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_crtcs =
{
  {
    -1,
    (gchar *) "crtcs",
    (gchar *) "a(uxiiiiiuaua{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_outputs =
{
  {
    -1,
    (gchar *) "outputs",
    (gchar *) "a(uxiausauaua{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_modes =
{
  {
    -1,
    (gchar *) "modes",
    (gchar *) "a(uxuudu)",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_max_screen_width =
{
  {
    -1,
    (gchar *) "max_screen_width",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_max_screen_height =
{
  {
    -1,
    (gchar *) "max_screen_height",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_serial.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_crtcs.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_outputs.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_modes.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_max_screen_width.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_max_screen_height.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_display_config_method_info_get_resources =
{
  {
    -1,
    (gchar *) "GetResources",
    NULL,
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_get_resources_OUT_ARG_pointers,
    NULL
  },
  "handle-get-resources",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_persistent =
{
  {
    -1,
    (gchar *) "persistent",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_crtcs =
{
  {
    -1,
    (gchar *) "crtcs",
    (gchar *) "a(uiiiuaua{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_outputs =
{
  {
    -1,
    (gchar *) "outputs",
    (gchar *) "a(ua{sv})",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_serial.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_persistent.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_crtcs.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_outputs.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_display_config_method_info_apply_configuration =
{
  {
    -1,
    (gchar *) "ApplyConfiguration",
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_apply_configuration_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-apply-configuration",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_output =
{
  {
    -1,
    (gchar *) "output",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_value =
{
  {
    -1,
    (gchar *) "value",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_serial.parent_struct,
  &_org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_output.parent_struct,
  &_org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_value.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_change_backlight_OUT_ARG_new_value =
{
  {
    -1,
    (gchar *) "new_value",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_change_backlight_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_change_backlight_OUT_ARG_new_value.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_display_config_method_info_change_backlight =
{
  {
    -1,
    (gchar *) "ChangeBacklight",
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_change_backlight_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_change_backlight_OUT_ARG_pointers,
    NULL
  },
  "handle-change-backlight",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_crtc_gamma_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_crtc_gamma_IN_ARG_crtc =
{
  {
    -1,
    (gchar *) "crtc",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_get_crtc_gamma_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_get_crtc_gamma_IN_ARG_serial.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_crtc_gamma_IN_ARG_crtc.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_red =
{
  {
    -1,
    (gchar *) "red",
    (gchar *) "aq",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_green =
{
  {
    -1,
    (gchar *) "green",
    (gchar *) "aq",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_blue =
{
  {
    -1,
    (gchar *) "blue",
    (gchar *) "aq",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_red.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_green.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_blue.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_display_config_method_info_get_crtc_gamma =
{
  {
    -1,
    (gchar *) "GetCrtcGamma",
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_get_crtc_gamma_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_get_crtc_gamma_OUT_ARG_pointers,
    NULL
  },
  "handle-get-crtc-gamma",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_crtc =
{
  {
    -1,
    (gchar *) "crtc",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_red =
{
  {
    -1,
    (gchar *) "red",
    (gchar *) "aq",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_green =
{
  {
    -1,
    (gchar *) "green",
    (gchar *) "aq",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_blue =
{
  {
    -1,
    (gchar *) "blue",
    (gchar *) "aq",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_serial.parent_struct,
  &_org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_crtc.parent_struct,
  &_org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_red.parent_struct,
  &_org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_green.parent_struct,
  &_org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_blue.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_display_config_method_info_set_crtc_gamma =
{
  {
    -1,
    (gchar *) "SetCrtcGamma",
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_set_crtc_gamma_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-set-crtc-gamma",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_monitors =
{
  {
    -1,
    (gchar *) "monitors",
    (gchar *) "a((ssss)a(siiddada{sv})a{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_logical_monitors =
{
  {
    -1,
    (gchar *) "logical_monitors",
    (gchar *) "a(iiduba(ssss)a{sv})",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_properties =
{
  {
    -1,
    (gchar *) "properties",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_serial.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_monitors.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_logical_monitors.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_properties.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_display_config_method_info_get_current_state =
{
  {
    -1,
    (gchar *) "GetCurrentState",
    NULL,
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_get_current_state_OUT_ARG_pointers,
    NULL
  },
  "handle-get-current-state",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_method =
{
  {
    -1,
    (gchar *) "method",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_logical_monitors =
{
  {
    -1,
    (gchar *) "logical_monitors",
    (gchar *) "a(iiduba(ssa{sv}))",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_properties =
{
  {
    -1,
    (gchar *) "properties",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_serial.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_method.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_logical_monitors.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_properties.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_display_config_method_info_apply_monitors_config =
{
  {
    -1,
    (gchar *) "ApplyMonitorsConfig",
    (GDBusArgInfo **) &_org_gnome_mutter_display_config_method_info_apply_monitors_config_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-apply-monitors-config",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_mutter_display_config_method_info_pointers[] =
{
  &_org_gnome_mutter_display_config_method_info_get_resources.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_configuration.parent_struct,
  &_org_gnome_mutter_display_config_method_info_change_backlight.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_crtc_gamma.parent_struct,
  &_org_gnome_mutter_display_config_method_info_set_crtc_gamma.parent_struct,
  &_org_gnome_mutter_display_config_method_info_get_current_state.parent_struct,
  &_org_gnome_mutter_display_config_method_info_apply_monitors_config.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_display_config_signal_info_monitors_changed =
{
  {
    -1,
    (gchar *) "MonitorsChanged",
    NULL,
    NULL
  },
  "monitors-changed"
};

static const GDBusSignalInfo * const _org_gnome_mutter_display_config_signal_info_pointers[] =
{
  &_org_gnome_mutter_display_config_signal_info_monitors_changed.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_display_config_property_info_power_save_mode =
{
  {
    -1,
    (gchar *) "PowerSaveMode",
    (gchar *) "i",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE | G_DBUS_PROPERTY_INFO_FLAGS_WRITABLE,
    NULL
  },
  "power-save-mode",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_mutter_display_config_property_info_pointers[] =
{
  &_org_gnome_mutter_display_config_property_info_power_save_mode.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_display_config_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.DisplayConfig",
    (GDBusMethodInfo **) &_org_gnome_mutter_display_config_method_info_pointers,
    (GDBusSignalInfo **) &_org_gnome_mutter_display_config_signal_info_pointers,
    (GDBusPropertyInfo **) &_org_gnome_mutter_display_config_property_info_pointers,
    NULL
  },
  "org-gnome-mutter-display-config",
};


/**
 * org_gnome_mutter_display_config_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-DisplayConfig.top_of_page">org.gnome.Mutter.DisplayConfig</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_display_config_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_display_config_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_display_config_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterDisplayConfig interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_display_config_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "power-save-mode");
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_display_config_signal_marshal_monitors_changed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_display_config_method_marshal_get_resources (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_display_config_method_marshal_apply_configuration (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_BOOLEAN_VARIANT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_display_config_method_marshal_change_backlight (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_display_config_method_marshal_get_crtc_gamma (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_display_config_method_marshal_set_crtc_gamma (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT_VARIANT_VARIANT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_display_config_method_marshal_get_current_state (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_display_config_method_marshal_apply_monitors_config (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_UINT_VARIANT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterDisplayConfig:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-DisplayConfig.top_of_page">org.gnome.Mutter.DisplayConfig</link>.
 */

/**
 * OrgGnomeMutterDisplayConfigIface:
 * @parent_iface: The parent interface.
 * @handle_apply_configuration: Handler for the #OrgGnomeMutterDisplayConfig::handle-apply-configuration signal.
 * @handle_apply_monitors_config: Handler for the #OrgGnomeMutterDisplayConfig::handle-apply-monitors-config signal.
 * @handle_change_backlight: Handler for the #OrgGnomeMutterDisplayConfig::handle-change-backlight signal.
 * @handle_get_crtc_gamma: Handler for the #OrgGnomeMutterDisplayConfig::handle-get-crtc-gamma signal.
 * @handle_get_current_state: Handler for the #OrgGnomeMutterDisplayConfig::handle-get-current-state signal.
 * @handle_get_resources: Handler for the #OrgGnomeMutterDisplayConfig::handle-get-resources signal.
 * @handle_set_crtc_gamma: Handler for the #OrgGnomeMutterDisplayConfig::handle-set-crtc-gamma signal.
 * @get_power_save_mode: Getter for the #OrgGnomeMutterDisplayConfig:power-save-mode property.
 * @monitors_changed: Handler for the #OrgGnomeMutterDisplayConfig::monitors-changed signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-DisplayConfig.top_of_page">org.gnome.Mutter.DisplayConfig</link>.
 */

typedef OrgGnomeMutterDisplayConfigIface OrgGnomeMutterDisplayConfigInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterDisplayConfig, org_gnome_mutter_display_config, G_TYPE_OBJECT)

static void
org_gnome_mutter_display_config_default_init (OrgGnomeMutterDisplayConfigIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeMutterDisplayConfig::handle-get-resources:
   * @object: A #OrgGnomeMutterDisplayConfig.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetResources">GetResources()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_display_config_complete_get_resources() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-resources",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, handle_get_resources),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_display_config_method_marshal_get_resources,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterDisplayConfig::handle-apply-configuration:
   * @object: A #OrgGnomeMutterDisplayConfig.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   * @arg_persistent: Argument passed by remote caller.
   * @arg_crtcs: Argument passed by remote caller.
   * @arg_outputs: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyConfiguration">ApplyConfiguration()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_display_config_complete_apply_configuration() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-apply-configuration",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, handle_apply_configuration),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_display_config_method_marshal_apply_configuration,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_BOOLEAN, G_TYPE_VARIANT, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterDisplayConfig::handle-change-backlight:
   * @object: A #OrgGnomeMutterDisplayConfig.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   * @arg_output: Argument passed by remote caller.
   * @arg_value: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ChangeBacklight">ChangeBacklight()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_display_config_complete_change_backlight() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-change-backlight",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, handle_change_backlight),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_display_config_method_marshal_change_backlight,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_INT);

  /**
   * OrgGnomeMutterDisplayConfig::handle-get-crtc-gamma:
   * @object: A #OrgGnomeMutterDisplayConfig.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   * @arg_crtc: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCrtcGamma">GetCrtcGamma()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_display_config_complete_get_crtc_gamma() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-crtc-gamma",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, handle_get_crtc_gamma),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_display_config_method_marshal_get_crtc_gamma,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_UINT);

  /**
   * OrgGnomeMutterDisplayConfig::handle-set-crtc-gamma:
   * @object: A #OrgGnomeMutterDisplayConfig.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   * @arg_crtc: Argument passed by remote caller.
   * @arg_red: Argument passed by remote caller.
   * @arg_green: Argument passed by remote caller.
   * @arg_blue: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.SetCrtcGamma">SetCrtcGamma()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_display_config_complete_set_crtc_gamma() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-crtc-gamma",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, handle_set_crtc_gamma),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_display_config_method_marshal_set_crtc_gamma,
    G_TYPE_BOOLEAN,
    6,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_VARIANT, G_TYPE_VARIANT, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterDisplayConfig::handle-get-current-state:
   * @object: A #OrgGnomeMutterDisplayConfig.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCurrentState">GetCurrentState()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_display_config_complete_get_current_state() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-current-state",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, handle_get_current_state),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_display_config_method_marshal_get_current_state,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterDisplayConfig::handle-apply-monitors-config:
   * @object: A #OrgGnomeMutterDisplayConfig.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   * @arg_method: Argument passed by remote caller.
   * @arg_logical_monitors: Argument passed by remote caller.
   * @arg_properties: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyMonitorsConfig">ApplyMonitorsConfig()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_display_config_complete_apply_monitors_config() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-apply-monitors-config",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, handle_apply_monitors_config),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_display_config_method_marshal_apply_monitors_config,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_VARIANT, G_TYPE_VARIANT);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGnomeMutterDisplayConfig::monitors-changed:
   * @object: A #OrgGnomeMutterDisplayConfig.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-DisplayConfig.MonitorsChanged">"MonitorsChanged"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_DISPLAY_CONFIG_SIGNALS[_ORG_GNOME_MUTTER_DISPLAY_CONFIG_MONITORS_CHANGED] =
    g_signal_new ("monitors-changed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterDisplayConfigIface, monitors_changed),
      NULL,
      NULL,
      org_gnome_mutter_display_config_signal_marshal_monitors_changed,
      G_TYPE_NONE,
      0);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeMutterDisplayConfig:power-save-mode:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-DisplayConfig.PowerSaveMode">"PowerSaveMode"</link>.
   *
   * Since the D-Bus property for this #GObject property is both readable and writable, it is meaningful to both read from it and write to it on both the service- and client-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int ("power-save-mode", "PowerSaveMode", "PowerSaveMode", G_MININT32, G_MAXINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_mutter_display_config_get_power_save_mode: (skip)
 * @object: A #OrgGnomeMutterDisplayConfig.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-DisplayConfig.PowerSaveMode">"PowerSaveMode"</link> D-Bus property.
 *
 * Since this D-Bus property is both readable and writable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint 
org_gnome_mutter_display_config_get_power_save_mode (OrgGnomeMutterDisplayConfig *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_DISPLAY_CONFIG (object), 0);

  return ORG_GNOME_MUTTER_DISPLAY_CONFIG_GET_IFACE (object)->get_power_save_mode (object);
}

/**
 * org_gnome_mutter_display_config_set_power_save_mode: (skip)
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-DisplayConfig.PowerSaveMode">"PowerSaveMode"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is both readable and writable, it is meaningful to use this function on both the client- and service-side.
 */
void
org_gnome_mutter_display_config_set_power_save_mode (OrgGnomeMutterDisplayConfig *object, gint value)
{
  g_object_set (G_OBJECT (object), "power-save-mode", value, NULL);
}

/**
 * org_gnome_mutter_display_config_emit_monitors_changed:
 * @object: A #OrgGnomeMutterDisplayConfig.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-DisplayConfig.MonitorsChanged">"MonitorsChanged"</link> D-Bus signal.
 */
void
org_gnome_mutter_display_config_emit_monitors_changed (
    OrgGnomeMutterDisplayConfig *object)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_DISPLAY_CONFIG_SIGNALS[_ORG_GNOME_MUTTER_DISPLAY_CONFIG_MONITORS_CHANGED], 0);
}

/**
 * org_gnome_mutter_display_config_call_get_resources:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetResources">GetResources()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_call_get_resources_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_call_get_resources_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_display_config_call_get_resources (
    OrgGnomeMutterDisplayConfig *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetResources",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_display_config_call_get_resources_finish:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @out_serial: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_crtcs: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_outputs: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_modes: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_max_screen_width: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_max_screen_height: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_call_get_resources().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_display_config_call_get_resources().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_get_resources_finish (
    OrgGnomeMutterDisplayConfig *proxy,
    guint* out_serial,
    GVariant ** out_crtcs,
    GVariant ** out_outputs,
    GVariant ** out_modes,
    gint* out_max_screen_width,
    gint* out_max_screen_height,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u@a(uxiiiiiuaua{sv})@a(uxiausauaua{sv})@a(uxuudu)ii)",
                 out_serial,
                 out_crtcs,
                 out_outputs,
                 out_modes,
                 out_max_screen_width,
                 out_max_screen_height);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_get_resources_sync:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @out_serial: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_crtcs: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_outputs: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_modes: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_max_screen_width: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_max_screen_height: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetResources">GetResources()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_call_get_resources() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_get_resources_sync (
    OrgGnomeMutterDisplayConfig *proxy,
    guint* out_serial,
    GVariant ** out_crtcs,
    GVariant ** out_outputs,
    GVariant ** out_modes,
    gint* out_max_screen_width,
    gint* out_max_screen_height,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetResources",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u@a(uxiiiiiuaua{sv})@a(uxiausauaua{sv})@a(uxuudu)ii)",
                 out_serial,
                 out_crtcs,
                 out_outputs,
                 out_modes,
                 out_max_screen_width,
                 out_max_screen_height);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_apply_configuration:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_persistent: Argument to pass with the method invocation.
 * @arg_crtcs: Argument to pass with the method invocation.
 * @arg_outputs: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyConfiguration">ApplyConfiguration()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_call_apply_configuration_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_call_apply_configuration_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_display_config_call_apply_configuration (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    gboolean arg_persistent,
    GVariant *arg_crtcs,
    GVariant *arg_outputs,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ApplyConfiguration",
    g_variant_new ("(ub@a(uiiiuaua{sv})@a(ua{sv}))",
                   arg_serial,
                   arg_persistent,
                   arg_crtcs,
                   arg_outputs),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_display_config_call_apply_configuration_finish:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_call_apply_configuration().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_display_config_call_apply_configuration().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_apply_configuration_finish (
    OrgGnomeMutterDisplayConfig *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_apply_configuration_sync:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_persistent: Argument to pass with the method invocation.
 * @arg_crtcs: Argument to pass with the method invocation.
 * @arg_outputs: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyConfiguration">ApplyConfiguration()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_call_apply_configuration() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_apply_configuration_sync (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    gboolean arg_persistent,
    GVariant *arg_crtcs,
    GVariant *arg_outputs,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ApplyConfiguration",
    g_variant_new ("(ub@a(uiiiuaua{sv})@a(ua{sv}))",
                   arg_serial,
                   arg_persistent,
                   arg_crtcs,
                   arg_outputs),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_change_backlight:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_output: Argument to pass with the method invocation.
 * @arg_value: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ChangeBacklight">ChangeBacklight()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_call_change_backlight_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_call_change_backlight_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_display_config_call_change_backlight (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_output,
    gint arg_value,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ChangeBacklight",
    g_variant_new ("(uui)",
                   arg_serial,
                   arg_output,
                   arg_value),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_display_config_call_change_backlight_finish:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @out_new_value: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_call_change_backlight().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_display_config_call_change_backlight().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_change_backlight_finish (
    OrgGnomeMutterDisplayConfig *proxy,
    gint* out_new_value,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(i)",
                 out_new_value);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_change_backlight_sync:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_output: Argument to pass with the method invocation.
 * @arg_value: Argument to pass with the method invocation.
 * @out_new_value: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ChangeBacklight">ChangeBacklight()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_call_change_backlight() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_change_backlight_sync (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_output,
    gint arg_value,
    gint* out_new_value,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ChangeBacklight",
    g_variant_new ("(uui)",
                   arg_serial,
                   arg_output,
                   arg_value),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(i)",
                 out_new_value);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_get_crtc_gamma:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_crtc: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCrtcGamma">GetCrtcGamma()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_call_get_crtc_gamma_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_call_get_crtc_gamma_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_display_config_call_get_crtc_gamma (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_crtc,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetCrtcGamma",
    g_variant_new ("(uu)",
                   arg_serial,
                   arg_crtc),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_display_config_call_get_crtc_gamma_finish:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @out_red: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_green: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_blue: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_call_get_crtc_gamma().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_display_config_call_get_crtc_gamma().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_get_crtc_gamma_finish (
    OrgGnomeMutterDisplayConfig *proxy,
    GVariant ** out_red,
    GVariant ** out_green,
    GVariant ** out_blue,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@aq@aq@aq)",
                 out_red,
                 out_green,
                 out_blue);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_get_crtc_gamma_sync:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_crtc: Argument to pass with the method invocation.
 * @out_red: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_green: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_blue: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCrtcGamma">GetCrtcGamma()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_call_get_crtc_gamma() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_get_crtc_gamma_sync (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_crtc,
    GVariant ** out_red,
    GVariant ** out_green,
    GVariant ** out_blue,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetCrtcGamma",
    g_variant_new ("(uu)",
                   arg_serial,
                   arg_crtc),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@aq@aq@aq)",
                 out_red,
                 out_green,
                 out_blue);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_set_crtc_gamma:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_crtc: Argument to pass with the method invocation.
 * @arg_red: Argument to pass with the method invocation.
 * @arg_green: Argument to pass with the method invocation.
 * @arg_blue: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.SetCrtcGamma">SetCrtcGamma()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_call_set_crtc_gamma_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_call_set_crtc_gamma_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_display_config_call_set_crtc_gamma (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_crtc,
    GVariant *arg_red,
    GVariant *arg_green,
    GVariant *arg_blue,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetCrtcGamma",
    g_variant_new ("(uu@aq@aq@aq)",
                   arg_serial,
                   arg_crtc,
                   arg_red,
                   arg_green,
                   arg_blue),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_display_config_call_set_crtc_gamma_finish:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_call_set_crtc_gamma().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_display_config_call_set_crtc_gamma().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_set_crtc_gamma_finish (
    OrgGnomeMutterDisplayConfig *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_set_crtc_gamma_sync:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_crtc: Argument to pass with the method invocation.
 * @arg_red: Argument to pass with the method invocation.
 * @arg_green: Argument to pass with the method invocation.
 * @arg_blue: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.SetCrtcGamma">SetCrtcGamma()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_call_set_crtc_gamma() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_set_crtc_gamma_sync (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_crtc,
    GVariant *arg_red,
    GVariant *arg_green,
    GVariant *arg_blue,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetCrtcGamma",
    g_variant_new ("(uu@aq@aq@aq)",
                   arg_serial,
                   arg_crtc,
                   arg_red,
                   arg_green,
                   arg_blue),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_get_current_state:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCurrentState">GetCurrentState()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_call_get_current_state_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_call_get_current_state_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_display_config_call_get_current_state (
    OrgGnomeMutterDisplayConfig *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetCurrentState",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_display_config_call_get_current_state_finish:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @out_serial: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_monitors: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_logical_monitors: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_properties: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_call_get_current_state().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_display_config_call_get_current_state().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_get_current_state_finish (
    OrgGnomeMutterDisplayConfig *proxy,
    guint* out_serial,
    GVariant ** out_monitors,
    GVariant ** out_logical_monitors,
    GVariant ** out_properties,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u@a((ssss)a(siiddada{sv})a{sv})@a(iiduba(ssss)a{sv})@a{sv})",
                 out_serial,
                 out_monitors,
                 out_logical_monitors,
                 out_properties);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_get_current_state_sync:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @out_serial: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_monitors: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_logical_monitors: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_properties: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCurrentState">GetCurrentState()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_call_get_current_state() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_get_current_state_sync (
    OrgGnomeMutterDisplayConfig *proxy,
    guint* out_serial,
    GVariant ** out_monitors,
    GVariant ** out_logical_monitors,
    GVariant ** out_properties,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetCurrentState",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u@a((ssss)a(siiddada{sv})a{sv})@a(iiduba(ssss)a{sv})@a{sv})",
                 out_serial,
                 out_monitors,
                 out_logical_monitors,
                 out_properties);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_apply_monitors_config:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_method: Argument to pass with the method invocation.
 * @arg_logical_monitors: Argument to pass with the method invocation.
 * @arg_properties: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyMonitorsConfig">ApplyMonitorsConfig()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_call_apply_monitors_config_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_call_apply_monitors_config_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_display_config_call_apply_monitors_config (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_method,
    GVariant *arg_logical_monitors,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ApplyMonitorsConfig",
    g_variant_new ("(uu@a(iiduba(ssa{sv}))@a{sv})",
                   arg_serial,
                   arg_method,
                   arg_logical_monitors,
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_display_config_call_apply_monitors_config_finish:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_call_apply_monitors_config().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_display_config_call_apply_monitors_config().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_apply_monitors_config_finish (
    OrgGnomeMutterDisplayConfig *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_call_apply_monitors_config_sync:
 * @proxy: A #OrgGnomeMutterDisplayConfigProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_method: Argument to pass with the method invocation.
 * @arg_logical_monitors: Argument to pass with the method invocation.
 * @arg_properties: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyMonitorsConfig">ApplyMonitorsConfig()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_call_apply_monitors_config() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_display_config_call_apply_monitors_config_sync (
    OrgGnomeMutterDisplayConfig *proxy,
    guint arg_serial,
    guint arg_method,
    GVariant *arg_logical_monitors,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ApplyMonitorsConfig",
    g_variant_new ("(uu@a(iiduba(ssa{sv}))@a{sv})",
                   arg_serial,
                   arg_method,
                   arg_logical_monitors,
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_display_config_complete_get_resources:
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @serial: Parameter to return.
 * @crtcs: Parameter to return.
 * @outputs: Parameter to return.
 * @modes: Parameter to return.
 * @max_screen_width: Parameter to return.
 * @max_screen_height: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetResources">GetResources()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_display_config_complete_get_resources (
    OrgGnomeMutterDisplayConfig *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    guint serial,
    GVariant *crtcs,
    GVariant *outputs,
    GVariant *modes,
    gint max_screen_width,
    gint max_screen_height)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(u@a(uxiiiiiuaua{sv})@a(uxiausauaua{sv})@a(uxuudu)ii)",
                   serial,
                   crtcs,
                   outputs,
                   modes,
                   max_screen_width,
                   max_screen_height));
}

/**
 * org_gnome_mutter_display_config_complete_apply_configuration:
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyConfiguration">ApplyConfiguration()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_display_config_complete_apply_configuration (
    OrgGnomeMutterDisplayConfig *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_display_config_complete_change_backlight:
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @new_value: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ChangeBacklight">ChangeBacklight()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_display_config_complete_change_backlight (
    OrgGnomeMutterDisplayConfig *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gint new_value)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(i)",
                   new_value));
}

/**
 * org_gnome_mutter_display_config_complete_get_crtc_gamma:
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @red: Parameter to return.
 * @green: Parameter to return.
 * @blue: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCrtcGamma">GetCrtcGamma()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_display_config_complete_get_crtc_gamma (
    OrgGnomeMutterDisplayConfig *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GVariant *red,
    GVariant *green,
    GVariant *blue)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@aq@aq@aq)",
                   red,
                   green,
                   blue));
}

/**
 * org_gnome_mutter_display_config_complete_set_crtc_gamma:
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.SetCrtcGamma">SetCrtcGamma()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_display_config_complete_set_crtc_gamma (
    OrgGnomeMutterDisplayConfig *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_display_config_complete_get_current_state:
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @serial: Parameter to return.
 * @monitors: Parameter to return.
 * @logical_monitors: Parameter to return.
 * @properties: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.GetCurrentState">GetCurrentState()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_display_config_complete_get_current_state (
    OrgGnomeMutterDisplayConfig *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    guint serial,
    GVariant *monitors,
    GVariant *logical_monitors,
    GVariant *properties)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(u@a((ssss)a(siiddada{sv})a{sv})@a(iiduba(ssss)a{sv})@a{sv})",
                   serial,
                   monitors,
                   logical_monitors,
                   properties));
}

/**
 * org_gnome_mutter_display_config_complete_apply_monitors_config:
 * @object: A #OrgGnomeMutterDisplayConfig.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-DisplayConfig.ApplyMonitorsConfig">ApplyMonitorsConfig()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_display_config_complete_apply_monitors_config (
    OrgGnomeMutterDisplayConfig *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterDisplayConfigProxy:
 *
 * The #OrgGnomeMutterDisplayConfigProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterDisplayConfigProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterDisplayConfigProxy.
 */

struct _OrgGnomeMutterDisplayConfigProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_display_config_proxy_iface_init (OrgGnomeMutterDisplayConfigIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterDisplayConfigProxy, org_gnome_mutter_display_config_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterDisplayConfigProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG, org_gnome_mutter_display_config_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterDisplayConfigProxy, org_gnome_mutter_display_config_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG, org_gnome_mutter_display_config_proxy_iface_init))

#endif
static void
org_gnome_mutter_display_config_proxy_finalize (GObject *object)
{
  OrgGnomeMutterDisplayConfigProxy *proxy = ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_display_config_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_display_config_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_display_config_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_mutter_display_config_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Mutter.DisplayConfig: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_mutter_display_config_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_display_config_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Mutter.DisplayConfig", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_mutter_display_config_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_mutter_display_config_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_display_config_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_display_config_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterDisplayConfigProxy *proxy = ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_display_config_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_display_config_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static gint 
org_gnome_mutter_display_config_proxy_get_power_save_mode (OrgGnomeMutterDisplayConfig *object)
{
  OrgGnomeMutterDisplayConfigProxy *proxy = ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY (object);
  GVariant *variant;
  gint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "PowerSaveMode");
  if (variant != NULL)
    {
      value = g_variant_get_int32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_gnome_mutter_display_config_proxy_init (OrgGnomeMutterDisplayConfigProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_display_config_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY, OrgGnomeMutterDisplayConfigProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_display_config_interface_info ());
}

static void
org_gnome_mutter_display_config_proxy_class_init (OrgGnomeMutterDisplayConfigProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_display_config_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_display_config_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_display_config_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_display_config_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_display_config_proxy_g_properties_changed;

  org_gnome_mutter_display_config_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterDisplayConfigProxyPrivate));
#endif
}

static void
org_gnome_mutter_display_config_proxy_iface_init (OrgGnomeMutterDisplayConfigIface *iface)
{
  iface->get_power_save_mode = org_gnome_mutter_display_config_proxy_get_power_save_mode;
}

/**
 * org_gnome_mutter_display_config_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-DisplayConfig.top_of_page">org.gnome.Mutter.DisplayConfig</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_display_config_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.DisplayConfig", NULL);
}

/**
 * org_gnome_mutter_display_config_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_display_config_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterDisplayConfigProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterDisplayConfig *
org_gnome_mutter_display_config_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_DISPLAY_CONFIG (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_display_config_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-DisplayConfig.top_of_page">org.gnome.Mutter.DisplayConfig</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterDisplayConfigProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterDisplayConfig *
org_gnome_mutter_display_config_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.DisplayConfig", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_DISPLAY_CONFIG (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_display_config_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_display_config_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_display_config_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_display_config_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_display_config_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.DisplayConfig", NULL);
}

/**
 * org_gnome_mutter_display_config_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_display_config_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_display_config_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterDisplayConfigProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterDisplayConfig *
org_gnome_mutter_display_config_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_DISPLAY_CONFIG (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_display_config_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_display_config_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_display_config_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterDisplayConfigProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterDisplayConfig *
org_gnome_mutter_display_config_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.DisplayConfig", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_DISPLAY_CONFIG (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterDisplayConfigSkeleton:
 *
 * The #OrgGnomeMutterDisplayConfigSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterDisplayConfigSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterDisplayConfigSkeleton.
 */

struct _OrgGnomeMutterDisplayConfigSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_display_config_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_display_config_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_display_config_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_display_config_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_display_config_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_display_config_skeleton_vtable =
{
  _org_gnome_mutter_display_config_skeleton_handle_method_call,
  _org_gnome_mutter_display_config_skeleton_handle_get_property,
  _org_gnome_mutter_display_config_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_display_config_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_display_config_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_display_config_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_display_config_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_display_config_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_display_config_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_display_config_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_display_config_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_display_config_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.DisplayConfig", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_mutter_display_config_emit_changed (gpointer user_data);

static void
org_gnome_mutter_display_config_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_mutter_display_config_emit_changed (skeleton);
}

static void
_org_gnome_mutter_display_config_on_signal_monitors_changed (
    OrgGnomeMutterDisplayConfig *object)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.DisplayConfig", "MonitorsChanged",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gnome_mutter_display_config_skeleton_iface_init (OrgGnomeMutterDisplayConfigIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterDisplayConfigSkeleton, org_gnome_mutter_display_config_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterDisplayConfigSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG, org_gnome_mutter_display_config_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterDisplayConfigSkeleton, org_gnome_mutter_display_config_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG, org_gnome_mutter_display_config_skeleton_iface_init))

#endif
static void
org_gnome_mutter_display_config_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (object);
  guint n;
  for (n = 0; n < 1; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_display_config_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_display_config_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_mutter_display_config_emit_changed (gpointer user_data)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Mutter.DisplayConfig",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_mutter_display_config_schedule_emit_changed (OrgGnomeMutterDisplayConfigSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_mutter_display_config_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_mutter_display_config_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_mutter_display_config_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_mutter_display_config_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_display_config_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_mutter_display_config_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_mutter_display_config_skeleton_init (OrgGnomeMutterDisplayConfigSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_display_config_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON, OrgGnomeMutterDisplayConfigSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 1);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_INT);
}

static gint 
org_gnome_mutter_display_config_skeleton_get_power_save_mode (OrgGnomeMutterDisplayConfig *object)
{
  OrgGnomeMutterDisplayConfigSkeleton *skeleton = ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON (object);
  gint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_mutter_display_config_skeleton_class_init (OrgGnomeMutterDisplayConfigSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_display_config_skeleton_finalize;
  gobject_class->get_property = org_gnome_mutter_display_config_skeleton_get_property;
  gobject_class->set_property = org_gnome_mutter_display_config_skeleton_set_property;
  gobject_class->notify       = org_gnome_mutter_display_config_skeleton_notify;


  org_gnome_mutter_display_config_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_display_config_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_display_config_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_display_config_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_display_config_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterDisplayConfigSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_display_config_skeleton_iface_init (OrgGnomeMutterDisplayConfigIface *iface)
{
  iface->monitors_changed = _org_gnome_mutter_display_config_on_signal_monitors_changed;
  iface->get_power_save_mode = org_gnome_mutter_display_config_skeleton_get_power_save_mode;
}

/**
 * org_gnome_mutter_display_config_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-DisplayConfig.top_of_page">org.gnome.Mutter.DisplayConfig</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterDisplayConfigSkeleton): The skeleton object.
 */
OrgGnomeMutterDisplayConfig *
org_gnome_mutter_display_config_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_DISPLAY_CONFIG (g_object_new (TYPE_ORG_GNOME_MUTTER_DISPLAY_CONFIG_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.InputCapture
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterInputCapture
 * @title: OrgGnomeMutterInputCapture
 * @short_description: Generated C code for the org.gnome.Mutter.InputCapture D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture.top_of_page">org.gnome.Mutter.InputCapture</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gnome.Mutter.InputCapture ---- */

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_method_info_create_session_IN_ARG_capabilities =
{
  {
    -1,
    (gchar *) "capabilities",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_method_info_create_session_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_method_info_create_session_IN_ARG_capabilities.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_method_info_create_session_OUT_ARG_session_path =
{
  {
    -1,
    (gchar *) "session_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_method_info_create_session_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_method_info_create_session_OUT_ARG_session_path.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_method_info_create_session =
{
  {
    -1,
    (gchar *) "CreateSession",
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_method_info_create_session_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_method_info_create_session_OUT_ARG_pointers,
    NULL
  },
  "handle-create-session",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_mutter_input_capture_method_info_pointers[] =
{
  &_org_gnome_mutter_input_capture_method_info_create_session.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_input_capture_property_info_supported_capabilities =
{
  {
    -1,
    (gchar *) "SupportedCapabilities",
    (gchar *) "u",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "supported-capabilities",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_mutter_input_capture_property_info_pointers[] =
{
  &_org_gnome_mutter_input_capture_property_info_supported_capabilities.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_input_capture_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.InputCapture",
    (GDBusMethodInfo **) &_org_gnome_mutter_input_capture_method_info_pointers,
    NULL,
    (GDBusPropertyInfo **) &_org_gnome_mutter_input_capture_property_info_pointers,
    NULL
  },
  "org-gnome-mutter-input-capture",
};


/**
 * org_gnome_mutter_input_capture_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture.top_of_page">org.gnome.Mutter.InputCapture</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_input_capture_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_input_capture_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterInputCapture interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_input_capture_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "supported-capabilities");
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_input_capture_method_marshal_create_session (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterInputCapture:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture.top_of_page">org.gnome.Mutter.InputCapture</link>.
 */

/**
 * OrgGnomeMutterInputCaptureIface:
 * @parent_iface: The parent interface.
 * @handle_create_session: Handler for the #OrgGnomeMutterInputCapture::handle-create-session signal.
 * @get_supported_capabilities: Getter for the #OrgGnomeMutterInputCapture:supported-capabilities property.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture.top_of_page">org.gnome.Mutter.InputCapture</link>.
 */

typedef OrgGnomeMutterInputCaptureIface OrgGnomeMutterInputCaptureInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterInputCapture, org_gnome_mutter_input_capture, G_TYPE_OBJECT)

static void
org_gnome_mutter_input_capture_default_init (OrgGnomeMutterInputCaptureIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeMutterInputCapture::handle-create-session:
   * @object: A #OrgGnomeMutterInputCapture.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_capabilities: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture.CreateSession">CreateSession()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_complete_create_session() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-create-session",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureIface, handle_create_session),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_method_marshal_create_session,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeMutterInputCapture:supported-capabilities:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-InputCapture.SupportedCapabilities">"SupportedCapabilities"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_uint ("supported-capabilities", "SupportedCapabilities", "SupportedCapabilities", 0, G_MAXUINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_mutter_input_capture_get_supported_capabilities: (skip)
 * @object: A #OrgGnomeMutterInputCapture.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-InputCapture.SupportedCapabilities">"SupportedCapabilities"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
guint 
org_gnome_mutter_input_capture_get_supported_capabilities (OrgGnomeMutterInputCapture *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_INPUT_CAPTURE (object), 0);

  return ORG_GNOME_MUTTER_INPUT_CAPTURE_GET_IFACE (object)->get_supported_capabilities (object);
}

/**
 * org_gnome_mutter_input_capture_set_supported_capabilities: (skip)
 * @object: A #OrgGnomeMutterInputCapture.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-InputCapture.SupportedCapabilities">"SupportedCapabilities"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_input_capture_set_supported_capabilities (OrgGnomeMutterInputCapture *object, guint value)
{
  g_object_set (G_OBJECT (object), "supported-capabilities", value, NULL);
}

/**
 * org_gnome_mutter_input_capture_call_create_session:
 * @proxy: A #OrgGnomeMutterInputCaptureProxy.
 * @arg_capabilities: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture.CreateSession">CreateSession()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_call_create_session_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_call_create_session_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_call_create_session (
    OrgGnomeMutterInputCapture *proxy,
    guint arg_capabilities,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "CreateSession",
    g_variant_new ("(u)",
                   arg_capabilities),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_call_create_session_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureProxy.
 * @out_session_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_call_create_session().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_call_create_session().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_call_create_session_finish (
    OrgGnomeMutterInputCapture *proxy,
    gchar ** out_session_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_session_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_call_create_session_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureProxy.
 * @arg_capabilities: Argument to pass with the method invocation.
 * @out_session_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture.CreateSession">CreateSession()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_call_create_session() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_call_create_session_sync (
    OrgGnomeMutterInputCapture *proxy,
    guint arg_capabilities,
    gchar ** out_session_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "CreateSession",
    g_variant_new ("(u)",
                   arg_capabilities),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_session_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_complete_create_session:
 * @object: A #OrgGnomeMutterInputCapture.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @session_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture.CreateSession">CreateSession()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_complete_create_session (
    OrgGnomeMutterInputCapture *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *session_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   session_path));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterInputCaptureProxy:
 *
 * The #OrgGnomeMutterInputCaptureProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterInputCaptureProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterInputCaptureProxy.
 */

struct _OrgGnomeMutterInputCaptureProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_input_capture_proxy_iface_init (OrgGnomeMutterInputCaptureIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureProxy, org_gnome_mutter_input_capture_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterInputCaptureProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE, org_gnome_mutter_input_capture_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureProxy, org_gnome_mutter_input_capture_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE, org_gnome_mutter_input_capture_proxy_iface_init))

#endif
static void
org_gnome_mutter_input_capture_proxy_finalize (GObject *object)
{
  OrgGnomeMutterInputCaptureProxy *proxy = ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_input_capture_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_input_capture_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_input_capture_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_mutter_input_capture_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Mutter.InputCapture: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_mutter_input_capture_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_input_capture_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Mutter.InputCapture", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_mutter_input_capture_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_mutter_input_capture_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_input_capture_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterInputCaptureProxy *proxy = ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static guint 
org_gnome_mutter_input_capture_proxy_get_supported_capabilities (OrgGnomeMutterInputCapture *object)
{
  OrgGnomeMutterInputCaptureProxy *proxy = ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY (object);
  GVariant *variant;
  guint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "SupportedCapabilities");
  if (variant != NULL)
    {
      value = g_variant_get_uint32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_gnome_mutter_input_capture_proxy_init (OrgGnomeMutterInputCaptureProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_input_capture_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY, OrgGnomeMutterInputCaptureProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_input_capture_interface_info ());
}

static void
org_gnome_mutter_input_capture_proxy_class_init (OrgGnomeMutterInputCaptureProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_input_capture_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_input_capture_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_input_capture_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_input_capture_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_input_capture_proxy_g_properties_changed;

  org_gnome_mutter_input_capture_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterInputCaptureProxyPrivate));
#endif
}

static void
org_gnome_mutter_input_capture_proxy_iface_init (OrgGnomeMutterInputCaptureIface *iface)
{
  iface->get_supported_capabilities = org_gnome_mutter_input_capture_proxy_get_supported_capabilities;
}

/**
 * org_gnome_mutter_input_capture_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture.top_of_page">org.gnome.Mutter.InputCapture</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_input_capture_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture", NULL);
}

/**
 * org_gnome_mutter_input_capture_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCapture *
org_gnome_mutter_input_capture_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_input_capture_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture.top_of_page">org.gnome.Mutter.InputCapture</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCapture *
org_gnome_mutter_input_capture_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_input_capture_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_input_capture_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_input_capture_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture", NULL);
}

/**
 * org_gnome_mutter_input_capture_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCapture *
org_gnome_mutter_input_capture_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_input_capture_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_input_capture_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCapture *
org_gnome_mutter_input_capture_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterInputCaptureSkeleton:
 *
 * The #OrgGnomeMutterInputCaptureSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterInputCaptureSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterInputCaptureSkeleton.
 */

struct _OrgGnomeMutterInputCaptureSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_input_capture_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_input_capture_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_input_capture_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_input_capture_skeleton_vtable =
{
  _org_gnome_mutter_input_capture_skeleton_handle_method_call,
  _org_gnome_mutter_input_capture_skeleton_handle_get_property,
  _org_gnome_mutter_input_capture_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_input_capture_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_input_capture_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_input_capture_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_input_capture_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_input_capture_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_input_capture_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_input_capture_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_input_capture_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_input_capture_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.InputCapture", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_mutter_input_capture_emit_changed (gpointer user_data);

static void
org_gnome_mutter_input_capture_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_mutter_input_capture_emit_changed (skeleton);
}

static void org_gnome_mutter_input_capture_skeleton_iface_init (OrgGnomeMutterInputCaptureIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureSkeleton, org_gnome_mutter_input_capture_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterInputCaptureSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE, org_gnome_mutter_input_capture_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureSkeleton, org_gnome_mutter_input_capture_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE, org_gnome_mutter_input_capture_skeleton_iface_init))

#endif
static void
org_gnome_mutter_input_capture_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (object);
  guint n;
  for (n = 0; n < 1; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_input_capture_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_input_capture_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_mutter_input_capture_emit_changed (gpointer user_data)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Mutter.InputCapture",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_mutter_input_capture_schedule_emit_changed (OrgGnomeMutterInputCaptureSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_mutter_input_capture_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_mutter_input_capture_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_mutter_input_capture_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_mutter_input_capture_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_input_capture_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_mutter_input_capture_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_mutter_input_capture_skeleton_init (OrgGnomeMutterInputCaptureSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_input_capture_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON, OrgGnomeMutterInputCaptureSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 1);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_UINT);
}

static guint 
org_gnome_mutter_input_capture_skeleton_get_supported_capabilities (OrgGnomeMutterInputCapture *object)
{
  OrgGnomeMutterInputCaptureSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON (object);
  guint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_uint (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_mutter_input_capture_skeleton_class_init (OrgGnomeMutterInputCaptureSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_input_capture_skeleton_finalize;
  gobject_class->get_property = org_gnome_mutter_input_capture_skeleton_get_property;
  gobject_class->set_property = org_gnome_mutter_input_capture_skeleton_set_property;
  gobject_class->notify       = org_gnome_mutter_input_capture_skeleton_notify;


  org_gnome_mutter_input_capture_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_input_capture_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_input_capture_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_input_capture_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_input_capture_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterInputCaptureSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_input_capture_skeleton_iface_init (OrgGnomeMutterInputCaptureIface *iface)
{
  iface->get_supported_capabilities = org_gnome_mutter_input_capture_skeleton_get_supported_capabilities;
}

/**
 * org_gnome_mutter_input_capture_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture.top_of_page">org.gnome.Mutter.InputCapture</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureSkeleton): The skeleton object.
 */
OrgGnomeMutterInputCapture *
org_gnome_mutter_input_capture_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_INPUT_CAPTURE (g_object_new (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.InputCapture.Session
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterInputCaptureSession
 * @title: OrgGnomeMutterInputCaptureSession
 * @short_description: Generated C code for the org.gnome.Mutter.InputCapture.Session D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture-Session.top_of_page">org.gnome.Mutter.InputCapture.Session</link> D-Bus interface in C.
 */

enum
{
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_ACTIVATED,
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_DEACTIVATED,
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_ZONES_CHANGED,
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_DISABLED,
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_CLOSED,
};

static unsigned _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[5] = { 0 };

/* ---- Introspection data for org.gnome.Mutter.InputCapture.Session ---- */

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_method_info_get_zones_OUT_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_method_info_get_zones_OUT_ARG_zones =
{
  {
    -1,
    (gchar *) "zones",
    (gchar *) "a(uuii)",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_session_method_info_get_zones_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_method_info_get_zones_OUT_ARG_serial.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_get_zones_OUT_ARG_zones.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_get_zones =
{
  {
    -1,
    (gchar *) "GetZones",
    NULL,
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_session_method_info_get_zones_OUT_ARG_pointers,
    NULL
  },
  "handle-get-zones",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_method_info_add_barrier_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_method_info_add_barrier_IN_ARG_position =
{
  {
    -1,
    (gchar *) "position",
    (gchar *) "(iiii)",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_session_method_info_add_barrier_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_method_info_add_barrier_IN_ARG_serial.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_add_barrier_IN_ARG_position.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_method_info_add_barrier_OUT_ARG_id =
{
  {
    -1,
    (gchar *) "id",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_session_method_info_add_barrier_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_method_info_add_barrier_OUT_ARG_id.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_add_barrier =
{
  {
    -1,
    (gchar *) "AddBarrier",
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_session_method_info_add_barrier_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_session_method_info_add_barrier_OUT_ARG_pointers,
    NULL
  },
  "handle-add-barrier",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_clear_barriers =
{
  {
    -1,
    (gchar *) "ClearBarriers",
    NULL,
    NULL,
    NULL
  },
  "handle-clear-barriers",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_enable =
{
  {
    -1,
    (gchar *) "Enable",
    NULL,
    NULL,
    NULL
  },
  "handle-enable",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_disable =
{
  {
    -1,
    (gchar *) "Disable",
    NULL,
    NULL,
    NULL
  },
  "handle-disable",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_method_info_connect_to_eis_OUT_ARG_fd =
{
  {
    -1,
    (gchar *) "fd",
    (gchar *) "h",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_session_method_info_connect_to_eis_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_method_info_connect_to_eis_OUT_ARG_fd.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_connect_to_eis =
{
  {
    -1,
    (gchar *) "ConnectToEIS",
    NULL,
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_session_method_info_connect_to_eis_OUT_ARG_pointers,
    NULL
  },
  "handle-connect-to-eis",
  TRUE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_method_info_release_IN_ARG_options =
{
  {
    -1,
    (gchar *) "options",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_session_method_info_release_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_method_info_release_IN_ARG_options.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_release =
{
  {
    -1,
    (gchar *) "Release",
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_session_method_info_release_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-release",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_input_capture_session_method_info_close =
{
  {
    -1,
    (gchar *) "Close",
    NULL,
    NULL,
    NULL
  },
  "handle-close",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_mutter_input_capture_session_method_info_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_method_info_get_zones.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_add_barrier.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_clear_barriers.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_enable.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_disable.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_connect_to_eis.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_release.parent_struct,
  &_org_gnome_mutter_input_capture_session_method_info_close.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_signal_info_activated_ARG_barrier_id =
{
  {
    -1,
    (gchar *) "barrier_id",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_signal_info_activated_ARG_activation_id =
{
  {
    -1,
    (gchar *) "activation_id",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_signal_info_activated_ARG_cursor_position =
{
  {
    -1,
    (gchar *) "cursor_position",
    (gchar *) "(dd)",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_session_signal_info_activated_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_signal_info_activated_ARG_barrier_id.parent_struct,
  &_org_gnome_mutter_input_capture_session_signal_info_activated_ARG_activation_id.parent_struct,
  &_org_gnome_mutter_input_capture_session_signal_info_activated_ARG_cursor_position.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_input_capture_session_signal_info_activated =
{
  {
    -1,
    (gchar *) "Activated",
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_session_signal_info_activated_ARG_pointers,
    NULL
  },
  "activated"
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_input_capture_session_signal_info_deactivated_ARG_activation_id =
{
  {
    -1,
    (gchar *) "activation_id",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_input_capture_session_signal_info_deactivated_ARG_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_signal_info_deactivated_ARG_activation_id.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_input_capture_session_signal_info_deactivated =
{
  {
    -1,
    (gchar *) "Deactivated",
    (GDBusArgInfo **) &_org_gnome_mutter_input_capture_session_signal_info_deactivated_ARG_pointers,
    NULL
  },
  "deactivated"
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_input_capture_session_signal_info_zones_changed =
{
  {
    -1,
    (gchar *) "ZonesChanged",
    NULL,
    NULL
  },
  "zones-changed"
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_input_capture_session_signal_info_disabled =
{
  {
    -1,
    (gchar *) "Disabled",
    NULL,
    NULL
  },
  "disabled"
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_input_capture_session_signal_info_closed =
{
  {
    -1,
    (gchar *) "Closed",
    NULL,
    NULL
  },
  "closed"
};

static const GDBusSignalInfo * const _org_gnome_mutter_input_capture_session_signal_info_pointers[] =
{
  &_org_gnome_mutter_input_capture_session_signal_info_activated.parent_struct,
  &_org_gnome_mutter_input_capture_session_signal_info_deactivated.parent_struct,
  &_org_gnome_mutter_input_capture_session_signal_info_zones_changed.parent_struct,
  &_org_gnome_mutter_input_capture_session_signal_info_disabled.parent_struct,
  &_org_gnome_mutter_input_capture_session_signal_info_closed.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_input_capture_session_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.InputCapture.Session",
    (GDBusMethodInfo **) &_org_gnome_mutter_input_capture_session_method_info_pointers,
    (GDBusSignalInfo **) &_org_gnome_mutter_input_capture_session_signal_info_pointers,
    NULL,
    NULL
  },
  "org-gnome-mutter-input-capture-session",
};


/**
 * org_gnome_mutter_input_capture_session_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture-Session.top_of_page">org.gnome.Mutter.InputCapture.Session</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_input_capture_session_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_session_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_input_capture_session_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterInputCaptureSession interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_input_capture_session_override_properties (GObjectClass *klass G_GNUC_UNUSED, guint property_id_begin)
{
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_input_capture_session_signal_marshal_activated (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_VOID__UINT_UINT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_signal_marshal_deactivated (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_signal_marshal_zones_changed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_signal_marshal_disabled (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_signal_marshal_closed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_get_zones (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_add_barrier (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_clear_barriers (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_enable (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_disable (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_connect_to_eis (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_release (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_input_capture_session_method_marshal_close (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterInputCaptureSession:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture-Session.top_of_page">org.gnome.Mutter.InputCapture.Session</link>.
 */

/**
 * OrgGnomeMutterInputCaptureSessionIface:
 * @parent_iface: The parent interface.
 * @handle_add_barrier: Handler for the #OrgGnomeMutterInputCaptureSession::handle-add-barrier signal.
 * @handle_clear_barriers: Handler for the #OrgGnomeMutterInputCaptureSession::handle-clear-barriers signal.
 * @handle_close: Handler for the #OrgGnomeMutterInputCaptureSession::handle-close signal.
 * @handle_connect_to_eis: Handler for the #OrgGnomeMutterInputCaptureSession::handle-connect-to-eis signal.
 * @handle_disable: Handler for the #OrgGnomeMutterInputCaptureSession::handle-disable signal.
 * @handle_enable: Handler for the #OrgGnomeMutterInputCaptureSession::handle-enable signal.
 * @handle_get_zones: Handler for the #OrgGnomeMutterInputCaptureSession::handle-get-zones signal.
 * @handle_release: Handler for the #OrgGnomeMutterInputCaptureSession::handle-release signal.
 * @activated: Handler for the #OrgGnomeMutterInputCaptureSession::activated signal.
 * @closed: Handler for the #OrgGnomeMutterInputCaptureSession::closed signal.
 * @deactivated: Handler for the #OrgGnomeMutterInputCaptureSession::deactivated signal.
 * @disabled: Handler for the #OrgGnomeMutterInputCaptureSession::disabled signal.
 * @zones_changed: Handler for the #OrgGnomeMutterInputCaptureSession::zones-changed signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture-Session.top_of_page">org.gnome.Mutter.InputCapture.Session</link>.
 */

typedef OrgGnomeMutterInputCaptureSessionIface OrgGnomeMutterInputCaptureSessionInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterInputCaptureSession, org_gnome_mutter_input_capture_session, G_TYPE_OBJECT)

static void
org_gnome_mutter_input_capture_session_default_init (OrgGnomeMutterInputCaptureSessionIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeMutterInputCaptureSession::handle-get-zones:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.GetZones">GetZones()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_get_zones() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-zones",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_get_zones),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_get_zones,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterInputCaptureSession::handle-add-barrier:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   * @arg_position: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.AddBarrier">AddBarrier()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_add_barrier() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-add-barrier",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_add_barrier),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_add_barrier,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterInputCaptureSession::handle-clear-barriers:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ClearBarriers">ClearBarriers()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_clear_barriers() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-clear-barriers",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_clear_barriers),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_clear_barriers,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterInputCaptureSession::handle-enable:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Enable">Enable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_enable() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-enable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_enable),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_enable,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterInputCaptureSession::handle-disable:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Disable">Disable()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_disable() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-disable",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_disable),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_disable,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterInputCaptureSession::handle-connect-to-eis:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   * @fd_list: (nullable): A #GUnixFDList or %NULL.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_connect_to_eis() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-connect-to-eis",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_connect_to_eis),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_connect_to_eis,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UNIX_FD_LIST);

  /**
   * OrgGnomeMutterInputCaptureSession::handle-release:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_options: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Release">Release()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_release() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-release",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_release),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_release,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterInputCaptureSession::handle-close:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Close">Close()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_input_capture_session_complete_close() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-close",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, handle_close),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_input_capture_session_method_marshal_close,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGnomeMutterInputCaptureSession::activated:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @arg_barrier_id: Argument.
   * @arg_activation_id: Argument.
   * @arg_cursor_position: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Activated">"Activated"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_ACTIVATED] =
    g_signal_new ("activated",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, activated),
      NULL,
      NULL,
      org_gnome_mutter_input_capture_session_signal_marshal_activated,
      G_TYPE_NONE,
      3, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterInputCaptureSession::deactivated:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   * @arg_activation_id: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Deactivated">"Deactivated"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_DEACTIVATED] =
    g_signal_new ("deactivated",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, deactivated),
      NULL,
      NULL,
      org_gnome_mutter_input_capture_session_signal_marshal_deactivated,
      G_TYPE_NONE,
      1, G_TYPE_UINT);

  /**
   * OrgGnomeMutterInputCaptureSession::zones-changed:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.ZonesChanged">"ZonesChanged"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_ZONES_CHANGED] =
    g_signal_new ("zones-changed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, zones_changed),
      NULL,
      NULL,
      org_gnome_mutter_input_capture_session_signal_marshal_zones_changed,
      G_TYPE_NONE,
      0);

  /**
   * OrgGnomeMutterInputCaptureSession::disabled:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Disabled">"Disabled"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_DISABLED] =
    g_signal_new ("disabled",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, disabled),
      NULL,
      NULL,
      org_gnome_mutter_input_capture_session_signal_marshal_disabled,
      G_TYPE_NONE,
      0);

  /**
   * OrgGnomeMutterInputCaptureSession::closed:
   * @object: A #OrgGnomeMutterInputCaptureSession.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Closed">"Closed"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_CLOSED] =
    g_signal_new ("closed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterInputCaptureSessionIface, closed),
      NULL,
      NULL,
      org_gnome_mutter_input_capture_session_signal_marshal_closed,
      G_TYPE_NONE,
      0);

}

/**
 * org_gnome_mutter_input_capture_session_emit_activated:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @arg_barrier_id: Argument to pass with the signal.
 * @arg_activation_id: Argument to pass with the signal.
 * @arg_cursor_position: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Activated">"Activated"</link> D-Bus signal.
 */
void
org_gnome_mutter_input_capture_session_emit_activated (
    OrgGnomeMutterInputCaptureSession *object,
    guint arg_barrier_id,
    guint arg_activation_id,
    GVariant *arg_cursor_position)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_ACTIVATED], 0, arg_barrier_id, arg_activation_id, arg_cursor_position);
}

/**
 * org_gnome_mutter_input_capture_session_emit_deactivated:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @arg_activation_id: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Deactivated">"Deactivated"</link> D-Bus signal.
 */
void
org_gnome_mutter_input_capture_session_emit_deactivated (
    OrgGnomeMutterInputCaptureSession *object,
    guint arg_activation_id)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_DEACTIVATED], 0, arg_activation_id);
}

/**
 * org_gnome_mutter_input_capture_session_emit_zones_changed:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.ZonesChanged">"ZonesChanged"</link> D-Bus signal.
 */
void
org_gnome_mutter_input_capture_session_emit_zones_changed (
    OrgGnomeMutterInputCaptureSession *object)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_ZONES_CHANGED], 0);
}

/**
 * org_gnome_mutter_input_capture_session_emit_disabled:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Disabled">"Disabled"</link> D-Bus signal.
 */
void
org_gnome_mutter_input_capture_session_emit_disabled (
    OrgGnomeMutterInputCaptureSession *object)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_DISABLED], 0);
}

/**
 * org_gnome_mutter_input_capture_session_emit_closed:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-InputCapture-Session.Closed">"Closed"</link> D-Bus signal.
 */
void
org_gnome_mutter_input_capture_session_emit_closed (
    OrgGnomeMutterInputCaptureSession *object)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SIGNALS[_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_CLOSED], 0);
}

/**
 * org_gnome_mutter_input_capture_session_call_get_zones:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.GetZones">GetZones()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_get_zones_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_get_zones_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_get_zones (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetZones",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_get_zones_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @out_serial: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_zones: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_get_zones().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_get_zones().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_get_zones_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    guint* out_serial,
    GVariant ** out_zones,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u@a(uuii))",
                 out_serial,
                 out_zones);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_get_zones_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @out_serial: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_zones: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.GetZones">GetZones()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_get_zones() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_get_zones_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    guint* out_serial,
    GVariant ** out_zones,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetZones",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u@a(uuii))",
                 out_serial,
                 out_zones);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_add_barrier:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_position: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.AddBarrier">AddBarrier()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_add_barrier_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_add_barrier_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_add_barrier (
    OrgGnomeMutterInputCaptureSession *proxy,
    guint arg_serial,
    GVariant *arg_position,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "AddBarrier",
    g_variant_new ("(u@(iiii))",
                   arg_serial,
                   arg_position),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_add_barrier_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @out_id: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_add_barrier().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_add_barrier().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_add_barrier_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    guint* out_id,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u)",
                 out_id);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_add_barrier_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_position: Argument to pass with the method invocation.
 * @out_id: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.AddBarrier">AddBarrier()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_add_barrier() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_add_barrier_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    guint arg_serial,
    GVariant *arg_position,
    guint* out_id,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "AddBarrier",
    g_variant_new ("(u@(iiii))",
                   arg_serial,
                   arg_position),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u)",
                 out_id);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_clear_barriers:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ClearBarriers">ClearBarriers()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_clear_barriers_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_clear_barriers_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_clear_barriers (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ClearBarriers",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_clear_barriers_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_clear_barriers().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_clear_barriers().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_clear_barriers_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_clear_barriers_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ClearBarriers">ClearBarriers()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_clear_barriers() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_clear_barriers_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ClearBarriers",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_enable:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Enable">Enable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_enable_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_enable_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_enable (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Enable",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_enable_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_enable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_enable().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_enable_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_enable_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Enable">Enable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_enable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_enable_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Enable",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_disable:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Disable">Disable()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_disable_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_disable_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_disable (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Disable",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_disable_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_disable().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_disable().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_disable_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_disable_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Disable">Disable()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_disable() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_disable_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Disable",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_connect_to_eis:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_connect_to_eis_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_connect_to_eis_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_connect_to_eis (
    OrgGnomeMutterInputCaptureSession *proxy,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call_with_unix_fd_list (G_DBUS_PROXY (proxy),
    "ConnectToEIS",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_connect_to_eis_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_connect_to_eis().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_connect_to_eis().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_connect_to_eis_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_finish (G_DBUS_PROXY (proxy), out_fd_list, res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_connect_to_eis_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_connect_to_eis() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_connect_to_eis_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    GUnixFDList  *fd_list,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_sync (G_DBUS_PROXY (proxy),
    "ConnectToEIS",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    out_fd_list,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_release:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Release">Release()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_release_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_release_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_release (
    OrgGnomeMutterInputCaptureSession *proxy,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Release",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_release_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_release().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_release().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_release_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_release_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Release">Release()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_release() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_release_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    GVariant *arg_options,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Release",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_close:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Close">Close()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_call_close_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_call_close_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_input_capture_session_call_close (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Close",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_input_capture_session_call_close_finish:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_call_close().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_call_close().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_close_finish (
    OrgGnomeMutterInputCaptureSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_call_close_sync:
 * @proxy: A #OrgGnomeMutterInputCaptureSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Close">Close()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_call_close() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_input_capture_session_call_close_sync (
    OrgGnomeMutterInputCaptureSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Close",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_input_capture_session_complete_get_zones:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @serial: Parameter to return.
 * @zones: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.GetZones">GetZones()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_get_zones (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    guint serial,
    GVariant *zones)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(u@a(uuii))",
                   serial,
                   zones));
}

/**
 * org_gnome_mutter_input_capture_session_complete_add_barrier:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @id: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.AddBarrier">AddBarrier()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_add_barrier (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    guint id)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(u)",
                   id));
}

/**
 * org_gnome_mutter_input_capture_session_complete_clear_barriers:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ClearBarriers">ClearBarriers()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_clear_barriers (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_input_capture_session_complete_enable:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Enable">Enable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_enable (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_input_capture_session_complete_disable:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Disable">Disable()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_disable (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_input_capture_session_complete_connect_to_eis:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @fd: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_connect_to_eis (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd)
{
  g_dbus_method_invocation_return_value_with_unix_fd_list (invocation,
    g_variant_new ("(@h)",
                   fd),
    fd_list);
}

/**
 * org_gnome_mutter_input_capture_session_complete_release:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Release">Release()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_release (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_input_capture_session_complete_close:
 * @object: A #OrgGnomeMutterInputCaptureSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-InputCapture-Session.Close">Close()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_input_capture_session_complete_close (
    OrgGnomeMutterInputCaptureSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterInputCaptureSessionProxy:
 *
 * The #OrgGnomeMutterInputCaptureSessionProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterInputCaptureSessionProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterInputCaptureSessionProxy.
 */

struct _OrgGnomeMutterInputCaptureSessionProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_input_capture_session_proxy_iface_init (OrgGnomeMutterInputCaptureSessionIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureSessionProxy, org_gnome_mutter_input_capture_session_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterInputCaptureSessionProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION, org_gnome_mutter_input_capture_session_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureSessionProxy, org_gnome_mutter_input_capture_session_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION, org_gnome_mutter_input_capture_session_proxy_iface_init))

#endif
static void
org_gnome_mutter_input_capture_session_proxy_finalize (GObject *object)
{
  OrgGnomeMutterInputCaptureSessionProxy *proxy = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_input_capture_session_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_input_capture_session_proxy_get_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  GValue       *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_mutter_input_capture_session_proxy_set_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  const GValue *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_mutter_input_capture_session_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_session_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_input_capture_session_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterInputCaptureSessionProxy *proxy = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_session_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_session_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
org_gnome_mutter_input_capture_session_proxy_init (OrgGnomeMutterInputCaptureSessionProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_input_capture_session_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_PROXY, OrgGnomeMutterInputCaptureSessionProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_input_capture_session_interface_info ());
}

static void
org_gnome_mutter_input_capture_session_proxy_class_init (OrgGnomeMutterInputCaptureSessionProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_input_capture_session_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_input_capture_session_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_input_capture_session_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_input_capture_session_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_input_capture_session_proxy_g_properties_changed;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterInputCaptureSessionProxyPrivate));
#endif
}

static void
org_gnome_mutter_input_capture_session_proxy_iface_init (OrgGnomeMutterInputCaptureSessionIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_mutter_input_capture_session_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture-Session.top_of_page">org.gnome.Mutter.InputCapture.Session</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_input_capture_session_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture.Session", NULL);
}

/**
 * org_gnome_mutter_input_capture_session_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCaptureSession *
org_gnome_mutter_input_capture_session_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_input_capture_session_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture-Session.top_of_page">org.gnome.Mutter.InputCapture.Session</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCaptureSession *
org_gnome_mutter_input_capture_session_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture.Session", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_input_capture_session_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_input_capture_session_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_input_capture_session_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_input_capture_session_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_input_capture_session_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture.Session", NULL);
}

/**
 * org_gnome_mutter_input_capture_session_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_input_capture_session_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_input_capture_session_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCaptureSession *
org_gnome_mutter_input_capture_session_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_input_capture_session_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_input_capture_session_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_input_capture_session_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterInputCaptureSession *
org_gnome_mutter_input_capture_session_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.InputCapture.Session", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterInputCaptureSessionSkeleton:
 *
 * The #OrgGnomeMutterInputCaptureSessionSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterInputCaptureSessionSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterInputCaptureSessionSkeleton.
 */

struct _OrgGnomeMutterInputCaptureSessionSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_input_capture_session_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_input_capture_session_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_session_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_input_capture_session_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_input_capture_session_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_input_capture_session_skeleton_vtable =
{
  _org_gnome_mutter_input_capture_session_skeleton_handle_method_call,
  _org_gnome_mutter_input_capture_session_skeleton_handle_get_property,
  _org_gnome_mutter_input_capture_session_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_input_capture_session_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_input_capture_session_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_input_capture_session_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_input_capture_session_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_input_capture_session_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_input_capture_session_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_input_capture_session_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_input_capture_session_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_input_capture_session_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.InputCapture.Session", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
org_gnome_mutter_input_capture_session_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton G_GNUC_UNUSED)
{
}

static void
_org_gnome_mutter_input_capture_session_on_signal_activated (
    OrgGnomeMutterInputCaptureSession *object,
    guint arg_barrier_id,
    guint arg_activation_id,
    GVariant *arg_cursor_position)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(uu@(dd))",
                   arg_barrier_id,
                   arg_activation_id,
                   arg_cursor_position));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.InputCapture.Session", "Activated",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_mutter_input_capture_session_on_signal_deactivated (
    OrgGnomeMutterInputCaptureSession *object,
    guint arg_activation_id)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(u)",
                   arg_activation_id));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.InputCapture.Session", "Deactivated",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_mutter_input_capture_session_on_signal_zones_changed (
    OrgGnomeMutterInputCaptureSession *object)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.InputCapture.Session", "ZonesChanged",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_mutter_input_capture_session_on_signal_disabled (
    OrgGnomeMutterInputCaptureSession *object)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.InputCapture.Session", "Disabled",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_mutter_input_capture_session_on_signal_closed (
    OrgGnomeMutterInputCaptureSession *object)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.InputCapture.Session", "Closed",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gnome_mutter_input_capture_session_skeleton_iface_init (OrgGnomeMutterInputCaptureSessionIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureSessionSkeleton, org_gnome_mutter_input_capture_session_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterInputCaptureSessionSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION, org_gnome_mutter_input_capture_session_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterInputCaptureSessionSkeleton, org_gnome_mutter_input_capture_session_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION, org_gnome_mutter_input_capture_session_skeleton_iface_init))

#endif
static void
org_gnome_mutter_input_capture_session_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterInputCaptureSessionSkeleton *skeleton = ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_input_capture_session_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_input_capture_session_skeleton_init (OrgGnomeMutterInputCaptureSessionSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_input_capture_session_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON, OrgGnomeMutterInputCaptureSessionSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
org_gnome_mutter_input_capture_session_skeleton_class_init (OrgGnomeMutterInputCaptureSessionSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_input_capture_session_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_input_capture_session_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_input_capture_session_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_input_capture_session_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_input_capture_session_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterInputCaptureSessionSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_input_capture_session_skeleton_iface_init (OrgGnomeMutterInputCaptureSessionIface *iface)
{
  iface->activated = _org_gnome_mutter_input_capture_session_on_signal_activated;
  iface->deactivated = _org_gnome_mutter_input_capture_session_on_signal_deactivated;
  iface->zones_changed = _org_gnome_mutter_input_capture_session_on_signal_zones_changed;
  iface->disabled = _org_gnome_mutter_input_capture_session_on_signal_disabled;
  iface->closed = _org_gnome_mutter_input_capture_session_on_signal_closed;
}

/**
 * org_gnome_mutter_input_capture_session_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-InputCapture-Session.top_of_page">org.gnome.Mutter.InputCapture.Session</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterInputCaptureSessionSkeleton): The skeleton object.
 */
OrgGnomeMutterInputCaptureSession *
org_gnome_mutter_input_capture_session_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION (g_object_new (TYPE_ORG_GNOME_MUTTER_INPUT_CAPTURE_SESSION_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.RemoteDesktop
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterRemoteDesktop
 * @title: OrgGnomeMutterRemoteDesktop
 * @short_description: Generated C code for the org.gnome.Mutter.RemoteDesktop D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop.top_of_page">org.gnome.Mutter.RemoteDesktop</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gnome.Mutter.RemoteDesktop ---- */

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_method_info_create_session_OUT_ARG_session_path =
{
  {
    -1,
    (gchar *) "session_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_method_info_create_session_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_method_info_create_session_OUT_ARG_session_path.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_method_info_create_session =
{
  {
    -1,
    (gchar *) "CreateSession",
    NULL,
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_method_info_create_session_OUT_ARG_pointers,
    NULL
  },
  "handle-create-session",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_mutter_remote_desktop_method_info_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_method_info_create_session.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_remote_desktop_property_info_supported_device_types =
{
  {
    -1,
    (gchar *) "SupportedDeviceTypes",
    (gchar *) "u",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "supported-device-types",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_remote_desktop_property_info_version =
{
  {
    -1,
    (gchar *) "Version",
    (gchar *) "i",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "version",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_mutter_remote_desktop_property_info_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_property_info_supported_device_types.parent_struct,
  &_org_gnome_mutter_remote_desktop_property_info_version.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_remote_desktop_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.RemoteDesktop",
    (GDBusMethodInfo **) &_org_gnome_mutter_remote_desktop_method_info_pointers,
    NULL,
    (GDBusPropertyInfo **) &_org_gnome_mutter_remote_desktop_property_info_pointers,
    NULL
  },
  "org-gnome-mutter-remote-desktop",
};


/**
 * org_gnome_mutter_remote_desktop_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop.top_of_page">org.gnome.Mutter.RemoteDesktop</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_remote_desktop_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_remote_desktop_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterRemoteDesktop interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_remote_desktop_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "supported-device-types");
  g_object_class_override_property (klass, property_id_begin++, "version");
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_remote_desktop_method_marshal_create_session (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterRemoteDesktop:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop.top_of_page">org.gnome.Mutter.RemoteDesktop</link>.
 */

/**
 * OrgGnomeMutterRemoteDesktopIface:
 * @parent_iface: The parent interface.
 * @handle_create_session: Handler for the #OrgGnomeMutterRemoteDesktop::handle-create-session signal.
 * @get_supported_device_types: Getter for the #OrgGnomeMutterRemoteDesktop:supported-device-types property.
 * @get_version: Getter for the #OrgGnomeMutterRemoteDesktop:version property.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop.top_of_page">org.gnome.Mutter.RemoteDesktop</link>.
 */

typedef OrgGnomeMutterRemoteDesktopIface OrgGnomeMutterRemoteDesktopInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterRemoteDesktop, org_gnome_mutter_remote_desktop, G_TYPE_OBJECT)

static void
org_gnome_mutter_remote_desktop_default_init (OrgGnomeMutterRemoteDesktopIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeMutterRemoteDesktop::handle-create-session:
   * @object: A #OrgGnomeMutterRemoteDesktop.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop.CreateSession">CreateSession()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_complete_create_session() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-create-session",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopIface, handle_create_session),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_method_marshal_create_session,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeMutterRemoteDesktop:supported-device-types:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop.SupportedDeviceTypes">"SupportedDeviceTypes"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_uint ("supported-device-types", "SupportedDeviceTypes", "SupportedDeviceTypes", 0, G_MAXUINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgGnomeMutterRemoteDesktop:version:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop.Version">"Version"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int ("version", "Version", "Version", G_MININT32, G_MAXINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_mutter_remote_desktop_get_supported_device_types: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktop.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop.SupportedDeviceTypes">"SupportedDeviceTypes"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
guint 
org_gnome_mutter_remote_desktop_get_supported_device_types (OrgGnomeMutterRemoteDesktop *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_REMOTE_DESKTOP (object), 0);

  return ORG_GNOME_MUTTER_REMOTE_DESKTOP_GET_IFACE (object)->get_supported_device_types (object);
}

/**
 * org_gnome_mutter_remote_desktop_set_supported_device_types: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktop.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop.SupportedDeviceTypes">"SupportedDeviceTypes"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_remote_desktop_set_supported_device_types (OrgGnomeMutterRemoteDesktop *object, guint value)
{
  g_object_set (G_OBJECT (object), "supported-device-types", value, NULL);
}

/**
 * org_gnome_mutter_remote_desktop_get_version: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktop.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop.Version">"Version"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint 
org_gnome_mutter_remote_desktop_get_version (OrgGnomeMutterRemoteDesktop *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_REMOTE_DESKTOP (object), 0);

  return ORG_GNOME_MUTTER_REMOTE_DESKTOP_GET_IFACE (object)->get_version (object);
}

/**
 * org_gnome_mutter_remote_desktop_set_version: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktop.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop.Version">"Version"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_remote_desktop_set_version (OrgGnomeMutterRemoteDesktop *object, gint value)
{
  g_object_set (G_OBJECT (object), "version", value, NULL);
}

/**
 * org_gnome_mutter_remote_desktop_call_create_session:
 * @proxy: A #OrgGnomeMutterRemoteDesktopProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop.CreateSession">CreateSession()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_call_create_session_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_call_create_session_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_call_create_session (
    OrgGnomeMutterRemoteDesktop *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "CreateSession",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_call_create_session_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopProxy.
 * @out_session_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_call_create_session().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_call_create_session().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_call_create_session_finish (
    OrgGnomeMutterRemoteDesktop *proxy,
    gchar ** out_session_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_session_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_call_create_session_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopProxy.
 * @out_session_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop.CreateSession">CreateSession()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_call_create_session() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_call_create_session_sync (
    OrgGnomeMutterRemoteDesktop *proxy,
    gchar ** out_session_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "CreateSession",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_session_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_complete_create_session:
 * @object: A #OrgGnomeMutterRemoteDesktop.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @session_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop.CreateSession">CreateSession()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_complete_create_session (
    OrgGnomeMutterRemoteDesktop *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *session_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   session_path));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterRemoteDesktopProxy:
 *
 * The #OrgGnomeMutterRemoteDesktopProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterRemoteDesktopProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterRemoteDesktopProxy.
 */

struct _OrgGnomeMutterRemoteDesktopProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_remote_desktop_proxy_iface_init (OrgGnomeMutterRemoteDesktopIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopProxy, org_gnome_mutter_remote_desktop_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterRemoteDesktopProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP, org_gnome_mutter_remote_desktop_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopProxy, org_gnome_mutter_remote_desktop_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP, org_gnome_mutter_remote_desktop_proxy_iface_init))

#endif
static void
org_gnome_mutter_remote_desktop_proxy_finalize (GObject *object)
{
  OrgGnomeMutterRemoteDesktopProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_remote_desktop_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_remote_desktop_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 2);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_remote_desktop_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_mutter_remote_desktop_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Mutter.RemoteDesktop: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_mutter_remote_desktop_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 2);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_remote_desktop_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Mutter.RemoteDesktop", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_mutter_remote_desktop_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_mutter_remote_desktop_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_remote_desktop_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterRemoteDesktopProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static guint 
org_gnome_mutter_remote_desktop_proxy_get_supported_device_types (OrgGnomeMutterRemoteDesktop *object)
{
  OrgGnomeMutterRemoteDesktopProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY (object);
  GVariant *variant;
  guint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "SupportedDeviceTypes");
  if (variant != NULL)
    {
      value = g_variant_get_uint32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gint 
org_gnome_mutter_remote_desktop_proxy_get_version (OrgGnomeMutterRemoteDesktop *object)
{
  OrgGnomeMutterRemoteDesktopProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY (object);
  GVariant *variant;
  gint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Version");
  if (variant != NULL)
    {
      value = g_variant_get_int32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_gnome_mutter_remote_desktop_proxy_init (OrgGnomeMutterRemoteDesktopProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_remote_desktop_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY, OrgGnomeMutterRemoteDesktopProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_remote_desktop_interface_info ());
}

static void
org_gnome_mutter_remote_desktop_proxy_class_init (OrgGnomeMutterRemoteDesktopProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_remote_desktop_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_remote_desktop_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_remote_desktop_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_remote_desktop_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_remote_desktop_proxy_g_properties_changed;

  org_gnome_mutter_remote_desktop_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterRemoteDesktopProxyPrivate));
#endif
}

static void
org_gnome_mutter_remote_desktop_proxy_iface_init (OrgGnomeMutterRemoteDesktopIface *iface)
{
  iface->get_supported_device_types = org_gnome_mutter_remote_desktop_proxy_get_supported_device_types;
  iface->get_version = org_gnome_mutter_remote_desktop_proxy_get_version;
}

/**
 * org_gnome_mutter_remote_desktop_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop.top_of_page">org.gnome.Mutter.RemoteDesktop</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_remote_desktop_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop", NULL);
}

/**
 * org_gnome_mutter_remote_desktop_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktop *
org_gnome_mutter_remote_desktop_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_remote_desktop_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop.top_of_page">org.gnome.Mutter.RemoteDesktop</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktop *
org_gnome_mutter_remote_desktop_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_remote_desktop_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_remote_desktop_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_remote_desktop_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop", NULL);
}

/**
 * org_gnome_mutter_remote_desktop_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktop *
org_gnome_mutter_remote_desktop_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_remote_desktop_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_remote_desktop_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktop *
org_gnome_mutter_remote_desktop_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterRemoteDesktopSkeleton:
 *
 * The #OrgGnomeMutterRemoteDesktopSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterRemoteDesktopSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterRemoteDesktopSkeleton.
 */

struct _OrgGnomeMutterRemoteDesktopSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_remote_desktop_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_remote_desktop_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_remote_desktop_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_remote_desktop_skeleton_vtable =
{
  _org_gnome_mutter_remote_desktop_skeleton_handle_method_call,
  _org_gnome_mutter_remote_desktop_skeleton_handle_get_property,
  _org_gnome_mutter_remote_desktop_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_remote_desktop_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_remote_desktop_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_remote_desktop_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_remote_desktop_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_remote_desktop_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_remote_desktop_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_remote_desktop_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_remote_desktop_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_remote_desktop_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.RemoteDesktop", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_mutter_remote_desktop_emit_changed (gpointer user_data);

static void
org_gnome_mutter_remote_desktop_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_mutter_remote_desktop_emit_changed (skeleton);
}

static void org_gnome_mutter_remote_desktop_skeleton_iface_init (OrgGnomeMutterRemoteDesktopIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopSkeleton, org_gnome_mutter_remote_desktop_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterRemoteDesktopSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP, org_gnome_mutter_remote_desktop_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopSkeleton, org_gnome_mutter_remote_desktop_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP, org_gnome_mutter_remote_desktop_skeleton_iface_init))

#endif
static void
org_gnome_mutter_remote_desktop_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (object);
  guint n;
  for (n = 0; n < 2; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_remote_desktop_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_remote_desktop_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 2);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_mutter_remote_desktop_emit_changed (gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Mutter.RemoteDesktop",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_mutter_remote_desktop_schedule_emit_changed (OrgGnomeMutterRemoteDesktopSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_mutter_remote_desktop_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_mutter_remote_desktop_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_mutter_remote_desktop_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_mutter_remote_desktop_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 2);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_remote_desktop_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_mutter_remote_desktop_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_mutter_remote_desktop_skeleton_init (OrgGnomeMutterRemoteDesktopSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_remote_desktop_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON, OrgGnomeMutterRemoteDesktopSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 2);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_UINT);
  g_value_init (&skeleton->priv->properties[1], G_TYPE_INT);
}

static guint 
org_gnome_mutter_remote_desktop_skeleton_get_supported_device_types (OrgGnomeMutterRemoteDesktop *object)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (object);
  guint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_uint (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gint 
org_gnome_mutter_remote_desktop_skeleton_get_version (OrgGnomeMutterRemoteDesktop *object)
{
  OrgGnomeMutterRemoteDesktopSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON (object);
  gint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int (&(skeleton->priv->properties[1]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_mutter_remote_desktop_skeleton_class_init (OrgGnomeMutterRemoteDesktopSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_remote_desktop_skeleton_finalize;
  gobject_class->get_property = org_gnome_mutter_remote_desktop_skeleton_get_property;
  gobject_class->set_property = org_gnome_mutter_remote_desktop_skeleton_set_property;
  gobject_class->notify       = org_gnome_mutter_remote_desktop_skeleton_notify;


  org_gnome_mutter_remote_desktop_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_remote_desktop_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_remote_desktop_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_remote_desktop_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_remote_desktop_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterRemoteDesktopSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_remote_desktop_skeleton_iface_init (OrgGnomeMutterRemoteDesktopIface *iface)
{
  iface->get_supported_device_types = org_gnome_mutter_remote_desktop_skeleton_get_supported_device_types;
  iface->get_version = org_gnome_mutter_remote_desktop_skeleton_get_version;
}

/**
 * org_gnome_mutter_remote_desktop_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop.top_of_page">org.gnome.Mutter.RemoteDesktop</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopSkeleton): The skeleton object.
 */
OrgGnomeMutterRemoteDesktop *
org_gnome_mutter_remote_desktop_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_REMOTE_DESKTOP (g_object_new (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.RemoteDesktop.Session
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterRemoteDesktopSession
 * @title: OrgGnomeMutterRemoteDesktopSession
 * @short_description: Generated C code for the org.gnome.Mutter.RemoteDesktop.Session D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop-Session.top_of_page">org.gnome.Mutter.RemoteDesktop.Session</link> D-Bus interface in C.
 */

enum
{
  _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_CLOSED,
  _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SELECTION_OWNER_CHANGED,
  _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SELECTION_TRANSFER,
};

static unsigned _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SIGNALS[3] = { 0 };

/* ---- Introspection data for org.gnome.Mutter.RemoteDesktop.Session ---- */

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_start =
{
  {
    -1,
    (gchar *) "Start",
    NULL,
    NULL,
    NULL
  },
  "handle-start",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_stop =
{
  {
    -1,
    (gchar *) "Stop",
    NULL,
    NULL,
    NULL
  },
  "handle-stop",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode_IN_ARG_keycode =
{
  {
    -1,
    (gchar *) "keycode",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode_IN_ARG_state =
{
  {
    -1,
    (gchar *) "state",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode_IN_ARG_keycode.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode_IN_ARG_state.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode =
{
  {
    -1,
    (gchar *) "NotifyKeyboardKeycode",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-keyboard-keycode",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym_IN_ARG_keysym =
{
  {
    -1,
    (gchar *) "keysym",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym_IN_ARG_state =
{
  {
    -1,
    (gchar *) "state",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym_IN_ARG_keysym.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym_IN_ARG_state.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym =
{
  {
    -1,
    (gchar *) "NotifyKeyboardKeysym",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-keyboard-keysym",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button_IN_ARG_button =
{
  {
    -1,
    (gchar *) "button",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button_IN_ARG_state =
{
  {
    -1,
    (gchar *) "state",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button_IN_ARG_button.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button_IN_ARG_state.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button =
{
  {
    -1,
    (gchar *) "NotifyPointerButton",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-pointer-button",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_dx =
{
  {
    -1,
    (gchar *) "dx",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_dy =
{
  {
    -1,
    (gchar *) "dy",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_flags =
{
  {
    -1,
    (gchar *) "flags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_dx.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_dy.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_flags.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis =
{
  {
    -1,
    (gchar *) "NotifyPointerAxis",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-pointer-axis",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete_IN_ARG_axis =
{
  {
    -1,
    (gchar *) "axis",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete_IN_ARG_steps =
{
  {
    -1,
    (gchar *) "steps",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete_IN_ARG_axis.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete_IN_ARG_steps.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete =
{
  {
    -1,
    (gchar *) "NotifyPointerAxisDiscrete",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-pointer-axis-discrete",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative_IN_ARG_dx =
{
  {
    -1,
    (gchar *) "dx",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative_IN_ARG_dy =
{
  {
    -1,
    (gchar *) "dy",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative_IN_ARG_dx.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative_IN_ARG_dy.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative =
{
  {
    -1,
    (gchar *) "NotifyPointerMotionRelative",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-pointer-motion-relative",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_stream =
{
  {
    -1,
    (gchar *) "stream",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_x =
{
  {
    -1,
    (gchar *) "x",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_y =
{
  {
    -1,
    (gchar *) "y",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_stream.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_x.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_y.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute =
{
  {
    -1,
    (gchar *) "NotifyPointerMotionAbsolute",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-pointer-motion-absolute",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_stream =
{
  {
    -1,
    (gchar *) "stream",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_slot =
{
  {
    -1,
    (gchar *) "slot",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_x =
{
  {
    -1,
    (gchar *) "x",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_y =
{
  {
    -1,
    (gchar *) "y",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_stream.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_slot.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_x.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_y.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down =
{
  {
    -1,
    (gchar *) "NotifyTouchDown",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-touch-down",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_stream =
{
  {
    -1,
    (gchar *) "stream",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_slot =
{
  {
    -1,
    (gchar *) "slot",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_x =
{
  {
    -1,
    (gchar *) "x",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_y =
{
  {
    -1,
    (gchar *) "y",
    (gchar *) "d",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_stream.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_slot.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_x.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_y.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion =
{
  {
    -1,
    (gchar *) "NotifyTouchMotion",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-touch-motion",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_up_IN_ARG_slot =
{
  {
    -1,
    (gchar *) "slot",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_up_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_up_IN_ARG_slot.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_notify_touch_up =
{
  {
    -1,
    (gchar *) "NotifyTouchUp",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_up_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-notify-touch-up",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_enable_clipboard_IN_ARG_options =
{
  {
    -1,
    (gchar *) "options",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_enable_clipboard_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_enable_clipboard_IN_ARG_options.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_enable_clipboard =
{
  {
    -1,
    (gchar *) "EnableClipboard",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_enable_clipboard_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-enable-clipboard",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_disable_clipboard =
{
  {
    -1,
    (gchar *) "DisableClipboard",
    NULL,
    NULL,
    NULL
  },
  "handle-disable-clipboard",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_set_selection_IN_ARG_options =
{
  {
    -1,
    (gchar *) "options",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_set_selection_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_set_selection_IN_ARG_options.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_set_selection =
{
  {
    -1,
    (gchar *) "SetSelection",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_set_selection_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-set-selection",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_write_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_selection_write_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_IN_ARG_serial.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_write_OUT_ARG_fd =
{
  {
    -1,
    (gchar *) "fd",
    (gchar *) "h",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_selection_write_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_OUT_ARG_fd.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_write =
{
  {
    -1,
    (gchar *) "SelectionWrite",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_OUT_ARG_pointers,
    NULL
  },
  "handle-selection-write",
  TRUE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_write_done_IN_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_write_done_IN_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_selection_write_done_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_done_IN_ARG_serial.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_done_IN_ARG_success.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_write_done =
{
  {
    -1,
    (gchar *) "SelectionWriteDone",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_done_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-selection-write-done",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_read_IN_ARG_mime_type =
{
  {
    -1,
    (gchar *) "mime_type",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_selection_read_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_read_IN_ARG_mime_type.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_read_OUT_ARG_fd =
{
  {
    -1,
    (gchar *) "fd",
    (gchar *) "h",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_selection_read_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_read_OUT_ARG_fd.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_selection_read =
{
  {
    -1,
    (gchar *) "SelectionRead",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_selection_read_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_selection_read_OUT_ARG_pointers,
    NULL
  },
  "handle-selection-read",
  TRUE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_IN_ARG_options =
{
  {
    -1,
    (gchar *) "options",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_IN_ARG_options.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_OUT_ARG_fd =
{
  {
    -1,
    (gchar *) "fd",
    (gchar *) "h",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_OUT_ARG_fd.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis =
{
  {
    -1,
    (gchar *) "ConnectToEIS",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis_OUT_ARG_pointers,
    NULL
  },
  "handle-connect-to-eis",
  TRUE
};

static const GDBusMethodInfo * const _org_gnome_mutter_remote_desktop_session_method_info_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_method_info_start.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_stop.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keycode.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_keyboard_keysym.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_button.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_axis_discrete.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_relative.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_pointer_motion_absolute.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_down.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_motion.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_notify_touch_up.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_enable_clipboard.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_disable_clipboard.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_set_selection.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_write.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_write_done.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_selection_read.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_method_info_connect_to_eis.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_remote_desktop_session_signal_info_closed =
{
  {
    -1,
    (gchar *) "Closed",
    NULL,
    NULL
  },
  "closed"
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_signal_info_selection_owner_changed_ARG_options =
{
  {
    -1,
    (gchar *) "options",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_signal_info_selection_owner_changed_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_signal_info_selection_owner_changed_ARG_options.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_remote_desktop_session_signal_info_selection_owner_changed =
{
  {
    -1,
    (gchar *) "SelectionOwnerChanged",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_signal_info_selection_owner_changed_ARG_pointers,
    NULL
  },
  "selection-owner-changed"
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer_ARG_mime_type =
{
  {
    -1,
    (gchar *) "mime_type",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer_ARG_serial =
{
  {
    -1,
    (gchar *) "serial",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer_ARG_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer_ARG_mime_type.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer_ARG_serial.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer =
{
  {
    -1,
    (gchar *) "SelectionTransfer",
    (GDBusArgInfo **) &_org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer_ARG_pointers,
    NULL
  },
  "selection-transfer"
};

static const GDBusSignalInfo * const _org_gnome_mutter_remote_desktop_session_signal_info_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_signal_info_closed.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_signal_info_selection_owner_changed.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_signal_info_selection_transfer.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_remote_desktop_session_property_info_session_id =
{
  {
    -1,
    (gchar *) "SessionId",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "session-id",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_remote_desktop_session_property_info_caps_lock_state =
{
  {
    -1,
    (gchar *) "CapsLockState",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "caps-lock-state",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_remote_desktop_session_property_info_num_lock_state =
{
  {
    -1,
    (gchar *) "NumLockState",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "num-lock-state",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_mutter_remote_desktop_session_property_info_pointers[] =
{
  &_org_gnome_mutter_remote_desktop_session_property_info_session_id.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_property_info_caps_lock_state.parent_struct,
  &_org_gnome_mutter_remote_desktop_session_property_info_num_lock_state.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_remote_desktop_session_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.RemoteDesktop.Session",
    (GDBusMethodInfo **) &_org_gnome_mutter_remote_desktop_session_method_info_pointers,
    (GDBusSignalInfo **) &_org_gnome_mutter_remote_desktop_session_signal_info_pointers,
    (GDBusPropertyInfo **) &_org_gnome_mutter_remote_desktop_session_property_info_pointers,
    NULL
  },
  "org-gnome-mutter-remote-desktop-session",
};


/**
 * org_gnome_mutter_remote_desktop_session_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop-Session.top_of_page">org.gnome.Mutter.RemoteDesktop.Session</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_remote_desktop_session_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_session_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_remote_desktop_session_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterRemoteDesktopSession interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_remote_desktop_session_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "session-id");
  g_object_class_override_property (klass, property_id_begin++, "caps-lock-state");
  g_object_class_override_property (klass, property_id_begin++, "num-lock-state");
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_remote_desktop_session_signal_marshal_closed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_signal_marshal_selection_owner_changed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_signal_marshal_selection_transfer (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_VOID__STRING_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_start (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_stop (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_keyboard_keycode (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_keyboard_keysym (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_button (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_axis (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_DOUBLE_DOUBLE_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_axis_discrete (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_motion_relative (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_DOUBLE_DOUBLE (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_motion_absolute (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_DOUBLE_DOUBLE (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_touch_down (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_UINT_DOUBLE_DOUBLE (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_touch_motion (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_UINT_DOUBLE_DOUBLE (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_notify_touch_up (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_enable_clipboard (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_disable_clipboard (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_set_selection (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_selection_write (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_selection_write_done (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT_BOOLEAN (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_selection_read (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_remote_desktop_session_method_marshal_connect_to_eis (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterRemoteDesktopSession:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop-Session.top_of_page">org.gnome.Mutter.RemoteDesktop.Session</link>.
 */

/**
 * OrgGnomeMutterRemoteDesktopSessionIface:
 * @parent_iface: The parent interface.
 * @handle_connect_to_eis: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-connect-to-eis signal.
 * @handle_disable_clipboard: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-disable-clipboard signal.
 * @handle_enable_clipboard: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-enable-clipboard signal.
 * @handle_notify_keyboard_keycode: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-keyboard-keycode signal.
 * @handle_notify_keyboard_keysym: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-keyboard-keysym signal.
 * @handle_notify_pointer_axis: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-axis signal.
 * @handle_notify_pointer_axis_discrete: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-axis-discrete signal.
 * @handle_notify_pointer_button: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-button signal.
 * @handle_notify_pointer_motion_absolute: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-motion-absolute signal.
 * @handle_notify_pointer_motion_relative: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-motion-relative signal.
 * @handle_notify_touch_down: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-touch-down signal.
 * @handle_notify_touch_motion: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-touch-motion signal.
 * @handle_notify_touch_up: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-notify-touch-up signal.
 * @handle_selection_read: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-selection-read signal.
 * @handle_selection_write: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-selection-write signal.
 * @handle_selection_write_done: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-selection-write-done signal.
 * @handle_set_selection: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-set-selection signal.
 * @handle_start: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-start signal.
 * @handle_stop: Handler for the #OrgGnomeMutterRemoteDesktopSession::handle-stop signal.
 * @get_caps_lock_state: Getter for the #OrgGnomeMutterRemoteDesktopSession:caps-lock-state property.
 * @get_num_lock_state: Getter for the #OrgGnomeMutterRemoteDesktopSession:num-lock-state property.
 * @get_session_id: Getter for the #OrgGnomeMutterRemoteDesktopSession:session-id property.
 * @closed: Handler for the #OrgGnomeMutterRemoteDesktopSession::closed signal.
 * @selection_owner_changed: Handler for the #OrgGnomeMutterRemoteDesktopSession::selection-owner-changed signal.
 * @selection_transfer: Handler for the #OrgGnomeMutterRemoteDesktopSession::selection-transfer signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop-Session.top_of_page">org.gnome.Mutter.RemoteDesktop.Session</link>.
 */

typedef OrgGnomeMutterRemoteDesktopSessionIface OrgGnomeMutterRemoteDesktopSessionInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterRemoteDesktopSession, org_gnome_mutter_remote_desktop_session, G_TYPE_OBJECT)

static void
org_gnome_mutter_remote_desktop_session_default_init (OrgGnomeMutterRemoteDesktopSessionIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-start:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Start">Start()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_start() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-start",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_start),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_start,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-stop:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Stop">Stop()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_stop() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-stop",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_stop),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_stop,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-keyboard-keycode:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_keycode: Argument passed by remote caller.
   * @arg_state: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeycode">NotifyKeyboardKeycode()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_keyboard_keycode() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-keyboard-keycode",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_keyboard_keycode),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_keyboard_keycode,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_BOOLEAN);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-keyboard-keysym:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_keysym: Argument passed by remote caller.
   * @arg_state: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeysym">NotifyKeyboardKeysym()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_keyboard_keysym() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-keyboard-keysym",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_keyboard_keysym),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_keyboard_keysym,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_BOOLEAN);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-button:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_button: Argument passed by remote caller.
   * @arg_state: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerButton">NotifyPointerButton()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_pointer_button() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-pointer-button",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_pointer_button),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_button,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT, G_TYPE_BOOLEAN);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-axis:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_dx: Argument passed by remote caller.
   * @arg_dy: Argument passed by remote caller.
   * @arg_flags: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxis">NotifyPointerAxis()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_pointer_axis() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-pointer-axis",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_pointer_axis),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_axis,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_DOUBLE, G_TYPE_DOUBLE, G_TYPE_UINT);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-axis-discrete:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_axis: Argument passed by remote caller.
   * @arg_steps: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxisDiscrete">NotifyPointerAxisDiscrete()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_pointer_axis_discrete() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-pointer-axis-discrete",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_pointer_axis_discrete),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_axis_discrete,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_INT);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-motion-relative:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_dx: Argument passed by remote caller.
   * @arg_dy: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionRelative">NotifyPointerMotionRelative()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_pointer_motion_relative() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-pointer-motion-relative",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_pointer_motion_relative),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_motion_relative,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_DOUBLE, G_TYPE_DOUBLE);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-pointer-motion-absolute:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_stream: Argument passed by remote caller.
   * @arg_x: Argument passed by remote caller.
   * @arg_y: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionAbsolute">NotifyPointerMotionAbsolute()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_pointer_motion_absolute() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-pointer-motion-absolute",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_pointer_motion_absolute),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_pointer_motion_absolute,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_DOUBLE, G_TYPE_DOUBLE);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-touch-down:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_stream: Argument passed by remote caller.
   * @arg_slot: Argument passed by remote caller.
   * @arg_x: Argument passed by remote caller.
   * @arg_y: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchDown">NotifyTouchDown()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_touch_down() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-touch-down",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_touch_down),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_touch_down,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_DOUBLE, G_TYPE_DOUBLE);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-touch-motion:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_stream: Argument passed by remote caller.
   * @arg_slot: Argument passed by remote caller.
   * @arg_x: Argument passed by remote caller.
   * @arg_y: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchMotion">NotifyTouchMotion()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_touch_motion() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-touch-motion",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_touch_motion),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_touch_motion,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_DOUBLE, G_TYPE_DOUBLE);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-notify-touch-up:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_slot: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchUp">NotifyTouchUp()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_notify_touch_up() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-notify-touch-up",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_notify_touch_up),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_notify_touch_up,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-enable-clipboard:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_options: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.EnableClipboard">EnableClipboard()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_enable_clipboard() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-enable-clipboard",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_enable_clipboard),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_enable_clipboard,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-disable-clipboard:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.DisableClipboard">DisableClipboard()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_disable_clipboard() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-disable-clipboard",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_disable_clipboard),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_disable_clipboard,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-set-selection:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_options: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SetSelection">SetSelection()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_set_selection() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-set-selection",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_set_selection),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_set_selection,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-selection-write:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @fd_list: (nullable): A #GUnixFDList or %NULL.
   * @arg_serial: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWrite">SelectionWrite()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_selection_write() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-selection-write",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_selection_write),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_selection_write,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UNIX_FD_LIST, G_TYPE_UINT);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-selection-write-done:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_serial: Argument passed by remote caller.
   * @arg_success: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWriteDone">SelectionWriteDone()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_selection_write_done() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-selection-write-done",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_selection_write_done),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_selection_write_done,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT, G_TYPE_BOOLEAN);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-selection-read:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @fd_list: (nullable): A #GUnixFDList or %NULL.
   * @arg_mime_type: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionRead">SelectionRead()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_selection_read() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-selection-read",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_selection_read),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_selection_read,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UNIX_FD_LIST, G_TYPE_STRING);

  /**
   * OrgGnomeMutterRemoteDesktopSession::handle-connect-to-eis:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @invocation: A #GDBusMethodInvocation.
   * @fd_list: (nullable): A #GUnixFDList or %NULL.
   * @arg_options: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_remote_desktop_session_complete_connect_to_eis() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-connect-to-eis",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, handle_connect_to_eis),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_remote_desktop_session_method_marshal_connect_to_eis,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UNIX_FD_LIST, G_TYPE_VARIANT);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGnomeMutterRemoteDesktopSession::closed:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-RemoteDesktop-Session.Closed">"Closed"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SIGNALS[_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_CLOSED] =
    g_signal_new ("closed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, closed),
      NULL,
      NULL,
      org_gnome_mutter_remote_desktop_session_signal_marshal_closed,
      G_TYPE_NONE,
      0);

  /**
   * OrgGnomeMutterRemoteDesktopSession::selection-owner-changed:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @arg_options: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-RemoteDesktop-Session.SelectionOwnerChanged">"SelectionOwnerChanged"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SIGNALS[_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SELECTION_OWNER_CHANGED] =
    g_signal_new ("selection-owner-changed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, selection_owner_changed),
      NULL,
      NULL,
      org_gnome_mutter_remote_desktop_session_signal_marshal_selection_owner_changed,
      G_TYPE_NONE,
      1, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterRemoteDesktopSession::selection-transfer:
   * @object: A #OrgGnomeMutterRemoteDesktopSession.
   * @arg_mime_type: Argument.
   * @arg_serial: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-RemoteDesktop-Session.SelectionTransfer">"SelectionTransfer"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SIGNALS[_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SELECTION_TRANSFER] =
    g_signal_new ("selection-transfer",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterRemoteDesktopSessionIface, selection_transfer),
      NULL,
      NULL,
      org_gnome_mutter_remote_desktop_session_signal_marshal_selection_transfer,
      G_TYPE_NONE,
      2, G_TYPE_STRING, G_TYPE_UINT);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeMutterRemoteDesktopSession:session-id:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.SessionId">"SessionId"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("session-id", "SessionId", "SessionId", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgGnomeMutterRemoteDesktopSession:caps-lock-state:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.CapsLockState">"CapsLockState"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("caps-lock-state", "CapsLockState", "CapsLockState", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgGnomeMutterRemoteDesktopSession:num-lock-state:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.NumLockState">"NumLockState"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("num-lock-state", "NumLockState", "NumLockState", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_mutter_remote_desktop_session_get_session_id: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.SessionId">"SessionId"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_gnome_mutter_remote_desktop_session_dup_session_id() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_gnome_mutter_remote_desktop_session_get_session_id (OrgGnomeMutterRemoteDesktopSession *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (object), NULL);

  return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_GET_IFACE (object)->get_session_id (object);
}

/**
 * org_gnome_mutter_remote_desktop_session_dup_session_id: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.SessionId">"SessionId"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_gnome_mutter_remote_desktop_session_dup_session_id (OrgGnomeMutterRemoteDesktopSession *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "session-id", &value, NULL);
  return value;
}

/**
 * org_gnome_mutter_remote_desktop_session_set_session_id: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.SessionId">"SessionId"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_remote_desktop_session_set_session_id (OrgGnomeMutterRemoteDesktopSession *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "session-id", value, NULL);
}

/**
 * org_gnome_mutter_remote_desktop_session_get_caps_lock_state: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.CapsLockState">"CapsLockState"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_gnome_mutter_remote_desktop_session_get_caps_lock_state (OrgGnomeMutterRemoteDesktopSession *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (object), FALSE);

  return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_GET_IFACE (object)->get_caps_lock_state (object);
}

/**
 * org_gnome_mutter_remote_desktop_session_set_caps_lock_state: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.CapsLockState">"CapsLockState"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_remote_desktop_session_set_caps_lock_state (OrgGnomeMutterRemoteDesktopSession *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "caps-lock-state", value, NULL);
}

/**
 * org_gnome_mutter_remote_desktop_session_get_num_lock_state: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.NumLockState">"NumLockState"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_gnome_mutter_remote_desktop_session_get_num_lock_state (OrgGnomeMutterRemoteDesktopSession *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (object), FALSE);

  return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_GET_IFACE (object)->get_num_lock_state (object);
}

/**
 * org_gnome_mutter_remote_desktop_session_set_num_lock_state: (skip)
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-RemoteDesktop-Session.NumLockState">"NumLockState"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_remote_desktop_session_set_num_lock_state (OrgGnomeMutterRemoteDesktopSession *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "num-lock-state", value, NULL);
}

/**
 * org_gnome_mutter_remote_desktop_session_emit_closed:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-RemoteDesktop-Session.Closed">"Closed"</link> D-Bus signal.
 */
void
org_gnome_mutter_remote_desktop_session_emit_closed (
    OrgGnomeMutterRemoteDesktopSession *object)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SIGNALS[_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_CLOSED], 0);
}

/**
 * org_gnome_mutter_remote_desktop_session_emit_selection_owner_changed:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @arg_options: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-RemoteDesktop-Session.SelectionOwnerChanged">"SelectionOwnerChanged"</link> D-Bus signal.
 */
void
org_gnome_mutter_remote_desktop_session_emit_selection_owner_changed (
    OrgGnomeMutterRemoteDesktopSession *object,
    GVariant *arg_options)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SIGNALS[_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SELECTION_OWNER_CHANGED], 0, arg_options);
}

/**
 * org_gnome_mutter_remote_desktop_session_emit_selection_transfer:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @arg_mime_type: Argument to pass with the signal.
 * @arg_serial: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-RemoteDesktop-Session.SelectionTransfer">"SelectionTransfer"</link> D-Bus signal.
 */
void
org_gnome_mutter_remote_desktop_session_emit_selection_transfer (
    OrgGnomeMutterRemoteDesktopSession *object,
    const gchar *arg_mime_type,
    guint arg_serial)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SIGNALS[_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SELECTION_TRANSFER], 0, arg_mime_type, arg_serial);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_start:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Start">Start()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_start_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_start_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_start (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Start",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_start_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_start().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_start().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_start_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_start_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Start">Start()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_start() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_start_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Start",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_stop:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Stop">Stop()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_stop_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_stop_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_stop (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Stop",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_stop_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_stop().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_stop().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_stop_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_stop_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Stop">Stop()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_stop() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_stop_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Stop",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_keycode: Argument to pass with the method invocation.
 * @arg_state: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeycode">NotifyKeyboardKeycode()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_keycode,
    gboolean arg_state,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyKeyboardKeycode",
    g_variant_new ("(ub)",
                   arg_keycode,
                   arg_state),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_keycode: Argument to pass with the method invocation.
 * @arg_state: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeycode">NotifyKeyboardKeycode()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keycode_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_keycode,
    gboolean arg_state,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyKeyboardKeycode",
    g_variant_new ("(ub)",
                   arg_keycode,
                   arg_state),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_keysym: Argument to pass with the method invocation.
 * @arg_state: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeysym">NotifyKeyboardKeysym()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_keysym,
    gboolean arg_state,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyKeyboardKeysym",
    g_variant_new ("(ub)",
                   arg_keysym,
                   arg_state),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_keysym: Argument to pass with the method invocation.
 * @arg_state: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeysym">NotifyKeyboardKeysym()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_keyboard_keysym_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_keysym,
    gboolean arg_state,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyKeyboardKeysym",
    g_variant_new ("(ub)",
                   arg_keysym,
                   arg_state),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_button:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_button: Argument to pass with the method invocation.
 * @arg_state: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerButton">NotifyPointerButton()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_pointer_button_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_button_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_pointer_button (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    gint arg_button,
    gboolean arg_state,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyPointerButton",
    g_variant_new ("(ib)",
                   arg_button,
                   arg_state),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_button_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_pointer_button().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_pointer_button().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_button_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_button_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_button: Argument to pass with the method invocation.
 * @arg_state: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerButton">NotifyPointerButton()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_button() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_button_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    gint arg_button,
    gboolean arg_state,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyPointerButton",
    g_variant_new ("(ib)",
                   arg_button,
                   arg_state),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_dx: Argument to pass with the method invocation.
 * @arg_dy: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxis">NotifyPointerAxis()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    gdouble arg_dx,
    gdouble arg_dy,
    guint arg_flags,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyPointerAxis",
    g_variant_new ("(ddu)",
                   arg_dx,
                   arg_dy,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_dx: Argument to pass with the method invocation.
 * @arg_dy: Argument to pass with the method invocation.
 * @arg_flags: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxis">NotifyPointerAxis()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    gdouble arg_dx,
    gdouble arg_dy,
    guint arg_flags,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyPointerAxis",
    g_variant_new ("(ddu)",
                   arg_dx,
                   arg_dy,
                   arg_flags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_axis: Argument to pass with the method invocation.
 * @arg_steps: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxisDiscrete">NotifyPointerAxisDiscrete()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_axis,
    gint arg_steps,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyPointerAxisDiscrete",
    g_variant_new ("(ui)",
                   arg_axis,
                   arg_steps),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_axis: Argument to pass with the method invocation.
 * @arg_steps: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxisDiscrete">NotifyPointerAxisDiscrete()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_axis_discrete_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_axis,
    gint arg_steps,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyPointerAxisDiscrete",
    g_variant_new ("(ui)",
                   arg_axis,
                   arg_steps),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_dx: Argument to pass with the method invocation.
 * @arg_dy: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionRelative">NotifyPointerMotionRelative()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    gdouble arg_dx,
    gdouble arg_dy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyPointerMotionRelative",
    g_variant_new ("(dd)",
                   arg_dx,
                   arg_dy),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_dx: Argument to pass with the method invocation.
 * @arg_dy: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionRelative">NotifyPointerMotionRelative()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_relative_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    gdouble arg_dx,
    gdouble arg_dy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyPointerMotionRelative",
    g_variant_new ("(dd)",
                   arg_dx,
                   arg_dy),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_stream: Argument to pass with the method invocation.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionAbsolute">NotifyPointerMotionAbsolute()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_stream,
    gdouble arg_x,
    gdouble arg_y,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyPointerMotionAbsolute",
    g_variant_new ("(sdd)",
                   arg_stream,
                   arg_x,
                   arg_y),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_stream: Argument to pass with the method invocation.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionAbsolute">NotifyPointerMotionAbsolute()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_pointer_motion_absolute_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_stream,
    gdouble arg_x,
    gdouble arg_y,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyPointerMotionAbsolute",
    g_variant_new ("(sdd)",
                   arg_stream,
                   arg_x,
                   arg_y),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_down:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_stream: Argument to pass with the method invocation.
 * @arg_slot: Argument to pass with the method invocation.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchDown">NotifyTouchDown()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_touch_down_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_touch_down_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_touch_down (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_stream,
    guint arg_slot,
    gdouble arg_x,
    gdouble arg_y,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyTouchDown",
    g_variant_new ("(sudd)",
                   arg_stream,
                   arg_slot,
                   arg_x,
                   arg_y),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_down_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_touch_down().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_touch_down().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_touch_down_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_down_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_stream: Argument to pass with the method invocation.
 * @arg_slot: Argument to pass with the method invocation.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchDown">NotifyTouchDown()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_touch_down() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_touch_down_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_stream,
    guint arg_slot,
    gdouble arg_x,
    gdouble arg_y,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyTouchDown",
    g_variant_new ("(sudd)",
                   arg_stream,
                   arg_slot,
                   arg_x,
                   arg_y),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_motion:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_stream: Argument to pass with the method invocation.
 * @arg_slot: Argument to pass with the method invocation.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchMotion">NotifyTouchMotion()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_touch_motion_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_touch_motion_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_touch_motion (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_stream,
    guint arg_slot,
    gdouble arg_x,
    gdouble arg_y,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyTouchMotion",
    g_variant_new ("(sudd)",
                   arg_stream,
                   arg_slot,
                   arg_x,
                   arg_y),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_motion_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_touch_motion().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_touch_motion().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_touch_motion_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_motion_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_stream: Argument to pass with the method invocation.
 * @arg_slot: Argument to pass with the method invocation.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchMotion">NotifyTouchMotion()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_touch_motion() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_touch_motion_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_stream,
    guint arg_slot,
    gdouble arg_x,
    gdouble arg_y,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyTouchMotion",
    g_variant_new ("(sudd)",
                   arg_stream,
                   arg_slot,
                   arg_x,
                   arg_y),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_up:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_slot: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchUp">NotifyTouchUp()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_notify_touch_up_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_touch_up_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_notify_touch_up (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_slot,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "NotifyTouchUp",
    g_variant_new ("(u)",
                   arg_slot),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_up_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_notify_touch_up().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_notify_touch_up().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_touch_up_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_notify_touch_up_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_slot: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchUp">NotifyTouchUp()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_notify_touch_up() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_notify_touch_up_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_slot,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "NotifyTouchUp",
    g_variant_new ("(u)",
                   arg_slot),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_enable_clipboard:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.EnableClipboard">EnableClipboard()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_enable_clipboard_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_enable_clipboard_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_enable_clipboard (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "EnableClipboard",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_enable_clipboard_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_enable_clipboard().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_enable_clipboard().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_enable_clipboard_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_enable_clipboard_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.EnableClipboard">EnableClipboard()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_enable_clipboard() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_enable_clipboard_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant *arg_options,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "EnableClipboard",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_disable_clipboard:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.DisableClipboard">DisableClipboard()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_disable_clipboard_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_disable_clipboard_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_disable_clipboard (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "DisableClipboard",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_disable_clipboard_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_disable_clipboard().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_disable_clipboard().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_disable_clipboard_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_disable_clipboard_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.DisableClipboard">DisableClipboard()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_disable_clipboard() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_disable_clipboard_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "DisableClipboard",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_set_selection:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SetSelection">SetSelection()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_set_selection_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_set_selection_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_set_selection (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SetSelection",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_set_selection_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_set_selection().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_set_selection().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_set_selection_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_set_selection_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SetSelection">SetSelection()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_set_selection() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_set_selection_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant *arg_options,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SetSelection",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_write:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWrite">SelectionWrite()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_selection_write_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_selection_write_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_selection_write (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_serial,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call_with_unix_fd_list (G_DBUS_PROXY (proxy),
    "SelectionWrite",
    g_variant_new ("(u)",
                   arg_serial),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_write_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_selection_write().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_selection_write().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_selection_write_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_finish (G_DBUS_PROXY (proxy), out_fd_list, res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_write_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWrite">SelectionWrite()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_selection_write() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_selection_write_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_serial,
    GUnixFDList  *fd_list,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_sync (G_DBUS_PROXY (proxy),
    "SelectionWrite",
    g_variant_new ("(u)",
                   arg_serial),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    out_fd_list,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_write_done:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_success: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWriteDone">SelectionWriteDone()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_selection_write_done_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_selection_write_done_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_selection_write_done (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_serial,
    gboolean arg_success,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SelectionWriteDone",
    g_variant_new ("(ub)",
                   arg_serial,
                   arg_success),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_write_done_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_selection_write_done().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_selection_write_done().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_selection_write_done_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_write_done_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_serial: Argument to pass with the method invocation.
 * @arg_success: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWriteDone">SelectionWriteDone()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_selection_write_done() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_selection_write_done_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    guint arg_serial,
    gboolean arg_success,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SelectionWriteDone",
    g_variant_new ("(ub)",
                   arg_serial,
                   arg_success),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_read:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_mime_type: Argument to pass with the method invocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionRead">SelectionRead()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_selection_read_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_selection_read_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_selection_read (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_mime_type,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call_with_unix_fd_list (G_DBUS_PROXY (proxy),
    "SelectionRead",
    g_variant_new ("(s)",
                   arg_mime_type),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_read_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_selection_read().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_selection_read().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_selection_read_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_finish (G_DBUS_PROXY (proxy), out_fd_list, res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_selection_read_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_mime_type: Argument to pass with the method invocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionRead">SelectionRead()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_selection_read() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_selection_read_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    const gchar *arg_mime_type,
    GUnixFDList  *fd_list,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_sync (G_DBUS_PROXY (proxy),
    "SelectionRead",
    g_variant_new ("(s)",
                   arg_mime_type),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    out_fd_list,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_connect_to_eis:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_call_connect_to_eis_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_call_connect_to_eis_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_remote_desktop_session_call_connect_to_eis (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant *arg_options,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call_with_unix_fd_list (G_DBUS_PROXY (proxy),
    "ConnectToEIS",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_remote_desktop_session_call_connect_to_eis_finish:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_call_connect_to_eis().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_call_connect_to_eis().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_connect_to_eis_finish (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_finish (G_DBUS_PROXY (proxy), out_fd_list, res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_call_connect_to_eis_sync:
 * @proxy: A #OrgGnomeMutterRemoteDesktopSessionProxy.
 * @arg_options: Argument to pass with the method invocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @out_fd: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_fd_list: (out) (optional) (nullable): Return location for a #GUnixFDList or %NULL.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_call_connect_to_eis() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_remote_desktop_session_call_connect_to_eis_sync (
    OrgGnomeMutterRemoteDesktopSession *proxy,
    GVariant *arg_options,
    GUnixFDList  *fd_list,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_with_unix_fd_list_sync (G_DBUS_PROXY (proxy),
    "ConnectToEIS",
    g_variant_new ("(@a{sv})",
                   arg_options),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    fd_list,
    out_fd_list,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@h)",
                 out_fd);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_start:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Start">Start()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_start (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_stop:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.Stop">Stop()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_stop (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_keyboard_keycode:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeycode">NotifyKeyboardKeycode()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_keyboard_keycode (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_keyboard_keysym:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyKeyboardKeysym">NotifyKeyboardKeysym()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_keyboard_keysym (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_pointer_button:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerButton">NotifyPointerButton()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_pointer_button (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_pointer_axis:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxis">NotifyPointerAxis()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_pointer_axis (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_pointer_axis_discrete:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerAxisDiscrete">NotifyPointerAxisDiscrete()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_pointer_axis_discrete (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_pointer_motion_relative:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionRelative">NotifyPointerMotionRelative()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_pointer_motion_relative (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_pointer_motion_absolute:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyPointerMotionAbsolute">NotifyPointerMotionAbsolute()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_pointer_motion_absolute (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_touch_down:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchDown">NotifyTouchDown()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_touch_down (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_touch_motion:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchMotion">NotifyTouchMotion()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_touch_motion (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_notify_touch_up:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.NotifyTouchUp">NotifyTouchUp()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_notify_touch_up (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_enable_clipboard:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.EnableClipboard">EnableClipboard()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_enable_clipboard (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_disable_clipboard:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.DisableClipboard">DisableClipboard()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_disable_clipboard (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_set_selection:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SetSelection">SetSelection()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_set_selection (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_selection_write:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @fd: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWrite">SelectionWrite()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_selection_write (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd)
{
  g_dbus_method_invocation_return_value_with_unix_fd_list (invocation,
    g_variant_new ("(@h)",
                   fd),
    fd_list);
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_selection_write_done:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionWriteDone">SelectionWriteDone()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_selection_write_done (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_selection_read:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @fd: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.SelectionRead">SelectionRead()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_selection_read (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd)
{
  g_dbus_method_invocation_return_value_with_unix_fd_list (invocation,
    g_variant_new ("(@h)",
                   fd),
    fd_list);
}

/**
 * org_gnome_mutter_remote_desktop_session_complete_connect_to_eis:
 * @object: A #OrgGnomeMutterRemoteDesktopSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @fd_list: (nullable): A #GUnixFDList or %NULL.
 * @fd: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-RemoteDesktop-Session.ConnectToEIS">ConnectToEIS()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_remote_desktop_session_complete_connect_to_eis (
    OrgGnomeMutterRemoteDesktopSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd)
{
  g_dbus_method_invocation_return_value_with_unix_fd_list (invocation,
    g_variant_new ("(@h)",
                   fd),
    fd_list);
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterRemoteDesktopSessionProxy:
 *
 * The #OrgGnomeMutterRemoteDesktopSessionProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterRemoteDesktopSessionProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterRemoteDesktopSessionProxy.
 */

struct _OrgGnomeMutterRemoteDesktopSessionProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_remote_desktop_session_proxy_iface_init (OrgGnomeMutterRemoteDesktopSessionIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopSessionProxy, org_gnome_mutter_remote_desktop_session_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterRemoteDesktopSessionProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION, org_gnome_mutter_remote_desktop_session_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopSessionProxy, org_gnome_mutter_remote_desktop_session_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION, org_gnome_mutter_remote_desktop_session_proxy_iface_init))

#endif
static void
org_gnome_mutter_remote_desktop_session_proxy_finalize (GObject *object)
{
  OrgGnomeMutterRemoteDesktopSessionProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_remote_desktop_session_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_remote_desktop_session_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_remote_desktop_session_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_mutter_remote_desktop_session_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Mutter.RemoteDesktop.Session: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_mutter_remote_desktop_session_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_remote_desktop_session_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Mutter.RemoteDesktop.Session", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_mutter_remote_desktop_session_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_mutter_remote_desktop_session_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_session_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_remote_desktop_session_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterRemoteDesktopSessionProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_session_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_session_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static const gchar *
org_gnome_mutter_remote_desktop_session_proxy_get_session_id (OrgGnomeMutterRemoteDesktopSession *object)
{
  OrgGnomeMutterRemoteDesktopSessionProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "SessionId");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
org_gnome_mutter_remote_desktop_session_proxy_get_caps_lock_state (OrgGnomeMutterRemoteDesktopSession *object)
{
  OrgGnomeMutterRemoteDesktopSessionProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "CapsLockState");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
org_gnome_mutter_remote_desktop_session_proxy_get_num_lock_state (OrgGnomeMutterRemoteDesktopSession *object)
{
  OrgGnomeMutterRemoteDesktopSessionProxy *proxy = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "NumLockState");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_gnome_mutter_remote_desktop_session_proxy_init (OrgGnomeMutterRemoteDesktopSessionProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_remote_desktop_session_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY, OrgGnomeMutterRemoteDesktopSessionProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_remote_desktop_session_interface_info ());
}

static void
org_gnome_mutter_remote_desktop_session_proxy_class_init (OrgGnomeMutterRemoteDesktopSessionProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_remote_desktop_session_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_remote_desktop_session_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_remote_desktop_session_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_remote_desktop_session_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_remote_desktop_session_proxy_g_properties_changed;

  org_gnome_mutter_remote_desktop_session_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterRemoteDesktopSessionProxyPrivate));
#endif
}

static void
org_gnome_mutter_remote_desktop_session_proxy_iface_init (OrgGnomeMutterRemoteDesktopSessionIface *iface)
{
  iface->get_session_id = org_gnome_mutter_remote_desktop_session_proxy_get_session_id;
  iface->get_caps_lock_state = org_gnome_mutter_remote_desktop_session_proxy_get_caps_lock_state;
  iface->get_num_lock_state = org_gnome_mutter_remote_desktop_session_proxy_get_num_lock_state;
}

/**
 * org_gnome_mutter_remote_desktop_session_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop-Session.top_of_page">org.gnome.Mutter.RemoteDesktop.Session</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_remote_desktop_session_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop.Session", NULL);
}

/**
 * org_gnome_mutter_remote_desktop_session_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktopSession *
org_gnome_mutter_remote_desktop_session_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop-Session.top_of_page">org.gnome.Mutter.RemoteDesktop.Session</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktopSession *
org_gnome_mutter_remote_desktop_session_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop.Session", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_remote_desktop_session_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_remote_desktop_session_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_remote_desktop_session_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_remote_desktop_session_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_remote_desktop_session_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop.Session", NULL);
}

/**
 * org_gnome_mutter_remote_desktop_session_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_remote_desktop_session_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_remote_desktop_session_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktopSession *
org_gnome_mutter_remote_desktop_session_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_remote_desktop_session_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_remote_desktop_session_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_remote_desktop_session_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterRemoteDesktopSession *
org_gnome_mutter_remote_desktop_session_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.RemoteDesktop.Session", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterRemoteDesktopSessionSkeleton:
 *
 * The #OrgGnomeMutterRemoteDesktopSessionSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterRemoteDesktopSessionSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterRemoteDesktopSessionSkeleton.
 */

struct _OrgGnomeMutterRemoteDesktopSessionSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_remote_desktop_session_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_remote_desktop_session_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_session_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_remote_desktop_session_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_remote_desktop_session_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_remote_desktop_session_skeleton_vtable =
{
  _org_gnome_mutter_remote_desktop_session_skeleton_handle_method_call,
  _org_gnome_mutter_remote_desktop_session_skeleton_handle_get_property,
  _org_gnome_mutter_remote_desktop_session_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_remote_desktop_session_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_remote_desktop_session_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_remote_desktop_session_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_remote_desktop_session_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_remote_desktop_session_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_remote_desktop_session_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.RemoteDesktop.Session", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_mutter_remote_desktop_session_emit_changed (gpointer user_data);

static void
org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_mutter_remote_desktop_session_emit_changed (skeleton);
}

static void
_org_gnome_mutter_remote_desktop_session_on_signal_closed (
    OrgGnomeMutterRemoteDesktopSession *object)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.RemoteDesktop.Session", "Closed",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_mutter_remote_desktop_session_on_signal_selection_owner_changed (
    OrgGnomeMutterRemoteDesktopSession *object,
    GVariant *arg_options)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(@a{sv})",
                   arg_options));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.RemoteDesktop.Session", "SelectionOwnerChanged",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_mutter_remote_desktop_session_on_signal_selection_transfer (
    OrgGnomeMutterRemoteDesktopSession *object,
    const gchar *arg_mime_type,
    guint arg_serial)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(su)",
                   arg_mime_type,
                   arg_serial));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.RemoteDesktop.Session", "SelectionTransfer",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gnome_mutter_remote_desktop_session_skeleton_iface_init (OrgGnomeMutterRemoteDesktopSessionIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopSessionSkeleton, org_gnome_mutter_remote_desktop_session_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterRemoteDesktopSessionSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION, org_gnome_mutter_remote_desktop_session_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterRemoteDesktopSessionSkeleton, org_gnome_mutter_remote_desktop_session_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION, org_gnome_mutter_remote_desktop_session_skeleton_iface_init))

#endif
static void
org_gnome_mutter_remote_desktop_session_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);
  guint n;
  for (n = 0; n < 3; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_remote_desktop_session_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_remote_desktop_session_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_mutter_remote_desktop_session_emit_changed (gpointer user_data)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Mutter.RemoteDesktop.Session",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_mutter_remote_desktop_session_schedule_emit_changed (OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_mutter_remote_desktop_session_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_mutter_remote_desktop_session_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_mutter_remote_desktop_session_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_mutter_remote_desktop_session_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_remote_desktop_session_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_mutter_remote_desktop_session_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_mutter_remote_desktop_session_skeleton_init (OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_remote_desktop_session_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON, OrgGnomeMutterRemoteDesktopSessionSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 3);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[1], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[2], G_TYPE_BOOLEAN);
}

static const gchar *
org_gnome_mutter_remote_desktop_session_skeleton_get_session_id (OrgGnomeMutterRemoteDesktopSession *object)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
org_gnome_mutter_remote_desktop_session_skeleton_get_caps_lock_state (OrgGnomeMutterRemoteDesktopSession *object)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[1]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
org_gnome_mutter_remote_desktop_session_skeleton_get_num_lock_state (OrgGnomeMutterRemoteDesktopSession *object)
{
  OrgGnomeMutterRemoteDesktopSessionSkeleton *skeleton = ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[2]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_mutter_remote_desktop_session_skeleton_class_init (OrgGnomeMutterRemoteDesktopSessionSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_remote_desktop_session_skeleton_finalize;
  gobject_class->get_property = org_gnome_mutter_remote_desktop_session_skeleton_get_property;
  gobject_class->set_property = org_gnome_mutter_remote_desktop_session_skeleton_set_property;
  gobject_class->notify       = org_gnome_mutter_remote_desktop_session_skeleton_notify;


  org_gnome_mutter_remote_desktop_session_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_remote_desktop_session_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterRemoteDesktopSessionSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_remote_desktop_session_skeleton_iface_init (OrgGnomeMutterRemoteDesktopSessionIface *iface)
{
  iface->closed = _org_gnome_mutter_remote_desktop_session_on_signal_closed;
  iface->selection_owner_changed = _org_gnome_mutter_remote_desktop_session_on_signal_selection_owner_changed;
  iface->selection_transfer = _org_gnome_mutter_remote_desktop_session_on_signal_selection_transfer;
  iface->get_session_id = org_gnome_mutter_remote_desktop_session_skeleton_get_session_id;
  iface->get_caps_lock_state = org_gnome_mutter_remote_desktop_session_skeleton_get_caps_lock_state;
  iface->get_num_lock_state = org_gnome_mutter_remote_desktop_session_skeleton_get_num_lock_state;
}

/**
 * org_gnome_mutter_remote_desktop_session_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-RemoteDesktop-Session.top_of_page">org.gnome.Mutter.RemoteDesktop.Session</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterRemoteDesktopSessionSkeleton): The skeleton object.
 */
OrgGnomeMutterRemoteDesktopSession *
org_gnome_mutter_remote_desktop_session_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION (g_object_new (TYPE_ORG_GNOME_MUTTER_REMOTE_DESKTOP_SESSION_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.ScreenCast
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterScreenCast
 * @title: OrgGnomeMutterScreenCast
 * @short_description: Generated C code for the org.gnome.Mutter.ScreenCast D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast.top_of_page">org.gnome.Mutter.ScreenCast</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gnome.Mutter.ScreenCast ---- */

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_method_info_create_session_IN_ARG_properties =
{
  {
    -1,
    (gchar *) "properties",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_method_info_create_session_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_method_info_create_session_IN_ARG_properties.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_method_info_create_session_OUT_ARG_session_path =
{
  {
    -1,
    (gchar *) "session_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_method_info_create_session_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_method_info_create_session_OUT_ARG_session_path.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_screen_cast_method_info_create_session =
{
  {
    -1,
    (gchar *) "CreateSession",
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_method_info_create_session_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_method_info_create_session_OUT_ARG_pointers,
    NULL
  },
  "handle-create-session",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_mutter_screen_cast_method_info_pointers[] =
{
  &_org_gnome_mutter_screen_cast_method_info_create_session.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_screen_cast_property_info_version =
{
  {
    -1,
    (gchar *) "Version",
    (gchar *) "i",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "version",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_mutter_screen_cast_property_info_pointers[] =
{
  &_org_gnome_mutter_screen_cast_property_info_version.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_screen_cast_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.ScreenCast",
    (GDBusMethodInfo **) &_org_gnome_mutter_screen_cast_method_info_pointers,
    NULL,
    (GDBusPropertyInfo **) &_org_gnome_mutter_screen_cast_property_info_pointers,
    NULL
  },
  "org-gnome-mutter-screen-cast",
};


/**
 * org_gnome_mutter_screen_cast_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast.top_of_page">org.gnome.Mutter.ScreenCast</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_screen_cast_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_screen_cast_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterScreenCast interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_screen_cast_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "version");
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_screen_cast_method_marshal_create_session (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterScreenCast:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast.top_of_page">org.gnome.Mutter.ScreenCast</link>.
 */

/**
 * OrgGnomeMutterScreenCastIface:
 * @parent_iface: The parent interface.
 * @handle_create_session: Handler for the #OrgGnomeMutterScreenCast::handle-create-session signal.
 * @get_version: Getter for the #OrgGnomeMutterScreenCast:version property.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast.top_of_page">org.gnome.Mutter.ScreenCast</link>.
 */

typedef OrgGnomeMutterScreenCastIface OrgGnomeMutterScreenCastInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterScreenCast, org_gnome_mutter_screen_cast, G_TYPE_OBJECT)

static void
org_gnome_mutter_screen_cast_default_init (OrgGnomeMutterScreenCastIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeMutterScreenCast::handle-create-session:
   * @object: A #OrgGnomeMutterScreenCast.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_properties: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast.CreateSession">CreateSession()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_screen_cast_complete_create_session() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-create-session",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterScreenCastIface, handle_create_session),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_screen_cast_method_marshal_create_session,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeMutterScreenCast:version:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-ScreenCast.Version">"Version"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_int ("version", "Version", "Version", G_MININT32, G_MAXINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_mutter_screen_cast_get_version: (skip)
 * @object: A #OrgGnomeMutterScreenCast.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-ScreenCast.Version">"Version"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gint 
org_gnome_mutter_screen_cast_get_version (OrgGnomeMutterScreenCast *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_SCREEN_CAST (object), 0);

  return ORG_GNOME_MUTTER_SCREEN_CAST_GET_IFACE (object)->get_version (object);
}

/**
 * org_gnome_mutter_screen_cast_set_version: (skip)
 * @object: A #OrgGnomeMutterScreenCast.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-ScreenCast.Version">"Version"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_screen_cast_set_version (OrgGnomeMutterScreenCast *object, gint value)
{
  g_object_set (G_OBJECT (object), "version", value, NULL);
}

/**
 * org_gnome_mutter_screen_cast_call_create_session:
 * @proxy: A #OrgGnomeMutterScreenCastProxy.
 * @arg_properties: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast.CreateSession">CreateSession()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_call_create_session_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_call_create_session_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_screen_cast_call_create_session (
    OrgGnomeMutterScreenCast *proxy,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "CreateSession",
    g_variant_new ("(@a{sv})",
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_screen_cast_call_create_session_finish:
 * @proxy: A #OrgGnomeMutterScreenCastProxy.
 * @out_session_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_call_create_session().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_call_create_session().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_call_create_session_finish (
    OrgGnomeMutterScreenCast *proxy,
    gchar ** out_session_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_session_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_call_create_session_sync:
 * @proxy: A #OrgGnomeMutterScreenCastProxy.
 * @arg_properties: Argument to pass with the method invocation.
 * @out_session_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast.CreateSession">CreateSession()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_call_create_session() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_call_create_session_sync (
    OrgGnomeMutterScreenCast *proxy,
    GVariant *arg_properties,
    gchar ** out_session_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "CreateSession",
    g_variant_new ("(@a{sv})",
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_session_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_complete_create_session:
 * @object: A #OrgGnomeMutterScreenCast.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @session_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast.CreateSession">CreateSession()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_screen_cast_complete_create_session (
    OrgGnomeMutterScreenCast *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *session_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   session_path));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterScreenCastProxy:
 *
 * The #OrgGnomeMutterScreenCastProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterScreenCastProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterScreenCastProxy.
 */

struct _OrgGnomeMutterScreenCastProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_screen_cast_proxy_iface_init (OrgGnomeMutterScreenCastIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastProxy, org_gnome_mutter_screen_cast_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterScreenCastProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST, org_gnome_mutter_screen_cast_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastProxy, org_gnome_mutter_screen_cast_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST, org_gnome_mutter_screen_cast_proxy_iface_init))

#endif
static void
org_gnome_mutter_screen_cast_proxy_finalize (GObject *object)
{
  OrgGnomeMutterScreenCastProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_screen_cast_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_screen_cast_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_screen_cast_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_mutter_screen_cast_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Mutter.ScreenCast: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_mutter_screen_cast_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_screen_cast_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Mutter.ScreenCast", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_mutter_screen_cast_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_mutter_screen_cast_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_SCREEN_CAST);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_screen_cast_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterScreenCastProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static gint 
org_gnome_mutter_screen_cast_proxy_get_version (OrgGnomeMutterScreenCast *object)
{
  OrgGnomeMutterScreenCastProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_PROXY (object);
  GVariant *variant;
  gint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Version");
  if (variant != NULL)
    {
      value = g_variant_get_int32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_gnome_mutter_screen_cast_proxy_init (OrgGnomeMutterScreenCastProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_screen_cast_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_PROXY, OrgGnomeMutterScreenCastProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_screen_cast_interface_info ());
}

static void
org_gnome_mutter_screen_cast_proxy_class_init (OrgGnomeMutterScreenCastProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_screen_cast_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_screen_cast_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_screen_cast_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_screen_cast_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_screen_cast_proxy_g_properties_changed;

  org_gnome_mutter_screen_cast_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterScreenCastProxyPrivate));
#endif
}

static void
org_gnome_mutter_screen_cast_proxy_iface_init (OrgGnomeMutterScreenCastIface *iface)
{
  iface->get_version = org_gnome_mutter_screen_cast_proxy_get_version;
}

/**
 * org_gnome_mutter_screen_cast_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast.top_of_page">org.gnome.Mutter.ScreenCast</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_screen_cast_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast", NULL);
}

/**
 * org_gnome_mutter_screen_cast_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCast *
org_gnome_mutter_screen_cast_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_screen_cast_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast.top_of_page">org.gnome.Mutter.ScreenCast</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCast *
org_gnome_mutter_screen_cast_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_screen_cast_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_screen_cast_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_screen_cast_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast", NULL);
}

/**
 * org_gnome_mutter_screen_cast_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCast *
org_gnome_mutter_screen_cast_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_screen_cast_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_screen_cast_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCast *
org_gnome_mutter_screen_cast_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterScreenCastSkeleton:
 *
 * The #OrgGnomeMutterScreenCastSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterScreenCastSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterScreenCastSkeleton.
 */

struct _OrgGnomeMutterScreenCastSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_screen_cast_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_SCREEN_CAST);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_screen_cast_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_screen_cast_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_screen_cast_skeleton_vtable =
{
  _org_gnome_mutter_screen_cast_skeleton_handle_method_call,
  _org_gnome_mutter_screen_cast_skeleton_handle_get_property,
  _org_gnome_mutter_screen_cast_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_screen_cast_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_screen_cast_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_screen_cast_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_screen_cast_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_screen_cast_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_screen_cast_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_screen_cast_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_screen_cast_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_screen_cast_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.ScreenCast", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_mutter_screen_cast_emit_changed (gpointer user_data);

static void
org_gnome_mutter_screen_cast_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_mutter_screen_cast_emit_changed (skeleton);
}

static void org_gnome_mutter_screen_cast_skeleton_iface_init (OrgGnomeMutterScreenCastIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastSkeleton, org_gnome_mutter_screen_cast_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterScreenCastSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST, org_gnome_mutter_screen_cast_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastSkeleton, org_gnome_mutter_screen_cast_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST, org_gnome_mutter_screen_cast_skeleton_iface_init))

#endif
static void
org_gnome_mutter_screen_cast_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (object);
  guint n;
  for (n = 0; n < 1; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_screen_cast_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_screen_cast_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_mutter_screen_cast_emit_changed (gpointer user_data)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Mutter.ScreenCast",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_mutter_screen_cast_schedule_emit_changed (OrgGnomeMutterScreenCastSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_mutter_screen_cast_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_mutter_screen_cast_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_mutter_screen_cast_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_mutter_screen_cast_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_screen_cast_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_mutter_screen_cast_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_mutter_screen_cast_skeleton_init (OrgGnomeMutterScreenCastSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_screen_cast_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON, OrgGnomeMutterScreenCastSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 1);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_INT);
}

static gint 
org_gnome_mutter_screen_cast_skeleton_get_version (OrgGnomeMutterScreenCast *object)
{
  OrgGnomeMutterScreenCastSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON (object);
  gint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_int (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_mutter_screen_cast_skeleton_class_init (OrgGnomeMutterScreenCastSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_screen_cast_skeleton_finalize;
  gobject_class->get_property = org_gnome_mutter_screen_cast_skeleton_get_property;
  gobject_class->set_property = org_gnome_mutter_screen_cast_skeleton_set_property;
  gobject_class->notify       = org_gnome_mutter_screen_cast_skeleton_notify;


  org_gnome_mutter_screen_cast_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_screen_cast_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_screen_cast_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_screen_cast_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_screen_cast_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterScreenCastSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_screen_cast_skeleton_iface_init (OrgGnomeMutterScreenCastIface *iface)
{
  iface->get_version = org_gnome_mutter_screen_cast_skeleton_get_version;
}

/**
 * org_gnome_mutter_screen_cast_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast.top_of_page">org.gnome.Mutter.ScreenCast</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastSkeleton): The skeleton object.
 */
OrgGnomeMutterScreenCast *
org_gnome_mutter_screen_cast_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_SCREEN_CAST (g_object_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.ScreenCast.Session
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterScreenCastSession
 * @title: OrgGnomeMutterScreenCastSession
 * @short_description: Generated C code for the org.gnome.Mutter.ScreenCast.Session D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Session.top_of_page">org.gnome.Mutter.ScreenCast.Session</link> D-Bus interface in C.
 */

enum
{
  _ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_CLOSED,
};

static unsigned _ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SIGNALS[1] = { 0 };

/* ---- Introspection data for org.gnome.Mutter.ScreenCast.Session ---- */

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_screen_cast_session_method_info_start =
{
  {
    -1,
    (gchar *) "Start",
    NULL,
    NULL,
    NULL
  },
  "handle-start",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_screen_cast_session_method_info_stop =
{
  {
    -1,
    (gchar *) "Stop",
    NULL,
    NULL,
    NULL
  },
  "handle-stop",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_monitor_IN_ARG_connector =
{
  {
    -1,
    (gchar *) "connector",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_monitor_IN_ARG_properties =
{
  {
    -1,
    (gchar *) "properties",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_monitor_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_monitor_IN_ARG_connector.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_monitor_IN_ARG_properties.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_monitor_OUT_ARG_stream_path =
{
  {
    -1,
    (gchar *) "stream_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_monitor_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_monitor_OUT_ARG_stream_path.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_screen_cast_session_method_info_record_monitor =
{
  {
    -1,
    (gchar *) "RecordMonitor",
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_monitor_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_monitor_OUT_ARG_pointers,
    NULL
  },
  "handle-record-monitor",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_window_IN_ARG_properties =
{
  {
    -1,
    (gchar *) "properties",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_window_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_window_IN_ARG_properties.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_window_OUT_ARG_stream_path =
{
  {
    -1,
    (gchar *) "stream_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_window_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_window_OUT_ARG_stream_path.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_screen_cast_session_method_info_record_window =
{
  {
    -1,
    (gchar *) "RecordWindow",
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_window_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_window_OUT_ARG_pointers,
    NULL
  },
  "handle-record-window",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_x =
{
  {
    -1,
    (gchar *) "x",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_y =
{
  {
    -1,
    (gchar *) "y",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_width =
{
  {
    -1,
    (gchar *) "width",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_height =
{
  {
    -1,
    (gchar *) "height",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_properties =
{
  {
    -1,
    (gchar *) "properties",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_x.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_y.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_width.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_height.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_properties.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_area_OUT_ARG_stream_path =
{
  {
    -1,
    (gchar *) "stream_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_area_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_area_OUT_ARG_stream_path.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_screen_cast_session_method_info_record_area =
{
  {
    -1,
    (gchar *) "RecordArea",
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_area_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_area_OUT_ARG_pointers,
    NULL
  },
  "handle-record-area",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_virtual_IN_ARG_properties =
{
  {
    -1,
    (gchar *) "properties",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_virtual_IN_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_virtual_IN_ARG_properties.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_session_method_info_record_virtual_OUT_ARG_stream_path =
{
  {
    -1,
    (gchar *) "stream_path",
    (gchar *) "o",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_session_method_info_record_virtual_OUT_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_record_virtual_OUT_ARG_stream_path.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_mutter_screen_cast_session_method_info_record_virtual =
{
  {
    -1,
    (gchar *) "RecordVirtual",
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_virtual_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_session_method_info_record_virtual_OUT_ARG_pointers,
    NULL
  },
  "handle-record-virtual",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_mutter_screen_cast_session_method_info_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_method_info_start.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_stop.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_monitor.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_window.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_area.parent_struct,
  &_org_gnome_mutter_screen_cast_session_method_info_record_virtual.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_screen_cast_session_signal_info_closed =
{
  {
    -1,
    (gchar *) "Closed",
    NULL,
    NULL
  },
  "closed"
};

static const GDBusSignalInfo * const _org_gnome_mutter_screen_cast_session_signal_info_pointers[] =
{
  &_org_gnome_mutter_screen_cast_session_signal_info_closed.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_screen_cast_session_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.ScreenCast.Session",
    (GDBusMethodInfo **) &_org_gnome_mutter_screen_cast_session_method_info_pointers,
    (GDBusSignalInfo **) &_org_gnome_mutter_screen_cast_session_signal_info_pointers,
    NULL,
    NULL
  },
  "org-gnome-mutter-screen-cast-session",
};


/**
 * org_gnome_mutter_screen_cast_session_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Session.top_of_page">org.gnome.Mutter.ScreenCast.Session</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_screen_cast_session_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_session_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_screen_cast_session_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterScreenCastSession interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_screen_cast_session_override_properties (GObjectClass *klass G_GNUC_UNUSED, guint property_id_begin)
{
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_screen_cast_session_signal_marshal_closed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_screen_cast_session_method_marshal_start (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_screen_cast_session_method_marshal_stop (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_screen_cast_session_method_marshal_record_monitor (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_screen_cast_session_method_marshal_record_window (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_screen_cast_session_method_marshal_record_area (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_INT_INT_INT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_mutter_screen_cast_session_method_marshal_record_virtual (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterScreenCastSession:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Session.top_of_page">org.gnome.Mutter.ScreenCast.Session</link>.
 */

/**
 * OrgGnomeMutterScreenCastSessionIface:
 * @parent_iface: The parent interface.
 * @handle_record_area: Handler for the #OrgGnomeMutterScreenCastSession::handle-record-area signal.
 * @handle_record_monitor: Handler for the #OrgGnomeMutterScreenCastSession::handle-record-monitor signal.
 * @handle_record_virtual: Handler for the #OrgGnomeMutterScreenCastSession::handle-record-virtual signal.
 * @handle_record_window: Handler for the #OrgGnomeMutterScreenCastSession::handle-record-window signal.
 * @handle_start: Handler for the #OrgGnomeMutterScreenCastSession::handle-start signal.
 * @handle_stop: Handler for the #OrgGnomeMutterScreenCastSession::handle-stop signal.
 * @closed: Handler for the #OrgGnomeMutterScreenCastSession::closed signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Session.top_of_page">org.gnome.Mutter.ScreenCast.Session</link>.
 */

typedef OrgGnomeMutterScreenCastSessionIface OrgGnomeMutterScreenCastSessionInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterScreenCastSession, org_gnome_mutter_screen_cast_session, G_TYPE_OBJECT)

static void
org_gnome_mutter_screen_cast_session_default_init (OrgGnomeMutterScreenCastSessionIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeMutterScreenCastSession::handle-start:
   * @object: A #OrgGnomeMutterScreenCastSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Start">Start()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_screen_cast_session_complete_start() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-start",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterScreenCastSessionIface, handle_start),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_screen_cast_session_method_marshal_start,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterScreenCastSession::handle-stop:
   * @object: A #OrgGnomeMutterScreenCastSession.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Stop">Stop()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_screen_cast_session_complete_stop() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-stop",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterScreenCastSessionIface, handle_stop),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_screen_cast_session_method_marshal_stop,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeMutterScreenCastSession::handle-record-monitor:
   * @object: A #OrgGnomeMutterScreenCastSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_connector: Argument passed by remote caller.
   * @arg_properties: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordMonitor">RecordMonitor()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_screen_cast_session_complete_record_monitor() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-record-monitor",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterScreenCastSessionIface, handle_record_monitor),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_screen_cast_session_method_marshal_record_monitor,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterScreenCastSession::handle-record-window:
   * @object: A #OrgGnomeMutterScreenCastSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_properties: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordWindow">RecordWindow()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_screen_cast_session_complete_record_window() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-record-window",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterScreenCastSessionIface, handle_record_window),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_screen_cast_session_method_marshal_record_window,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterScreenCastSession::handle-record-area:
   * @object: A #OrgGnomeMutterScreenCastSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_x: Argument passed by remote caller.
   * @arg_y: Argument passed by remote caller.
   * @arg_width: Argument passed by remote caller.
   * @arg_height: Argument passed by remote caller.
   * @arg_properties: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordArea">RecordArea()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_screen_cast_session_complete_record_area() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-record-area",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterScreenCastSessionIface, handle_record_area),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_screen_cast_session_method_marshal_record_area,
    G_TYPE_BOOLEAN,
    6,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_VARIANT);

  /**
   * OrgGnomeMutterScreenCastSession::handle-record-virtual:
   * @object: A #OrgGnomeMutterScreenCastSession.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_properties: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordVirtual">RecordVirtual()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_mutter_screen_cast_session_complete_record_virtual() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-record-virtual",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeMutterScreenCastSessionIface, handle_record_virtual),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_mutter_screen_cast_session_method_marshal_record_virtual,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGnomeMutterScreenCastSession::closed:
   * @object: A #OrgGnomeMutterScreenCastSession.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-ScreenCast-Session.Closed">"Closed"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SIGNALS[_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_CLOSED] =
    g_signal_new ("closed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterScreenCastSessionIface, closed),
      NULL,
      NULL,
      org_gnome_mutter_screen_cast_session_signal_marshal_closed,
      G_TYPE_NONE,
      0);

}

/**
 * org_gnome_mutter_screen_cast_session_emit_closed:
 * @object: A #OrgGnomeMutterScreenCastSession.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-ScreenCast-Session.Closed">"Closed"</link> D-Bus signal.
 */
void
org_gnome_mutter_screen_cast_session_emit_closed (
    OrgGnomeMutterScreenCastSession *object)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SIGNALS[_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_CLOSED], 0);
}

/**
 * org_gnome_mutter_screen_cast_session_call_start:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Start">Start()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_call_start_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_call_start_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_screen_cast_session_call_start (
    OrgGnomeMutterScreenCastSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Start",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_screen_cast_session_call_start_finish:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_call_start().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_call_start().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_start_finish (
    OrgGnomeMutterScreenCastSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_start_sync:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Start">Start()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_call_start() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_start_sync (
    OrgGnomeMutterScreenCastSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Start",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_stop:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Stop">Stop()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_call_stop_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_call_stop_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_screen_cast_session_call_stop (
    OrgGnomeMutterScreenCastSession *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Stop",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_screen_cast_session_call_stop_finish:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_call_stop().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_call_stop().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_stop_finish (
    OrgGnomeMutterScreenCastSession *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_stop_sync:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Stop">Stop()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_call_stop() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_stop_sync (
    OrgGnomeMutterScreenCastSession *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Stop",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_monitor:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_connector: Argument to pass with the method invocation.
 * @arg_properties: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordMonitor">RecordMonitor()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_call_record_monitor_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_monitor_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_screen_cast_session_call_record_monitor (
    OrgGnomeMutterScreenCastSession *proxy,
    const gchar *arg_connector,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RecordMonitor",
    g_variant_new ("(s@a{sv})",
                   arg_connector,
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_monitor_finish:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_call_record_monitor().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_call_record_monitor().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_monitor_finish (
    OrgGnomeMutterScreenCastSession *proxy,
    gchar ** out_stream_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_monitor_sync:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_connector: Argument to pass with the method invocation.
 * @arg_properties: Argument to pass with the method invocation.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordMonitor">RecordMonitor()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_monitor() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_monitor_sync (
    OrgGnomeMutterScreenCastSession *proxy,
    const gchar *arg_connector,
    GVariant *arg_properties,
    gchar ** out_stream_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RecordMonitor",
    g_variant_new ("(s@a{sv})",
                   arg_connector,
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_window:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_properties: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordWindow">RecordWindow()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_call_record_window_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_window_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_screen_cast_session_call_record_window (
    OrgGnomeMutterScreenCastSession *proxy,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RecordWindow",
    g_variant_new ("(@a{sv})",
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_window_finish:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_call_record_window().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_call_record_window().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_window_finish (
    OrgGnomeMutterScreenCastSession *proxy,
    gchar ** out_stream_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_window_sync:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_properties: Argument to pass with the method invocation.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordWindow">RecordWindow()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_window() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_window_sync (
    OrgGnomeMutterScreenCastSession *proxy,
    GVariant *arg_properties,
    gchar ** out_stream_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RecordWindow",
    g_variant_new ("(@a{sv})",
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_area:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @arg_width: Argument to pass with the method invocation.
 * @arg_height: Argument to pass with the method invocation.
 * @arg_properties: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordArea">RecordArea()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_call_record_area_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_area_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_screen_cast_session_call_record_area (
    OrgGnomeMutterScreenCastSession *proxy,
    gint arg_x,
    gint arg_y,
    gint arg_width,
    gint arg_height,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RecordArea",
    g_variant_new ("(iiii@a{sv})",
                   arg_x,
                   arg_y,
                   arg_width,
                   arg_height,
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_area_finish:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_call_record_area().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_call_record_area().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_area_finish (
    OrgGnomeMutterScreenCastSession *proxy,
    gchar ** out_stream_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_area_sync:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @arg_width: Argument to pass with the method invocation.
 * @arg_height: Argument to pass with the method invocation.
 * @arg_properties: Argument to pass with the method invocation.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordArea">RecordArea()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_area() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_area_sync (
    OrgGnomeMutterScreenCastSession *proxy,
    gint arg_x,
    gint arg_y,
    gint arg_width,
    gint arg_height,
    GVariant *arg_properties,
    gchar ** out_stream_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RecordArea",
    g_variant_new ("(iiii@a{sv})",
                   arg_x,
                   arg_y,
                   arg_width,
                   arg_height,
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_virtual:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_properties: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordVirtual">RecordVirtual()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_call_record_virtual_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_virtual_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_mutter_screen_cast_session_call_record_virtual (
    OrgGnomeMutterScreenCastSession *proxy,
    GVariant *arg_properties,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RecordVirtual",
    g_variant_new ("(@a{sv})",
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_virtual_finish:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_call_record_virtual().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_call_record_virtual().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_virtual_finish (
    OrgGnomeMutterScreenCastSession *proxy,
    gchar ** out_stream_path,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_call_record_virtual_sync:
 * @proxy: A #OrgGnomeMutterScreenCastSessionProxy.
 * @arg_properties: Argument to pass with the method invocation.
 * @out_stream_path: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordVirtual">RecordVirtual()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_call_record_virtual() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_mutter_screen_cast_session_call_record_virtual_sync (
    OrgGnomeMutterScreenCastSession *proxy,
    GVariant *arg_properties,
    gchar ** out_stream_path,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RecordVirtual",
    g_variant_new ("(@a{sv})",
                   arg_properties),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(o)",
                 out_stream_path);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_complete_start:
 * @object: A #OrgGnomeMutterScreenCastSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Start">Start()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_screen_cast_session_complete_start (
    OrgGnomeMutterScreenCastSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_screen_cast_session_complete_stop:
 * @object: A #OrgGnomeMutterScreenCastSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.Stop">Stop()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_screen_cast_session_complete_stop (
    OrgGnomeMutterScreenCastSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_mutter_screen_cast_session_complete_record_monitor:
 * @object: A #OrgGnomeMutterScreenCastSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @stream_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordMonitor">RecordMonitor()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_screen_cast_session_complete_record_monitor (
    OrgGnomeMutterScreenCastSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *stream_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   stream_path));
}

/**
 * org_gnome_mutter_screen_cast_session_complete_record_window:
 * @object: A #OrgGnomeMutterScreenCastSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @stream_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordWindow">RecordWindow()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_screen_cast_session_complete_record_window (
    OrgGnomeMutterScreenCastSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *stream_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   stream_path));
}

/**
 * org_gnome_mutter_screen_cast_session_complete_record_area:
 * @object: A #OrgGnomeMutterScreenCastSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @stream_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordArea">RecordArea()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_screen_cast_session_complete_record_area (
    OrgGnomeMutterScreenCastSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *stream_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   stream_path));
}

/**
 * org_gnome_mutter_screen_cast_session_complete_record_virtual:
 * @object: A #OrgGnomeMutterScreenCastSession.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @stream_path: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Mutter-ScreenCast-Session.RecordVirtual">RecordVirtual()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_mutter_screen_cast_session_complete_record_virtual (
    OrgGnomeMutterScreenCastSession *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    const gchar *stream_path)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(o)",
                   stream_path));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterScreenCastSessionProxy:
 *
 * The #OrgGnomeMutterScreenCastSessionProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterScreenCastSessionProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterScreenCastSessionProxy.
 */

struct _OrgGnomeMutterScreenCastSessionProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_screen_cast_session_proxy_iface_init (OrgGnomeMutterScreenCastSessionIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastSessionProxy, org_gnome_mutter_screen_cast_session_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterScreenCastSessionProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION, org_gnome_mutter_screen_cast_session_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastSessionProxy, org_gnome_mutter_screen_cast_session_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION, org_gnome_mutter_screen_cast_session_proxy_iface_init))

#endif
static void
org_gnome_mutter_screen_cast_session_proxy_finalize (GObject *object)
{
  OrgGnomeMutterScreenCastSessionProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_screen_cast_session_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_screen_cast_session_proxy_get_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  GValue       *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_mutter_screen_cast_session_proxy_set_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  const GValue *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_mutter_screen_cast_session_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_session_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_screen_cast_session_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterScreenCastSessionProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_session_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_session_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
org_gnome_mutter_screen_cast_session_proxy_init (OrgGnomeMutterScreenCastSessionProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_screen_cast_session_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_PROXY, OrgGnomeMutterScreenCastSessionProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_screen_cast_session_interface_info ());
}

static void
org_gnome_mutter_screen_cast_session_proxy_class_init (OrgGnomeMutterScreenCastSessionProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_screen_cast_session_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_screen_cast_session_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_screen_cast_session_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_screen_cast_session_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_screen_cast_session_proxy_g_properties_changed;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterScreenCastSessionProxyPrivate));
#endif
}

static void
org_gnome_mutter_screen_cast_session_proxy_iface_init (OrgGnomeMutterScreenCastSessionIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_mutter_screen_cast_session_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Session.top_of_page">org.gnome.Mutter.ScreenCast.Session</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_screen_cast_session_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Session", NULL);
}

/**
 * org_gnome_mutter_screen_cast_session_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastSession *
org_gnome_mutter_screen_cast_session_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_SESSION (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Session.top_of_page">org.gnome.Mutter.ScreenCast.Session</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastSession *
org_gnome_mutter_screen_cast_session_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Session", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_SESSION (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_screen_cast_session_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_screen_cast_session_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_session_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_session_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_screen_cast_session_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Session", NULL);
}

/**
 * org_gnome_mutter_screen_cast_session_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_session_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_session_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastSession *
org_gnome_mutter_screen_cast_session_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_SESSION (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_screen_cast_session_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_screen_cast_session_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_session_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastSessionProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastSession *
org_gnome_mutter_screen_cast_session_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Session", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_SESSION (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterScreenCastSessionSkeleton:
 *
 * The #OrgGnomeMutterScreenCastSessionSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterScreenCastSessionSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterScreenCastSessionSkeleton.
 */

struct _OrgGnomeMutterScreenCastSessionSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_screen_cast_session_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastSessionSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_screen_cast_session_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastSessionSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_session_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_screen_cast_session_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastSessionSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_session_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_screen_cast_session_skeleton_vtable =
{
  _org_gnome_mutter_screen_cast_session_skeleton_handle_method_call,
  _org_gnome_mutter_screen_cast_session_skeleton_handle_get_property,
  _org_gnome_mutter_screen_cast_session_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_screen_cast_session_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_screen_cast_session_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterScreenCastSessionSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_screen_cast_session_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_screen_cast_session_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_screen_cast_session_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_screen_cast_session_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.ScreenCast.Session", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton G_GNUC_UNUSED)
{
}

static void
_org_gnome_mutter_screen_cast_session_on_signal_closed (
    OrgGnomeMutterScreenCastSession *object)
{
  OrgGnomeMutterScreenCastSessionSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.ScreenCast.Session", "Closed",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gnome_mutter_screen_cast_session_skeleton_iface_init (OrgGnomeMutterScreenCastSessionIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastSessionSkeleton, org_gnome_mutter_screen_cast_session_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterScreenCastSessionSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION, org_gnome_mutter_screen_cast_session_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastSessionSkeleton, org_gnome_mutter_screen_cast_session_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION, org_gnome_mutter_screen_cast_session_skeleton_iface_init))

#endif
static void
org_gnome_mutter_screen_cast_session_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterScreenCastSessionSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_screen_cast_session_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_screen_cast_session_skeleton_init (OrgGnomeMutterScreenCastSessionSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_screen_cast_session_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON, OrgGnomeMutterScreenCastSessionSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
org_gnome_mutter_screen_cast_session_skeleton_class_init (OrgGnomeMutterScreenCastSessionSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_screen_cast_session_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_screen_cast_session_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterScreenCastSessionSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_screen_cast_session_skeleton_iface_init (OrgGnomeMutterScreenCastSessionIface *iface)
{
  iface->closed = _org_gnome_mutter_screen_cast_session_on_signal_closed;
}

/**
 * org_gnome_mutter_screen_cast_session_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Session.top_of_page">org.gnome.Mutter.ScreenCast.Session</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastSessionSkeleton): The skeleton object.
 */
OrgGnomeMutterScreenCastSession *
org_gnome_mutter_screen_cast_session_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_SCREEN_CAST_SESSION (g_object_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_SESSION_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Mutter.ScreenCast.Stream
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeMutterScreenCastStream
 * @title: OrgGnomeMutterScreenCastStream
 * @short_description: Generated C code for the org.gnome.Mutter.ScreenCast.Stream D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Stream.top_of_page">org.gnome.Mutter.ScreenCast.Stream</link> D-Bus interface in C.
 */

enum
{
  _ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PIPEWIRE_STREAM_ADDED,
};

static unsigned _ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SIGNALS[1] = { 0 };

/* ---- Introspection data for org.gnome.Mutter.ScreenCast.Stream ---- */

static const _ExtendedGDBusArgInfo _org_gnome_mutter_screen_cast_stream_signal_info_pipewire_stream_added_ARG_node_id =
{
  {
    -1,
    (gchar *) "node_id",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_mutter_screen_cast_stream_signal_info_pipewire_stream_added_ARG_pointers[] =
{
  &_org_gnome_mutter_screen_cast_stream_signal_info_pipewire_stream_added_ARG_node_id.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_mutter_screen_cast_stream_signal_info_pipewire_stream_added =
{
  {
    -1,
    (gchar *) "PipeWireStreamAdded",
    (GDBusArgInfo **) &_org_gnome_mutter_screen_cast_stream_signal_info_pipewire_stream_added_ARG_pointers,
    NULL
  },
  "pipewire-stream-added"
};

static const GDBusSignalInfo * const _org_gnome_mutter_screen_cast_stream_signal_info_pointers[] =
{
  &_org_gnome_mutter_screen_cast_stream_signal_info_pipewire_stream_added.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_mutter_screen_cast_stream_property_info_parameters =
{
  {
    -1,
    (gchar *) "Parameters",
    (gchar *) "a{sv}",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "parameters",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_mutter_screen_cast_stream_property_info_pointers[] =
{
  &_org_gnome_mutter_screen_cast_stream_property_info_parameters.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_mutter_screen_cast_stream_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Mutter.ScreenCast.Stream",
    NULL,
    (GDBusSignalInfo **) &_org_gnome_mutter_screen_cast_stream_signal_info_pointers,
    (GDBusPropertyInfo **) &_org_gnome_mutter_screen_cast_stream_property_info_pointers,
    NULL
  },
  "org-gnome-mutter-screen-cast-stream",
};


/**
 * org_gnome_mutter_screen_cast_stream_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Stream.top_of_page">org.gnome.Mutter.ScreenCast.Stream</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_mutter_screen_cast_stream_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_stream_interface_info.parent_struct;
}

/**
 * org_gnome_mutter_screen_cast_stream_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeMutterScreenCastStream interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_mutter_screen_cast_stream_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "parameters");
  return property_id_begin - 1;
}


inline static void
org_gnome_mutter_screen_cast_stream_signal_marshal_pipewire_stream_added (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeMutterScreenCastStream:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Stream.top_of_page">org.gnome.Mutter.ScreenCast.Stream</link>.
 */

/**
 * OrgGnomeMutterScreenCastStreamIface:
 * @parent_iface: The parent interface.
 * @get_parameters: Getter for the #OrgGnomeMutterScreenCastStream:parameters property.
 * @pipewire_stream_added: Handler for the #OrgGnomeMutterScreenCastStream::pipewire-stream-added signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Stream.top_of_page">org.gnome.Mutter.ScreenCast.Stream</link>.
 */

typedef OrgGnomeMutterScreenCastStreamIface OrgGnomeMutterScreenCastStreamInterface;
G_DEFINE_INTERFACE (OrgGnomeMutterScreenCastStream, org_gnome_mutter_screen_cast_stream, G_TYPE_OBJECT)

static void
org_gnome_mutter_screen_cast_stream_default_init (OrgGnomeMutterScreenCastStreamIface *iface)
{
  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGnomeMutterScreenCastStream::pipewire-stream-added:
   * @object: A #OrgGnomeMutterScreenCastStream.
   * @arg_node_id: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Mutter-ScreenCast-Stream.PipeWireStreamAdded">"PipeWireStreamAdded"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SIGNALS[_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PIPEWIRE_STREAM_ADDED] =
    g_signal_new ("pipewire-stream-added",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeMutterScreenCastStreamIface, pipewire_stream_added),
      NULL,
      NULL,
      org_gnome_mutter_screen_cast_stream_signal_marshal_pipewire_stream_added,
      G_TYPE_NONE,
      1, G_TYPE_UINT);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeMutterScreenCastStream:parameters:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Mutter-ScreenCast-Stream.Parameters">"Parameters"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_variant ("parameters", "Parameters", "Parameters", G_VARIANT_TYPE ("a{sv}"), NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_mutter_screen_cast_stream_get_parameters: (skip)
 * @object: A #OrgGnomeMutterScreenCastStream.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Mutter-ScreenCast-Stream.Parameters">"Parameters"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_gnome_mutter_screen_cast_stream_dup_parameters() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
GVariant *
org_gnome_mutter_screen_cast_stream_get_parameters (OrgGnomeMutterScreenCastStream *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM (object), NULL);

  return ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_GET_IFACE (object)->get_parameters (object);
}

/**
 * org_gnome_mutter_screen_cast_stream_dup_parameters: (skip)
 * @object: A #OrgGnomeMutterScreenCastStream.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-gnome-Mutter-ScreenCast-Stream.Parameters">"Parameters"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_variant_unref().
 */
GVariant *
org_gnome_mutter_screen_cast_stream_dup_parameters (OrgGnomeMutterScreenCastStream *object)
{
  GVariant *value;
  g_object_get (G_OBJECT (object), "parameters", &value, NULL);
  return value;
}

/**
 * org_gnome_mutter_screen_cast_stream_set_parameters: (skip)
 * @object: A #OrgGnomeMutterScreenCastStream.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Mutter-ScreenCast-Stream.Parameters">"Parameters"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_mutter_screen_cast_stream_set_parameters (OrgGnomeMutterScreenCastStream *object, GVariant *value)
{
  g_object_set (G_OBJECT (object), "parameters", value, NULL);
}

/**
 * org_gnome_mutter_screen_cast_stream_emit_pipewire_stream_added:
 * @object: A #OrgGnomeMutterScreenCastStream.
 * @arg_node_id: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Mutter-ScreenCast-Stream.PipeWireStreamAdded">"PipeWireStreamAdded"</link> D-Bus signal.
 */
void
org_gnome_mutter_screen_cast_stream_emit_pipewire_stream_added (
    OrgGnomeMutterScreenCastStream *object,
    guint arg_node_id)
{
  g_signal_emit (object, _ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SIGNALS[_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PIPEWIRE_STREAM_ADDED], 0, arg_node_id);
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterScreenCastStreamProxy:
 *
 * The #OrgGnomeMutterScreenCastStreamProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterScreenCastStreamProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterScreenCastStreamProxy.
 */

struct _OrgGnomeMutterScreenCastStreamProxyPrivate
{
  GData *qdata;
};

static void org_gnome_mutter_screen_cast_stream_proxy_iface_init (OrgGnomeMutterScreenCastStreamIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastStreamProxy, org_gnome_mutter_screen_cast_stream_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeMutterScreenCastStreamProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM, org_gnome_mutter_screen_cast_stream_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastStreamProxy, org_gnome_mutter_screen_cast_stream_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM, org_gnome_mutter_screen_cast_stream_proxy_iface_init))

#endif
static void
org_gnome_mutter_screen_cast_stream_proxy_finalize (GObject *object)
{
  OrgGnomeMutterScreenCastStreamProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_mutter_screen_cast_stream_proxy_parent_class)->finalize (object);
}

static void
org_gnome_mutter_screen_cast_stream_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_screen_cast_stream_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_mutter_screen_cast_stream_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Mutter.ScreenCast.Stream: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_mutter_screen_cast_stream_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_screen_cast_stream_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Mutter.ScreenCast.Stream", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_mutter_screen_cast_stream_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_mutter_screen_cast_stream_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_stream_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_mutter_screen_cast_stream_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeMutterScreenCastStreamProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_stream_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_stream_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static GVariant *
org_gnome_mutter_screen_cast_stream_proxy_get_parameters (OrgGnomeMutterScreenCastStream *object)
{
  OrgGnomeMutterScreenCastStreamProxy *proxy = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY (object);
  GVariant *variant;
  GVariant *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Parameters");
  value = variant;
  if (variant != NULL)
    g_variant_unref (variant);
  return value;
}

static void
org_gnome_mutter_screen_cast_stream_proxy_init (OrgGnomeMutterScreenCastStreamProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_mutter_screen_cast_stream_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY, OrgGnomeMutterScreenCastStreamProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_mutter_screen_cast_stream_interface_info ());
}

static void
org_gnome_mutter_screen_cast_stream_proxy_class_init (OrgGnomeMutterScreenCastStreamProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_mutter_screen_cast_stream_proxy_finalize;
  gobject_class->get_property = org_gnome_mutter_screen_cast_stream_proxy_get_property;
  gobject_class->set_property = org_gnome_mutter_screen_cast_stream_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_mutter_screen_cast_stream_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_mutter_screen_cast_stream_proxy_g_properties_changed;

  org_gnome_mutter_screen_cast_stream_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterScreenCastStreamProxyPrivate));
#endif
}

static void
org_gnome_mutter_screen_cast_stream_proxy_iface_init (OrgGnomeMutterScreenCastStreamIface *iface)
{
  iface->get_parameters = org_gnome_mutter_screen_cast_stream_proxy_get_parameters;
}

/**
 * org_gnome_mutter_screen_cast_stream_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Stream.top_of_page">org.gnome.Mutter.ScreenCast.Stream</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_stream_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_stream_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_screen_cast_stream_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Stream", NULL);
}

/**
 * org_gnome_mutter_screen_cast_stream_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_stream_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_stream_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastStreamProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastStream *
org_gnome_mutter_screen_cast_stream_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_STREAM (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_screen_cast_stream_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Stream.top_of_page">org.gnome.Mutter.ScreenCast.Stream</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_stream_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastStreamProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastStream *
org_gnome_mutter_screen_cast_stream_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Stream", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_STREAM (ret);
  else
    return NULL;
}


/**
 * org_gnome_mutter_screen_cast_stream_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_mutter_screen_cast_stream_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_mutter_screen_cast_stream_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_mutter_screen_cast_stream_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_mutter_screen_cast_stream_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Stream", NULL);
}

/**
 * org_gnome_mutter_screen_cast_stream_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_mutter_screen_cast_stream_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_mutter_screen_cast_stream_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastStreamProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastStream *
org_gnome_mutter_screen_cast_stream_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_STREAM (ret);
  else
    return NULL;
}

/**
 * org_gnome_mutter_screen_cast_stream_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_mutter_screen_cast_stream_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_mutter_screen_cast_stream_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastStreamProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeMutterScreenCastStream *
org_gnome_mutter_screen_cast_stream_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Mutter.ScreenCast.Stream", NULL);
  if (ret != NULL)
    return ORG_GNOME_MUTTER_SCREEN_CAST_STREAM (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeMutterScreenCastStreamSkeleton:
 *
 * The #OrgGnomeMutterScreenCastStreamSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeMutterScreenCastStreamSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeMutterScreenCastStreamSkeleton.
 */

struct _OrgGnomeMutterScreenCastStreamSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_mutter_screen_cast_stream_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_mutter_screen_cast_stream_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_stream_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_mutter_screen_cast_stream_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_mutter_screen_cast_stream_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_mutter_screen_cast_stream_skeleton_vtable =
{
  _org_gnome_mutter_screen_cast_stream_skeleton_handle_method_call,
  _org_gnome_mutter_screen_cast_stream_skeleton_handle_get_property,
  _org_gnome_mutter_screen_cast_stream_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_mutter_screen_cast_stream_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_mutter_screen_cast_stream_skeleton_vtable;
}

static GVariant *
org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_mutter_screen_cast_stream_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_mutter_screen_cast_stream_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_mutter_screen_cast_stream_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_mutter_screen_cast_stream_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.ScreenCast.Stream", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_mutter_screen_cast_stream_emit_changed (gpointer user_data);

static void
org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_mutter_screen_cast_stream_emit_changed (skeleton);
}

static void
_org_gnome_mutter_screen_cast_stream_on_signal_pipewire_stream_added (
    OrgGnomeMutterScreenCastStream *object,
    guint arg_node_id)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(u)",
                   arg_node_id));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Mutter.ScreenCast.Stream", "PipeWireStreamAdded",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gnome_mutter_screen_cast_stream_skeleton_iface_init (OrgGnomeMutterScreenCastStreamIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastStreamSkeleton, org_gnome_mutter_screen_cast_stream_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeMutterScreenCastStreamSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM, org_gnome_mutter_screen_cast_stream_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeMutterScreenCastStreamSkeleton, org_gnome_mutter_screen_cast_stream_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM, org_gnome_mutter_screen_cast_stream_skeleton_iface_init))

#endif
static void
org_gnome_mutter_screen_cast_stream_skeleton_finalize (GObject *object)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (object);
  guint n;
  for (n = 0; n < 1; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_mutter_screen_cast_stream_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_mutter_screen_cast_stream_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_mutter_screen_cast_stream_emit_changed (gpointer user_data)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Mutter.ScreenCast.Stream",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_mutter_screen_cast_stream_schedule_emit_changed (OrgGnomeMutterScreenCastStreamSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_mutter_screen_cast_stream_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_mutter_screen_cast_stream_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_mutter_screen_cast_stream_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_mutter_screen_cast_stream_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 1);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_mutter_screen_cast_stream_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_mutter_screen_cast_stream_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_mutter_screen_cast_stream_skeleton_init (OrgGnomeMutterScreenCastStreamSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_mutter_screen_cast_stream_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON, OrgGnomeMutterScreenCastStreamSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 1);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_VARIANT);
}

static GVariant *
org_gnome_mutter_screen_cast_stream_skeleton_get_parameters (OrgGnomeMutterScreenCastStream *object)
{
  OrgGnomeMutterScreenCastStreamSkeleton *skeleton = ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON (object);
  GVariant *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_variant (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_mutter_screen_cast_stream_skeleton_class_init (OrgGnomeMutterScreenCastStreamSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_mutter_screen_cast_stream_skeleton_finalize;
  gobject_class->get_property = org_gnome_mutter_screen_cast_stream_skeleton_get_property;
  gobject_class->set_property = org_gnome_mutter_screen_cast_stream_skeleton_set_property;
  gobject_class->notify       = org_gnome_mutter_screen_cast_stream_skeleton_notify;


  org_gnome_mutter_screen_cast_stream_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_mutter_screen_cast_stream_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeMutterScreenCastStreamSkeletonPrivate));
#endif
}

static void
org_gnome_mutter_screen_cast_stream_skeleton_iface_init (OrgGnomeMutterScreenCastStreamIface *iface)
{
  iface->pipewire_stream_added = _org_gnome_mutter_screen_cast_stream_on_signal_pipewire_stream_added;
  iface->get_parameters = org_gnome_mutter_screen_cast_stream_skeleton_get_parameters;
}

/**
 * org_gnome_mutter_screen_cast_stream_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Mutter-ScreenCast-Stream.top_of_page">org.gnome.Mutter.ScreenCast.Stream</link>.
 *
 * Returns: (transfer full) (type OrgGnomeMutterScreenCastStreamSkeleton): The skeleton object.
 */
OrgGnomeMutterScreenCastStream *
org_gnome_mutter_screen_cast_stream_skeleton_new (void)
{
  return ORG_GNOME_MUTTER_SCREEN_CAST_STREAM (g_object_new (TYPE_ORG_GNOME_MUTTER_SCREEN_CAST_STREAM_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Shell.Introspect
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeShellIntrospect
 * @title: OrgGnomeShellIntrospect
 * @short_description: Generated C code for the org.gnome.Shell.Introspect D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Shell-Introspect.top_of_page">org.gnome.Shell.Introspect</link> D-Bus interface in C.
 */

enum
{
  _ORG_GNOME_SHELL_INTROSPECT_RUNNING_APPLICATIONS_CHANGED,
  _ORG_GNOME_SHELL_INTROSPECT_WINDOWS_CHANGED,
};

static unsigned _ORG_GNOME_SHELL_INTROSPECT_SIGNALS[2] = { 0 };

/* ---- Introspection data for org.gnome.Shell.Introspect ---- */

static const _ExtendedGDBusArgInfo _org_gnome_shell_introspect_method_info_get_running_applications_OUT_ARG_apps =
{
  {
    -1,
    (gchar *) "apps",
    (gchar *) "a{sa{sv}}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_introspect_method_info_get_running_applications_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_introspect_method_info_get_running_applications_OUT_ARG_apps.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_introspect_method_info_get_running_applications =
{
  {
    -1,
    (gchar *) "GetRunningApplications",
    NULL,
    (GDBusArgInfo **) &_org_gnome_shell_introspect_method_info_get_running_applications_OUT_ARG_pointers,
    NULL
  },
  "handle-get-running-applications",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_introspect_method_info_get_windows_OUT_ARG_windows =
{
  {
    -1,
    (gchar *) "windows",
    (gchar *) "a{ta{sv}}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_introspect_method_info_get_windows_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_introspect_method_info_get_windows_OUT_ARG_windows.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_introspect_method_info_get_windows =
{
  {
    -1,
    (gchar *) "GetWindows",
    NULL,
    (GDBusArgInfo **) &_org_gnome_shell_introspect_method_info_get_windows_OUT_ARG_pointers,
    NULL
  },
  "handle-get-windows",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_shell_introspect_method_info_pointers[] =
{
  &_org_gnome_shell_introspect_method_info_get_running_applications.parent_struct,
  &_org_gnome_shell_introspect_method_info_get_windows.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_shell_introspect_signal_info_running_applications_changed =
{
  {
    -1,
    (gchar *) "RunningApplicationsChanged",
    NULL,
    NULL
  },
  "running-applications-changed"
};

static const _ExtendedGDBusSignalInfo _org_gnome_shell_introspect_signal_info_windows_changed =
{
  {
    -1,
    (gchar *) "WindowsChanged",
    NULL,
    NULL
  },
  "windows-changed"
};

static const GDBusSignalInfo * const _org_gnome_shell_introspect_signal_info_pointers[] =
{
  &_org_gnome_shell_introspect_signal_info_running_applications_changed.parent_struct,
  &_org_gnome_shell_introspect_signal_info_windows_changed.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_shell_introspect_property_info_animations_enabled =
{
  {
    -1,
    (gchar *) "AnimationsEnabled",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "animations-enabled",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_gnome_shell_introspect_property_info_screen_size =
{
  {
    -1,
    (gchar *) "ScreenSize",
    (gchar *) "(ii)",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "screen-size",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_gnome_shell_introspect_property_info_version =
{
  {
    -1,
    (gchar *) "version",
    (gchar *) "u",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "version",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_shell_introspect_property_info_pointers[] =
{
  &_org_gnome_shell_introspect_property_info_animations_enabled.parent_struct,
  &_org_gnome_shell_introspect_property_info_screen_size.parent_struct,
  &_org_gnome_shell_introspect_property_info_version.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_shell_introspect_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Shell.Introspect",
    (GDBusMethodInfo **) &_org_gnome_shell_introspect_method_info_pointers,
    (GDBusSignalInfo **) &_org_gnome_shell_introspect_signal_info_pointers,
    (GDBusPropertyInfo **) &_org_gnome_shell_introspect_property_info_pointers,
    NULL
  },
  "org-gnome-shell-introspect",
};


/**
 * org_gnome_shell_introspect_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Shell-Introspect.top_of_page">org.gnome.Shell.Introspect</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_shell_introspect_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_shell_introspect_interface_info.parent_struct;
}

/**
 * org_gnome_shell_introspect_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeShellIntrospect interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_shell_introspect_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "animations-enabled");
  g_object_class_override_property (klass, property_id_begin++, "screen-size");
  g_object_class_override_property (klass, property_id_begin++, "version");
  return property_id_begin - 1;
}


inline static void
org_gnome_shell_introspect_signal_marshal_running_applications_changed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_introspect_signal_marshal_windows_changed (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  g_cclosure_marshal_VOID__VOID (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_introspect_method_marshal_get_running_applications (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_introspect_method_marshal_get_windows (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeShellIntrospect:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Introspect.top_of_page">org.gnome.Shell.Introspect</link>.
 */

/**
 * OrgGnomeShellIntrospectIface:
 * @parent_iface: The parent interface.
 * @handle_get_running_applications: Handler for the #OrgGnomeShellIntrospect::handle-get-running-applications signal.
 * @handle_get_windows: Handler for the #OrgGnomeShellIntrospect::handle-get-windows signal.
 * @get_animations_enabled: Getter for the #OrgGnomeShellIntrospect:animations-enabled property.
 * @get_screen_size: Getter for the #OrgGnomeShellIntrospect:screen-size property.
 * @get_version: Getter for the #OrgGnomeShellIntrospect:version property.
 * @running_applications_changed: Handler for the #OrgGnomeShellIntrospect::running-applications-changed signal.
 * @windows_changed: Handler for the #OrgGnomeShellIntrospect::windows-changed signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Introspect.top_of_page">org.gnome.Shell.Introspect</link>.
 */

typedef OrgGnomeShellIntrospectIface OrgGnomeShellIntrospectInterface;
G_DEFINE_INTERFACE (OrgGnomeShellIntrospect, org_gnome_shell_introspect, G_TYPE_OBJECT)

static void
org_gnome_shell_introspect_default_init (OrgGnomeShellIntrospectIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeShellIntrospect::handle-get-running-applications:
   * @object: A #OrgGnomeShellIntrospect.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetRunningApplications">GetRunningApplications()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_introspect_complete_get_running_applications() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-running-applications",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIntrospectIface, handle_get_running_applications),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_introspect_method_marshal_get_running_applications,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeShellIntrospect::handle-get-windows:
   * @object: A #OrgGnomeShellIntrospect.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetWindows">GetWindows()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_introspect_complete_get_windows() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-get-windows",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIntrospectIface, handle_get_windows),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_introspect_method_marshal_get_windows,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGnomeShellIntrospect::running-applications-changed:
   * @object: A #OrgGnomeShellIntrospect.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Shell-Introspect.RunningApplicationsChanged">"RunningApplicationsChanged"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_SHELL_INTROSPECT_SIGNALS[_ORG_GNOME_SHELL_INTROSPECT_RUNNING_APPLICATIONS_CHANGED] =
    g_signal_new ("running-applications-changed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeShellIntrospectIface, running_applications_changed),
      NULL,
      NULL,
      org_gnome_shell_introspect_signal_marshal_running_applications_changed,
      G_TYPE_NONE,
      0);

  /**
   * OrgGnomeShellIntrospect::windows-changed:
   * @object: A #OrgGnomeShellIntrospect.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Shell-Introspect.WindowsChanged">"WindowsChanged"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_SHELL_INTROSPECT_SIGNALS[_ORG_GNOME_SHELL_INTROSPECT_WINDOWS_CHANGED] =
    g_signal_new ("windows-changed",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeShellIntrospectIface, windows_changed),
      NULL,
      NULL,
      org_gnome_shell_introspect_signal_marshal_windows_changed,
      G_TYPE_NONE,
      0);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeShellIntrospect:animations-enabled:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Shell-Introspect.AnimationsEnabled">"AnimationsEnabled"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("animations-enabled", "AnimationsEnabled", "AnimationsEnabled", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgGnomeShellIntrospect:screen-size:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Shell-Introspect.ScreenSize">"ScreenSize"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_variant ("screen-size", "ScreenSize", "ScreenSize", G_VARIANT_TYPE ("(ii)"), NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgGnomeShellIntrospect:version:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Shell-Introspect.version">"version"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_uint ("version", "version", "version", 0, G_MAXUINT32, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_shell_introspect_get_animations_enabled: (skip)
 * @object: A #OrgGnomeShellIntrospect.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Shell-Introspect.AnimationsEnabled">"AnimationsEnabled"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_gnome_shell_introspect_get_animations_enabled (OrgGnomeShellIntrospect *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_SHELL_INTROSPECT (object), FALSE);

  return ORG_GNOME_SHELL_INTROSPECT_GET_IFACE (object)->get_animations_enabled (object);
}

/**
 * org_gnome_shell_introspect_set_animations_enabled: (skip)
 * @object: A #OrgGnomeShellIntrospect.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Shell-Introspect.AnimationsEnabled">"AnimationsEnabled"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_shell_introspect_set_animations_enabled (OrgGnomeShellIntrospect *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "animations-enabled", value, NULL);
}

/**
 * org_gnome_shell_introspect_get_screen_size: (skip)
 * @object: A #OrgGnomeShellIntrospect.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Shell-Introspect.ScreenSize">"ScreenSize"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_gnome_shell_introspect_dup_screen_size() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
GVariant *
org_gnome_shell_introspect_get_screen_size (OrgGnomeShellIntrospect *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_SHELL_INTROSPECT (object), NULL);

  return ORG_GNOME_SHELL_INTROSPECT_GET_IFACE (object)->get_screen_size (object);
}

/**
 * org_gnome_shell_introspect_dup_screen_size: (skip)
 * @object: A #OrgGnomeShellIntrospect.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-gnome-Shell-Introspect.ScreenSize">"ScreenSize"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_variant_unref().
 */
GVariant *
org_gnome_shell_introspect_dup_screen_size (OrgGnomeShellIntrospect *object)
{
  GVariant *value;
  g_object_get (G_OBJECT (object), "screen-size", &value, NULL);
  return value;
}

/**
 * org_gnome_shell_introspect_set_screen_size: (skip)
 * @object: A #OrgGnomeShellIntrospect.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Shell-Introspect.ScreenSize">"ScreenSize"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_shell_introspect_set_screen_size (OrgGnomeShellIntrospect *object, GVariant *value)
{
  g_object_set (G_OBJECT (object), "screen-size", value, NULL);
}

/**
 * org_gnome_shell_introspect_get_version: (skip)
 * @object: A #OrgGnomeShellIntrospect.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Shell-Introspect.version">"version"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
guint 
org_gnome_shell_introspect_get_version (OrgGnomeShellIntrospect *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_SHELL_INTROSPECT (object), 0);

  return ORG_GNOME_SHELL_INTROSPECT_GET_IFACE (object)->get_version (object);
}

/**
 * org_gnome_shell_introspect_set_version: (skip)
 * @object: A #OrgGnomeShellIntrospect.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Shell-Introspect.version">"version"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_shell_introspect_set_version (OrgGnomeShellIntrospect *object, guint value)
{
  g_object_set (G_OBJECT (object), "version", value, NULL);
}

/**
 * org_gnome_shell_introspect_emit_running_applications_changed:
 * @object: A #OrgGnomeShellIntrospect.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Shell-Introspect.RunningApplicationsChanged">"RunningApplicationsChanged"</link> D-Bus signal.
 */
void
org_gnome_shell_introspect_emit_running_applications_changed (
    OrgGnomeShellIntrospect *object)
{
  g_signal_emit (object, _ORG_GNOME_SHELL_INTROSPECT_SIGNALS[_ORG_GNOME_SHELL_INTROSPECT_RUNNING_APPLICATIONS_CHANGED], 0);
}

/**
 * org_gnome_shell_introspect_emit_windows_changed:
 * @object: A #OrgGnomeShellIntrospect.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Shell-Introspect.WindowsChanged">"WindowsChanged"</link> D-Bus signal.
 */
void
org_gnome_shell_introspect_emit_windows_changed (
    OrgGnomeShellIntrospect *object)
{
  g_signal_emit (object, _ORG_GNOME_SHELL_INTROSPECT_SIGNALS[_ORG_GNOME_SHELL_INTROSPECT_WINDOWS_CHANGED], 0);
}

/**
 * org_gnome_shell_introspect_call_get_running_applications:
 * @proxy: A #OrgGnomeShellIntrospectProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetRunningApplications">GetRunningApplications()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_introspect_call_get_running_applications_finish() to get the result of the operation.
 *
 * See org_gnome_shell_introspect_call_get_running_applications_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_introspect_call_get_running_applications (
    OrgGnomeShellIntrospect *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetRunningApplications",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_introspect_call_get_running_applications_finish:
 * @proxy: A #OrgGnomeShellIntrospectProxy.
 * @out_apps: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_introspect_call_get_running_applications().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_introspect_call_get_running_applications().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_introspect_call_get_running_applications_finish (
    OrgGnomeShellIntrospect *proxy,
    GVariant ** out_apps,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{sa{sv}})",
                 out_apps);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_introspect_call_get_running_applications_sync:
 * @proxy: A #OrgGnomeShellIntrospectProxy.
 * @out_apps: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetRunningApplications">GetRunningApplications()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_introspect_call_get_running_applications() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_introspect_call_get_running_applications_sync (
    OrgGnomeShellIntrospect *proxy,
    GVariant ** out_apps,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetRunningApplications",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{sa{sv}})",
                 out_apps);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_introspect_call_get_windows:
 * @proxy: A #OrgGnomeShellIntrospectProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetWindows">GetWindows()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_introspect_call_get_windows_finish() to get the result of the operation.
 *
 * See org_gnome_shell_introspect_call_get_windows_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_introspect_call_get_windows (
    OrgGnomeShellIntrospect *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GetWindows",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_introspect_call_get_windows_finish:
 * @proxy: A #OrgGnomeShellIntrospectProxy.
 * @out_windows: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_introspect_call_get_windows().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_introspect_call_get_windows().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_introspect_call_get_windows_finish (
    OrgGnomeShellIntrospect *proxy,
    GVariant ** out_windows,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{ta{sv}})",
                 out_windows);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_introspect_call_get_windows_sync:
 * @proxy: A #OrgGnomeShellIntrospectProxy.
 * @out_windows: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetWindows">GetWindows()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_introspect_call_get_windows() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_introspect_call_get_windows_sync (
    OrgGnomeShellIntrospect *proxy,
    GVariant ** out_windows,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GetWindows",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{ta{sv}})",
                 out_windows);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_introspect_complete_get_running_applications:
 * @object: A #OrgGnomeShellIntrospect.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @apps: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetRunningApplications">GetRunningApplications()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_introspect_complete_get_running_applications (
    OrgGnomeShellIntrospect *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GVariant *apps)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a{sa{sv}})",
                   apps));
}

/**
 * org_gnome_shell_introspect_complete_get_windows:
 * @object: A #OrgGnomeShellIntrospect.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @windows: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Introspect.GetWindows">GetWindows()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_introspect_complete_get_windows (
    OrgGnomeShellIntrospect *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GVariant *windows)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a{ta{sv}})",
                   windows));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeShellIntrospectProxy:
 *
 * The #OrgGnomeShellIntrospectProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeShellIntrospectProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeShellIntrospectProxy.
 */

struct _OrgGnomeShellIntrospectProxyPrivate
{
  GData *qdata;
};

static void org_gnome_shell_introspect_proxy_iface_init (OrgGnomeShellIntrospectIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellIntrospectProxy, org_gnome_shell_introspect_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeShellIntrospectProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_INTROSPECT, org_gnome_shell_introspect_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellIntrospectProxy, org_gnome_shell_introspect_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_INTROSPECT, org_gnome_shell_introspect_proxy_iface_init))

#endif
static void
org_gnome_shell_introspect_proxy_finalize (GObject *object)
{
  OrgGnomeShellIntrospectProxy *proxy = ORG_GNOME_SHELL_INTROSPECT_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_shell_introspect_proxy_parent_class)->finalize (object);
}

static void
org_gnome_shell_introspect_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_shell_introspect_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_shell_introspect_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Shell.Introspect: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_shell_introspect_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_shell_introspect_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Shell.Introspect", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_shell_introspect_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_shell_introspect_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_shell_introspect_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_SHELL_INTROSPECT);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SHELL_INTROSPECT);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_shell_introspect_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeShellIntrospectProxy *proxy = ORG_GNOME_SHELL_INTROSPECT_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_introspect_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_introspect_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static gboolean 
org_gnome_shell_introspect_proxy_get_animations_enabled (OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectProxy *proxy = ORG_GNOME_SHELL_INTROSPECT_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "AnimationsEnabled");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static GVariant *
org_gnome_shell_introspect_proxy_get_screen_size (OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectProxy *proxy = ORG_GNOME_SHELL_INTROSPECT_PROXY (object);
  GVariant *variant;
  GVariant *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "ScreenSize");
  value = variant;
  if (variant != NULL)
    g_variant_unref (variant);
  return value;
}

static guint 
org_gnome_shell_introspect_proxy_get_version (OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectProxy *proxy = ORG_GNOME_SHELL_INTROSPECT_PROXY (object);
  GVariant *variant;
  guint value = 0;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "version");
  if (variant != NULL)
    {
      value = g_variant_get_uint32 (variant);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_gnome_shell_introspect_proxy_init (OrgGnomeShellIntrospectProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_shell_introspect_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_SHELL_INTROSPECT_PROXY, OrgGnomeShellIntrospectProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_shell_introspect_interface_info ());
}

static void
org_gnome_shell_introspect_proxy_class_init (OrgGnomeShellIntrospectProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_shell_introspect_proxy_finalize;
  gobject_class->get_property = org_gnome_shell_introspect_proxy_get_property;
  gobject_class->set_property = org_gnome_shell_introspect_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_shell_introspect_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_shell_introspect_proxy_g_properties_changed;

  org_gnome_shell_introspect_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeShellIntrospectProxyPrivate));
#endif
}

static void
org_gnome_shell_introspect_proxy_iface_init (OrgGnomeShellIntrospectIface *iface)
{
  iface->get_animations_enabled = org_gnome_shell_introspect_proxy_get_animations_enabled;
  iface->get_screen_size = org_gnome_shell_introspect_proxy_get_screen_size;
  iface->get_version = org_gnome_shell_introspect_proxy_get_version;
}

/**
 * org_gnome_shell_introspect_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Introspect.top_of_page">org.gnome.Shell.Introspect</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_introspect_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_shell_introspect_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_shell_introspect_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SHELL_INTROSPECT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Introspect", NULL);
}

/**
 * org_gnome_shell_introspect_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_introspect_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_shell_introspect_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeShellIntrospectProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellIntrospect *
org_gnome_shell_introspect_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SHELL_INTROSPECT (ret);
  else
    return NULL;
}

/**
 * org_gnome_shell_introspect_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Introspect.top_of_page">org.gnome.Shell.Introspect</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_introspect_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeShellIntrospectProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellIntrospect *
org_gnome_shell_introspect_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SHELL_INTROSPECT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Introspect", NULL);
  if (ret != NULL)
    return ORG_GNOME_SHELL_INTROSPECT (ret);
  else
    return NULL;
}


/**
 * org_gnome_shell_introspect_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_shell_introspect_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_introspect_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_shell_introspect_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_shell_introspect_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SHELL_INTROSPECT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Introspect", NULL);
}

/**
 * org_gnome_shell_introspect_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_introspect_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_shell_introspect_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeShellIntrospectProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellIntrospect *
org_gnome_shell_introspect_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SHELL_INTROSPECT (ret);
  else
    return NULL;
}

/**
 * org_gnome_shell_introspect_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_shell_introspect_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_introspect_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeShellIntrospectProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellIntrospect *
org_gnome_shell_introspect_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SHELL_INTROSPECT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Introspect", NULL);
  if (ret != NULL)
    return ORG_GNOME_SHELL_INTROSPECT (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeShellIntrospectSkeleton:
 *
 * The #OrgGnomeShellIntrospectSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeShellIntrospectSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeShellIntrospectSkeleton.
 */

struct _OrgGnomeShellIntrospectSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_shell_introspect_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_SHELL_INTROSPECT);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SHELL_INTROSPECT);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_shell_introspect_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_introspect_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_shell_introspect_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_introspect_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_shell_introspect_skeleton_vtable =
{
  _org_gnome_shell_introspect_skeleton_handle_method_call,
  _org_gnome_shell_introspect_skeleton_handle_get_property,
  _org_gnome_shell_introspect_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_shell_introspect_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_shell_introspect_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_shell_introspect_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_shell_introspect_skeleton_vtable;
}

static GVariant *
org_gnome_shell_introspect_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_shell_introspect_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_shell_introspect_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_shell_introspect_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_shell_introspect_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Shell.Introspect", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_shell_introspect_emit_changed (gpointer user_data);

static void
org_gnome_shell_introspect_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_shell_introspect_emit_changed (skeleton);
}

static void
_org_gnome_shell_introspect_on_signal_running_applications_changed (
    OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Shell.Introspect", "RunningApplicationsChanged",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_shell_introspect_on_signal_windows_changed (
    OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("()"));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Shell.Introspect", "WindowsChanged",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gnome_shell_introspect_skeleton_iface_init (OrgGnomeShellIntrospectIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellIntrospectSkeleton, org_gnome_shell_introspect_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeShellIntrospectSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_INTROSPECT, org_gnome_shell_introspect_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellIntrospectSkeleton, org_gnome_shell_introspect_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_INTROSPECT, org_gnome_shell_introspect_skeleton_iface_init))

#endif
static void
org_gnome_shell_introspect_skeleton_finalize (GObject *object)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);
  guint n;
  for (n = 0; n < 3; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_shell_introspect_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_shell_introspect_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_shell_introspect_emit_changed (gpointer user_data)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Shell.Introspect",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_shell_introspect_schedule_emit_changed (OrgGnomeShellIntrospectSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_shell_introspect_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_shell_introspect_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_shell_introspect_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_shell_introspect_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_shell_introspect_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_shell_introspect_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_shell_introspect_skeleton_init (OrgGnomeShellIntrospectSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_shell_introspect_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_SHELL_INTROSPECT_SKELETON, OrgGnomeShellIntrospectSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 3);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[1], G_TYPE_VARIANT);
  g_value_init (&skeleton->priv->properties[2], G_TYPE_UINT);
}

static gboolean 
org_gnome_shell_introspect_skeleton_get_animations_enabled (OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static GVariant *
org_gnome_shell_introspect_skeleton_get_screen_size (OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);
  GVariant *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_variant (&(skeleton->priv->properties[1]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static guint 
org_gnome_shell_introspect_skeleton_get_version (OrgGnomeShellIntrospect *object)
{
  OrgGnomeShellIntrospectSkeleton *skeleton = ORG_GNOME_SHELL_INTROSPECT_SKELETON (object);
  guint value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_uint (&(skeleton->priv->properties[2]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_shell_introspect_skeleton_class_init (OrgGnomeShellIntrospectSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_shell_introspect_skeleton_finalize;
  gobject_class->get_property = org_gnome_shell_introspect_skeleton_get_property;
  gobject_class->set_property = org_gnome_shell_introspect_skeleton_set_property;
  gobject_class->notify       = org_gnome_shell_introspect_skeleton_notify;


  org_gnome_shell_introspect_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_shell_introspect_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_shell_introspect_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_shell_introspect_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_shell_introspect_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeShellIntrospectSkeletonPrivate));
#endif
}

static void
org_gnome_shell_introspect_skeleton_iface_init (OrgGnomeShellIntrospectIface *iface)
{
  iface->running_applications_changed = _org_gnome_shell_introspect_on_signal_running_applications_changed;
  iface->windows_changed = _org_gnome_shell_introspect_on_signal_windows_changed;
  iface->get_animations_enabled = org_gnome_shell_introspect_skeleton_get_animations_enabled;
  iface->get_screen_size = org_gnome_shell_introspect_skeleton_get_screen_size;
  iface->get_version = org_gnome_shell_introspect_skeleton_get_version;
}

/**
 * org_gnome_shell_introspect_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Introspect.top_of_page">org.gnome.Shell.Introspect</link>.
 *
 * Returns: (transfer full) (type OrgGnomeShellIntrospectSkeleton): The skeleton object.
 */
OrgGnomeShellIntrospect *
org_gnome_shell_introspect_skeleton_new (void)
{
  return ORG_GNOME_SHELL_INTROSPECT (g_object_new (TYPE_ORG_GNOME_SHELL_INTROSPECT_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Shell.Screenshot
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeShellScreenshot
 * @title: OrgGnomeShellScreenshot
 * @short_description: Generated C code for the org.gnome.Shell.Screenshot D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Shell-Screenshot.top_of_page">org.gnome.Shell.Screenshot</link> D-Bus interface in C.
 */

/* ---- Introspection data for org.gnome.Shell.Screenshot ---- */

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_interactive_screenshot_OUT_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_interactive_screenshot_OUT_ARG_uri =
{
  {
    -1,
    (gchar *) "uri",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_interactive_screenshot_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_interactive_screenshot_OUT_ARG_success.parent_struct,
  &_org_gnome_shell_screenshot_method_info_interactive_screenshot_OUT_ARG_uri.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_screenshot_method_info_interactive_screenshot =
{
  {
    -1,
    (gchar *) "InteractiveScreenshot",
    NULL,
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_interactive_screenshot_OUT_ARG_pointers,
    NULL
  },
  "handle-interactive-screenshot",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_include_cursor =
{
  {
    -1,
    (gchar *) "include_cursor",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_flash =
{
  {
    -1,
    (gchar *) "flash",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_filename =
{
  {
    -1,
    (gchar *) "filename",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_include_cursor.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_flash.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_filename.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_OUT_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_OUT_ARG_filename_used =
{
  {
    -1,
    (gchar *) "filename_used",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_screenshot_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_screenshot_OUT_ARG_success.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_OUT_ARG_filename_used.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_screenshot_method_info_screenshot =
{
  {
    -1,
    (gchar *) "Screenshot",
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_screenshot_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_screenshot_OUT_ARG_pointers,
    NULL
  },
  "handle-screenshot",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_include_frame =
{
  {
    -1,
    (gchar *) "include_frame",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_include_cursor =
{
  {
    -1,
    (gchar *) "include_cursor",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_flash =
{
  {
    -1,
    (gchar *) "flash",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_filename =
{
  {
    -1,
    (gchar *) "filename",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_include_frame.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_include_cursor.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_flash.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_filename.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_window_OUT_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_window_OUT_ARG_filename_used =
{
  {
    -1,
    (gchar *) "filename_used",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_screenshot_window_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_screenshot_window_OUT_ARG_success.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_window_OUT_ARG_filename_used.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_screenshot_method_info_screenshot_window =
{
  {
    -1,
    (gchar *) "ScreenshotWindow",
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_screenshot_window_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_screenshot_window_OUT_ARG_pointers,
    NULL
  },
  "handle-screenshot-window",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_x =
{
  {
    -1,
    (gchar *) "x",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_y =
{
  {
    -1,
    (gchar *) "y",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_width =
{
  {
    -1,
    (gchar *) "width",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_height =
{
  {
    -1,
    (gchar *) "height",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_flash =
{
  {
    -1,
    (gchar *) "flash",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_filename =
{
  {
    -1,
    (gchar *) "filename",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_x.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_y.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_width.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_height.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_flash.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_filename.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_OUT_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_screenshot_area_OUT_ARG_filename_used =
{
  {
    -1,
    (gchar *) "filename_used",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_screenshot_area_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_screenshot_area_OUT_ARG_success.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_area_OUT_ARG_filename_used.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_screenshot_method_info_screenshot_area =
{
  {
    -1,
    (gchar *) "ScreenshotArea",
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_screenshot_area_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_screenshot_area_OUT_ARG_pointers,
    NULL
  },
  "handle-screenshot-area",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_x =
{
  {
    -1,
    (gchar *) "x",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_y =
{
  {
    -1,
    (gchar *) "y",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_width =
{
  {
    -1,
    (gchar *) "width",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_height =
{
  {
    -1,
    (gchar *) "height",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_x.parent_struct,
  &_org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_y.parent_struct,
  &_org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_width.parent_struct,
  &_org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_height.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_screenshot_method_info_flash_area =
{
  {
    -1,
    (gchar *) "FlashArea",
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_flash_area_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-flash-area",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_x =
{
  {
    -1,
    (gchar *) "x",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_y =
{
  {
    -1,
    (gchar *) "y",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_width =
{
  {
    -1,
    (gchar *) "width",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_height =
{
  {
    -1,
    (gchar *) "height",
    (gchar *) "i",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_x.parent_struct,
  &_org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_y.parent_struct,
  &_org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_width.parent_struct,
  &_org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_height.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_screenshot_method_info_select_area =
{
  {
    -1,
    (gchar *) "SelectArea",
    NULL,
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_select_area_OUT_ARG_pointers,
    NULL
  },
  "handle-select-area",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_screenshot_method_info_pick_color_OUT_ARG_result =
{
  {
    -1,
    (gchar *) "result",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_screenshot_method_info_pick_color_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_pick_color_OUT_ARG_result.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_screenshot_method_info_pick_color =
{
  {
    -1,
    (gchar *) "PickColor",
    NULL,
    (GDBusArgInfo **) &_org_gnome_shell_screenshot_method_info_pick_color_OUT_ARG_pointers,
    NULL
  },
  "handle-pick-color",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_shell_screenshot_method_info_pointers[] =
{
  &_org_gnome_shell_screenshot_method_info_interactive_screenshot.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_window.parent_struct,
  &_org_gnome_shell_screenshot_method_info_screenshot_area.parent_struct,
  &_org_gnome_shell_screenshot_method_info_flash_area.parent_struct,
  &_org_gnome_shell_screenshot_method_info_select_area.parent_struct,
  &_org_gnome_shell_screenshot_method_info_pick_color.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_shell_screenshot_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Shell.Screenshot",
    (GDBusMethodInfo **) &_org_gnome_shell_screenshot_method_info_pointers,
    NULL,
    NULL,
    NULL
  },
  "org-gnome-shell-screenshot",
};


/**
 * org_gnome_shell_screenshot_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Shell-Screenshot.top_of_page">org.gnome.Shell.Screenshot</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_shell_screenshot_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_shell_screenshot_interface_info.parent_struct;
}

/**
 * org_gnome_shell_screenshot_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeShellScreenshot interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_shell_screenshot_override_properties (GObjectClass *klass G_GNUC_UNUSED, guint property_id_begin)
{
  return property_id_begin - 1;
}


inline static void
org_gnome_shell_screenshot_method_marshal_interactive_screenshot (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_screenshot_method_marshal_screenshot (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN_BOOLEAN_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_screenshot_method_marshal_screenshot_window (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_BOOLEAN_BOOLEAN_BOOLEAN_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_screenshot_method_marshal_screenshot_area (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_INT_INT_INT_BOOLEAN_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_screenshot_method_marshal_flash_area (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_INT_INT_INT_INT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_screenshot_method_marshal_select_area (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_screenshot_method_marshal_pick_color (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeShellScreenshot:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Screenshot.top_of_page">org.gnome.Shell.Screenshot</link>.
 */

/**
 * OrgGnomeShellScreenshotIface:
 * @parent_iface: The parent interface.
 * @handle_flash_area: Handler for the #OrgGnomeShellScreenshot::handle-flash-area signal.
 * @handle_interactive_screenshot: Handler for the #OrgGnomeShellScreenshot::handle-interactive-screenshot signal.
 * @handle_pick_color: Handler for the #OrgGnomeShellScreenshot::handle-pick-color signal.
 * @handle_screenshot: Handler for the #OrgGnomeShellScreenshot::handle-screenshot signal.
 * @handle_screenshot_area: Handler for the #OrgGnomeShellScreenshot::handle-screenshot-area signal.
 * @handle_screenshot_window: Handler for the #OrgGnomeShellScreenshot::handle-screenshot-window signal.
 * @handle_select_area: Handler for the #OrgGnomeShellScreenshot::handle-select-area signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Screenshot.top_of_page">org.gnome.Shell.Screenshot</link>.
 */

typedef OrgGnomeShellScreenshotIface OrgGnomeShellScreenshotInterface;
G_DEFINE_INTERFACE (OrgGnomeShellScreenshot, org_gnome_shell_screenshot, G_TYPE_OBJECT)

static void
org_gnome_shell_screenshot_default_init (OrgGnomeShellScreenshotIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeShellScreenshot::handle-interactive-screenshot:
   * @object: A #OrgGnomeShellScreenshot.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.InteractiveScreenshot">InteractiveScreenshot()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_screenshot_complete_interactive_screenshot() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-interactive-screenshot",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellScreenshotIface, handle_interactive_screenshot),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_screenshot_method_marshal_interactive_screenshot,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeShellScreenshot::handle-screenshot:
   * @object: A #OrgGnomeShellScreenshot.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_include_cursor: Argument passed by remote caller.
   * @arg_flash: Argument passed by remote caller.
   * @arg_filename: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.Screenshot">Screenshot()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_screenshot_complete_screenshot() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-screenshot",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellScreenshotIface, handle_screenshot),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_screenshot_method_marshal_screenshot,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_STRING);

  /**
   * OrgGnomeShellScreenshot::handle-screenshot-window:
   * @object: A #OrgGnomeShellScreenshot.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_include_frame: Argument passed by remote caller.
   * @arg_include_cursor: Argument passed by remote caller.
   * @arg_flash: Argument passed by remote caller.
   * @arg_filename: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotWindow">ScreenshotWindow()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_screenshot_complete_screenshot_window() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-screenshot-window",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellScreenshotIface, handle_screenshot_window),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_screenshot_method_marshal_screenshot_window,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_STRING);

  /**
   * OrgGnomeShellScreenshot::handle-screenshot-area:
   * @object: A #OrgGnomeShellScreenshot.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_x: Argument passed by remote caller.
   * @arg_y: Argument passed by remote caller.
   * @arg_width: Argument passed by remote caller.
   * @arg_height: Argument passed by remote caller.
   * @arg_flash: Argument passed by remote caller.
   * @arg_filename: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotArea">ScreenshotArea()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_screenshot_complete_screenshot_area() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-screenshot-area",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellScreenshotIface, handle_screenshot_area),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_screenshot_method_marshal_screenshot_area,
    G_TYPE_BOOLEAN,
    7,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_BOOLEAN, G_TYPE_STRING);

  /**
   * OrgGnomeShellScreenshot::handle-flash-area:
   * @object: A #OrgGnomeShellScreenshot.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_x: Argument passed by remote caller.
   * @arg_y: Argument passed by remote caller.
   * @arg_width: Argument passed by remote caller.
   * @arg_height: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.FlashArea">FlashArea()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_screenshot_complete_flash_area() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-flash-area",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellScreenshotIface, handle_flash_area),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_screenshot_method_marshal_flash_area,
    G_TYPE_BOOLEAN,
    5,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

  /**
   * OrgGnomeShellScreenshot::handle-select-area:
   * @object: A #OrgGnomeShellScreenshot.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.SelectArea">SelectArea()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_screenshot_complete_select_area() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-select-area",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellScreenshotIface, handle_select_area),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_screenshot_method_marshal_select_area,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeShellScreenshot::handle-pick-color:
   * @object: A #OrgGnomeShellScreenshot.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.PickColor">PickColor()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_screenshot_complete_pick_color() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-pick-color",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellScreenshotIface, handle_pick_color),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_screenshot_method_marshal_pick_color,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

}

/**
 * org_gnome_shell_screenshot_call_interactive_screenshot:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.InteractiveScreenshot">InteractiveScreenshot()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_call_interactive_screenshot_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_call_interactive_screenshot_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_screenshot_call_interactive_screenshot (
    OrgGnomeShellScreenshot *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "InteractiveScreenshot",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_screenshot_call_interactive_screenshot_finish:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_uri: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_call_interactive_screenshot().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_screenshot_call_interactive_screenshot().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_interactive_screenshot_finish (
    OrgGnomeShellScreenshot *proxy,
    gboolean* out_success,
    gchar ** out_uri,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_uri);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_interactive_screenshot_sync:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_uri: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.InteractiveScreenshot">InteractiveScreenshot()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_call_interactive_screenshot() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_interactive_screenshot_sync (
    OrgGnomeShellScreenshot *proxy,
    gboolean* out_success,
    gchar ** out_uri,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "InteractiveScreenshot",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_uri);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_screenshot:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_include_cursor: Argument to pass with the method invocation.
 * @arg_flash: Argument to pass with the method invocation.
 * @arg_filename: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.Screenshot">Screenshot()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_call_screenshot_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_call_screenshot_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_screenshot_call_screenshot (
    OrgGnomeShellScreenshot *proxy,
    gboolean arg_include_cursor,
    gboolean arg_flash,
    const gchar *arg_filename,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Screenshot",
    g_variant_new ("(bbs)",
                   arg_include_cursor,
                   arg_flash,
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_screenshot_call_screenshot_finish:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_filename_used: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_call_screenshot().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_screenshot_call_screenshot().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_screenshot_finish (
    OrgGnomeShellScreenshot *proxy,
    gboolean* out_success,
    gchar ** out_filename_used,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_filename_used);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_screenshot_sync:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_include_cursor: Argument to pass with the method invocation.
 * @arg_flash: Argument to pass with the method invocation.
 * @arg_filename: Argument to pass with the method invocation.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_filename_used: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.Screenshot">Screenshot()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_call_screenshot() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_screenshot_sync (
    OrgGnomeShellScreenshot *proxy,
    gboolean arg_include_cursor,
    gboolean arg_flash,
    const gchar *arg_filename,
    gboolean* out_success,
    gchar ** out_filename_used,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Screenshot",
    g_variant_new ("(bbs)",
                   arg_include_cursor,
                   arg_flash,
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_filename_used);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_screenshot_window:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_include_frame: Argument to pass with the method invocation.
 * @arg_include_cursor: Argument to pass with the method invocation.
 * @arg_flash: Argument to pass with the method invocation.
 * @arg_filename: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotWindow">ScreenshotWindow()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_call_screenshot_window_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_call_screenshot_window_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_screenshot_call_screenshot_window (
    OrgGnomeShellScreenshot *proxy,
    gboolean arg_include_frame,
    gboolean arg_include_cursor,
    gboolean arg_flash,
    const gchar *arg_filename,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ScreenshotWindow",
    g_variant_new ("(bbbs)",
                   arg_include_frame,
                   arg_include_cursor,
                   arg_flash,
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_screenshot_call_screenshot_window_finish:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_filename_used: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_call_screenshot_window().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_screenshot_call_screenshot_window().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_screenshot_window_finish (
    OrgGnomeShellScreenshot *proxy,
    gboolean* out_success,
    gchar ** out_filename_used,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_filename_used);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_screenshot_window_sync:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_include_frame: Argument to pass with the method invocation.
 * @arg_include_cursor: Argument to pass with the method invocation.
 * @arg_flash: Argument to pass with the method invocation.
 * @arg_filename: Argument to pass with the method invocation.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_filename_used: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotWindow">ScreenshotWindow()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_call_screenshot_window() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_screenshot_window_sync (
    OrgGnomeShellScreenshot *proxy,
    gboolean arg_include_frame,
    gboolean arg_include_cursor,
    gboolean arg_flash,
    const gchar *arg_filename,
    gboolean* out_success,
    gchar ** out_filename_used,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ScreenshotWindow",
    g_variant_new ("(bbbs)",
                   arg_include_frame,
                   arg_include_cursor,
                   arg_flash,
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_filename_used);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_screenshot_area:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @arg_width: Argument to pass with the method invocation.
 * @arg_height: Argument to pass with the method invocation.
 * @arg_flash: Argument to pass with the method invocation.
 * @arg_filename: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotArea">ScreenshotArea()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_call_screenshot_area_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_call_screenshot_area_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_screenshot_call_screenshot_area (
    OrgGnomeShellScreenshot *proxy,
    gint arg_x,
    gint arg_y,
    gint arg_width,
    gint arg_height,
    gboolean arg_flash,
    const gchar *arg_filename,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ScreenshotArea",
    g_variant_new ("(iiiibs)",
                   arg_x,
                   arg_y,
                   arg_width,
                   arg_height,
                   arg_flash,
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_screenshot_call_screenshot_area_finish:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_filename_used: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_call_screenshot_area().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_screenshot_call_screenshot_area().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_screenshot_area_finish (
    OrgGnomeShellScreenshot *proxy,
    gboolean* out_success,
    gchar ** out_filename_used,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_filename_used);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_screenshot_area_sync:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @arg_width: Argument to pass with the method invocation.
 * @arg_height: Argument to pass with the method invocation.
 * @arg_flash: Argument to pass with the method invocation.
 * @arg_filename: Argument to pass with the method invocation.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_filename_used: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotArea">ScreenshotArea()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_call_screenshot_area() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_screenshot_area_sync (
    OrgGnomeShellScreenshot *proxy,
    gint arg_x,
    gint arg_y,
    gint arg_width,
    gint arg_height,
    gboolean arg_flash,
    const gchar *arg_filename,
    gboolean* out_success,
    gchar ** out_filename_used,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ScreenshotArea",
    g_variant_new ("(iiiibs)",
                   arg_x,
                   arg_y,
                   arg_width,
                   arg_height,
                   arg_flash,
                   arg_filename),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_filename_used);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_flash_area:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @arg_width: Argument to pass with the method invocation.
 * @arg_height: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.FlashArea">FlashArea()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_call_flash_area_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_call_flash_area_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_screenshot_call_flash_area (
    OrgGnomeShellScreenshot *proxy,
    gint arg_x,
    gint arg_y,
    gint arg_width,
    gint arg_height,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "FlashArea",
    g_variant_new ("(iiii)",
                   arg_x,
                   arg_y,
                   arg_width,
                   arg_height),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_screenshot_call_flash_area_finish:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_call_flash_area().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_screenshot_call_flash_area().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_flash_area_finish (
    OrgGnomeShellScreenshot *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_flash_area_sync:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @arg_x: Argument to pass with the method invocation.
 * @arg_y: Argument to pass with the method invocation.
 * @arg_width: Argument to pass with the method invocation.
 * @arg_height: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.FlashArea">FlashArea()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_call_flash_area() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_flash_area_sync (
    OrgGnomeShellScreenshot *proxy,
    gint arg_x,
    gint arg_y,
    gint arg_width,
    gint arg_height,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "FlashArea",
    g_variant_new ("(iiii)",
                   arg_x,
                   arg_y,
                   arg_width,
                   arg_height),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_select_area:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.SelectArea">SelectArea()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_call_select_area_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_call_select_area_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_screenshot_call_select_area (
    OrgGnomeShellScreenshot *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "SelectArea",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_screenshot_call_select_area_finish:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_x: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_y: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_width: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_height: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_call_select_area().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_screenshot_call_select_area().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_select_area_finish (
    OrgGnomeShellScreenshot *proxy,
    gint* out_x,
    gint* out_y,
    gint* out_width,
    gint* out_height,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(iiii)",
                 out_x,
                 out_y,
                 out_width,
                 out_height);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_select_area_sync:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_x: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_y: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_width: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_height: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.SelectArea">SelectArea()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_call_select_area() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_select_area_sync (
    OrgGnomeShellScreenshot *proxy,
    gint* out_x,
    gint* out_y,
    gint* out_width,
    gint* out_height,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "SelectArea",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(iiii)",
                 out_x,
                 out_y,
                 out_width,
                 out_height);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_pick_color:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.PickColor">PickColor()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_call_pick_color_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_call_pick_color_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_screenshot_call_pick_color (
    OrgGnomeShellScreenshot *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "PickColor",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_screenshot_call_pick_color_finish:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_result: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_call_pick_color().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_screenshot_call_pick_color().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_pick_color_finish (
    OrgGnomeShellScreenshot *proxy,
    GVariant ** out_result,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{sv})",
                 out_result);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_call_pick_color_sync:
 * @proxy: A #OrgGnomeShellScreenshotProxy.
 * @out_result: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.PickColor">PickColor()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_call_pick_color() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_screenshot_call_pick_color_sync (
    OrgGnomeShellScreenshot *proxy,
    GVariant ** out_result,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "PickColor",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@a{sv})",
                 out_result);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_screenshot_complete_interactive_screenshot:
 * @object: A #OrgGnomeShellScreenshot.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @success: Parameter to return.
 * @uri: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.InteractiveScreenshot">InteractiveScreenshot()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_screenshot_complete_interactive_screenshot (
    OrgGnomeShellScreenshot *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gboolean success,
    const gchar *uri)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bs)",
                   success,
                   uri));
}

/**
 * org_gnome_shell_screenshot_complete_screenshot:
 * @object: A #OrgGnomeShellScreenshot.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @success: Parameter to return.
 * @filename_used: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.Screenshot">Screenshot()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_screenshot_complete_screenshot (
    OrgGnomeShellScreenshot *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gboolean success,
    const gchar *filename_used)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bs)",
                   success,
                   filename_used));
}

/**
 * org_gnome_shell_screenshot_complete_screenshot_window:
 * @object: A #OrgGnomeShellScreenshot.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @success: Parameter to return.
 * @filename_used: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotWindow">ScreenshotWindow()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_screenshot_complete_screenshot_window (
    OrgGnomeShellScreenshot *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gboolean success,
    const gchar *filename_used)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bs)",
                   success,
                   filename_used));
}

/**
 * org_gnome_shell_screenshot_complete_screenshot_area:
 * @object: A #OrgGnomeShellScreenshot.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @success: Parameter to return.
 * @filename_used: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.ScreenshotArea">ScreenshotArea()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_screenshot_complete_screenshot_area (
    OrgGnomeShellScreenshot *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gboolean success,
    const gchar *filename_used)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bs)",
                   success,
                   filename_used));
}

/**
 * org_gnome_shell_screenshot_complete_flash_area:
 * @object: A #OrgGnomeShellScreenshot.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.FlashArea">FlashArea()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_screenshot_complete_flash_area (
    OrgGnomeShellScreenshot *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_shell_screenshot_complete_select_area:
 * @object: A #OrgGnomeShellScreenshot.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @x: Parameter to return.
 * @y: Parameter to return.
 * @width: Parameter to return.
 * @height: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.SelectArea">SelectArea()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_screenshot_complete_select_area (
    OrgGnomeShellScreenshot *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gint x,
    gint y,
    gint width,
    gint height)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(iiii)",
                   x,
                   y,
                   width,
                   height));
}

/**
 * org_gnome_shell_screenshot_complete_pick_color:
 * @object: A #OrgGnomeShellScreenshot.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @result: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell-Screenshot.PickColor">PickColor()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_screenshot_complete_pick_color (
    OrgGnomeShellScreenshot *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GVariant *result)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@a{sv})",
                   result));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeShellScreenshotProxy:
 *
 * The #OrgGnomeShellScreenshotProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeShellScreenshotProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeShellScreenshotProxy.
 */

struct _OrgGnomeShellScreenshotProxyPrivate
{
  GData *qdata;
};

static void org_gnome_shell_screenshot_proxy_iface_init (OrgGnomeShellScreenshotIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellScreenshotProxy, org_gnome_shell_screenshot_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeShellScreenshotProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_SCREENSHOT, org_gnome_shell_screenshot_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellScreenshotProxy, org_gnome_shell_screenshot_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_SCREENSHOT, org_gnome_shell_screenshot_proxy_iface_init))

#endif
static void
org_gnome_shell_screenshot_proxy_finalize (GObject *object)
{
  OrgGnomeShellScreenshotProxy *proxy = ORG_GNOME_SHELL_SCREENSHOT_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_shell_screenshot_proxy_parent_class)->finalize (object);
}

static void
org_gnome_shell_screenshot_proxy_get_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  GValue       *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_shell_screenshot_proxy_set_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  const GValue *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gnome_shell_screenshot_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_shell_screenshot_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_SHELL_SCREENSHOT);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SHELL_SCREENSHOT);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_shell_screenshot_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeShellScreenshotProxy *proxy = ORG_GNOME_SHELL_SCREENSHOT_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_screenshot_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_screenshot_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
org_gnome_shell_screenshot_proxy_init (OrgGnomeShellScreenshotProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_shell_screenshot_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_SHELL_SCREENSHOT_PROXY, OrgGnomeShellScreenshotProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_shell_screenshot_interface_info ());
}

static void
org_gnome_shell_screenshot_proxy_class_init (OrgGnomeShellScreenshotProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_shell_screenshot_proxy_finalize;
  gobject_class->get_property = org_gnome_shell_screenshot_proxy_get_property;
  gobject_class->set_property = org_gnome_shell_screenshot_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_shell_screenshot_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_shell_screenshot_proxy_g_properties_changed;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeShellScreenshotProxyPrivate));
#endif
}

static void
org_gnome_shell_screenshot_proxy_iface_init (OrgGnomeShellScreenshotIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_shell_screenshot_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Screenshot.top_of_page">org.gnome.Shell.Screenshot</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_shell_screenshot_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SHELL_SCREENSHOT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Screenshot", NULL);
}

/**
 * org_gnome_shell_screenshot_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_shell_screenshot_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeShellScreenshotProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellScreenshot *
org_gnome_shell_screenshot_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SHELL_SCREENSHOT (ret);
  else
    return NULL;
}

/**
 * org_gnome_shell_screenshot_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Screenshot.top_of_page">org.gnome.Shell.Screenshot</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeShellScreenshotProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellScreenshot *
org_gnome_shell_screenshot_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SHELL_SCREENSHOT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Screenshot", NULL);
  if (ret != NULL)
    return ORG_GNOME_SHELL_SCREENSHOT (ret);
  else
    return NULL;
}


/**
 * org_gnome_shell_screenshot_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_shell_screenshot_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_screenshot_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_shell_screenshot_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_shell_screenshot_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SHELL_SCREENSHOT_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Screenshot", NULL);
}

/**
 * org_gnome_shell_screenshot_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_screenshot_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_shell_screenshot_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeShellScreenshotProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellScreenshot *
org_gnome_shell_screenshot_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SHELL_SCREENSHOT (ret);
  else
    return NULL;
}

/**
 * org_gnome_shell_screenshot_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_shell_screenshot_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_screenshot_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeShellScreenshotProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShellScreenshot *
org_gnome_shell_screenshot_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SHELL_SCREENSHOT_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell.Screenshot", NULL);
  if (ret != NULL)
    return ORG_GNOME_SHELL_SCREENSHOT (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeShellScreenshotSkeleton:
 *
 * The #OrgGnomeShellScreenshotSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeShellScreenshotSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeShellScreenshotSkeleton.
 */

struct _OrgGnomeShellScreenshotSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_shell_screenshot_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeShellScreenshotSkeleton *skeleton = ORG_GNOME_SHELL_SCREENSHOT_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_SHELL_SCREENSHOT);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SHELL_SCREENSHOT);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_shell_screenshot_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeShellScreenshotSkeleton *skeleton = ORG_GNOME_SHELL_SCREENSHOT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_screenshot_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_shell_screenshot_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeShellScreenshotSkeleton *skeleton = ORG_GNOME_SHELL_SCREENSHOT_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_screenshot_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_shell_screenshot_skeleton_vtable =
{
  _org_gnome_shell_screenshot_skeleton_handle_method_call,
  _org_gnome_shell_screenshot_skeleton_handle_get_property,
  _org_gnome_shell_screenshot_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_shell_screenshot_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_shell_screenshot_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_shell_screenshot_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_shell_screenshot_skeleton_vtable;
}

static GVariant *
org_gnome_shell_screenshot_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeShellScreenshotSkeleton *skeleton = ORG_GNOME_SHELL_SCREENSHOT_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_shell_screenshot_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_shell_screenshot_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_shell_screenshot_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_shell_screenshot_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Shell.Screenshot", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
org_gnome_shell_screenshot_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton G_GNUC_UNUSED)
{
}

static void org_gnome_shell_screenshot_skeleton_iface_init (OrgGnomeShellScreenshotIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellScreenshotSkeleton, org_gnome_shell_screenshot_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeShellScreenshotSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_SCREENSHOT, org_gnome_shell_screenshot_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellScreenshotSkeleton, org_gnome_shell_screenshot_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL_SCREENSHOT, org_gnome_shell_screenshot_skeleton_iface_init))

#endif
static void
org_gnome_shell_screenshot_skeleton_finalize (GObject *object)
{
  OrgGnomeShellScreenshotSkeleton *skeleton = ORG_GNOME_SHELL_SCREENSHOT_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_shell_screenshot_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_shell_screenshot_skeleton_init (OrgGnomeShellScreenshotSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_shell_screenshot_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_SHELL_SCREENSHOT_SKELETON, OrgGnomeShellScreenshotSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
org_gnome_shell_screenshot_skeleton_class_init (OrgGnomeShellScreenshotSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_shell_screenshot_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_shell_screenshot_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_shell_screenshot_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_shell_screenshot_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_shell_screenshot_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeShellScreenshotSkeletonPrivate));
#endif
}

static void
org_gnome_shell_screenshot_skeleton_iface_init (OrgGnomeShellScreenshotIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gnome_shell_screenshot_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell-Screenshot.top_of_page">org.gnome.Shell.Screenshot</link>.
 *
 * Returns: (transfer full) (type OrgGnomeShellScreenshotSkeleton): The skeleton object.
 */
OrgGnomeShellScreenshot *
org_gnome_shell_screenshot_skeleton_new (void)
{
  return ORG_GNOME_SHELL_SCREENSHOT (g_object_new (TYPE_ORG_GNOME_SHELL_SCREENSHOT_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gnome.Shell
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGnomeShell
 * @title: OrgGnomeShell
 * @short_description: Generated C code for the org.gnome.Shell D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gnome-Shell.top_of_page">org.gnome.Shell</link> D-Bus interface in C.
 */

enum
{
  _ORG_GNOME_SHELL_ACCELERATOR_ACTIVATED,
  _ORG_GNOME_SHELL_ACCELERATOR_DEACTIVATED,
};

static unsigned _ORG_GNOME_SHELL_SIGNALS[2] = { 0 };

/* ---- Introspection data for org.gnome.Shell ---- */

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_eval_IN_ARG_script =
{
  {
    -1,
    (gchar *) "script",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_eval_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_eval_IN_ARG_script.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_eval_OUT_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_eval_OUT_ARG_result =
{
  {
    -1,
    (gchar *) "result",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_eval_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_eval_OUT_ARG_success.parent_struct,
  &_org_gnome_shell_method_info_eval_OUT_ARG_result.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_eval =
{
  {
    -1,
    (gchar *) "Eval",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_eval_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_method_info_eval_OUT_ARG_pointers,
    NULL
  },
  "handle-eval",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_focus_search =
{
  {
    -1,
    (gchar *) "FocusSearch",
    NULL,
    NULL,
    NULL
  },
  "handle-focus-search",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_show_osd_IN_ARG_params =
{
  {
    -1,
    (gchar *) "params",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_show_osd_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_show_osd_IN_ARG_params.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_show_osd =
{
  {
    -1,
    (gchar *) "ShowOSD",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_show_osd_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-show-osd",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_show_monitor_labels_IN_ARG_params =
{
  {
    -1,
    (gchar *) "params",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_show_monitor_labels_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_show_monitor_labels_IN_ARG_params.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_show_monitor_labels =
{
  {
    -1,
    (gchar *) "ShowMonitorLabels",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_show_monitor_labels_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-show-monitor-labels",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_hide_monitor_labels =
{
  {
    -1,
    (gchar *) "HideMonitorLabels",
    NULL,
    NULL,
    NULL
  },
  "handle-hide-monitor-labels",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_focus_app_IN_ARG_id =
{
  {
    -1,
    (gchar *) "id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_focus_app_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_focus_app_IN_ARG_id.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_focus_app =
{
  {
    -1,
    (gchar *) "FocusApp",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_focus_app_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-focus-app",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_show_applications =
{
  {
    -1,
    (gchar *) "ShowApplications",
    NULL,
    NULL,
    NULL
  },
  "handle-show-applications",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_grab_accelerator_IN_ARG_accelerator =
{
  {
    -1,
    (gchar *) "accelerator",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_grab_accelerator_IN_ARG_modeFlags =
{
  {
    -1,
    (gchar *) "modeFlags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_grab_accelerator_IN_ARG_grabFlags =
{
  {
    -1,
    (gchar *) "grabFlags",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_grab_accelerator_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_grab_accelerator_IN_ARG_accelerator.parent_struct,
  &_org_gnome_shell_method_info_grab_accelerator_IN_ARG_modeFlags.parent_struct,
  &_org_gnome_shell_method_info_grab_accelerator_IN_ARG_grabFlags.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_grab_accelerator_OUT_ARG_action =
{
  {
    -1,
    (gchar *) "action",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_grab_accelerator_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_grab_accelerator_OUT_ARG_action.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_grab_accelerator =
{
  {
    -1,
    (gchar *) "GrabAccelerator",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_grab_accelerator_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_method_info_grab_accelerator_OUT_ARG_pointers,
    NULL
  },
  "handle-grab-accelerator",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_grab_accelerators_IN_ARG_accelerators =
{
  {
    -1,
    (gchar *) "accelerators",
    (gchar *) "a(suu)",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_grab_accelerators_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_grab_accelerators_IN_ARG_accelerators.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_grab_accelerators_OUT_ARG_actions =
{
  {
    -1,
    (gchar *) "actions",
    (gchar *) "au",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_grab_accelerators_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_grab_accelerators_OUT_ARG_actions.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_grab_accelerators =
{
  {
    -1,
    (gchar *) "GrabAccelerators",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_grab_accelerators_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_method_info_grab_accelerators_OUT_ARG_pointers,
    NULL
  },
  "handle-grab-accelerators",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_ungrab_accelerator_IN_ARG_action =
{
  {
    -1,
    (gchar *) "action",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_ungrab_accelerator_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_ungrab_accelerator_IN_ARG_action.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_ungrab_accelerator_OUT_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_ungrab_accelerator_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_ungrab_accelerator_OUT_ARG_success.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_ungrab_accelerator =
{
  {
    -1,
    (gchar *) "UngrabAccelerator",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_ungrab_accelerator_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_method_info_ungrab_accelerator_OUT_ARG_pointers,
    NULL
  },
  "handle-ungrab-accelerator",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_ungrab_accelerators_IN_ARG_action =
{
  {
    -1,
    (gchar *) "action",
    (gchar *) "au",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_ungrab_accelerators_IN_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_ungrab_accelerators_IN_ARG_action.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_method_info_ungrab_accelerators_OUT_ARG_success =
{
  {
    -1,
    (gchar *) "success",
    (gchar *) "b",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_method_info_ungrab_accelerators_OUT_ARG_pointers[] =
{
  &_org_gnome_shell_method_info_ungrab_accelerators_OUT_ARG_success.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_ungrab_accelerators =
{
  {
    -1,
    (gchar *) "UngrabAccelerators",
    (GDBusArgInfo **) &_org_gnome_shell_method_info_ungrab_accelerators_IN_ARG_pointers,
    (GDBusArgInfo **) &_org_gnome_shell_method_info_ungrab_accelerators_OUT_ARG_pointers,
    NULL
  },
  "handle-ungrab-accelerators",
  FALSE
};

static const _ExtendedGDBusMethodInfo _org_gnome_shell_method_info_screen_transition =
{
  {
    -1,
    (gchar *) "ScreenTransition",
    NULL,
    NULL,
    NULL
  },
  "handle-screen-transition",
  FALSE
};

static const GDBusMethodInfo * const _org_gnome_shell_method_info_pointers[] =
{
  &_org_gnome_shell_method_info_eval.parent_struct,
  &_org_gnome_shell_method_info_focus_search.parent_struct,
  &_org_gnome_shell_method_info_show_osd.parent_struct,
  &_org_gnome_shell_method_info_show_monitor_labels.parent_struct,
  &_org_gnome_shell_method_info_hide_monitor_labels.parent_struct,
  &_org_gnome_shell_method_info_focus_app.parent_struct,
  &_org_gnome_shell_method_info_show_applications.parent_struct,
  &_org_gnome_shell_method_info_grab_accelerator.parent_struct,
  &_org_gnome_shell_method_info_grab_accelerators.parent_struct,
  &_org_gnome_shell_method_info_ungrab_accelerator.parent_struct,
  &_org_gnome_shell_method_info_ungrab_accelerators.parent_struct,
  &_org_gnome_shell_method_info_screen_transition.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_signal_info_accelerator_activated_ARG_action =
{
  {
    -1,
    (gchar *) "action",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_signal_info_accelerator_activated_ARG_parameters =
{
  {
    -1,
    (gchar *) "parameters",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_signal_info_accelerator_activated_ARG_pointers[] =
{
  &_org_gnome_shell_signal_info_accelerator_activated_ARG_action.parent_struct,
  &_org_gnome_shell_signal_info_accelerator_activated_ARG_parameters.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_shell_signal_info_accelerator_activated =
{
  {
    -1,
    (gchar *) "AcceleratorActivated",
    (GDBusArgInfo **) &_org_gnome_shell_signal_info_accelerator_activated_ARG_pointers,
    NULL
  },
  "accelerator-activated"
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_signal_info_accelerator_deactivated_ARG_action =
{
  {
    -1,
    (gchar *) "action",
    (gchar *) "u",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gnome_shell_signal_info_accelerator_deactivated_ARG_parameters =
{
  {
    -1,
    (gchar *) "parameters",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gnome_shell_signal_info_accelerator_deactivated_ARG_pointers[] =
{
  &_org_gnome_shell_signal_info_accelerator_deactivated_ARG_action.parent_struct,
  &_org_gnome_shell_signal_info_accelerator_deactivated_ARG_parameters.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gnome_shell_signal_info_accelerator_deactivated =
{
  {
    -1,
    (gchar *) "AcceleratorDeactivated",
    (GDBusArgInfo **) &_org_gnome_shell_signal_info_accelerator_deactivated_ARG_pointers,
    NULL
  },
  "accelerator-deactivated"
};

static const GDBusSignalInfo * const _org_gnome_shell_signal_info_pointers[] =
{
  &_org_gnome_shell_signal_info_accelerator_activated.parent_struct,
  &_org_gnome_shell_signal_info_accelerator_deactivated.parent_struct,
  NULL
};

static const _ExtendedGDBusPropertyInfo _org_gnome_shell_property_info_mode =
{
  {
    -1,
    (gchar *) "Mode",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "mode",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_gnome_shell_property_info_overview_active =
{
  {
    -1,
    (gchar *) "OverviewActive",
    (gchar *) "b",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE | G_DBUS_PROPERTY_INFO_FLAGS_WRITABLE,
    NULL
  },
  "overview-active",
  FALSE,
  TRUE
};

static const _ExtendedGDBusPropertyInfo _org_gnome_shell_property_info_shell_version =
{
  {
    -1,
    (gchar *) "ShellVersion",
    (gchar *) "s",
    G_DBUS_PROPERTY_INFO_FLAGS_READABLE,
    NULL
  },
  "shell-version",
  FALSE,
  TRUE
};

static const GDBusPropertyInfo * const _org_gnome_shell_property_info_pointers[] =
{
  &_org_gnome_shell_property_info_mode.parent_struct,
  &_org_gnome_shell_property_info_overview_active.parent_struct,
  &_org_gnome_shell_property_info_shell_version.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gnome_shell_interface_info =
{
  {
    -1,
    (gchar *) "org.gnome.Shell",
    (GDBusMethodInfo **) &_org_gnome_shell_method_info_pointers,
    (GDBusSignalInfo **) &_org_gnome_shell_signal_info_pointers,
    (GDBusPropertyInfo **) &_org_gnome_shell_property_info_pointers,
    NULL
  },
  "org-gnome-shell",
};


/**
 * org_gnome_shell_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gnome-Shell.top_of_page">org.gnome.Shell</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gnome_shell_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gnome_shell_interface_info.parent_struct;
}

/**
 * org_gnome_shell_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGnomeShell interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gnome_shell_override_properties (GObjectClass *klass, guint property_id_begin)
{
  g_object_class_override_property (klass, property_id_begin++, "mode");
  g_object_class_override_property (klass, property_id_begin++, "overview-active");
  g_object_class_override_property (klass, property_id_begin++, "shell-version");
  return property_id_begin - 1;
}


inline static void
org_gnome_shell_signal_marshal_accelerator_activated (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_VOID__UINT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_signal_marshal_accelerator_deactivated (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_VOID__UINT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_eval (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_focus_search (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_show_osd (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_show_monitor_labels (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_hide_monitor_labels (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_focus_app (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_show_applications (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_grab_accelerator (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_UINT_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_grab_accelerators (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_ungrab_accelerator (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_UINT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_ungrab_accelerators (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gnome_shell_method_marshal_screen_transition (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGnomeShell:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell.top_of_page">org.gnome.Shell</link>.
 */

/**
 * OrgGnomeShellIface:
 * @parent_iface: The parent interface.
 * @handle_eval: Handler for the #OrgGnomeShell::handle-eval signal.
 * @handle_focus_app: Handler for the #OrgGnomeShell::handle-focus-app signal.
 * @handle_focus_search: Handler for the #OrgGnomeShell::handle-focus-search signal.
 * @handle_grab_accelerator: Handler for the #OrgGnomeShell::handle-grab-accelerator signal.
 * @handle_grab_accelerators: Handler for the #OrgGnomeShell::handle-grab-accelerators signal.
 * @handle_hide_monitor_labels: Handler for the #OrgGnomeShell::handle-hide-monitor-labels signal.
 * @handle_screen_transition: Handler for the #OrgGnomeShell::handle-screen-transition signal.
 * @handle_show_applications: Handler for the #OrgGnomeShell::handle-show-applications signal.
 * @handle_show_monitor_labels: Handler for the #OrgGnomeShell::handle-show-monitor-labels signal.
 * @handle_show_osd: Handler for the #OrgGnomeShell::handle-show-osd signal.
 * @handle_ungrab_accelerator: Handler for the #OrgGnomeShell::handle-ungrab-accelerator signal.
 * @handle_ungrab_accelerators: Handler for the #OrgGnomeShell::handle-ungrab-accelerators signal.
 * @get_mode: Getter for the #OrgGnomeShell:mode property.
 * @get_overview_active: Getter for the #OrgGnomeShell:overview-active property.
 * @get_shell_version: Getter for the #OrgGnomeShell:shell-version property.
 * @accelerator_activated: Handler for the #OrgGnomeShell::accelerator-activated signal.
 * @accelerator_deactivated: Handler for the #OrgGnomeShell::accelerator-deactivated signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell.top_of_page">org.gnome.Shell</link>.
 */

typedef OrgGnomeShellIface OrgGnomeShellInterface;
G_DEFINE_INTERFACE (OrgGnomeShell, org_gnome_shell, G_TYPE_OBJECT)

static void
org_gnome_shell_default_init (OrgGnomeShellIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGnomeShell::handle-eval:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_script: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.Eval">Eval()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_eval() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-eval",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_eval),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_eval,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgGnomeShell::handle-focus-search:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.FocusSearch">FocusSearch()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_focus_search() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-focus-search",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_focus_search),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_focus_search,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeShell::handle-show-osd:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_params: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.ShowOSD">ShowOSD()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_show_osd() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-show-osd",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_show_osd),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_show_osd,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeShell::handle-show-monitor-labels:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_params: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.ShowMonitorLabels">ShowMonitorLabels()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_show_monitor_labels() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-show-monitor-labels",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_show_monitor_labels),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_show_monitor_labels,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeShell::handle-hide-monitor-labels:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.HideMonitorLabels">HideMonitorLabels()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_hide_monitor_labels() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-hide-monitor-labels",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_hide_monitor_labels),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_hide_monitor_labels,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeShell::handle-focus-app:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_id: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.FocusApp">FocusApp()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_focus_app() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-focus-app",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_focus_app),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_focus_app,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING);

  /**
   * OrgGnomeShell::handle-show-applications:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.ShowApplications">ShowApplications()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_show_applications() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-show-applications",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_show_applications),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_show_applications,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /**
   * OrgGnomeShell::handle-grab-accelerator:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_accelerator: Argument passed by remote caller.
   * @arg_modeFlags: Argument passed by remote caller.
   * @arg_grabFlags: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerator">GrabAccelerator()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_grab_accelerator() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-grab-accelerator",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_grab_accelerator),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_grab_accelerator,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_UINT);

  /**
   * OrgGnomeShell::handle-grab-accelerators:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_accelerators: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerators">GrabAccelerators()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_grab_accelerators() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-grab-accelerators",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_grab_accelerators),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_grab_accelerators,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeShell::handle-ungrab-accelerator:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_action: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerator">UngrabAccelerator()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_ungrab_accelerator() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-ungrab-accelerator",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_ungrab_accelerator),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_ungrab_accelerator,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_UINT);

  /**
   * OrgGnomeShell::handle-ungrab-accelerators:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_action: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerators">UngrabAccelerators()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_ungrab_accelerators() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-ungrab-accelerators",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_ungrab_accelerators),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_ungrab_accelerators,
    G_TYPE_BOOLEAN,
    2,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_VARIANT);

  /**
   * OrgGnomeShell::handle-screen-transition:
   * @object: A #OrgGnomeShell.
   * @invocation: A #GDBusMethodInvocation.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gnome-Shell.ScreenTransition">ScreenTransition()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gnome_shell_complete_screen_transition() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-screen-transition",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGnomeShellIface, handle_screen_transition),
    g_signal_accumulator_true_handled,
    NULL,
      org_gnome_shell_method_marshal_screen_transition,
    G_TYPE_BOOLEAN,
    1,
    G_TYPE_DBUS_METHOD_INVOCATION);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGnomeShell::accelerator-activated:
   * @object: A #OrgGnomeShell.
   * @arg_action: Argument.
   * @arg_parameters: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Shell.AcceleratorActivated">"AcceleratorActivated"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_SHELL_SIGNALS[_ORG_GNOME_SHELL_ACCELERATOR_ACTIVATED] =
    g_signal_new ("accelerator-activated",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeShellIface, accelerator_activated),
      NULL,
      NULL,
      org_gnome_shell_signal_marshal_accelerator_activated,
      G_TYPE_NONE,
      2, G_TYPE_UINT, G_TYPE_VARIANT);

  /**
   * OrgGnomeShell::accelerator-deactivated:
   * @object: A #OrgGnomeShell.
   * @arg_action: Argument.
   * @arg_parameters: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gnome-Shell.AcceleratorDeactivated">"AcceleratorDeactivated"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GNOME_SHELL_SIGNALS[_ORG_GNOME_SHELL_ACCELERATOR_DEACTIVATED] =
    g_signal_new ("accelerator-deactivated",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGnomeShellIface, accelerator_deactivated),
      NULL,
      NULL,
      org_gnome_shell_signal_marshal_accelerator_deactivated,
      G_TYPE_NONE,
      2, G_TYPE_UINT, G_TYPE_VARIANT);

  /* GObject properties for D-Bus properties: */
  /**
   * OrgGnomeShell:mode:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Shell.Mode">"Mode"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("mode", "Mode", "Mode", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgGnomeShell:overview-active:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Shell.OverviewActive">"OverviewActive"</link>.
   *
   * Since the D-Bus property for this #GObject property is both readable and writable, it is meaningful to both read from it and write to it on both the service- and client-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_boolean ("overview-active", "OverviewActive", "OverviewActive", FALSE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  /**
   * OrgGnomeShell:shell-version:
   *
   * Represents the D-Bus property <link linkend="gdbus-property-org-gnome-Shell.ShellVersion">"ShellVersion"</link>.
   *
   * Since the D-Bus property for this #GObject property is readable but not writable, it is meaningful to read from it on both the client- and service-side. It is only meaningful, however, to write to it on the service-side.
   */
  g_object_interface_install_property (iface,
    g_param_spec_string ("shell-version", "ShellVersion", "ShellVersion", NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

/**
 * org_gnome_shell_get_mode: (skip)
 * @object: A #OrgGnomeShell.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Shell.Mode">"Mode"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_gnome_shell_dup_mode() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_gnome_shell_get_mode (OrgGnomeShell *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_SHELL (object), NULL);

  return ORG_GNOME_SHELL_GET_IFACE (object)->get_mode (object);
}

/**
 * org_gnome_shell_dup_mode: (skip)
 * @object: A #OrgGnomeShell.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-gnome-Shell.Mode">"Mode"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_gnome_shell_dup_mode (OrgGnomeShell *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "mode", &value, NULL);
  return value;
}

/**
 * org_gnome_shell_set_mode: (skip)
 * @object: A #OrgGnomeShell.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Shell.Mode">"Mode"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_shell_set_mode (OrgGnomeShell *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "mode", value, NULL);
}

/**
 * org_gnome_shell_get_overview_active: (skip)
 * @object: A #OrgGnomeShell.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Shell.OverviewActive">"OverviewActive"</link> D-Bus property.
 *
 * Since this D-Bus property is both readable and writable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: The property value.
 */
gboolean 
org_gnome_shell_get_overview_active (OrgGnomeShell *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_SHELL (object), FALSE);

  return ORG_GNOME_SHELL_GET_IFACE (object)->get_overview_active (object);
}

/**
 * org_gnome_shell_set_overview_active: (skip)
 * @object: A #OrgGnomeShell.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Shell.OverviewActive">"OverviewActive"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is both readable and writable, it is meaningful to use this function on both the client- and service-side.
 */
void
org_gnome_shell_set_overview_active (OrgGnomeShell *object, gboolean value)
{
  g_object_set (G_OBJECT (object), "overview-active", value, NULL);
}

/**
 * org_gnome_shell_get_shell_version: (skip)
 * @object: A #OrgGnomeShell.
 *
 * Gets the value of the <link linkend="gdbus-property-org-gnome-Shell.ShellVersion">"ShellVersion"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * The returned value is only valid until the property changes so on the client-side it is only safe to use this function on the thread where @object was constructed. Use org_gnome_shell_dup_shell_version() if on another thread.
 *
 * Returns: (transfer none) (nullable): The property value or %NULL if the property is not set. Do not free the returned value, it belongs to @object.
 */
const gchar *
org_gnome_shell_get_shell_version (OrgGnomeShell *object)
{
  g_return_val_if_fail (IS_ORG_GNOME_SHELL (object), NULL);

  return ORG_GNOME_SHELL_GET_IFACE (object)->get_shell_version (object);
}

/**
 * org_gnome_shell_dup_shell_version: (skip)
 * @object: A #OrgGnomeShell.
 *
 * Gets a copy of the <link linkend="gdbus-property-org-gnome-Shell.ShellVersion">"ShellVersion"</link> D-Bus property.
 *
 * Since this D-Bus property is readable, it is meaningful to use this function on both the client- and service-side.
 *
 * Returns: (transfer full) (nullable): The property value or %NULL if the property is not set. The returned value should be freed with g_free().
 */
gchar *
org_gnome_shell_dup_shell_version (OrgGnomeShell *object)
{
  gchar *value;
  g_object_get (G_OBJECT (object), "shell-version", &value, NULL);
  return value;
}

/**
 * org_gnome_shell_set_shell_version: (skip)
 * @object: A #OrgGnomeShell.
 * @value: The value to set.
 *
 * Sets the <link linkend="gdbus-property-org-gnome-Shell.ShellVersion">"ShellVersion"</link> D-Bus property to @value.
 *
 * Since this D-Bus property is not writable, it is only meaningful to use this function on the service-side.
 */
void
org_gnome_shell_set_shell_version (OrgGnomeShell *object, const gchar *value)
{
  g_object_set (G_OBJECT (object), "shell-version", value, NULL);
}

/**
 * org_gnome_shell_emit_accelerator_activated:
 * @object: A #OrgGnomeShell.
 * @arg_action: Argument to pass with the signal.
 * @arg_parameters: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Shell.AcceleratorActivated">"AcceleratorActivated"</link> D-Bus signal.
 */
void
org_gnome_shell_emit_accelerator_activated (
    OrgGnomeShell *object,
    guint arg_action,
    GVariant *arg_parameters)
{
  g_signal_emit (object, _ORG_GNOME_SHELL_SIGNALS[_ORG_GNOME_SHELL_ACCELERATOR_ACTIVATED], 0, arg_action, arg_parameters);
}

/**
 * org_gnome_shell_emit_accelerator_deactivated:
 * @object: A #OrgGnomeShell.
 * @arg_action: Argument to pass with the signal.
 * @arg_parameters: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gnome-Shell.AcceleratorDeactivated">"AcceleratorDeactivated"</link> D-Bus signal.
 */
void
org_gnome_shell_emit_accelerator_deactivated (
    OrgGnomeShell *object,
    guint arg_action,
    GVariant *arg_parameters)
{
  g_signal_emit (object, _ORG_GNOME_SHELL_SIGNALS[_ORG_GNOME_SHELL_ACCELERATOR_DEACTIVATED], 0, arg_action, arg_parameters);
}

/**
 * org_gnome_shell_call_eval:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_script: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.Eval">Eval()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_eval_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_eval_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_eval (
    OrgGnomeShell *proxy,
    const gchar *arg_script,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "Eval",
    g_variant_new ("(s)",
                   arg_script),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_eval_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_result: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_eval().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_eval().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_eval_finish (
    OrgGnomeShell *proxy,
    gboolean* out_success,
    gchar ** out_result,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_result);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_eval_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_script: Argument to pass with the method invocation.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @out_result: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.Eval">Eval()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_eval() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_eval_sync (
    OrgGnomeShell *proxy,
    const gchar *arg_script,
    gboolean* out_success,
    gchar ** out_result,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "Eval",
    g_variant_new ("(s)",
                   arg_script),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(bs)",
                 out_success,
                 out_result);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_focus_search:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.FocusSearch">FocusSearch()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_focus_search_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_focus_search_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_focus_search (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "FocusSearch",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_focus_search_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_focus_search().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_focus_search().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_focus_search_finish (
    OrgGnomeShell *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_focus_search_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.FocusSearch">FocusSearch()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_focus_search() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_focus_search_sync (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "FocusSearch",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_show_osd:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_params: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ShowOSD">ShowOSD()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_show_osd_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_show_osd_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_show_osd (
    OrgGnomeShell *proxy,
    GVariant *arg_params,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ShowOSD",
    g_variant_new ("(@a{sv})",
                   arg_params),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_show_osd_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_show_osd().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_show_osd().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_show_osd_finish (
    OrgGnomeShell *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_show_osd_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_params: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ShowOSD">ShowOSD()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_show_osd() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_show_osd_sync (
    OrgGnomeShell *proxy,
    GVariant *arg_params,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ShowOSD",
    g_variant_new ("(@a{sv})",
                   arg_params),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_show_monitor_labels:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_params: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ShowMonitorLabels">ShowMonitorLabels()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_show_monitor_labels_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_show_monitor_labels_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_show_monitor_labels (
    OrgGnomeShell *proxy,
    GVariant *arg_params,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ShowMonitorLabels",
    g_variant_new ("(@a{sv})",
                   arg_params),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_show_monitor_labels_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_show_monitor_labels().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_show_monitor_labels().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_show_monitor_labels_finish (
    OrgGnomeShell *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_show_monitor_labels_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_params: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ShowMonitorLabels">ShowMonitorLabels()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_show_monitor_labels() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_show_monitor_labels_sync (
    OrgGnomeShell *proxy,
    GVariant *arg_params,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ShowMonitorLabels",
    g_variant_new ("(@a{sv})",
                   arg_params),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_hide_monitor_labels:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.HideMonitorLabels">HideMonitorLabels()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_hide_monitor_labels_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_hide_monitor_labels_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_hide_monitor_labels (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "HideMonitorLabels",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_hide_monitor_labels_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_hide_monitor_labels().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_hide_monitor_labels().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_hide_monitor_labels_finish (
    OrgGnomeShell *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_hide_monitor_labels_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.HideMonitorLabels">HideMonitorLabels()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_hide_monitor_labels() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_hide_monitor_labels_sync (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "HideMonitorLabels",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_focus_app:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_id: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.FocusApp">FocusApp()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_focus_app_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_focus_app_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_focus_app (
    OrgGnomeShell *proxy,
    const gchar *arg_id,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "FocusApp",
    g_variant_new ("(s)",
                   arg_id),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_focus_app_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_focus_app().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_focus_app().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_focus_app_finish (
    OrgGnomeShell *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_focus_app_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_id: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.FocusApp">FocusApp()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_focus_app() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_focus_app_sync (
    OrgGnomeShell *proxy,
    const gchar *arg_id,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "FocusApp",
    g_variant_new ("(s)",
                   arg_id),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_show_applications:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ShowApplications">ShowApplications()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_show_applications_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_show_applications_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_show_applications (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ShowApplications",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_show_applications_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_show_applications().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_show_applications().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_show_applications_finish (
    OrgGnomeShell *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_show_applications_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ShowApplications">ShowApplications()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_show_applications() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_show_applications_sync (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ShowApplications",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_grab_accelerator:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_accelerator: Argument to pass with the method invocation.
 * @arg_modeFlags: Argument to pass with the method invocation.
 * @arg_grabFlags: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerator">GrabAccelerator()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_grab_accelerator_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_grab_accelerator_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_grab_accelerator (
    OrgGnomeShell *proxy,
    const gchar *arg_accelerator,
    guint arg_modeFlags,
    guint arg_grabFlags,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GrabAccelerator",
    g_variant_new ("(suu)",
                   arg_accelerator,
                   arg_modeFlags,
                   arg_grabFlags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_grab_accelerator_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @out_action: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_grab_accelerator().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_grab_accelerator().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_grab_accelerator_finish (
    OrgGnomeShell *proxy,
    guint* out_action,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u)",
                 out_action);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_grab_accelerator_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_accelerator: Argument to pass with the method invocation.
 * @arg_modeFlags: Argument to pass with the method invocation.
 * @arg_grabFlags: Argument to pass with the method invocation.
 * @out_action: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerator">GrabAccelerator()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_grab_accelerator() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_grab_accelerator_sync (
    OrgGnomeShell *proxy,
    const gchar *arg_accelerator,
    guint arg_modeFlags,
    guint arg_grabFlags,
    guint* out_action,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GrabAccelerator",
    g_variant_new ("(suu)",
                   arg_accelerator,
                   arg_modeFlags,
                   arg_grabFlags),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(u)",
                 out_action);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_grab_accelerators:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_accelerators: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerators">GrabAccelerators()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_grab_accelerators_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_grab_accelerators_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_grab_accelerators (
    OrgGnomeShell *proxy,
    GVariant *arg_accelerators,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "GrabAccelerators",
    g_variant_new ("(@a(suu))",
                   arg_accelerators),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_grab_accelerators_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @out_actions: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_grab_accelerators().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_grab_accelerators().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_grab_accelerators_finish (
    OrgGnomeShell *proxy,
    GVariant ** out_actions,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@au)",
                 out_actions);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_grab_accelerators_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_accelerators: Argument to pass with the method invocation.
 * @out_actions: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerators">GrabAccelerators()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_grab_accelerators() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_grab_accelerators_sync (
    OrgGnomeShell *proxy,
    GVariant *arg_accelerators,
    GVariant ** out_actions,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "GrabAccelerators",
    g_variant_new ("(@a(suu))",
                   arg_accelerators),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(@au)",
                 out_actions);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_ungrab_accelerator:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_action: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerator">UngrabAccelerator()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_ungrab_accelerator_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_ungrab_accelerator_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_ungrab_accelerator (
    OrgGnomeShell *proxy,
    guint arg_action,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "UngrabAccelerator",
    g_variant_new ("(u)",
                   arg_action),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_ungrab_accelerator_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_ungrab_accelerator().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_ungrab_accelerator().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_ungrab_accelerator_finish (
    OrgGnomeShell *proxy,
    gboolean* out_success,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(b)",
                 out_success);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_ungrab_accelerator_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_action: Argument to pass with the method invocation.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerator">UngrabAccelerator()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_ungrab_accelerator() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_ungrab_accelerator_sync (
    OrgGnomeShell *proxy,
    guint arg_action,
    gboolean* out_success,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "UngrabAccelerator",
    g_variant_new ("(u)",
                   arg_action),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(b)",
                 out_success);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_ungrab_accelerators:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_action: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerators">UngrabAccelerators()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_ungrab_accelerators_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_ungrab_accelerators_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_ungrab_accelerators (
    OrgGnomeShell *proxy,
    GVariant *arg_action,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "UngrabAccelerators",
    g_variant_new ("(@au)",
                   arg_action),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_ungrab_accelerators_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_ungrab_accelerators().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_ungrab_accelerators().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_ungrab_accelerators_finish (
    OrgGnomeShell *proxy,
    gboolean* out_success,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(b)",
                 out_success);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_ungrab_accelerators_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @arg_action: Argument to pass with the method invocation.
 * @out_success: (out) (optional): Return location for return parameter or %NULL to ignore.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerators">UngrabAccelerators()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_ungrab_accelerators() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_ungrab_accelerators_sync (
    OrgGnomeShell *proxy,
    GVariant *arg_action,
    gboolean* out_success,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "UngrabAccelerators",
    g_variant_new ("(@au)",
                   arg_action),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "(b)",
                 out_success);
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_screen_transition:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ScreenTransition">ScreenTransition()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_call_screen_transition_finish() to get the result of the operation.
 *
 * See org_gnome_shell_call_screen_transition_sync() for the synchronous, blocking version of this method.
 */
void
org_gnome_shell_call_screen_transition (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "ScreenTransition",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gnome_shell_call_screen_transition_finish:
 * @proxy: A #OrgGnomeShellProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_call_screen_transition().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gnome_shell_call_screen_transition().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_screen_transition_finish (
    OrgGnomeShell *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_call_screen_transition_sync:
 * @proxy: A #OrgGnomeShellProxy.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gnome-Shell.ScreenTransition">ScreenTransition()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_call_screen_transition() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gnome_shell_call_screen_transition_sync (
    OrgGnomeShell *proxy,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "ScreenTransition",
    g_variant_new ("()"),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gnome_shell_complete_eval:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @success: Parameter to return.
 * @result: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.Eval">Eval()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_eval (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gboolean success,
    const gchar *result)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(bs)",
                   success,
                   result));
}

/**
 * org_gnome_shell_complete_focus_search:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.FocusSearch">FocusSearch()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_focus_search (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_shell_complete_show_osd:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.ShowOSD">ShowOSD()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_show_osd (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_shell_complete_show_monitor_labels:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.ShowMonitorLabels">ShowMonitorLabels()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_show_monitor_labels (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_shell_complete_hide_monitor_labels:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.HideMonitorLabels">HideMonitorLabels()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_hide_monitor_labels (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_shell_complete_focus_app:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.FocusApp">FocusApp()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_focus_app (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_shell_complete_show_applications:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.ShowApplications">ShowApplications()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_show_applications (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gnome_shell_complete_grab_accelerator:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @action: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerator">GrabAccelerator()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_grab_accelerator (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    guint action)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(u)",
                   action));
}

/**
 * org_gnome_shell_complete_grab_accelerators:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @actions: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.GrabAccelerators">GrabAccelerators()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_grab_accelerators (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    GVariant *actions)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(@au)",
                   actions));
}

/**
 * org_gnome_shell_complete_ungrab_accelerator:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @success: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerator">UngrabAccelerator()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_ungrab_accelerator (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gboolean success)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(b)",
                   success));
}

/**
 * org_gnome_shell_complete_ungrab_accelerators:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 * @success: Parameter to return.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.UngrabAccelerators">UngrabAccelerators()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_ungrab_accelerators (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation,
    gboolean success)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("(b)",
                   success));
}

/**
 * org_gnome_shell_complete_screen_transition:
 * @object: A #OrgGnomeShell.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gnome-Shell.ScreenTransition">ScreenTransition()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gnome_shell_complete_screen_transition (
    OrgGnomeShell *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeShellProxy:
 *
 * The #OrgGnomeShellProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeShellProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeShellProxy.
 */

struct _OrgGnomeShellProxyPrivate
{
  GData *qdata;
};

static void org_gnome_shell_proxy_iface_init (OrgGnomeShellIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellProxy, org_gnome_shell_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGnomeShellProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL, org_gnome_shell_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellProxy, org_gnome_shell_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL, org_gnome_shell_proxy_iface_init))

#endif
static void
org_gnome_shell_proxy_finalize (GObject *object)
{
  OrgGnomeShellProxy *proxy = ORG_GNOME_SHELL_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gnome_shell_proxy_parent_class)->finalize (object);
}

static void
org_gnome_shell_proxy_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_shell_property_info_pointers[prop_id - 1];
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (object), info->parent_struct.name);
  if (info->use_gvariant)
    {
      g_value_set_variant (value, variant);
    }
  else
    {
      if (variant != NULL)
        g_dbus_gvariant_to_gvalue (variant, value);
    }
  if (variant != NULL)
    g_variant_unref (variant);
}

static void
org_gnome_shell_proxy_set_property_cb (GDBusProxy *proxy,
  GAsyncResult *res,
  gpointer      user_data)
{
  const _ExtendedGDBusPropertyInfo *info = user_data;
  GError *error;
  GVariant *_ret;
  error = NULL;
  _ret = g_dbus_proxy_call_finish (proxy, res, &error);
  if (!_ret)
    {
      g_warning ("Error setting property '%s' on interface org.gnome.Shell: %s (%s, %d)",
                 info->parent_struct.name, 
                 error->message, g_quark_to_string (error->domain), error->code);
      g_error_free (error);
    }
  else
    {
      g_variant_unref (_ret);
    }
}

static void
org_gnome_shell_proxy_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  const _ExtendedGDBusPropertyInfo *info;
  GVariant *variant;
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_shell_property_info_pointers[prop_id - 1];
  variant = g_dbus_gvalue_to_gvariant (value, G_VARIANT_TYPE (info->parent_struct.signature));
  g_dbus_proxy_call (G_DBUS_PROXY (object),
    "org.freedesktop.DBus.Properties.Set",
    g_variant_new ("(ssv)", "org.gnome.Shell", info->parent_struct.name, variant),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    NULL, (GAsyncReadyCallback) org_gnome_shell_proxy_set_property_cb, (GDBusPropertyInfo *) &info->parent_struct);
  g_variant_unref (variant);
}

static void
org_gnome_shell_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gnome_shell_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GNOME_SHELL);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SHELL);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gnome_shell_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGnomeShellProxy *proxy = ORG_GNOME_SHELL_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static const gchar *
org_gnome_shell_proxy_get_mode (OrgGnomeShell *object)
{
  OrgGnomeShellProxy *proxy = ORG_GNOME_SHELL_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "Mode");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static gboolean 
org_gnome_shell_proxy_get_overview_active (OrgGnomeShell *object)
{
  OrgGnomeShellProxy *proxy = ORG_GNOME_SHELL_PROXY (object);
  GVariant *variant;
  gboolean value = FALSE;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "OverviewActive");
  if (variant != NULL)
    {
      value = g_variant_get_boolean (variant);
      g_variant_unref (variant);
    }
  return value;
}

static const gchar *
org_gnome_shell_proxy_get_shell_version (OrgGnomeShell *object)
{
  OrgGnomeShellProxy *proxy = ORG_GNOME_SHELL_PROXY (object);
  GVariant *variant;
  const gchar *value = NULL;
  variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (proxy), "ShellVersion");
  if (variant != NULL)
    {
      value = g_variant_get_string (variant, NULL);
      g_variant_unref (variant);
    }
  return value;
}

static void
org_gnome_shell_proxy_init (OrgGnomeShellProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gnome_shell_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GNOME_SHELL_PROXY, OrgGnomeShellProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gnome_shell_interface_info ());
}

static void
org_gnome_shell_proxy_class_init (OrgGnomeShellProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gnome_shell_proxy_finalize;
  gobject_class->get_property = org_gnome_shell_proxy_get_property;
  gobject_class->set_property = org_gnome_shell_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gnome_shell_proxy_g_signal;
  proxy_class->g_properties_changed = org_gnome_shell_proxy_g_properties_changed;

  org_gnome_shell_override_properties (gobject_class, 1);

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeShellProxyPrivate));
#endif
}

static void
org_gnome_shell_proxy_iface_init (OrgGnomeShellIface *iface)
{
  iface->get_mode = org_gnome_shell_proxy_get_mode;
  iface->get_overview_active = org_gnome_shell_proxy_get_overview_active;
  iface->get_shell_version = org_gnome_shell_proxy_get_shell_version;
}

/**
 * org_gnome_shell_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell.top_of_page">org.gnome.Shell</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_proxy_new_finish() to get the result of the operation.
 *
 * See org_gnome_shell_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_shell_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SHELL_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell", NULL);
}

/**
 * org_gnome_shell_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_shell_proxy_new().
 *
 * Returns: (transfer full) (type OrgGnomeShellProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShell *
org_gnome_shell_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SHELL (ret);
  else
    return NULL;
}

/**
 * org_gnome_shell_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell.top_of_page">org.gnome.Shell</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeShellProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShell *
org_gnome_shell_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SHELL_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell", NULL);
  if (ret != NULL)
    return ORG_GNOME_SHELL (ret);
  else
    return NULL;
}


/**
 * org_gnome_shell_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gnome_shell_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gnome_shell_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gnome_shell_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gnome_shell_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GNOME_SHELL_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell", NULL);
}

/**
 * org_gnome_shell_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gnome_shell_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gnome_shell_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGnomeShellProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShell *
org_gnome_shell_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GNOME_SHELL (ret);
  else
    return NULL;
}

/**
 * org_gnome_shell_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gnome_shell_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gnome_shell_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGnomeShellProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGnomeShell *
org_gnome_shell_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GNOME_SHELL_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gnome.Shell", NULL);
  if (ret != NULL)
    return ORG_GNOME_SHELL (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGnomeShellSkeleton:
 *
 * The #OrgGnomeShellSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGnomeShellSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGnomeShellSkeleton.
 */

struct _OrgGnomeShellSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gnome_shell_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GNOME_SHELL);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GNOME_SHELL);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gnome_shell_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gnome_shell_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gnome_shell_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gnome_shell_skeleton_vtable =
{
  _org_gnome_shell_skeleton_handle_method_call,
  _org_gnome_shell_skeleton_handle_get_property,
  _org_gnome_shell_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gnome_shell_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gnome_shell_interface_info ();
}

static GDBusInterfaceVTable *
org_gnome_shell_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gnome_shell_skeleton_vtable;
}

static GVariant *
org_gnome_shell_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gnome_shell_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gnome_shell_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gnome_shell_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gnome_shell_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Shell", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static gboolean _org_gnome_shell_emit_changed (gpointer user_data);

static void
org_gnome_shell_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (_skeleton);
  gboolean emit_changed = FALSE;

  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties_idle_source != NULL)
    {
      g_source_destroy (skeleton->priv->changed_properties_idle_source);
      skeleton->priv->changed_properties_idle_source = NULL;
      emit_changed = TRUE;
    }
  g_mutex_unlock (&skeleton->priv->lock);

  if (emit_changed)
    _org_gnome_shell_emit_changed (skeleton);
}

static void
_org_gnome_shell_on_signal_accelerator_activated (
    OrgGnomeShell *object,
    guint arg_action,
    GVariant *arg_parameters)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(u@a{sv})",
                   arg_action,
                   arg_parameters));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Shell", "AcceleratorActivated",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void
_org_gnome_shell_on_signal_accelerator_deactivated (
    OrgGnomeShell *object,
    guint arg_action,
    GVariant *arg_parameters)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(u@a{sv})",
                   arg_action,
                   arg_parameters));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gnome.Shell", "AcceleratorDeactivated",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gnome_shell_skeleton_iface_init (OrgGnomeShellIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellSkeleton, org_gnome_shell_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGnomeShellSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL, org_gnome_shell_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGnomeShellSkeleton, org_gnome_shell_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GNOME_SHELL, org_gnome_shell_skeleton_iface_init))

#endif
static void
org_gnome_shell_skeleton_finalize (GObject *object)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);
  guint n;
  for (n = 0; n < 3; n++)
    g_value_unset (&skeleton->priv->properties[n]);
  g_free (skeleton->priv->properties);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gnome_shell_skeleton_parent_class)->finalize (object);
}

static void
org_gnome_shell_skeleton_get_property (GObject      *object,
  guint         prop_id,
  GValue       *value,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  g_mutex_lock (&skeleton->priv->lock);
  g_value_copy (&skeleton->priv->properties[prop_id - 1], value);
  g_mutex_unlock (&skeleton->priv->lock);
}

static gboolean
_org_gnome_shell_emit_changed (gpointer user_data)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (user_data);
  GList *l;
  GVariantBuilder builder;
  GVariantBuilder invalidated_builder;
  guint num_changes;

  g_mutex_lock (&skeleton->priv->lock);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init_static (&invalidated_builder, G_VARIANT_TYPE ("as"));
#else
  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
  g_variant_builder_init (&invalidated_builder, G_VARIANT_TYPE ("as"));
#endif
  for (l = skeleton->priv->changed_properties, num_changes = 0; l != NULL; l = l->next)
    {
      ChangedProperty *cp = l->data;
      GVariant *variant;
      const GValue *cur_value;

      cur_value = &skeleton->priv->properties[cp->prop_id - 1];
      if (!_g_value_equal (cur_value, &cp->orig_value))
        {
          variant = g_dbus_gvalue_to_gvariant (cur_value, G_VARIANT_TYPE (cp->info->parent_struct.signature));
          g_variant_builder_add (&builder, "{sv}", cp->info->parent_struct.name, variant);
          g_variant_unref (variant);
          num_changes++;
        }
    }
  if (num_changes > 0)
    {
      GList *connections, *ll;
      GVariant *signal_variant;
      signal_variant = g_variant_ref_sink (g_variant_new ("(sa{sv}as)", "org.gnome.Shell",
                                           &builder, &invalidated_builder));
      connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));
      for (ll = connections; ll != NULL; ll = ll->next)
        {
          GDBusConnection *connection = ll->data;

          g_dbus_connection_emit_signal (connection,
                                         NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)),
                                         "org.freedesktop.DBus.Properties",
                                         "PropertiesChanged",
                                         signal_variant,
                                         NULL);
        }
      g_variant_unref (signal_variant);
      g_list_free_full (connections, g_object_unref);
    }
  else
    {
      g_variant_builder_clear (&builder);
      g_variant_builder_clear (&invalidated_builder);
    }
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
  skeleton->priv->changed_properties = NULL;
  skeleton->priv->changed_properties_idle_source = NULL;
  g_mutex_unlock (&skeleton->priv->lock);
  return FALSE;
}

static void
_org_gnome_shell_schedule_emit_changed (OrgGnomeShellSkeleton *skeleton, const _ExtendedGDBusPropertyInfo *info, guint prop_id, const GValue *orig_value)
{
  ChangedProperty *cp;
  GList *l;
  cp = NULL;
  for (l = skeleton->priv->changed_properties; l != NULL; l = l->next)
    {
      ChangedProperty *i_cp = l->data;
      if (i_cp->info == info)
        {
          cp = i_cp;
          break;
        }
    }
  if (cp == NULL)
    {
      cp = g_new0 (ChangedProperty, 1);
      cp->prop_id = prop_id;
      cp->info = info;
      skeleton->priv->changed_properties = g_list_prepend (skeleton->priv->changed_properties, cp);
      g_value_init (&cp->orig_value, G_VALUE_TYPE (orig_value));
      g_value_copy (orig_value, &cp->orig_value);
    }
}

static void
org_gnome_shell_skeleton_notify (GObject      *object,
  GParamSpec *pspec G_GNUC_UNUSED)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);
  g_mutex_lock (&skeleton->priv->lock);
  if (skeleton->priv->changed_properties != NULL &&
      skeleton->priv->changed_properties_idle_source == NULL)
    {
      skeleton->priv->changed_properties_idle_source = g_idle_source_new ();
      g_source_set_priority (skeleton->priv->changed_properties_idle_source, G_PRIORITY_DEFAULT);
      g_source_set_callback (skeleton->priv->changed_properties_idle_source, _org_gnome_shell_emit_changed, g_object_ref (skeleton), (GDestroyNotify) g_object_unref);
      g_source_set_name (skeleton->priv->changed_properties_idle_source, "[generated] _org_gnome_shell_emit_changed");
      g_source_attach (skeleton->priv->changed_properties_idle_source, skeleton->priv->context);
      g_source_unref (skeleton->priv->changed_properties_idle_source);
    }
  g_mutex_unlock (&skeleton->priv->lock);
}

static void
org_gnome_shell_skeleton_set_property (GObject      *object,
  guint         prop_id,
  const GValue *value,
  GParamSpec   *pspec)
{
  const _ExtendedGDBusPropertyInfo *info;
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);
  g_assert (prop_id != 0 && prop_id - 1 < 3);
  info = (const _ExtendedGDBusPropertyInfo *) _org_gnome_shell_property_info_pointers[prop_id - 1];
  g_mutex_lock (&skeleton->priv->lock);
  g_object_freeze_notify (object);
  if (!_g_value_equal (value, &skeleton->priv->properties[prop_id - 1]))
    {
      if (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)) != NULL &&
          info->emits_changed_signal)
        _org_gnome_shell_schedule_emit_changed (skeleton, info, prop_id, &skeleton->priv->properties[prop_id - 1]);
      g_value_copy (value, &skeleton->priv->properties[prop_id - 1]);
      g_object_notify_by_pspec (object, pspec);
    }
  g_mutex_unlock (&skeleton->priv->lock);
  g_object_thaw_notify (object);
}

static void
org_gnome_shell_skeleton_init (OrgGnomeShellSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gnome_shell_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GNOME_SHELL_SKELETON, OrgGnomeShellSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
  skeleton->priv->properties = g_new0 (GValue, 3);
  g_value_init (&skeleton->priv->properties[0], G_TYPE_STRING);
  g_value_init (&skeleton->priv->properties[1], G_TYPE_BOOLEAN);
  g_value_init (&skeleton->priv->properties[2], G_TYPE_STRING);
}

static const gchar *
org_gnome_shell_skeleton_get_mode (OrgGnomeShell *object)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[0]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static gboolean 
org_gnome_shell_skeleton_get_overview_active (OrgGnomeShell *object)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);
  gboolean value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_boolean (&(skeleton->priv->properties[1]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static const gchar *
org_gnome_shell_skeleton_get_shell_version (OrgGnomeShell *object)
{
  OrgGnomeShellSkeleton *skeleton = ORG_GNOME_SHELL_SKELETON (object);
  const gchar *value;
  g_mutex_lock (&skeleton->priv->lock);
  value = g_marshal_value_peek_string (&(skeleton->priv->properties[2]));
  g_mutex_unlock (&skeleton->priv->lock);
  return value;
}

static void
org_gnome_shell_skeleton_class_init (OrgGnomeShellSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gnome_shell_skeleton_finalize;
  gobject_class->get_property = org_gnome_shell_skeleton_get_property;
  gobject_class->set_property = org_gnome_shell_skeleton_set_property;
  gobject_class->notify       = org_gnome_shell_skeleton_notify;


  org_gnome_shell_override_properties (gobject_class, 1);

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gnome_shell_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gnome_shell_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gnome_shell_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gnome_shell_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGnomeShellSkeletonPrivate));
#endif
}

static void
org_gnome_shell_skeleton_iface_init (OrgGnomeShellIface *iface)
{
  iface->accelerator_activated = _org_gnome_shell_on_signal_accelerator_activated;
  iface->accelerator_deactivated = _org_gnome_shell_on_signal_accelerator_deactivated;
  iface->get_mode = org_gnome_shell_skeleton_get_mode;
  iface->get_overview_active = org_gnome_shell_skeleton_get_overview_active;
  iface->get_shell_version = org_gnome_shell_skeleton_get_shell_version;
}

/**
 * org_gnome_shell_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gnome-Shell.top_of_page">org.gnome.Shell</link>.
 *
 * Returns: (transfer full) (type OrgGnomeShellSkeleton): The skeleton object.
 */
OrgGnomeShell *
org_gnome_shell_skeleton_new (void)
{
  return ORG_GNOME_SHELL (g_object_new (TYPE_ORG_GNOME_SHELL_SKELETON, NULL));
}

/* ------------------------------------------------------------------------
 * Code for interface org.gtk.Notifications
 * ------------------------------------------------------------------------
 */

/**
 * SECTION:OrgGtkNotifications
 * @title: OrgGtkNotifications
 * @short_description: Generated C code for the org.gtk.Notifications D-Bus interface
 *
 * This section contains code for working with the <link linkend="gdbus-interface-org-gtk-Notifications.top_of_page">org.gtk.Notifications</link> D-Bus interface in C.
 */

enum
{
  _ORG_GTK_NOTIFICATIONS_ACTION_INVOKED,
};

static unsigned _ORG_GTK_NOTIFICATIONS_SIGNALS[1] = { 0 };

/* ---- Introspection data for org.gtk.Notifications ---- */

static const _ExtendedGDBusArgInfo _org_gtk_notifications_method_info_add_notification_IN_ARG_app_id =
{
  {
    -1,
    (gchar *) "app_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_method_info_add_notification_IN_ARG_id =
{
  {
    -1,
    (gchar *) "id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_method_info_add_notification_IN_ARG_notification =
{
  {
    -1,
    (gchar *) "notification",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gtk_notifications_method_info_add_notification_IN_ARG_pointers[] =
{
  &_org_gtk_notifications_method_info_add_notification_IN_ARG_app_id.parent_struct,
  &_org_gtk_notifications_method_info_add_notification_IN_ARG_id.parent_struct,
  &_org_gtk_notifications_method_info_add_notification_IN_ARG_notification.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gtk_notifications_method_info_add_notification =
{
  {
    -1,
    (gchar *) "AddNotification",
    (GDBusArgInfo **) &_org_gtk_notifications_method_info_add_notification_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-add-notification",
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_method_info_remove_notification_IN_ARG_app_id =
{
  {
    -1,
    (gchar *) "app_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_method_info_remove_notification_IN_ARG_id =
{
  {
    -1,
    (gchar *) "id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gtk_notifications_method_info_remove_notification_IN_ARG_pointers[] =
{
  &_org_gtk_notifications_method_info_remove_notification_IN_ARG_app_id.parent_struct,
  &_org_gtk_notifications_method_info_remove_notification_IN_ARG_id.parent_struct,
  NULL
};

static const _ExtendedGDBusMethodInfo _org_gtk_notifications_method_info_remove_notification =
{
  {
    -1,
    (gchar *) "RemoveNotification",
    (GDBusArgInfo **) &_org_gtk_notifications_method_info_remove_notification_IN_ARG_pointers,
    NULL,
    NULL
  },
  "handle-remove-notification",
  FALSE
};

static const GDBusMethodInfo * const _org_gtk_notifications_method_info_pointers[] =
{
  &_org_gtk_notifications_method_info_add_notification.parent_struct,
  &_org_gtk_notifications_method_info_remove_notification.parent_struct,
  NULL
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_signal_info_action_invoked_ARG_app_id =
{
  {
    -1,
    (gchar *) "app_id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_signal_info_action_invoked_ARG_id =
{
  {
    -1,
    (gchar *) "id",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_signal_info_action_invoked_ARG_action =
{
  {
    -1,
    (gchar *) "action",
    (gchar *) "s",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_signal_info_action_invoked_ARG_parameter =
{
  {
    -1,
    (gchar *) "parameter",
    (gchar *) "av",
    NULL
  },
  FALSE
};

static const _ExtendedGDBusArgInfo _org_gtk_notifications_signal_info_action_invoked_ARG_platform_data =
{
  {
    -1,
    (gchar *) "platform_data",
    (gchar *) "a{sv}",
    NULL
  },
  FALSE
};

static const GDBusArgInfo * const _org_gtk_notifications_signal_info_action_invoked_ARG_pointers[] =
{
  &_org_gtk_notifications_signal_info_action_invoked_ARG_app_id.parent_struct,
  &_org_gtk_notifications_signal_info_action_invoked_ARG_id.parent_struct,
  &_org_gtk_notifications_signal_info_action_invoked_ARG_action.parent_struct,
  &_org_gtk_notifications_signal_info_action_invoked_ARG_parameter.parent_struct,
  &_org_gtk_notifications_signal_info_action_invoked_ARG_platform_data.parent_struct,
  NULL
};

static const _ExtendedGDBusSignalInfo _org_gtk_notifications_signal_info_action_invoked =
{
  {
    -1,
    (gchar *) "ActionInvoked",
    (GDBusArgInfo **) &_org_gtk_notifications_signal_info_action_invoked_ARG_pointers,
    NULL
  },
  "action-invoked"
};

static const GDBusSignalInfo * const _org_gtk_notifications_signal_info_pointers[] =
{
  &_org_gtk_notifications_signal_info_action_invoked.parent_struct,
  NULL
};

static const _ExtendedGDBusInterfaceInfo _org_gtk_notifications_interface_info =
{
  {
    -1,
    (gchar *) "org.gtk.Notifications",
    (GDBusMethodInfo **) &_org_gtk_notifications_method_info_pointers,
    (GDBusSignalInfo **) &_org_gtk_notifications_signal_info_pointers,
    NULL,
    NULL
  },
  "org-gtk-notifications",
};


/**
 * org_gtk_notifications_interface_info:
 *
 * Gets a machine-readable description of the <link linkend="gdbus-interface-org-gtk-Notifications.top_of_page">org.gtk.Notifications</link> D-Bus interface.
 *
 * Returns: (transfer none): A #GDBusInterfaceInfo. Do not free.
 */
GDBusInterfaceInfo *
org_gtk_notifications_interface_info (void)
{
  return (GDBusInterfaceInfo *) &_org_gtk_notifications_interface_info.parent_struct;
}

/**
 * org_gtk_notifications_override_properties:
 * @klass: The class structure for a #GObject derived class.
 * @property_id_begin: The property id to assign to the first overridden property.
 *
 * Overrides all #GObject properties in the #OrgGtkNotifications interface for a concrete class.
 * The properties are overridden in the order they are defined.
 *
 * Returns: The last property id.
 */
guint
org_gtk_notifications_override_properties (GObjectClass *klass G_GNUC_UNUSED, guint property_id_begin)
{
  return property_id_begin - 1;
}


inline static void
org_gtk_notifications_signal_marshal_action_invoked (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_VOID__STRING_STRING_STRING_VARIANT_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gtk_notifications_method_marshal_add_notification (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING_VARIANT (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}

inline static void
org_gtk_notifications_method_marshal_remove_notification (
    GClosure     *closure,
    GValue       *return_value,
    unsigned int  n_param_values,
    const GValue *param_values,
    void         *invocation_hint,
    void         *marshal_data)
{
  _g_dbus_codegen_marshal_BOOLEAN__OBJECT_STRING_STRING (closure,
    return_value, n_param_values, param_values, invocation_hint, marshal_data);
}


/**
 * OrgGtkNotifications:
 *
 * Abstract interface type for the D-Bus interface <link linkend="gdbus-interface-org-gtk-Notifications.top_of_page">org.gtk.Notifications</link>.
 */

/**
 * OrgGtkNotificationsIface:
 * @parent_iface: The parent interface.
 * @handle_add_notification: Handler for the #OrgGtkNotifications::handle-add-notification signal.
 * @handle_remove_notification: Handler for the #OrgGtkNotifications::handle-remove-notification signal.
 * @action_invoked: Handler for the #OrgGtkNotifications::action-invoked signal.
 *
 * Virtual table for the D-Bus interface <link linkend="gdbus-interface-org-gtk-Notifications.top_of_page">org.gtk.Notifications</link>.
 */

typedef OrgGtkNotificationsIface OrgGtkNotificationsInterface;
G_DEFINE_INTERFACE (OrgGtkNotifications, org_gtk_notifications, G_TYPE_OBJECT)

static void
org_gtk_notifications_default_init (OrgGtkNotificationsIface *iface)
{
  /* GObject signals for incoming D-Bus method calls: */
  /**
   * OrgGtkNotifications::handle-add-notification:
   * @object: A #OrgGtkNotifications.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_app_id: Argument passed by remote caller.
   * @arg_id: Argument passed by remote caller.
   * @arg_notification: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-Notifications.AddNotification">AddNotification()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gtk_notifications_complete_add_notification() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-add-notification",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGtkNotificationsIface, handle_add_notification),
    g_signal_accumulator_true_handled,
    NULL,
      org_gtk_notifications_method_marshal_add_notification,
    G_TYPE_BOOLEAN,
    4,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_VARIANT);

  /**
   * OrgGtkNotifications::handle-remove-notification:
   * @object: A #OrgGtkNotifications.
   * @invocation: A #GDBusMethodInvocation.
   * @arg_app_id: Argument passed by remote caller.
   * @arg_id: Argument passed by remote caller.
   *
   * Signal emitted when a remote caller is invoking the <link linkend="gdbus-method-org-gtk-Notifications.RemoveNotification">RemoveNotification()</link> D-Bus method.
   *
   * If a signal handler returns %TRUE, it means the signal handler will handle the invocation (e.g. take a reference to @invocation and eventually call org_gtk_notifications_complete_remove_notification() or e.g. g_dbus_method_invocation_return_error() on it) and no other signal handlers will run. If no signal handler handles the invocation, the %G_DBUS_ERROR_UNKNOWN_METHOD error is returned.
   *
   * Returns: %G_DBUS_METHOD_INVOCATION_HANDLED or %TRUE if the invocation was handled, %G_DBUS_METHOD_INVOCATION_UNHANDLED or %FALSE to let other signal handlers run.
   */
  g_signal_new ("handle-remove-notification",
    G_TYPE_FROM_INTERFACE (iface),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET (OrgGtkNotificationsIface, handle_remove_notification),
    g_signal_accumulator_true_handled,
    NULL,
      org_gtk_notifications_method_marshal_remove_notification,
    G_TYPE_BOOLEAN,
    3,
    G_TYPE_DBUS_METHOD_INVOCATION, G_TYPE_STRING, G_TYPE_STRING);

  /* GObject signals for received D-Bus signals: */
  /**
   * OrgGtkNotifications::action-invoked:
   * @object: A #OrgGtkNotifications.
   * @arg_app_id: Argument.
   * @arg_id: Argument.
   * @arg_action: Argument.
   * @arg_parameter: Argument.
   * @arg_platform_data: Argument.
   *
   * On the client-side, this signal is emitted whenever the D-Bus signal <link linkend="gdbus-signal-org-gtk-Notifications.ActionInvoked">"ActionInvoked"</link> is received.
   *
   * On the service-side, this signal can be used with e.g. g_signal_emit_by_name() to make the object emit the D-Bus signal.
   */
  _ORG_GTK_NOTIFICATIONS_SIGNALS[_ORG_GTK_NOTIFICATIONS_ACTION_INVOKED] =
    g_signal_new ("action-invoked",
      G_TYPE_FROM_INTERFACE (iface),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (OrgGtkNotificationsIface, action_invoked),
      NULL,
      NULL,
      org_gtk_notifications_signal_marshal_action_invoked,
      G_TYPE_NONE,
      5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_VARIANT, G_TYPE_VARIANT);

}

/**
 * org_gtk_notifications_emit_action_invoked:
 * @object: A #OrgGtkNotifications.
 * @arg_app_id: Argument to pass with the signal.
 * @arg_id: Argument to pass with the signal.
 * @arg_action: Argument to pass with the signal.
 * @arg_parameter: Argument to pass with the signal.
 * @arg_platform_data: Argument to pass with the signal.
 *
 * Emits the <link linkend="gdbus-signal-org-gtk-Notifications.ActionInvoked">"ActionInvoked"</link> D-Bus signal.
 */
void
org_gtk_notifications_emit_action_invoked (
    OrgGtkNotifications *object,
    const gchar *arg_app_id,
    const gchar *arg_id,
    const gchar *arg_action,
    GVariant *arg_parameter,
    GVariant *arg_platform_data)
{
  g_signal_emit (object, _ORG_GTK_NOTIFICATIONS_SIGNALS[_ORG_GTK_NOTIFICATIONS_ACTION_INVOKED], 0, arg_app_id, arg_id, arg_action, arg_parameter, arg_platform_data);
}

/**
 * org_gtk_notifications_call_add_notification:
 * @proxy: A #OrgGtkNotificationsProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_id: Argument to pass with the method invocation.
 * @arg_notification: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-Notifications.AddNotification">AddNotification()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gtk_notifications_call_add_notification_finish() to get the result of the operation.
 *
 * See org_gtk_notifications_call_add_notification_sync() for the synchronous, blocking version of this method.
 */
void
org_gtk_notifications_call_add_notification (
    OrgGtkNotifications *proxy,
    const gchar *arg_app_id,
    const gchar *arg_id,
    GVariant *arg_notification,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "AddNotification",
    g_variant_new ("(ss@a{sv})",
                   arg_app_id,
                   arg_id,
                   arg_notification),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gtk_notifications_call_add_notification_finish:
 * @proxy: A #OrgGtkNotificationsProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gtk_notifications_call_add_notification().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gtk_notifications_call_add_notification().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gtk_notifications_call_add_notification_finish (
    OrgGtkNotifications *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gtk_notifications_call_add_notification_sync:
 * @proxy: A #OrgGtkNotificationsProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_id: Argument to pass with the method invocation.
 * @arg_notification: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-Notifications.AddNotification">AddNotification()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gtk_notifications_call_add_notification() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gtk_notifications_call_add_notification_sync (
    OrgGtkNotifications *proxy,
    const gchar *arg_app_id,
    const gchar *arg_id,
    GVariant *arg_notification,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "AddNotification",
    g_variant_new ("(ss@a{sv})",
                   arg_app_id,
                   arg_id,
                   arg_notification),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gtk_notifications_call_remove_notification:
 * @proxy: A #OrgGtkNotificationsProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_id: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied or %NULL.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously invokes the <link linkend="gdbus-method-org-gtk-Notifications.RemoveNotification">RemoveNotification()</link> D-Bus method on @proxy.
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gtk_notifications_call_remove_notification_finish() to get the result of the operation.
 *
 * See org_gtk_notifications_call_remove_notification_sync() for the synchronous, blocking version of this method.
 */
void
org_gtk_notifications_call_remove_notification (
    OrgGtkNotifications *proxy,
    const gchar *arg_app_id,
    const gchar *arg_id,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data)
{
  g_dbus_proxy_call (G_DBUS_PROXY (proxy),
    "RemoveNotification",
    g_variant_new ("(ss)",
                   arg_app_id,
                   arg_id),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    callback,
    user_data);
}

/**
 * org_gtk_notifications_call_remove_notification_finish:
 * @proxy: A #OrgGtkNotificationsProxy.
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gtk_notifications_call_remove_notification().
 * @error: Return location for error or %NULL.
 *
 * Finishes an operation started with org_gtk_notifications_call_remove_notification().
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gtk_notifications_call_remove_notification_finish (
    OrgGtkNotifications *proxy,
    GAsyncResult *res,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (proxy), res, error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gtk_notifications_call_remove_notification_sync:
 * @proxy: A #OrgGtkNotificationsProxy.
 * @arg_app_id: Argument to pass with the method invocation.
 * @arg_id: Argument to pass with the method invocation.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL.
 *
 * Synchronously invokes the <link linkend="gdbus-method-org-gtk-Notifications.RemoveNotification">RemoveNotification()</link> D-Bus method on @proxy. The calling thread is blocked until a reply is received.
 *
 * See org_gtk_notifications_call_remove_notification() for the asynchronous version of this method.
 *
 * Returns: (skip): %TRUE if the call succeeded, %FALSE if @error is set.
 */
gboolean
org_gtk_notifications_call_remove_notification_sync (
    OrgGtkNotifications *proxy,
    const gchar *arg_app_id,
    const gchar *arg_id,
    GCancellable *cancellable,
    GError **error)
{
  GVariant *_ret;
  _ret = g_dbus_proxy_call_sync (G_DBUS_PROXY (proxy),
    "RemoveNotification",
    g_variant_new ("(ss)",
                   arg_app_id,
                   arg_id),
    G_DBUS_CALL_FLAGS_NONE,
    -1,
    cancellable,
    error);
  if (_ret == NULL)
    goto _out;
  g_variant_get (_ret,
                 "()");
  g_variant_unref (_ret);
_out:
  return _ret != NULL;
}

/**
 * org_gtk_notifications_complete_add_notification:
 * @object: A #OrgGtkNotifications.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-Notifications.AddNotification">AddNotification()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gtk_notifications_complete_add_notification (
    OrgGtkNotifications *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/**
 * org_gtk_notifications_complete_remove_notification:
 * @object: A #OrgGtkNotifications.
 * @invocation: (transfer full): A #GDBusMethodInvocation.
 *
 * Helper function used in service implementations to finish handling invocations of the <link linkend="gdbus-method-org-gtk-Notifications.RemoveNotification">RemoveNotification()</link> D-Bus method. If you instead want to finish handling an invocation by returning an error, use g_dbus_method_invocation_return_error() or similar.
 *
 * This method will free @invocation, you cannot use it afterwards.
 */
void
org_gtk_notifications_complete_remove_notification (
    OrgGtkNotifications *object G_GNUC_UNUSED,
    GDBusMethodInvocation *invocation)
{
  g_dbus_method_invocation_return_value (invocation,
    g_variant_new ("()"));
}

/* ------------------------------------------------------------------------ */

/**
 * OrgGtkNotificationsProxy:
 *
 * The #OrgGtkNotificationsProxy structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGtkNotificationsProxyClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGtkNotificationsProxy.
 */

struct _OrgGtkNotificationsProxyPrivate
{
  GData *qdata;
};

static void org_gtk_notifications_proxy_iface_init (OrgGtkNotificationsIface *iface);

#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGtkNotificationsProxy, org_gtk_notifications_proxy, G_TYPE_DBUS_PROXY,
                         G_ADD_PRIVATE (OrgGtkNotificationsProxy)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GTK_NOTIFICATIONS, org_gtk_notifications_proxy_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGtkNotificationsProxy, org_gtk_notifications_proxy, G_TYPE_DBUS_PROXY,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GTK_NOTIFICATIONS, org_gtk_notifications_proxy_iface_init))

#endif
static void
org_gtk_notifications_proxy_finalize (GObject *object)
{
  OrgGtkNotificationsProxy *proxy = ORG_GTK_NOTIFICATIONS_PROXY (object);
  g_datalist_clear (&proxy->priv->qdata);
  G_OBJECT_CLASS (org_gtk_notifications_proxy_parent_class)->finalize (object);
}

static void
org_gtk_notifications_proxy_get_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  GValue       *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gtk_notifications_proxy_set_property (GObject      *object G_GNUC_UNUSED,
  guint         prop_id G_GNUC_UNUSED,
  const GValue *value G_GNUC_UNUSED,
  GParamSpec   *pspec G_GNUC_UNUSED)
{
}

static void
org_gtk_notifications_proxy_g_signal (GDBusProxy *proxy,
  const gchar *sender_name G_GNUC_UNUSED,
  const gchar *signal_name,
  GVariant *parameters)
{
  _ExtendedGDBusSignalInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  gsize n;
  guint signal_id;
  info = (_ExtendedGDBusSignalInfo *) g_dbus_interface_info_lookup_signal ((GDBusInterfaceInfo *) &_org_gtk_notifications_interface_info.parent_struct, signal_name);
  if (info == NULL)
    return;
  num_params = g_variant_n_children (parameters);
  paramv = g_new0 (GValue, num_params + 1);
  g_value_init (&paramv[0], TYPE_ORG_GTK_NOTIFICATIONS);
  g_value_set_object (&paramv[0], proxy);
  g_variant_iter_init (&iter, parameters);
  n = 1;
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.args[n - 1];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GTK_NOTIFICATIONS);
  g_signal_emitv (paramv, signal_id, 0, NULL);
  for (n = 0; n < num_params + 1; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static void
org_gtk_notifications_proxy_g_properties_changed (GDBusProxy *_proxy,
  GVariant *changed_properties,
  const gchar *const *invalidated_properties)
{
  OrgGtkNotificationsProxy *proxy = ORG_GTK_NOTIFICATIONS_PROXY (_proxy);
  guint n;
  const gchar *key;
  GVariantIter *iter;
  _ExtendedGDBusPropertyInfo *info;
  g_variant_get (changed_properties, "a{sv}", &iter);
  while (g_variant_iter_next (iter, "{&sv}", &key, NULL))
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gtk_notifications_interface_info.parent_struct, key);
      g_datalist_remove_data (&proxy->priv->qdata, key);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
  g_variant_iter_free (iter);
  for (n = 0; invalidated_properties[n] != NULL; n++)
    {
      info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gtk_notifications_interface_info.parent_struct, invalidated_properties[n]);
      g_datalist_remove_data (&proxy->priv->qdata, invalidated_properties[n]);
      if (info != NULL)
        g_object_notify (G_OBJECT (proxy), info->hyphen_name);
    }
}

static void
org_gtk_notifications_proxy_init (OrgGtkNotificationsProxy *proxy)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  proxy->priv = org_gtk_notifications_proxy_get_instance_private (proxy);
#else
  proxy->priv = G_TYPE_INSTANCE_GET_PRIVATE (proxy, TYPE_ORG_GTK_NOTIFICATIONS_PROXY, OrgGtkNotificationsProxyPrivate);
#endif

  g_dbus_proxy_set_interface_info (G_DBUS_PROXY (proxy), org_gtk_notifications_interface_info ());
}

static void
org_gtk_notifications_proxy_class_init (OrgGtkNotificationsProxyClass *klass)
{
  GObjectClass *gobject_class;
  GDBusProxyClass *proxy_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize     = org_gtk_notifications_proxy_finalize;
  gobject_class->get_property = org_gtk_notifications_proxy_get_property;
  gobject_class->set_property = org_gtk_notifications_proxy_set_property;

  proxy_class = G_DBUS_PROXY_CLASS (klass);
  proxy_class->g_signal = org_gtk_notifications_proxy_g_signal;
  proxy_class->g_properties_changed = org_gtk_notifications_proxy_g_properties_changed;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGtkNotificationsProxyPrivate));
#endif
}

static void
org_gtk_notifications_proxy_iface_init (OrgGtkNotificationsIface *iface G_GNUC_UNUSED)
{
}

/**
 * org_gtk_notifications_proxy_new:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Asynchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-Notifications.top_of_page">org.gtk.Notifications</link>. See g_dbus_proxy_new() for more details.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gtk_notifications_proxy_new_finish() to get the result of the operation.
 *
 * See org_gtk_notifications_proxy_new_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gtk_notifications_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GTK_NOTIFICATIONS_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.Notifications", NULL);
}

/**
 * org_gtk_notifications_proxy_new_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gtk_notifications_proxy_new().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gtk_notifications_proxy_new().
 *
 * Returns: (transfer full) (type OrgGtkNotificationsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGtkNotifications *
org_gtk_notifications_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GTK_NOTIFICATIONS (ret);
  else
    return NULL;
}

/**
 * org_gtk_notifications_proxy_new_sync:
 * @connection: A #GDBusConnection.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: (nullable): A bus name (well-known or unique) or %NULL if @connection is not a message bus connection.
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Synchronously creates a proxy for the D-Bus interface <link linkend="gdbus-interface-org-gtk-Notifications.top_of_page">org.gtk.Notifications</link>. See g_dbus_proxy_new_sync() for more details.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gtk_notifications_proxy_new() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGtkNotificationsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGtkNotifications *
org_gtk_notifications_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GTK_NOTIFICATIONS_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-connection", connection, "g-object-path", object_path, "g-interface-name", "org.gtk.Notifications", NULL);
  if (ret != NULL)
    return ORG_GTK_NOTIFICATIONS (ret);
  else
    return NULL;
}


/**
 * org_gtk_notifications_proxy_new_for_bus:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @callback: A #GAsyncReadyCallback to call when the request is satisfied.
 * @user_data: User data to pass to @callback.
 *
 * Like org_gtk_notifications_proxy_new() but takes a #GBusType instead of a #GDBusConnection.
 *
 * When the operation is finished, @callback will be invoked in the thread-default main loop of the thread you are calling this method from (see g_main_context_push_thread_default()).
 * You can then call org_gtk_notifications_proxy_new_for_bus_finish() to get the result of the operation.
 *
 * See org_gtk_notifications_proxy_new_for_bus_sync() for the synchronous, blocking version of this constructor.
 */
void
org_gtk_notifications_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data)
{
  g_async_initable_new_async (TYPE_ORG_GTK_NOTIFICATIONS_PROXY, G_PRIORITY_DEFAULT, cancellable, callback, user_data, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.Notifications", NULL);
}

/**
 * org_gtk_notifications_proxy_new_for_bus_finish:
 * @res: The #GAsyncResult obtained from the #GAsyncReadyCallback passed to org_gtk_notifications_proxy_new_for_bus().
 * @error: Return location for error or %NULL
 *
 * Finishes an operation started with org_gtk_notifications_proxy_new_for_bus().
 *
 * Returns: (transfer full) (type OrgGtkNotificationsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGtkNotifications *
org_gtk_notifications_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error)
{
  GObject *ret;
  GObject *source_object;
  source_object = g_async_result_get_source_object (res);
  ret = g_async_initable_new_finish (G_ASYNC_INITABLE (source_object), res, error);
  g_object_unref (source_object);
  if (ret != NULL)
    return ORG_GTK_NOTIFICATIONS (ret);
  else
    return NULL;
}

/**
 * org_gtk_notifications_proxy_new_for_bus_sync:
 * @bus_type: A #GBusType.
 * @flags: Flags from the #GDBusProxyFlags enumeration.
 * @name: A bus name (well-known or unique).
 * @object_path: An object path.
 * @cancellable: (nullable): A #GCancellable or %NULL.
 * @error: Return location for error or %NULL
 *
 * Like org_gtk_notifications_proxy_new_sync() but takes a #GBusType instead of a #GDBusConnection.
 *
 * The calling thread is blocked until a reply is received.
 *
 * See org_gtk_notifications_proxy_new_for_bus() for the asynchronous version of this constructor.
 *
 * Returns: (transfer full) (type OrgGtkNotificationsProxy): The constructed proxy object or %NULL if @error is set.
 */
OrgGtkNotifications *
org_gtk_notifications_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error)
{
  GInitable *ret;
  ret = g_initable_new (TYPE_ORG_GTK_NOTIFICATIONS_PROXY, cancellable, error, "g-flags", flags, "g-name", name, "g-bus-type", bus_type, "g-object-path", object_path, "g-interface-name", "org.gtk.Notifications", NULL);
  if (ret != NULL)
    return ORG_GTK_NOTIFICATIONS (ret);
  else
    return NULL;
}


/* ------------------------------------------------------------------------ */

/**
 * OrgGtkNotificationsSkeleton:
 *
 * The #OrgGtkNotificationsSkeleton structure contains only private data and should only be accessed using the provided API.
 */

/**
 * OrgGtkNotificationsSkeletonClass:
 * @parent_class: The parent class.
 *
 * Class structure for #OrgGtkNotificationsSkeleton.
 */

struct _OrgGtkNotificationsSkeletonPrivate
{
  GValue *properties;
  GList *changed_properties;
  GSource *changed_properties_idle_source;
  GMainContext *context;
  GMutex lock;
};

static void
_org_gtk_notifications_skeleton_handle_method_call (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name,
  const gchar *method_name,
  GVariant *parameters,
  GDBusMethodInvocation *invocation,
  gpointer user_data)
{
  OrgGtkNotificationsSkeleton *skeleton = ORG_GTK_NOTIFICATIONS_SKELETON (user_data);
  _ExtendedGDBusMethodInfo *info;
  GVariantIter iter;
  GVariant *child;
  GValue *paramv;
  gsize num_params;
  guint num_extra;
  gsize n;
  guint signal_id;
  GValue return_value = G_VALUE_INIT;
  info = (_ExtendedGDBusMethodInfo *) g_dbus_method_invocation_get_method_info (invocation);
  g_assert (info != NULL);
  num_params = g_variant_n_children (parameters);
  num_extra = info->pass_fdlist ? 3 : 2;  paramv = g_new0 (GValue, num_params + num_extra);
  n = 0;
  g_value_init (&paramv[n], TYPE_ORG_GTK_NOTIFICATIONS);
  g_value_set_object (&paramv[n++], skeleton);
  g_value_init (&paramv[n], G_TYPE_DBUS_METHOD_INVOCATION);
  g_value_set_object (&paramv[n++], invocation);
  if (info->pass_fdlist)
    {
#ifdef G_OS_UNIX
      g_value_init (&paramv[n], G_TYPE_UNIX_FD_LIST);
      g_value_set_object (&paramv[n++], g_dbus_message_get_unix_fd_list (g_dbus_method_invocation_get_message (invocation)));
#else
      g_assert_not_reached ();
#endif
    }
  g_variant_iter_init (&iter, parameters);
  while ((child = g_variant_iter_next_value (&iter)) != NULL)
    {
      _ExtendedGDBusArgInfo *arg_info = (_ExtendedGDBusArgInfo *) info->parent_struct.in_args[n - num_extra];
      if (arg_info->use_gvariant)
        {
          g_value_init (&paramv[n], G_TYPE_VARIANT);
          g_value_set_variant (&paramv[n], child);
          n++;
        }
      else
        g_dbus_gvariant_to_gvalue (child, &paramv[n++]);
      g_variant_unref (child);
    }
  signal_id = g_signal_lookup (info->signal_name, TYPE_ORG_GTK_NOTIFICATIONS);
  g_value_init (&return_value, G_TYPE_BOOLEAN);
  g_signal_emitv (paramv, signal_id, 0, &return_value);
  if (!g_value_get_boolean (&return_value))
    g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Method %s is not implemented on interface %s", method_name, interface_name);
  g_value_unset (&return_value);
  for (n = 0; n < num_params + num_extra; n++)
    g_value_unset (&paramv[n]);
  g_free (paramv);
}

static GVariant *
_org_gtk_notifications_skeleton_handle_get_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GError **error,
  gpointer user_data)
{
  OrgGtkNotificationsSkeleton *skeleton = ORG_GTK_NOTIFICATIONS_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  GVariant *ret;
  ret = NULL;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gtk_notifications_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      g_value_init (&value, pspec->value_type);
      g_object_get_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      ret = g_dbus_gvalue_to_gvariant (&value, G_VARIANT_TYPE (info->parent_struct.signature));
      g_value_unset (&value);
    }
  return ret;
}

static gboolean
_org_gtk_notifications_skeleton_handle_set_property (
  GDBusConnection *connection G_GNUC_UNUSED,
  const gchar *sender G_GNUC_UNUSED,
  const gchar *object_path G_GNUC_UNUSED,
  const gchar *interface_name G_GNUC_UNUSED,
  const gchar *property_name,
  GVariant *variant,
  GError **error,
  gpointer user_data)
{
  OrgGtkNotificationsSkeleton *skeleton = ORG_GTK_NOTIFICATIONS_SKELETON (user_data);
  GValue value = G_VALUE_INIT;
  GParamSpec *pspec;
  _ExtendedGDBusPropertyInfo *info;
  gboolean ret;
  ret = FALSE;
  info = (_ExtendedGDBusPropertyInfo *) g_dbus_interface_info_lookup_property ((GDBusInterfaceInfo *) &_org_gtk_notifications_interface_info.parent_struct, property_name);
  g_assert (info != NULL);
  pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (skeleton), info->hyphen_name);
  if (pspec == NULL)
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "No property with name %s", property_name);
    }
  else
    {
      if (info->use_gvariant)
        g_value_set_variant (&value, variant);
      else
        g_dbus_gvariant_to_gvalue (variant, &value);
      g_object_set_property (G_OBJECT (skeleton), info->hyphen_name, &value);
      g_value_unset (&value);
      ret = TRUE;
    }
  return ret;
}

static const GDBusInterfaceVTable _org_gtk_notifications_skeleton_vtable =
{
  _org_gtk_notifications_skeleton_handle_method_call,
  _org_gtk_notifications_skeleton_handle_get_property,
  _org_gtk_notifications_skeleton_handle_set_property,
  {NULL}
};

static GDBusInterfaceInfo *
org_gtk_notifications_skeleton_dbus_interface_get_info (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return org_gtk_notifications_interface_info ();
}

static GDBusInterfaceVTable *
org_gtk_notifications_skeleton_dbus_interface_get_vtable (GDBusInterfaceSkeleton *skeleton G_GNUC_UNUSED)
{
  return (GDBusInterfaceVTable *) &_org_gtk_notifications_skeleton_vtable;
}

static GVariant *
org_gtk_notifications_skeleton_dbus_interface_get_properties (GDBusInterfaceSkeleton *_skeleton)
{
  OrgGtkNotificationsSkeleton *skeleton = ORG_GTK_NOTIFICATIONS_SKELETON (_skeleton);

  GVariantBuilder builder;
  guint n;
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_84
  g_variant_builder_init_static (&builder, G_VARIANT_TYPE ("a{sv}"));
#else
  g_variant_builder_init(&builder, G_VARIANT_TYPE ("a{sv}"));
#endif
  if (_org_gtk_notifications_interface_info.parent_struct.properties == NULL)
    goto out;
  for (n = 0; _org_gtk_notifications_interface_info.parent_struct.properties[n] != NULL; n++)
    {
      GDBusPropertyInfo *info = _org_gtk_notifications_interface_info.parent_struct.properties[n];
      if (info->flags & G_DBUS_PROPERTY_INFO_FLAGS_READABLE)
        {
          GVariant *value;
          value = _org_gtk_notifications_skeleton_handle_get_property (g_dbus_interface_skeleton_get_connection (G_DBUS_INTERFACE_SKELETON (skeleton)), NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.Notifications", info->name, NULL, skeleton);
          if (value != NULL)
            {
              g_variant_take_ref (value);
              g_variant_builder_add (&builder, "{sv}", info->name, value);
              g_variant_unref (value);
            }
        }
    }
out:
  return g_variant_builder_end (&builder);
}

static void
org_gtk_notifications_skeleton_dbus_interface_flush (GDBusInterfaceSkeleton *_skeleton G_GNUC_UNUSED)
{
}

static void
_org_gtk_notifications_on_signal_action_invoked (
    OrgGtkNotifications *object,
    const gchar *arg_app_id,
    const gchar *arg_id,
    const gchar *arg_action,
    GVariant *arg_parameter,
    GVariant *arg_platform_data)
{
  OrgGtkNotificationsSkeleton *skeleton = ORG_GTK_NOTIFICATIONS_SKELETON (object);

  GList      *connections, *l;
  GVariant   *signal_variant;
  connections = g_dbus_interface_skeleton_get_connections (G_DBUS_INTERFACE_SKELETON (skeleton));

  signal_variant = g_variant_ref_sink (g_variant_new ("(sss@av@a{sv})",
                   arg_app_id,
                   arg_id,
                   arg_action,
                   arg_parameter,
                   arg_platform_data));
  for (l = connections; l != NULL; l = l->next)
    {
      GDBusConnection *connection = l->data;
      g_dbus_connection_emit_signal (connection,
        NULL, g_dbus_interface_skeleton_get_object_path (G_DBUS_INTERFACE_SKELETON (skeleton)), "org.gtk.Notifications", "ActionInvoked",
        signal_variant, NULL);
    }
  g_variant_unref (signal_variant);
  g_list_free_full (connections, g_object_unref);
}

static void org_gtk_notifications_skeleton_iface_init (OrgGtkNotificationsIface *iface);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
G_DEFINE_TYPE_WITH_CODE (OrgGtkNotificationsSkeleton, org_gtk_notifications_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_ADD_PRIVATE (OrgGtkNotificationsSkeleton)
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GTK_NOTIFICATIONS, org_gtk_notifications_skeleton_iface_init))

#else
G_DEFINE_TYPE_WITH_CODE (OrgGtkNotificationsSkeleton, org_gtk_notifications_skeleton, G_TYPE_DBUS_INTERFACE_SKELETON,
                         G_IMPLEMENT_INTERFACE (TYPE_ORG_GTK_NOTIFICATIONS, org_gtk_notifications_skeleton_iface_init))

#endif
static void
org_gtk_notifications_skeleton_finalize (GObject *object)
{
  OrgGtkNotificationsSkeleton *skeleton = ORG_GTK_NOTIFICATIONS_SKELETON (object);
  g_list_free_full (skeleton->priv->changed_properties, (GDestroyNotify) _changed_property_free);
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  /* coverity[missing_lock : SUPPRESS] */
  g_clear_pointer (&skeleton->priv->changed_properties_idle_source, g_source_destroy);
#else
  if (skeleton->priv->changed_properties_idle_source != NULL)
    g_source_destroy (skeleton->priv->changed_properties_idle_source);
skeleton->priv->changed_properties_idle_source = NULL;
#endif
  g_main_context_unref (skeleton->priv->context);
  g_mutex_clear (&skeleton->priv->lock);
  G_OBJECT_CLASS (org_gtk_notifications_skeleton_parent_class)->finalize (object);
}

static void
org_gtk_notifications_skeleton_init (OrgGtkNotificationsSkeleton *skeleton)
{
#if GLIB_VERSION_MAX_ALLOWED >= GLIB_VERSION_2_38
  skeleton->priv = org_gtk_notifications_skeleton_get_instance_private (skeleton);
#else
  skeleton->priv = G_TYPE_INSTANCE_GET_PRIVATE (skeleton, TYPE_ORG_GTK_NOTIFICATIONS_SKELETON, OrgGtkNotificationsSkeletonPrivate);
#endif

  g_mutex_init (&skeleton->priv->lock);
  skeleton->priv->context = g_main_context_ref_thread_default ();
}

static void
org_gtk_notifications_skeleton_class_init (OrgGtkNotificationsSkeletonClass *klass)
{
  GObjectClass *gobject_class;
  GDBusInterfaceSkeletonClass *skeleton_class;

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = org_gtk_notifications_skeleton_finalize;

  skeleton_class = G_DBUS_INTERFACE_SKELETON_CLASS (klass);
  skeleton_class->get_info = org_gtk_notifications_skeleton_dbus_interface_get_info;
  skeleton_class->get_properties = org_gtk_notifications_skeleton_dbus_interface_get_properties;
  skeleton_class->flush = org_gtk_notifications_skeleton_dbus_interface_flush;
  skeleton_class->get_vtable = org_gtk_notifications_skeleton_dbus_interface_get_vtable;

#if GLIB_VERSION_MAX_ALLOWED < GLIB_VERSION_2_38
  g_type_class_add_private (klass, sizeof (OrgGtkNotificationsSkeletonPrivate));
#endif
}

static void
org_gtk_notifications_skeleton_iface_init (OrgGtkNotificationsIface *iface)
{
  iface->action_invoked = _org_gtk_notifications_on_signal_action_invoked;
}

/**
 * org_gtk_notifications_skeleton_new:
 *
 * Creates a skeleton object for the D-Bus interface <link linkend="gdbus-interface-org-gtk-Notifications.top_of_page">org.gtk.Notifications</link>.
 *
 * Returns: (transfer full) (type OrgGtkNotificationsSkeleton): The skeleton object.
 */
OrgGtkNotifications *
org_gtk_notifications_skeleton_new (void)
{
  return ORG_GTK_NOTIFICATIONS (g_object_new (TYPE_ORG_GTK_NOTIFICATIONS_SKELETON, NULL));
}

